#include "stdafx.h"
#include "Texture Font GeneratorDlg.h"
#include "Texture Font Generator.h"
#include "TextureFont.h"
#include <fstream>
#include <set>
#include <vector>
using namespace std;

static TextureFont *g_pTextureFont = nullptr;


IMPLEMENT_DYNAMIC(CTextureFontGeneratorDlg, CDialog);
CTextureFontGeneratorDlg::CTextureFontGeneratorDlg(CWnd* pParent)
	: CDialog(IDD, pParent), m_bUpdateFontNeeded(false),
	  m_bUpdateFontViewAndCloseUpNeeded(false)
{
	m_hIcon = AfxGetApp()->LoadIcon(IDR_MAINFRAME);

	g_pTextureFont = new TextureFont;
}

CTextureFontGeneratorDlg::~CTextureFontGeneratorDlg()
{
	delete g_pTextureFont;
}

void CTextureFontGeneratorDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	DDX_Control(pDX, IDC_PICTURE, m_FontView);
	DDX_Control(pDX, IDC_SHOWN_PAGE, m_ShownPage);
	DDX_Control(pDX, IDC_FAMILY_LIST, m_FamilyList);
	DDX_Control(pDX, IDC_TEXT_OVERLAP, m_TextOverlap);
	DDX_Control(pDX, IDC_FONT_SIZE, m_FontSize);
	DDX_Control(pDX, IDC_ERROR_OR_WARNING, m_ErrorOrWarning);
	DDX_Control(pDX, IDC_CLOSEUP, m_CloseUp);
	DDX_Control(pDX, IDC_SPIN_TOP, m_SpinTop);
	DDX_Control(pDX, IDC_SPIN_BASELINE, m_SpinBaseline);
	DDX_Control(pDX, IDC_PADDING, m_Padding);
}

BEGIN_MESSAGE_MAP(CTextureFontGeneratorDlg, CDialog)
	ON_WM_DESTROY()
	ON_WM_CLOSE()
	ON_WM_PAINT()
	ON_WM_QUERYDRAGICON()
	ON_WM_MOUSEWHEEL()
	//}}AFX_MSG_MAP
	ON_CBN_SELCHANGE(IDC_SHOWN_PAGE, OnCbnSelchangeShownPage)
	ON_CBN_SELCHANGE(IDC_FAMILY_LIST, OnCbnSelchangeFamilyList)
	ON_EN_CHANGE(IDC_FONT_SIZE, OnEnChangeFontSize)
	ON_COMMAND(ID_STYLE_BOLD, OnStyleBold)
	ON_COMMAND(ID_STYLE_ITALIC, OnStyleItalic)
	ON_COMMAND(ID_STYLE_ANTIALIASED, OnStyleAntialiased)
	ON_NOTIFY(UDN_DELTAPOS, IDC_SPIN_TOP, OnDeltaposSpinTop)
	ON_NOTIFY(UDN_DELTAPOS, IDC_SPIN_BASELINE, OnDeltaposSpinBaseline)
	ON_EN_CHANGE(IDC_PADDING, OnEnChangePadding)
	ON_COMMAND(ID_FILE_SAVE, OnFileSave)
	ON_COMMAND(ID_FILE_EXIT, OnFileExit)
	ON_COMMAND(ID_ACCELERATOR_ITALIC, OnStyleItalic)
	ON_COMMAND(ID_ACCELERATOR_BOLD, OnStyleBold)
	ON_COMMAND(ID_OPTIONS_DOUBLERES, &CTextureFontGeneratorDlg::OnOptionsDoubleres)
	ON_COMMAND(ID_OPTIONS_EXPORTSTROKETEMPLATES, &CTextureFontGeneratorDlg::OnOptionsExportstroketemplates)
	ON_COMMAND(ID_OPTIONS_NUMBERSONLY, &CTextureFontGeneratorDlg::OnOptionsNumbersonly)
END_MESSAGE_MAP()

// TODO: separate those HUGE tables to new file
// TODO: split GBK/2 table at 0x0094BB
/*
	TODO: rename GBK tables
	symbol-gbk1 -> symbol-gb
	kanji-gbk2 -> kanji-gb1 and kanji-gb2
	kanji-gbk3 -> kanji-???
	kanji-gbk4 -> kanji-???
	symbol-gbk5 -> symbol-gbk?
*/

static const unsigned map_main[] = {
	0x003000, 0x003001, 0x003002, 0x003003, 0x003004, 0x003005, 0x003006, 0x003007, 0x003008, 0x003009, 0x00300A, 0x00300B, 0x00300C, 0x00300D, 0x00300E, 0x00300F,
	0x003010, 0x003011, 0x003012, 0x003013, 0x003014, 0x003015, 0x003016, 0x003017, 0x003018, 0x003019, 0x00301A, 0x00301B, 0x00301C, 0x00301D, 0x00301E, 0x00301F,
	0x003020, 0x003021, 0x003022, 0x003023, 0x003024, 0x003025, 0x003026, 0x003027, 0x003028, 0x003029, 0x00302A, 0x00302B, 0x00302C, 0x00302D, 0x00302E, 0x00302F,
	0x003030, 0x003031, 0x003032, 0x003033, 0x003034, 0x003035, 0x003036, 0x003037, 0x003038, 0x003039, 0x00303A, 0x00303B, 0x00303C, 0x00303D, 0x00303E, 0x00303F,
	0x003040, 0x003041, 0x003042, 0x003043, 0x003044, 0x003045, 0x003046, 0x003047, 0x003048, 0x003049, 0x00304A, 0x00304B, 0x00304C, 0x00304D, 0x00304E, 0x00304F,
	0x003050, 0x003051, 0x003052, 0x003053, 0x003054, 0x003055, 0x003056, 0x003057, 0x003058, 0x003059, 0x00305A, 0x00305B, 0x00305C, 0x00305D, 0x00305E, 0x00305F,
	0x003060, 0x003061, 0x003062, 0x003063, 0x003064, 0x003065, 0x003066, 0x003067, 0x003068, 0x003069, 0x00306A, 0x00306B, 0x00306C, 0x00306D, 0x00306E, 0x00306F,
	0x003070, 0x003071, 0x003072, 0x003073, 0x003074, 0x003075, 0x003076, 0x003077, 0x003078, 0x003079, 0x00307A, 0x00307B, 0x00307C, 0x00307D, 0x00307E, 0x00307F,
	0x003080, 0x003081, 0x003082, 0x003083, 0x003084, 0x003085, 0x003086, 0x003087, 0x003088, 0x003089, 0x00308A, 0x00308B, 0x00308C, 0x00308D, 0x00308E, 0x00308F,
	0x003090, 0x003091, 0x003092, 0x003093, 0x003094, 0x003095, 0x003096, 0x003097, 0x003098, 0x003099, 0x00309A, 0x00309B, 0x00309C, 0x00309D, 0x00309E, 0x00309F,
	0x0030A0, 0x0030A1, 0x0030A2, 0x0030A3, 0x0030A4, 0x0030A5, 0x0030A6, 0x0030A7, 0x0030A8, 0x0030A9, 0x0030AA, 0x0030AB, 0x0030AC, 0x0030AD, 0x0030AE, 0x0030AF,
	0x0030B0, 0x0030B1, 0x0030B2, 0x0030B3, 0x0030B4, 0x0030B5, 0x0030B6, 0x0030B7, 0x0030B8, 0x0030B9, 0x0030BA, 0x0030BB, 0x0030BC, 0x0030BD, 0x0030BE, 0x0030BF,
	0x0030C0, 0x0030C1, 0x0030C2, 0x0030C3, 0x0030C4, 0x0030C5, 0x0030C6, 0x0030C7, 0x0030C8, 0x0030C9, 0x0030CA, 0x0030CB, 0x0030CC, 0x0030CD, 0x0030CE, 0x0030CF,
	0x0030D0, 0x0030D1, 0x0030D2, 0x0030D3, 0x0030D4, 0x0030D5, 0x0030D6, 0x0030D7, 0x0030D8, 0x0030D9, 0x0030DA, 0x0030DB, 0x0030DC, 0x0030DD, 0x0030DE, 0x0030DF,
	0x0030E0, 0x0030E1, 0x0030E2, 0x0030E3, 0x0030E4, 0x0030E5, 0x0030E6, 0x0030E7, 0x0030E8, 0x0030E9, 0x0030EA, 0x0030EB, 0x0030EC, 0x0030ED, 0x0030EE, 0x0030EF,
	0x0030F0, 0x0030F1, 0x0030F2, 0x0030F3, 0x0030F4, 0x0030F5, 0x0030F6, 0x0030F7, 0x0030F8, 0x0030F9, 0x0030FA, 0x0030FB, 0x0030FC, 0x0030FD, 0x0030FE, 0x0030FF,
	0
};

static const unsigned map_ank[] = {
	0x000020, 0x000021, 0x000022, 0x000023, 0x000024, 0x000025, 0x000026, 0x000027, 0x000028, 0x000029, 0x00002A, 0x00002B, 0x00002C, 0x00002D, 0x00002E, 0x00002F,
	0x000030, 0x000031, 0x000032, 0x000033, 0x000034, 0x000035, 0x000036, 0x000037, 0x000038, 0x000039, 0x00003A, 0x00003B, 0x00003C, 0x00003D, 0x00003E, 0x00003F,
	0x000040, 0x000041, 0x000042, 0x000043, 0x000044, 0x000045, 0x000046, 0x000047, 0x000048, 0x000049, 0x00004A, 0x00004B, 0x00004C, 0x00004D, 0x00004E, 0x00004F,
	0x000050, 0x000051, 0x000052, 0x000053, 0x000054, 0x000055, 0x000056, 0x000057, 0x000058, 0x000059, 0x00005A, 0x00005B, 0x00005C, 0x00005D, 0x00005E, 0x00005F,
	0x000060, 0x000061, 0x000062, 0x000063, 0x000064, 0x000065, 0x000066, 0x000067, 0x000068, 0x000069, 0x00006A, 0x00006B, 0x00006C, 0x00006D, 0x00006E, 0x00006F,
	0x000070, 0x000071, 0x000072, 0x000073, 0x000074, 0x000075, 0x000076, 0x000077, 0x000078, 0x000079, 0x00007A, 0x00007B, 0x00007C, 0x00007D, 0x00007E, 0x00007F,
	0x00FF61, 0x00FF62, 0x00FF63, 0x00FF64, 0x00FF65, 0x00FF66, 0x00FF67, 0x00FF68, 0x00FF69, 0x00FF6A, 0x00FF6B, 0x00FF6C, 0x00FF6D, 0x00FF6E, 0x00FF6F, 0x00FF70,
	0x00FF71, 0x00FF72, 0x00FF73, 0x00FF74, 0x00FF75, 0x00FF76, 0x00FF77, 0x00FF78, 0x00FF79, 0x00FF7A, 0x00FF7B, 0x00FF7C, 0x00FF7D, 0x00FF7E, 0x00FF7F, 0x00FF80,
	0x00FF81, 0x00FF82, 0x00FF83, 0x00FF84, 0x00FF85, 0x00FF86, 0x00FF87, 0x00FF88, 0x00FF89, 0x00FF8A, 0x00FF8B, 0x00FF8C, 0x00FF8D, 0x00FF8E, 0x00FF8F, 0x00FF90,
	0x00FF91, 0x00FF92, 0x00FF93, 0x00FF94, 0x00FF95, 0x00FF96, 0x00FF97, 0x00FF98, 0x00FF99, 0x00FF9A, 0x00FF9B, 0x00FF9C, 0x00FF9D, 0x00FF9E, 0x00FF9F,
	0
};

/*
// foonmix's kanji 1-9, no longer used
static const unsigned map_kanji_1[] = {
	0x0096E3, 0x006613, 0x005B9A, 0x005909, 0x0066F4, 0x0062BC, 0x00753B, 0x0066F2,
	0x0073FE, 0x005728, 0x009078, 0x00629E, 0x009805, 0x0076EE, 0x00540C, 0x008A2D,
	0x00958B, 0x009589, 0x005185, 0x0089A7, 0x008868, 0x00793A, 0x008A73, 0x007D30,
	0x0053CA, 0x004F5C, 0x0077E2, 0x005370, 0x005834, 0x008B66, 0x00544A, 0x0053EF,
	0x0080FD, 0x006027, 0x0051E6, 0x007406, 0x00901F, 0x006570, 0x005024, 0x008D85,
	0x0052B9, 0x004F7F, 0x007279, 0x0056FA, 0x007701, 0x007565, 0x0051FA, 0x008A5E,
	0x0064CD, 0x005168, 0x006F14, 0x00594F, 0x009032, 0x0070B9, 0x006A5F, 0x0079FB,
	0x005DE6, 0x0053F3, 0x005909, 0x0066F4, 0x004EE5, 0x005916, 0x0080CC, 0x00666F,
	0
};

static const unsigned map_kanji_2[] = {
	0x007FA9, 0x0080CC, 0x00666F, 0x005371, 0x00967A, 0x0072B6, 0x00614B, 0x007FD2,
	0x008B5C, 0x005224, 0x006E1B, 0x005C11, 0x007387, 0x009069, 0x005897, 0x0052A0,
	0x007121, 0x007D42, 0x004E86, 0x0095A2, 0x004FC2, 0x0053C2, 0x008003, 0x004ED6,
	0x008A00, 0x005909, 0x0063DB, 0x008A55, 0x004FA1, 0x006761, 0x004EF6, 0x007A81,
	0x004E8B, 0x008FFD, 0x004E0D, 0x005207, 0x005E38, 0x009664, 0x006B21, 0x007B49,
	0x004F4E, 0x00623B, 0x0053B3, 0x006B8B, 0x0057FA, 0x009014, 0x00975E, 0x005236,
	0x009650, 0x00914D, 0x004FDD, 0x005B58, 0x0079D2, 0x0053CD, 0x005FDC, 0x005C02,
	0x008AA4, 0x009632, 0x006574, 0x00592E, 0x0096A0, 0x0050D5, 0x0063F4, 0x008D77,
	0
};

static const unsigned map_kanji_3[] = {
	0x0076F8, 0x0064B2, 0x005C4B, 0x004F55, 0x007A74, 0x00540D, 0x004F8B, 0x005302,
	0x005FC3, 0x0063FA, 0x009F3B, 0x005F53, 0x008179, 0x009CF4, 0x0075E2, 0x004FBF,
	0x005C3B, 0x0062ED, 0x0092FC, 0x0096EB, 0x006697, 0x009ED2, 0x008FF7, 0x008DEF,
	0x0063A2, 0x00691C, 0x00524D, 0x0054E1, 0x006D6E, 0x0099D0, 0x0063A5, 0x007D9A,
	0x00985E, 0x0058EB, 0x008AF8, 0x00541B, 0x0054C1, 0x007269, 0x0056F0, 0x0058F0,
	0x0053E3, 0x008ABF, 0x0098DF, 0x00515A, 0x007D50, 0x005A5A, 0x008EAB, 0x004E45,
	0x0053CB, 0x0096FB, 0x008A71, 0x005EC3, 0x004F11, 0x009152, 0x0098F2, 0x005F15,
	0x007BED, 0x005974, 0x005B66, 0x008077, 0x00624B, 0x008449, 0x005BBF, 0x0099AC,
	0
};

static const unsigned map_kanji_4[] = {
	0x004EAC, 0x005BBF, 0x006D6A, 0x0066AE, 0x006A2A, 0x005DE3, 0x009D28, 0x00512A,
	0x005C45, 0x00601D, 0x00624B, 0x007DDA, 0x0056DE, 0x007D9A, 0x005206, 0x007136,
	0x0053D6, 0x009054, 0x008074, 0x006700, 0x005B57, 0x008F9E, 0x0066F8, 0x0072EC,
	0x005B63, 0x007BC0, 0x004FFA, 0x006B63, 0x005BFE, 0x008857, 0x00753A, 0x006E0B,
	0x0059FF, 0x007B11, 0x0096A3, 0x0075C5, 0x009662, 0x00958B, 0x009802, 0x0079D1,
	0x005148, 0x00771E, 0x005F13, 0x006CA2, 0x006C37, 0x00606D, 0x0056FD, 0x008033,
	0x004ECA, 0x00901D, 0x005143, 0x004E88, 0x005831, 0x005206, 0x0079D8, 0x005BC6,
	0x007269, 0x007F70, 0x00671B, 0x005CF0, 0x005DEB, 0x005DE5, 0x00696D, 0x00793E,
	0
};

static const unsigned map_kanji_5[] = {
	0x00FF5E, 0x0003C6, 0x00221E, 0x00FF1A, 0x005802, 0x0059B9, 0x00702C, 0x005DDD,
	0x0098DB, 0x005DFB, 0x006A5F, 0x005C71, 0x00689D, 0x006D66, 0x0068EE, 0x0054E1,
	0x00540C, 0x006E05, 0x006C34, 0x005EA7, 0x0052C9, 0x005F37, 0x0096C6, 0x005408,
	0x006025, 0x006B62, 0x0051FA, 0x005185, 0x005247, 0x004E16, 0x00754C, 0x0079D2,
	0x006570, 0x005FB3, 0x009999, 0x00767E, 0x005343, 0x008449, 0x007D17, 0x009759,
	0x00661F, 0x007C73, 0x005C0B, 0x00614E, 0x00543E, 0x005CA1, 0x007AF9, 0x0068A8,
	0x0067F4, 0x0077E2, 0x009B4D, 0x009B4E, 0x00904A, 0x006483, 0x00968A, 0x0091CC,
	0x008CC0, 0x0073B2, 0x00798F, 0x0082B3,
	0
};

static const unsigned map_kanji_6[] = {
	0x005CF6, 0x00793C, 0x006238, 0x006843, 0x009EA6, 0x004EC1, 0x00667A, 0x008D64,
	0x008ECA, 0x006B4C, 0x0077E5, 0x0083EF, 0x00738B, 0x005148, 0x00884C, 0x008005,
	0x009858, 0x00611B, 0x005504, 0x007D50, 0x005A5A, 0x007D20, 0x0076F4, 0x00967D,
	0x005B87, 0x005B99, 0x006D77, 0x008CCA, 0x005192, 0x00967A, 0x009580, 0x005916,
	0x0065E9, 0x0091CD, 0x00756A, 0x00969B, 0x00901A, 0x00518D, 0x0050CF, 0x009BAE,
	0x0091CF, 0x005FC5, 0x008981, 0x007A2E, 0x005F85, 0x005177, 0x008CEA, 0x006D88,
	0x0053BB, 0x005BAE, 0x008A08, 0x007B97, 0x00964D, 0x008C61, 0x005FD8, 0x006F15,
	0x006EC5, 0x00554F, 0x008239, 0x004E57, 0x0065C5, 0x0078BA, 0x00904E, 0x0058CA,
	0
};

static const unsigned map_kanji_7[] = {
	0x006DF1, 0x0090FD, 0x007BC9, 0x00934B, 0x00884C, 0x00811A, 0x007363, 0x00767B,
	0x0096E8, 0x0059CB, 0x007D04, 0x00675F, 0x006240, 0x008679, 0x005F69, 0x006F84,
	0x009803, 0x005E74, 0x005DE1, 0x004EA4, 0x005E7C, 0x004F9B, 0x005E74, 0x008A70,
	0x008349, 0x008A93, 0x005B64, 0x008A69, 0x00610F, 0x005473, 0x006392, 0x006E9D,
	0x00523B, 0x004ED8, 0x00826F, 0x008D70, 0x006211, 0x008B33, 0x007684, 0x00652F,
	0x0053F0, 0x005339, 0x006771, 0x008584, 0x005473, 0x00564C, 0x00629C, 0x006C41,
	0x005F79, 0x00718A, 0x0060DA, 0x007247, 0x006301, 0x006821, 0x0053D7, 0x009A13,
	0x00805E, 0x007AF6, 0x004E89, 0x006FC0, 0x0070BA, 0x0061F8, 0x00683C, 0x00767D,
	0
};

static const unsigned map_kanji_8[] = {
	0x004F59, 0x007B54, 0x00767A, 0x005BC2, 0x005352, 0x0089AA, 0x005B5D, 0x007ACB,
	0x006D3E, 0x005C40, 0x007E01, 0x008A6E, 0x006ED1, 0x0062DD, 0x00559C, 0x006BC0,
	0x005E72, 0x0061F2, 0x0053F7, 0x006557, 0x00582A, 0x007D76, 0x006307, 0x004ED5,
	0x0080C6, 0x004E9B, 0x007D30, 0x006162, 0x009996, 0x0050BE, 0x005DF1, 0x004E3B,
	0x00632F, 0x008FD4, 0x005FAE, 0x005999, 0x0096F7, 0x005869, 0x005E7D, 0x00970A,
	0x00697D, 0x0056E3, 0x009060, 0x006368, 0x0082E6, 0x00611F, 0x005A01, 0x006563,
	0x008CED, 0x006469, 0x00674E, 0x00592B, 0x00821E, 0x009DFA, 0x009808, 0x0096F2,
	0x006D41, 0x008AB0, 0x008EE2, 0x007559, 0x0099C4, 0x006E90, 0x006559, 0x00547C,
	0
};

static const unsigned map_kanji_9[] = {
	0x009006, 0x008DB3, 0x007DF4, 0x009055, 0x007167, 0x005E3D, 0x00534A, 0x008DDD,
	0x0096E2, 0x005F1F, 0x008107, 0x005144, 0x005B8C, 0x0074A7, 0x006551, 0x008A98,
	0x0062D0, 0x006280, 0x008E0F, 0x0096EA, 0x008D8A, 0x007DD1, 0x0062BC, 0x008D6B,
	0x0067D3, 0x005FA1, 0x005ACC, 0x007CDE, 0x00533B, 0x007642, 0x0067FB, 0x009451,
	0x008CDE, 0x007DE8, 0x00524A, 0x009818, 0x005411, 0x006BB5, 0x00968E, 0x006B8A,
	0x00679C, 0x007D1A, 0x005411, 0x008155, 0x00900F, 0x00524D, 0x005FA9, 0x004F4D,
	0x005531, 0x009752, 0x009854, 0x0053EB, 0x0063CF, 0x006BBF, 0x006383, 0x00622F,
	0x006016, 0x005E95, 0x006CB3, 0x00795D, 0x005301, 0x005B9D, 0x0066A6,
	0
};
*/

// Symbol from foonmix
static const unsigned map_symbol[] = {
	0x00FF06, 0x00FF08, 0x00FF09, 0x00FF1F, 0x00FF01, 0x00FF0F, 0x002025, 0x00FF3E, 0x002026, 0x00FF20, 0x00FF04, 0x00FF03, 0x00201D, 0x00FF0B, 0x00FF1D, 0x00FFE5, 0x00FF1B, 0x00FF1C, 0x00FF1E, 0x0030FB, 0x00300C, 0x00300D, 0x00FF5B, 0x00FF5D, 0x00FF05, 0x00FF0A,
	0x00FF41, 0x00FF42, 0x00FF43, 0x00FF44, 0x00FF45, 0x00FF46, 0x00FF47, 0x00FF48, 0x00FF49, 0x00FF4A, 0x00FF4B, 0x00FF4C, 0x00FF4D, 0x00FF4E, 0x00FF4F, 0x00FF50, 0x00FF51, 0x00FF52, 0x00FF53, 0x00FF54, 0x00FF55, 0x00FF56, 0x00FF57, 0x00FF58, 0x00FF59, 0x00FF5A,
	0x00FF21, 0x00FF22, 0x00FF23, 0x00FF24, 0x00FF25, 0x00FF26, 0x00FF27, 0x00FF28, 0x00FF29, 0x00FF2A, 0x00FF2B, 0x00FF2C, 0x00FF2D, 0x00FF2E, 0x00FF2F, 0x00FF30, 0x00FF31, 0x00FF32, 0x00FF33, 0x00FF34, 0x00FF35, 0x00FF36, 0x00FF37, 0x00FF38, 0x00FF39, 0x00FF3A,
	0
};

// JIS X 0208 Symbol
static const unsigned map_symbol2[] = {
	0x00FF0C, 0x00FF0E, 0x00FF1A, 0x0000B4, 0x00FF40, 0x0000A8, 0x00203E, 0x00FF3F, 0x004EDD, 0x002014, 0x002010, 0x00FF3C, 0x002016, 0x00FF5C, 0x002018, 0x002019,
	0x00201C, 0x00FF3B, 0x00FF3D, 0x002212, 0x0000B1, 0x0000D7, 0x0000F7, 0x002260, 0x002266, 0x002267, 0x00221E, 0x002234, 0x002642, 0x002640, 0x0000B0, 0x002032,
	0x002033, 0x002103, 0x0000A2, 0x0000A3, 0x0000A7, 0x002606, 0x002605, 0x0025CB, 0x0025CF, 0x0025CE, 0x0025C7, 0x0025C6, 0x0025A1, 0x0025A0, 0x0025B3, 0x0025B2,
	0x0025BD, 0x0025BC, 0x00203B, 0x002192, 0x002190, 0x002191, 0x002193, 0x002208, 0x00220B, 0x002286, 0x002287, 0x002282, 0x002283, 0x00222A, 0x002229, 0x002227,
	0x002228, 0x0000AC, 0x0021D2, 0x0021D4, 0x002200, 0x002203, 0x002220, 0x0022A5, 0x002312, 0x002202, 0x002207, 0x002261, 0x002252, 0x00226A, 0x00226B, 0x00221A,
	0x00223D, 0x00221D, 0x002235, 0x00222B, 0x00222C, 0x00212B, 0x002030, 0x00266F, 0x00266D, 0x00266A, 0x002020, 0x002021, 0x0000B6, 0x0025EF, 0x00FF10, 0x00FF11,
	0x00FF12, 0x00FF13, 0x00FF14, 0x00FF15, 0x00FF16, 0x00FF17, 0x00FF18, 0x00FF19, 0x000391, 0x000392, 0x000393, 0x000394, 0x000395, 0x000396, 0x000397, 0x000398,
	0x000399, 0x00039A, 0x00039B, 0x00039C, 0x00039D, 0x00039E, 0x00039F, 0x0003A0, 0x0003A1, 0x0003A3, 0x0003A4, 0x0003A5, 0x0003A6, 0x0003A7, 0x0003A8, 0x0003A9,
	0x0003B1, 0x0003B2, 0x0003B3, 0x0003B4, 0x0003B5, 0x0003B6, 0x0003B7, 0x0003B8, 0x0003B9, 0x0003BA, 0x0003BB, 0x0003BC, 0x0003BD, 0x0003BE, 0x0003BF, 0x0003C0,
	0x0003C1, 0x0003C3, 0x0003C4, 0x0003C5, 0x0003C7, 0x0003C8, 0x0003C9, 0x000410, 0x000411, 0x000412, 0x000413, 0x000414, 0x000415, 0x000401, 0x000416, 0x000417,
	0x000418, 0x000419, 0x00041A, 0x00041B, 0x00041C, 0x00041D, 0x00041E, 0x00041F, 0x000420, 0x000421, 0x000422, 0x000423, 0x000424, 0x000425, 0x000426, 0x000427,
	0x000428, 0x000429, 0x00042A, 0x00042B, 0x00042C, 0x00042D, 0x00042E, 0x00042F, 0x000430, 0x000431, 0x000432, 0x000433, 0x000434, 0x000435, 0x000451, 0x000436,
	0x000437, 0x000438, 0x000439, 0x00043A, 0x00043B, 0x00043C, 0x00043D, 0x00043E, 0x00043F, 0x000440, 0x000441, 0x000442, 0x000443, 0x000444, 0x000445, 0x000446,
	0x000447, 0x000448, 0x000449, 0x00044A, 0x00044B, 0x00044C, 0x00044D, 0x00044E, 0x00044F, 0x002500, 0x002502, 0x00250C, 0x002510, 0x002518, 0x002514, 0x00251C,
	0x00252C, 0x002524, 0x002534, 0x00253C, 0x002501, 0x002503, 0x00250F, 0x002513, 0x00251B, 0x002517, 0x002523, 0x002533, 0x00252B, 0x00253B, 0x00254B, 0x002520,
	0x00252F, 0x002528, 0x002537, 0x00253F, 0x00251D, 0x002530, 0x002525, 0x002538, 0x002542,
	0
};

// NEC Symbol
static const unsigned map_symbol3[] = {
	0x002460, 0x002461, 0x002462, 0x002463, 0x002464, 0x002465, 0x002466, 0x002467, 0x002468, 0x002469, 0x00246A, 0x00246B, 0x00246C, 0x00246D, 0x00246E, 0x00246F,
	0x002470, 0x002471, 0x002472, 0x002473, 0x002160, 0x002161, 0x002162, 0x002163, 0x002164, 0x002165, 0x002166, 0x002167, 0x002168, 0x002169, 0x003349, 0x003314,
	0x003322, 0x00334D, 0x003318, 0x003327, 0x003303, 0x003336, 0x003351, 0x003357, 0x00330D, 0x003326, 0x003323, 0x00332B, 0x00334A, 0x00333B, 0x00339C, 0x00339D,
	0x00339E, 0x00338E, 0x00338F, 0x0033C4, 0x0033A1, 0x00337B, 0x002116, 0x0033CD, 0x002121, 0x0032A4, 0x0032A5, 0x0032A6, 0x0032A7, 0x0032A8, 0x003231, 0x003232,
	0x003239, 0x00337E, 0x00337D, 0x00337C, 0x00222E, 0x00221F, 0x0022BF, 0x002170, 0x002171, 0x002172, 0x002173, 0x002174, 0x002175, 0x002176, 0x002177, 0x002178,
	0x002179, 0x00FFE4, 0x00FF07, 0x00FF02,
	0
};

// JIS X 0213 Symbol
static const unsigned map_symbol4[] = {
	0x002284, 0x002285, 0x00228A, 0x00228B, 0x002209, 0x002205, 0x002305, 0x002306, 0x002295, 0x002296, 0x002297, 0x002225, 0x002226, 0x00FF5F, 0x00FF60, 0x002262,
	0x002243, 0x002245, 0x002248, 0x002276, 0x002277, 0x002194, 0x00266E, 0x00266B, 0x00266C, 0x002669, 0x0025B7, 0x0025B6, 0x0025C1, 0x0025C0, 0x002197, 0x002198,
	0x002196, 0x002199, 0x0021C4, 0x0021E8, 0x0021E6, 0x0021E7, 0x0021E9, 0x002934, 0x002935, 0x0029BF, 0x0025C9, 0x00FE46, 0x00FE45, 0x0025E6, 0x002022, 0x002213,
	0x002135, 0x00210F, 0x0033CB, 0x002113, 0x002127, 0x002013, 0x0029FA, 0x0029FB, 0x002664, 0x002660, 0x002662, 0x002666, 0x002661, 0x002665, 0x002667, 0x002663,
	0x0003C2, 0x0024F5, 0x0024F6, 0x0024F7, 0x0024F8, 0x0024F9, 0x0024FA, 0x0024FB, 0x0024FC, 0x0024FD, 0x0024FE, 0x002616, 0x002617, 0x00260E, 0x002600, 0x002601,
	0x002602, 0x002603, 0x002668, 0x0025B1, 0x0031F0, 0x0031F1, 0x0031F2, 0x0031F3, 0x0031F4, 0x0031F5, 0x0031F6, 0x0031F7, 0x0031F8, 0x0031F9, 0x0031FA, 0x0031FB,
	0x0031FC, 0x0031FD, 0x0031FE, 0x0031FF, 0x0023BE, 0x0023BF, 0x0023C0, 0x0023C1, 0x0023C2, 0x0023C3, 0x0023C4, 0x0023C5, 0x0023C6, 0x0023C7, 0x0023C8, 0x0023C9,
	0x0023CA, 0x0023CB, 0x0023CC, 0x0022DA, 0x0022DB, 0x002153, 0x002154, 0x002155, 0x002713, 0x002318, 0x002423, 0x0023CE,
	0
};

static const unsigned map_symbol5[] = {
	0x003251, 0x003252, 0x003253, 0x003254, 0x003255, 0x003256, 0x003257, 0x003258, 0x003259, 0x00325A, 0x00325B, 0x00325C, 0x00325D, 0x00325E, 0x00325F, 0x0032B1, 0x0032B2, 0x0032B3, 0x0032B4, 0x0032B5, 0x0032B6, 0x0032B7, 0x0032B8, 0x0032B9, 0x0032BA, 0x0032BB, 0x0032BC, 0x0032BD, 0x0032BE, 0x0032BF, 0x0025D0, 0x0025D1,
	0x0025D2, 0x0025D3, 0x00203C, 0x002047, 0x002048, 0x002049, 0x0001CD, 0x0001CE, 0x0001D0, 0x001E3E, 0x001E3F, 0x0001F8, 0x0001F9, 0x0001D1, 0x0001D2, 0x0001D4, 0x0001D6, 0x0001D8, 0x0001DA, 0x0001DC, 0x0020AC, 0x0000A0, 0x0000A1, 0x0000A4, 0x0000A6, 0x0000A9, 0x0000AA, 0x0000AB, 0x0000AD, 0x0000AE, 0x0000AF, 0x0000B2,
	0x0000B3, 0x0000B7, 0x0000B8, 0x0000B9, 0x0000BA, 0x0000BB, 0x0000BC, 0x0000BD, 0x0000BE, 0x0000BF, 0x0000C0, 0x0000C1, 0x0000C2, 0x0000C3, 0x0000C4, 0x0000C5, 0x0000C6, 0x0000C7, 0x0000C8, 0x0000C9, 0x0000CA, 0x0000CB, 0x0000CC, 0x0000CD, 0x0000CE, 0x0000CF, 0x0000D0, 0x0000D1, 0x0000D2, 0x0000D3, 0x0000D4, 0x0000D5,
	0x0000D6, 0x0000D8, 0x0000D9, 0x0000DA, 0x0000DB, 0x0000DC, 0x0000DD, 0x0000DE, 0x0000DF, 0x0000E0, 0x0000E1, 0x0000E2, 0x0000E3, 0x0000E4, 0x0000E5, 0x0000E6, 0x0000E7, 0x0000E8, 0x0000E9, 0x0000EA, 0x0000EB, 0x0000EC, 0x0000ED, 0x0000EE, 0x0000EF, 0x0000F0, 0x0000F1, 0x0000F2, 0x0000F3, 0x0000F4, 0x0000F5, 0x0000F6,
	0x0000F8, 0x0000F9, 0x0000FA, 0x0000FB, 0x0000FC, 0x0000FD, 0x0000FE, 0x0000FF, 0x000100, 0x00012A, 0x00016A, 0x000112, 0x00014C, 0x000101, 0x00012B, 0x00016B, 0x000113, 0x00014D, 0x000104, 0x0002D8, 0x000141, 0x00013D, 0x00015A, 0x000160, 0x00015E, 0x000164, 0x000179, 0x00017D, 0x00017B, 0x000105, 0x0002DB, 0x000142,
	0x00013E, 0x00015B, 0x0002C7, 0x000161, 0x00015F, 0x000165, 0x00017A, 0x0002DD, 0x00017E, 0x00017C, 0x000154, 0x000102, 0x000139, 0x000106, 0x00010C, 0x000118, 0x00011A, 0x00010E, 0x000143, 0x000147, 0x000150, 0x000158, 0x00016E, 0x000170, 0x000162, 0x000155, 0x000103, 0x00013A, 0x000107, 0x00010D, 0x000119, 0x00011B,
	0x00010F, 0x000111, 0x000144, 0x000148, 0x000151, 0x000159, 0x00016F, 0x000171, 0x000163, 0x0002D9, 0x000108, 0x00011C, 0x000124, 0x000134, 0x00015C, 0x00016C, 0x000109, 0x00011D, 0x000125, 0x000135, 0x00015D, 0x00016D, 0x000271, 0x00028B, 0x002776, 0x002777, 0x002778, 0x002779, 0x00277A, 0x00277B, 0x00277C, 0x00277D,
	0x00277E, 0x00277F, 0x0024EB, 0x0024EC, 0x0024ED, 0x0024EE, 0x0024EF, 0x0024F0, 0x0024F1, 0x0024F2, 0x0024F3, 0x0024F4, 0x00217A, 0x00217B, 0x0024D0, 0x0024D1, 0x0024D2, 0x0024D3, 0x0024D4, 0x0024D5, 0x0024D6, 0x0024D7, 0x0024D8, 0x0024D9, 0x0024DA, 0x0024DB, 0x0024DC, 0x0024DD, 0x0024DE, 0x0024DF, 0x0024E0, 0x0024E1,
	0x0024E2, 0x0024E3, 0x0024E4, 0x0024E5, 0x0024E6, 0x0024E7, 0x0024E8, 0x0024E9, 0x0032D0, 0x0032D1, 0x0032D2, 0x0032D3, 0x0032D4, 0x0032D5, 0x0032D6, 0x0032D7, 0x0032D8, 0x0032D9, 0x0032DA, 0x0032DB, 0x0032DC, 0x0032DD, 0x0032DE, 0x0032DF, 0x0032E0, 0x0032E1, 0x0032E2, 0x0032E3, 0x0032FA, 0x0032E9, 0x0032E5, 0x0032ED,
	0x0032EC, 0x002051, 0x002042, 0x00216A, 0x00216B, 0x002756, 0x00261E,
	0
};

// JIS X 0212 Symbol
static const unsigned map_symbol_hojo[] = {
	0x0002DA, 0x000384, 0x000385, 0x002122, 0x000386, 0x000388, 0x000389, 0x00038A, 0x0003AA, 0x00038C, 0x00038E, 0x0003AB, 0x00038F, 0x0003AC, 0x0003AD, 0x0003AE,
	0x0003AF, 0x0003CA, 0x000390, 0x0003CC, 0x0003CD, 0x0003CB, 0x0003B0, 0x0003CE, 0x000402, 0x000403, 0x000404, 0x000405, 0x000406, 0x000407, 0x000408, 0x000409,
	0x00040A, 0x00040B, 0x00040C, 0x00040E, 0x00040F, 0x000452, 0x000453, 0x000454, 0x000455, 0x000456, 0x000457, 0x000458, 0x000459, 0x00045A, 0x00045B, 0x00045C,
	0x00045E, 0x00045F, 0x000110, 0x000126, 0x000132, 0x00013F, 0x00014A, 0x000152, 0x000166, 0x000127, 0x000131, 0x000133, 0x000138, 0x000140, 0x000149, 0x00014B,
	0x000153, 0x000167, 0x00010A, 0x000116, 0x00011E, 0x000122, 0x000120, 0x0001CF, 0x000130, 0x00012E, 0x000128, 0x000136, 0x00013B, 0x000145, 0x000156, 0x0001D3,
	0x000172, 0x000168, 0x0001D7, 0x0001DB, 0x0001D9, 0x0001D5, 0x000174, 0x000178, 0x000176, 0x00010B, 0x000117, 0x0001F5, 0x00011F, 0x000121, 0x00012F, 0x000129,
	0x000137, 0x00013C, 0x000146, 0x000157, 0x000173, 0x000169, 0x000175, 0x000177,
	0
};

// For compatibility
static const unsigned map_compatible[] = {
	0x00FFE3, 0x002015, 0x00FF0D, 0x00FF5E,
	0x0000A5, 0x00FFE0, 0x00FFE1, 0x00FFE2,
	0
};

// Jōyō Kanji 1981, not 2010
static const unsigned map_kanji_regular_1[] = {
	0x004E9C, 0x0054C0, 0x00611B, 0x0060AA, 0x0063E1, 0x005727, 0x006271, 0x005B89, 0x006697, 0x006848, 0x004EE5, 0x004F4D, 0x004F9D, 0x005049, 0x0056F2, 0x0059D4, 0x005A01, 0x005C09, 0x00610F, 0x006170, 0x006613, 0x0070BA, 0x007570, 0x0079FB, 0x007DAD, 0x007DEF, 0x0080C3, 0x008863, 0x009055, 0x00907A, 0x00533B, 0x004E95,
	0x0057DF, 0x0080B2, 0x004E00, 0x0058F1, 0x009038, 0x007A32, 0x00828B, 0x005370, 0x0054E1, 0x0056E0, 0x0059FB, 0x005F15, 0x0098F2, 0x009662, 0x009670, 0x0096A0, 0x0097FB, 0x0053F3, 0x005B87, 0x007FBD, 0x0096E8, 0x006E26, 0x006D66, 0x00904B, 0x0096F2, 0x0055B6, 0x005F71, 0x006620, 0x006804, 0x006C38, 0x006CF3, 0x0082F1,
	0x00885B, 0x008A60, 0x0092ED, 0x006DB2, 0x0075AB, 0x0076CA, 0x0099C5, 0x0060A6, 0x008B01, 0x008D8A, 0x0095B2, 0x005186, 0x005712, 0x005BB4, 0x005EF6, 0x0063F4, 0x006CBF, 0x006F14, 0x00708E, 0x007159, 0x00733F, 0x007E01, 0x009060, 0x00925B, 0x005869, 0x006C5A, 0x0051F9, 0x00592E, 0x005965, 0x005F80, 0x005FDC, 0x0062BC,
	0x006A2A, 0x006B27, 0x006BB4, 0x00738B, 0x007FC1, 0x009EC4, 0x006C96, 0x005104, 0x005C4B, 0x0061B6, 0x004E59, 0x005378, 0x006069, 0x006E29, 0x007A4F, 0x0097F3, 0x004E0B, 0x005316, 0x004EEE, 0x004F55, 0x004FA1, 0x004F73, 0x0052A0, 0x0053EF, 0x00590F, 0x005AC1, 0x005BB6, 0x005BE1, 0x0079D1, 0x006687, 0x00679C, 0x0067B6,
	0x006B4C, 0x006CB3, 0x00706B, 0x00798D, 0x007A3C, 0x007B87, 0x0082B1, 0x008377, 0x0083EF, 0x0083D3, 0x008AB2, 0x008CA8, 0x00904E, 0x00868A, 0x006211, 0x00753B, 0x0082BD, 0x008CC0, 0x0096C5, 0x009913, 0x004ECB, 0x004F1A, 0x0089E3, 0x0056DE, 0x00584A, 0x0058CA, 0x005FEB, 0x00602A, 0x006094, 0x0061D0, 0x006212, 0x0062D0,
	0x006539, 0x0068B0, 0x006D77, 0x007070, 0x00754C, 0x007686, 0x007D75, 0x00958B, 0x00968E, 0x008C9D, 0x0052BE, 0x005916, 0x005BB3, 0x006168, 0x006982, 0x006DAF, 0x008857, 0x008A72, 0x0057A3, 0x005687, 0x005404, 0x0062E1, 0x00683C, 0x006838, 0x006BBB, 0x007372, 0x0078BA, 0x007A6B, 0x00899A, 0x0089D2, 0x008F03, 0x0090ED,
	0x0095A3, 0x009694, 0x009769, 0x005B66, 0x005CB3, 0x00697D, 0x00984D, 0x00639B, 0x006F5F, 0x005272, 0x00559D, 0x0062EC, 0x006D3B, 0x006E07, 0x006ED1, 0x008910, 0x008F44, 0x004E14, 0x00682A, 0x005208, 0x004E7E, 0x0051A0, 0x005BD2, 0x00520A, 0x0052D8, 0x0052E7, 0x005DFB, 0x00559A, 0x00582A, 0x005B8C, 0x005B98, 0x005BDB,
	0x005E72, 0x005E79, 0x0060A3, 0x00611F, 0x006163, 0x0061BE, 0x0063DB, 0x006562, 0x0068FA, 0x006B3E, 0x006B53, 0x006C57, 0x006F22, 0x0074B0, 0x007518, 0x0076E3, 0x00770B, 0x007BA1, 0x007C21, 0x007DE9, 0x007F36, 0x00809D, 0x008266, 0x0089B3, 0x008CAB, 0x009084, 0x009451, 0x009593, 0x009591, 0x0095A2, 0x009665, 0x009928,
	0x004E38, 0x00542B, 0x005CB8, 0x00773C, 0x005CA9, 0x009811, 0x009854, 0x009858, 0x004F01, 0x005371, 0x00559C, 0x005668, 0x0057FA, 0x005947, 0x005BC4, 0x005C90, 0x005E0C, 0x005E7E, 0x005FCC, 0x0063EE, 0x00673A, 0x0065D7, 0x0065E2, 0x00671F, 0x0068CB, 0x0068C4, 0x006A5F, 0x005E30, 0x006C17, 0x006C7D, 0x007948, 0x005B63,
	0x007D00, 0x00898F, 0x008A18, 0x008CB4, 0x008D77, 0x008ECC, 0x008F1D, 0x0098E2, 0x009A0E, 0x009B3C, 0x00507D, 0x005100, 0x005B9C, 0x00622F, 0x006280, 0x0064EC, 0x006B3A, 0x0072A0, 0x007591, 0x007FA9, 0x008B70, 0x0083CA, 0x005409, 0x0055AB, 0x008A70, 0x005374, 0x005BA2, 0x00811A, 0x008650, 0x009006, 0x004E18, 0x004E45,
	0x004F11, 0x0053CA, 0x005438, 0x005BAE, 0x005F13, 0x006025, 0x006551, 0x00673D, 0x006C42, 0x006CE3, 0x007403, 0x007A76, 0x007AAE, 0x007D1A, 0x007CFE, 0x007D66, 0x0065E7, 0x00725B, 0x0053BB, 0x005C45, 0x005DE8, 0x0062D2, 0x0062E0, 0x006319, 0x00865A, 0x008A31, 0x008DDD, 0x006F01, 0x009B5A, 0x004EAB, 0x004EAC, 0x004F9B,
	0x007AF6, 0x005171, 0x0051F6, 0x005354, 0x0053EB, 0x005883, 0x005CE1, 0x005F37, 0x006050, 0x00606D, 0x00631F, 0x006559, 0x006A4B, 0x006CC1, 0x0072C2, 0x0072ED, 0x0077EF, 0x0080F8, 0x008105, 0x008208, 0x0090F7, 0x0093E1, 0x0097FF, 0x009A5A, 0x004EF0, 0x0051DD, 0x006681, 0x00696D, 0x005C40, 0x0066F2, 0x006975, 0x007389,
	0x0052E4, 0x005747, 0x0065A4, 0x007434, 0x007981, 0x007B4B, 0x007DCA, 0x0083CC, 0x00895F, 0x008B39, 0x008FD1, 0x0091D1, 0x00541F, 0x009280, 0x004E5D, 0x0053E5, 0x00533A, 0x0082E6, 0x0099C6, 0x005177, 0x00611A, 0x00865E, 0x007A7A, 0x005076, 0x009047, 0x009685, 0x005C48, 0x006398, 0x009774, 0x007E70, 0x006851, 0x0052F2,
	0x00541B, 0x0085AB, 0x008A13, 0x007FA4, 0x008ECD, 0x0090E1, 0x004FC2, 0x0050BE, 0x005211, 0x005144, 0x005553, 0x00578B, 0x005951, 0x005F62, 0x005F84, 0x006075, 0x006176, 0x0061A9, 0x0063B2, 0x00643A, 0x00656C, 0x00666F, 0x006E13, 0x007CFB, 0x007D4C, 0x007D99, 0x00830E, 0x0086CD, 0x008A08, 0x008B66, 0x008EFD, 0x009D8F,
	0x0082B8, 0x008FCE, 0x009BE8, 0x005287, 0x006483, 0x006FC0, 0x005091, 0x006B20, 0x006C7A, 0x006F54, 0x007A74, 0x007D50, 0x008840, 0x006708, 0x004EF6, 0x005039, 0x005065, 0x00517C, 0x005238, 0x005263, 0x00570F, 0x005805, 0x005ACC, 0x005EFA, 0x0061B2, 0x0061F8, 0x00691C, 0x006A29, 0x0072AC, 0x00732E, 0x007814, 0x007D79,
	0x00770C, 0x0080A9, 0x00898B, 0x008B19, 0x008CE2, 0x008ED2, 0x009063, 0x00967A, 0x009855, 0x009A13, 0x005143, 0x00539F, 0x0053B3, 0x005E7B, 0x005F26, 0x006E1B, 0x006E90, 0x007384, 0x0073FE, 0x008A00, 0x009650, 0x00500B, 0x0053E4, 0x00547C, 0x0056FA, 0x005B64, 0x005DF1, 0x005EAB, 0x005F27, 0x006238, 0x006545, 0x0067AF,
	0x006E56, 0x008A87, 0x0096C7, 0x009867, 0x009F13, 0x004E94, 0x004E92, 0x005348, 0x005449, 0x005A2F, 0x005F8C, 0x005FA1, 0x00609F, 0x007881, 0x008A9E, 0x008AA4, 0x008B77, 0x004EA4, 0x004FAF, 0x005019, 0x005149, 0x00516C, 0x00529F, 0x0052B9, 0x00539A, 0x0053E3, 0x005411, 0x00540E, 0x005751, 0x00597D, 0x005B54, 0x005B5D,
	0x005DE5, 0x005DE7, 0x005E78, 0x005E83, 0x005EB7, 0x006052, 0x00614C, 0x006297, 0x0062D8, 0x0063A7, 0x00653B, 0x0066F4, 0x006821, 0x0069CB, 0x006C5F, 0x006D2A, 0x006E2F, 0x006E9D, 0x007532, 0x007687, 0x00786C, 0x007A3F, 0x007D05, 0x007D5E, 0x007DB1, 0x008015, 0x008003, 0x0080AF, 0x00822A, 0x008352, 0x00884C, 0x008861,
	0x008B1B, 0x008CA2, 0x008CFC, 0x0090CA, 0x009175, 0x009271, 0x0092FC, 0x00964D, 0x009805, 0x009999, 0x009AD8, 0x00525B, 0x0053F7, 0x005408, 0x0062F7, 0x008C6A, 0x00514B, 0x00523B, 0x00544A, 0x0056FD, 0x007A40, 0x009177, 0x009ED2, 0x007344, 0x008170, 0x009AA8, 0x008FBC, 0x004ECA, 0x0056F0, 0x0058BE, 0x005A5A, 0x006068,
	0x0061C7, 0x006606, 0x006839, 0x006DF7, 0x007D3A, 0x009B42, 0x004F50, 0x005506, 0x005DE6, 0x005DEE, 0x0067FB, 0x007802, 0x008A50, 0x009396, 0x005EA7, 0x0050B5, 0x0050AC, 0x00518D, 0x006700, 0x0059BB, 0x005BB0, 0x005F69, 0x00624D, 0x0063A1, 0x00683D, 0x006B73, 0x006E08, 0x00707D, 0x007815, 0x00796D, 0x00658E, 0x007D30,
	0x0083DC, 0x0088C1, 0x008F09, 0x00969B, 0x005264, 0x005728, 0x006750, 0x007F6A, 0x008CA1, 0x005742, 0x0054B2, 0x005D0E, 0x004F5C, 0x00524A, 0x00643E, 0x006628, 0x007B56, 0x007D22, 0x00932F, 0x00685C, 0x00518A, 0x005237, 0x005BDF, 0x0064AE, 0x0064E6, 0x00672D, 0x006BBA, 0x0096D1, 0x0076BF, 0x004E09, 0x005098, 0x0053C2,
	0x005C71, 0x0060E8, 0x006563, 0x00685F, 0x007523, 0x007B97, 0x008695, 0x008CDB, 0x009178, 0x0066AB, 0x006B8B, 0x004ED5, 0x004F3A, 0x004F7F, 0x00523A, 0x0053F8, 0x0053F2, 0x0055E3, 0x0056DB, 0x0058EB, 0x0059CB, 0x0059C9, 0x0059FF, 0x005B50, 0x005E02, 0x005E2B, 0x005FD7, 0x00601D, 0x006307, 0x00652F, 0x0065BD, 0x0065E8,
	0x00679D, 0x006B62, 0x006B7B, 0x006C0F, 0x007949, 0x0079C1, 0x007CF8, 0x007D19, 0x007D2B, 0x0080A2, 0x008102, 0x0081F3, 0x008996, 0x008A5E, 0x008A69, 0x008A66, 0x008A8C, 0x008AEE, 0x008CC7, 0x008CDC, 0x0096CC, 0x0098FC, 0x006B6F, 0x004E8B, 0x004F3C, 0x004F8D, 0x005150, 0x005B57, 0x005BFA, 0x006148, 0x006301, 0x006642,
	0x006B21, 0x006ECB, 0x006CBB, 0x0074BD, 0x0078C1, 0x00793A, 0x008033, 0x0081EA, 0x008F9E, 0x005F0F, 0x008B58, 0x008EF8, 0x004E03, 0x0057F7, 0x005931, 0x005BA4, 0x006E7F, 0x006F06, 0x0075BE, 0x008CEA, 0x005B9F, 0x00829D, 0x00820E, 0x005199, 0x005C04, 0x006368, 0x008D66, 0x00659C, 0x00716E, 0x00793E, 0x008005, 0x008B1D,
	0x008ECA, 0x00906E, 0x0086C7, 0x0090AA, 0x00501F, 0x0052FA, 0x005C3A, 0x007235, 0x00914C, 0x0091C8, 0x0082E5, 0x005BC2, 0x005F31, 0x004E3B, 0x0053D6, 0x005B88, 0x00624B, 0x006731, 0x006B8A, 0x0072E9, 0x0073E0, 0x007A2E, 0x008DA3, 0x009152, 0x009996, 0x005112, 0x0053D7, 0x005BFF, 0x006388, 0x006A39, 0x009700, 0x0056DA,
	0x0053CE, 0x005468, 0x005B97, 0x005C31, 0x005DDE, 0x004FEE, 0x006101, 0x0062FE, 0x0079C0, 0x0079CB, 0x007D42, 0x007FD2, 0x0081ED, 0x00821F, 0x008846, 0x008972, 0x009031, 0x00916C, 0x0096C6, 0x00919C, 0x004F4F, 0x005145, 0x005341, 0x005F93, 0x0067D4, 0x006C41, 0x006E0B, 0x007363, 0x007E26, 0x0091CD, 0x009283, 0x0053D4,
	0x005BBF, 0x006DD1, 0x00795D, 0x007E2E, 0x007C9B, 0x00587E, 0x00719F, 0x0051FA, 0x008853, 0x008FF0, 0x004FCA, 0x006625, 0x0077AC, 0x0051C6, 0x005FAA, 0x0065EC, 0x006B89, 0x006E96, 0x006F64, 0x0076FE, 0x007D14, 0x005DE1, 0x009075, 0x009806, 0x0051E6, 0x00521D, 0x006240, 0x006691, 0x005EB6, 0x007DD2, 0x007F72, 0x0066F8,
	0x008AF8, 0x0052A9, 0x0053D9, 0x005973, 0x005E8F, 0x005F90, 0x009664, 0x0050B7, 0x00511F, 0x0052DD, 0x005320, 0x005347, 0x0053EC, 0x005546, 0x005531, 0x005968, 0x005BB5, 0x005C06, 0x005C0F, 0x005C11, 0x005C1A, 0x005E8A, 0x005F70, 0x00627F, 0x006284, 0x0062DB, 0x00638C, 0x006607, 0x00662D, 0x006676, 0x00677E, 0x006CBC,
	0x006D88, 0x006E09, 0x00713C, 0x007126, 0x007167, 0x0075C7, 0x007701, 0x00785D, 0x007901, 0x007965, 0x0079F0, 0x007AE0, 0x007B11, 0x007CA7, 0x007D39, 0x008096, 0x00885D, 0x008A1F, 0x008A3C, 0x008A54, 0x008A73, 0x008C61, 0x008CDE, 0x009418, 0x00969C, 0x004E0A, 0x004E08, 0x004E57, 0x005197, 0x005270, 0x0057CE, 0x005834,
	0x0058CC, 0x005B22, 0x005E38, 0x0060C5, 0x006761, 0x006D44, 0x0072B6, 0x007573, 0x0084B8, 0x008B72, 0x0091B8, 0x009320, 0x005631, 0x0098FE, 0x00690D, 0x006B96, 0x007E54, 0x008077, 0x008272, 0x0089E6, 0x0098DF, 0x008FB1, 0x004F38, 0x004FE1, 0x004FB5, 0x005507, 0x005A20, 0x005BDD, 0x005BE9, 0x005FC3, 0x00614E, 0x00632F,
	0x0065B0, 0x0068EE, 0x006D78, 0x006DF1, 0x007533, 0x00771F, 0x00795E, 0x007D33, 0x0081E3, 0x0085AA, 0x0089AA, 0x008A3A, 0x008EAB, 0x008F9B, 0x009032, 0x0091DD, 0x009707, 0x004EBA, 0x004EC1, 0x005203, 0x005C0B, 0x00751A, 0x005C3D, 0x008FC5, 0x009663, 0x009162, 0x0056F3, 0x005439, 0x005782, 0x005E25, 0x0063A8, 0x006C34,
	0x00708A, 0x007761, 0x007C8B, 0x008870, 0x009042, 0x009154, 0x009318, 0x00968F, 0x009AC4, 0x005D07, 0x006570, 0x0067A2, 0x00636E, 0x006749, 0x006F84, 0x005BF8, 0x004E16, 0x00702C, 0x00755D, 0x00662F, 0x005236, 0x0052E2, 0x0059D3, 0x005F81, 0x006027, 0x006210, 0x00653F, 0x006574, 0x00661F, 0x006674, 0x006B63, 0x006E05,
	0
};

static const unsigned map_kanji_regular_2[] = {
	0x007272, 0x00751F, 0x0076DB, 0x007CBE, 0x008056, 0x0058F0, 0x0088FD, 0x00897F, 0x008AA0, 0x008A93, 0x008ACB, 0x00901D, 0x009752, 0x009759, 0x006589, 0x007A0E, 0x0096BB, 0x005E2D, 0x0060DC, 0x0065A5, 0x006614, 0x006790, 0x0077F3, 0x007A4D, 0x007C4D, 0x007E3E, 0x008CAC, 0x008D64, 0x008DE1, 0x005207, 0x0062D9, 0x0063A5,
	0x006442, 0x006298, 0x008A2D, 0x007A83, 0x007BC0, 0x008AAC, 0x0096EA, 0x007D76, 0x00820C, 0x004ED9, 0x005148, 0x005343, 0x005360, 0x005BA3, 0x005C02, 0x005DDD, 0x006226, 0x006247, 0x006813, 0x006CC9, 0x006D45, 0x006D17, 0x0067D3, 0x006F5C, 0x0065CB, 0x007DDA, 0x007E4A, 0x008239, 0x0085A6, 0x008DF5, 0x009078, 0x009077,
	0x0092AD, 0x009291, 0x009BAE, 0x00524D, 0x005584, 0x006F38, 0x007136, 0x005168, 0x007985, 0x007E55, 0x005851, 0x0063AA, 0x00758E, 0x00790E, 0x007956, 0x0079DF, 0x007C97, 0x007D20, 0x007D44, 0x008A34, 0x00963B, 0x0050E7, 0x005275, 0x0053CC, 0x005009, 0x0055AA, 0x0058EE, 0x00594F, 0x005C64, 0x0060F3, 0x00635C, 0x006383,
	0x00633F, 0x0064CD, 0x0065E9, 0x0066F9, 0x005DE3, 0x0069FD, 0x0071E5, 0x004E89, 0x0076F8, 0x007A93, 0x007DCF, 0x008349, 0x008358, 0x00846C, 0x0085FB, 0x0088C5, 0x008D70, 0x009001, 0x00906D, 0x00971C, 0x009A12, 0x0050CF, 0x005897, 0x00618E, 0x0081D3, 0x008535, 0x008D08, 0x009020, 0x004FC3, 0x005074, 0x005247, 0x005373,
	0x00606F, 0x00675F, 0x006E2C, 0x008DB3, 0x00901F, 0x004FD7, 0x005C5E, 0x008CCA, 0x0065CF, 0x007D9A, 0x005352, 0x005B58, 0x005B6B, 0x005C0A, 0x00640D, 0x006751, 0x004ED6, 0x00591A, 0x00592A, 0x005815, 0x0059A5, 0x0060F0, 0x006253, 0x0099C4, 0x004F53, 0x005BFE, 0x008010, 0x005E2F, 0x005F85, 0x006020, 0x00614B, 0x0066FF,
	0x006CF0, 0x006EDE, 0x0080CE, 0x00888B, 0x008CB8, 0x009000, 0x00902E, 0x00968A, 0x004EE3, 0x0053F0, 0x005927, 0x007B2C, 0x00984C, 0x006EDD, 0x005353, 0x005B85, 0x00629E, 0x0062D3, 0x006CA2, 0x006FEF, 0x008A17, 0x006FC1, 0x008AFE, 0x004F46, 0x009054, 0x00596A, 0x008131, 0x0068DA, 0x008C37, 0x004E39, 0x005358, 0x005606,
	0x0062C5, 0x0063A2, 0x006DE1, 0x0070AD, 0x0077ED, 0x007AEF, 0x0080C6, 0x008A95, 0x00935B, 0x0056E3, 0x0058C7, 0x005F3E, 0x0065AD, 0x006696, 0x006BB5, 0x007537, 0x008AC7, 0x005024, 0x0077E5, 0x005730, 0x006065, 0x006C60, 0x0075F4, 0x007A1A, 0x007F6E, 0x0081F4, 0x009045, 0x007BC9, 0x00755C, 0x007AF9, 0x0084C4, 0x009010,
	0x0079E9, 0x007A92, 0x008336, 0x005AE1, 0x007740, 0x004E2D, 0x004EF2, 0x005B99, 0x005FE0, 0x0062BD, 0x00663C, 0x0067F1, 0x006CE8, 0x00866B, 0x008877, 0x0092F3, 0x0099D0, 0x008457, 0x008CAF, 0x004E01, 0x005146, 0x005E33, 0x005E81, 0x005F14, 0x005F35, 0x005F6B, 0x005FB4, 0x0061F2, 0x006311, 0x00671D, 0x006F6E, 0x00753A,
	0x00773A, 0x008074, 0x008139, 0x008178, 0x008ABF, 0x008D85, 0x008DF3, 0x009577, 0x009802, 0x009CE5, 0x0052C5, 0x0076F4, 0x006715, 0x006C88, 0x0073CD, 0x008CC3, 0x0093AE, 0x009673, 0x006D25, 0x00589C, 0x008FFD, 0x0075DB, 0x00901A, 0x00585A, 0x006F2C, 0x00576A, 0x0091E3, 0x004EAD, 0x004F4E, 0x00505C, 0x005075, 0x008C9E,
	0x005448, 0x005824, 0x005B9A, 0x005E1D, 0x005E95, 0x005EAD, 0x005EF7, 0x005F1F, 0x0062B5, 0x0063D0, 0x007A0B, 0x007DE0, 0x008247, 0x008A02, 0x009013, 0x0090B8, 0x006CE5, 0x006458, 0x006575, 0x006EF4, 0x007684, 0x007B1B, 0x009069, 0x0054F2, 0x005FB9, 0x0064A4, 0x008FED, 0x009244, 0x005178, 0x005929, 0x005C55, 0x005E97,
	0x006DFB, 0x008EE2, 0x0070B9, 0x004F1D, 0x006BBF, 0x007530, 0x0096FB, 0x005410, 0x005857, 0x005F92, 0x006597, 0x006E21, 0x00767B, 0x009014, 0x0090FD, 0x0052AA, 0x005EA6, 0x00571F, 0x005974, 0x006012, 0x005012, 0x00515A, 0x0051AC, 0x0051CD, 0x005200, 0x005510, 0x005854, 0x005CF6, 0x0060BC, 0x006295, 0x00642D, 0x006771,
	0x006843, 0x0068DF, 0x0076D7, 0x006E6F, 0x00706F, 0x005F53, 0x0075D8, 0x007B49, 0x007B54, 0x007B52, 0x007CD6, 0x007D71, 0x005230, 0x008A0E, 0x008B04, 0x008C46, 0x008E0F, 0x009003, 0x00900F, 0x009676, 0x00982D, 0x009A30, 0x0095D8, 0x0050CD, 0x0052D5, 0x00540C, 0x005802, 0x005C0E, 0x006D1E, 0x007AE5, 0x0080F4, 0x009053,
	0x009285, 0x005CE0, 0x00533F, 0x005F97, 0x005FB3, 0x007279, 0x007763, 0x007BE4, 0x006BD2, 0x0072EC, 0x008AAD, 0x0051F8, 0x007A81, 0x005C4A, 0x005C6F, 0x008C5A, 0x0066C7, 0x00920D, 0x005185, 0x007E04, 0x005357, 0x008EDF, 0x0096E3, 0x004E8C, 0x005C3C, 0x005F10, 0x008089, 0x0065E5, 0x004E73, 0x005165, 0x005982, 0x005C3F,
	0x004EFB, 0x00598A, 0x005FCD, 0x008A8D, 0x005BE7, 0x00732B, 0x0071B1, 0x005E74, 0x005FF5, 0x0071C3, 0x007C98, 0x0060A9, 0x006FC3, 0x007D0D, 0x0080FD, 0x008133, 0x008FB2, 0x00628A, 0x008987, 0x006CE2, 0x006D3E, 0x007834, 0x005A46, 0x0099AC, 0x004FF3, 0x005EC3, 0x0062DD, 0x006392, 0x006557, 0x00676F, 0x0080CC, 0x0080BA,
	0x008F29, 0x00914D, 0x00500D, 0x0057F9, 0x005A92, 0x006885, 0x008CB7, 0x0058F2, 0x008CE0, 0x00966A, 0x004F2F, 0x00535A, 0x0062CD, 0x006CCA, 0x00767D, 0x008236, 0x008584, 0x008FEB, 0x006F20, 0x007206, 0x007E1B, 0x009EA6, 0x007BB1, 0x00808C, 0x007551, 0x00516B, 0x009262, 0x00767A, 0x009AEA, 0x004F10, 0x007F70, 0x00629C,
	0x0095A5, 0x004F34, 0x005224, 0x00534A, 0x0053CD, 0x005E06, 0x00642C, 0x00677F, 0x007248, 0x0072AF, 0x0073ED, 0x007554, 0x007E41, 0x00822C, 0x0085E9, 0x008CA9, 0x007BC4, 0x007169, 0x009812, 0x0098EF, 0x006669, 0x00756A, 0x0076E4, 0x0086EE, 0x005351, 0x005426, 0x005983, 0x005F7C, 0x0060B2, 0x006249, 0x006279, 0x0062AB,
	0x006BD4, 0x006CCC, 0x0075B2, 0x0076AE, 0x007891, 0x0079D8, 0x007F77, 0x0080A5, 0x0088AB, 0x008CBB, 0x00907F, 0x00975E, 0x0098DB, 0x005099, 0x005C3E, 0x005FAE, 0x007F8E, 0x009F3B, 0x005339, 0x005FC5, 0x007B46, 0x0059EB, 0x00767E, 0x004FF5, 0x006A19, 0x006C37, 0x006F02, 0x007968, 0x008868, 0x008A55, 0x0063CF, 0x0075C5,
	0x0079D2, 0x0082D7, 0x0054C1, 0x006D5C, 0x008CA7, 0x008CD3, 0x00983B, 0x00654F, 0x0074F6, 0x004E0D, 0x004ED8, 0x00592B, 0x005A66, 0x005BCC, 0x005E03, 0x005E9C, 0x006016, 0x006276, 0x006577, 0x00666E, 0x006D6E, 0x007236, 0x007B26, 0x008150, 0x00819A, 0x008B5C, 0x008CA0, 0x008CE6, 0x008D74, 0x009644, 0x004FAE, 0x006B66,
	0x00821E, 0x0090E8, 0x005C01, 0x0098A8, 0x004F0F, 0x00526F, 0x005FA9, 0x005E45, 0x00670D, 0x00798F, 0x008179, 0x008907, 0x008986, 0x006255, 0x006CB8, 0x004ECF, 0x007269, 0x005206, 0x005674, 0x0058B3, 0x0061A4, 0x00596E, 0x007C89, 0x007D1B, 0x0096F0, 0x006587, 0x00805E, 0x004E19, 0x004F75, 0x005175, 0x005840, 0x005E63,
	0x005E73, 0x005F0A, 0x0067C4, 0x004E26, 0x009589, 0x00965B, 0x007C73, 0x0058C1, 0x007656, 0x005225, 0x00504F, 0x005909, 0x007247, 0x007DE8, 0x008FBA, 0x008FD4, 0x00904D, 0x004FBF, 0x0052C9, 0x005F01, 0x004FDD, 0x008217, 0x006355, 0x006B69, 0x0088DC, 0x007A42, 0x0052DF, 0x005893, 0x006155, 0x0066AE, 0x006BCD, 0x007C3F,
	0x005023, 0x004FF8, 0x005305, 0x005831, 0x005949, 0x005B9D, 0x005CF0, 0x005D29, 0x0062B1, 0x00653E, 0x0065B9, 0x006CD5, 0x006CE1, 0x007832, 0x007E2B, 0x0080DE, 0x0082B3, 0x008912, 0x008A2A, 0x008C4A, 0x0090A6, 0x0098FD, 0x004E4F, 0x004EA1, 0x00508D, 0x005256, 0x00574A, 0x0059A8, 0x005E3D, 0x005FD8, 0x005FD9, 0x00623F,
	0x0066B4, 0x00671B, 0x0067D0, 0x0068D2, 0x005192, 0x007D21, 0x0080AA, 0x0081A8, 0x008B00, 0x008CBF, 0x009632, 0x005317, 0x0050D5, 0x0058A8, 0x0064B2, 0x006734, 0x007267, 0x006CA1, 0x005800, 0x005954, 0x00672C, 0x007FFB, 0x0051E1, 0x0076C6, 0x006469, 0x0078E8, 0x009B54, 0x009EBB, 0x0057CB, 0x0059B9, 0x00679A, 0x006BCE,
	0x005E55, 0x00819C, 0x0053C8, 0x0062B9, 0x00672B, 0x007E6D, 0x004E07, 0x006162, 0x006E80, 0x006F2B, 0x005473, 0x00672A, 0x009B45, 0x005CAC, 0x005BC6, 0x008108, 0x005999, 0x006C11, 0x007720, 0x0052D9, 0x005922, 0x007121, 0x0077DB, 0x009727, 0x005A7F, 0x005A18, 0x00540D, 0x00547D, 0x00660E, 0x0076DF, 0x008FF7, 0x009298,
	0x009CF4, 0x006EC5, 0x00514D, 0x007DBF, 0x009762, 0x006A21, 0x008302, 0x005984, 0x006BDB, 0x00731B, 0x0076F2, 0x007DB2, 0x008017, 0x006728, 0x009ED9, 0x0076EE, 0x00623B, 0x00554F, 0x007D0B, 0x009580, 0x005301, 0x00591C, 0x0091CE, 0x0077E2, 0x005384, 0x005F79, 0x007D04, 0x0085AC, 0x008A33, 0x008E8D, 0x0067F3, 0x006109,
	0x006CB9, 0x007652, 0x008AED, 0x008F38, 0x00552F, 0x00512A, 0x0052C7, 0x0053CB, 0x005E7D, 0x0060A0, 0x006182, 0x006709, 0x007336, 0x007531, 0x0088D5, 0x008A98, 0x00904A, 0x0090F5, 0x0096C4, 0x00878D, 0x005915, 0x004E88, 0x004F59, 0x004E0E, 0x008A89, 0x009810, 0x005E7C, 0x005BB9, 0x005EB8, 0x0063DA, 0x0063FA, 0x0064C1,
	0x0066DC, 0x0069D8, 0x006D0B, 0x006EB6, 0x007528, 0x007AAF, 0x007F8A, 0x008449, 0x008981, 0x008B21, 0x008E0A, 0x00967D, 0x00990A, 0x006291, 0x006B32, 0x006D74, 0x007FCC, 0x007FFC, 0x007F85, 0x0088F8, 0x006765, 0x00983C, 0x0096F7, 0x007D61, 0x00843D, 0x00916A, 0x004E71, 0x005375, 0x006B04, 0x006FEB, 0x0089A7, 0x005229,
	0x00540F, 0x005C65, 0x007406, 0x0075E2, 0x0088CF, 0x0091CC, 0x0096E2, 0x009678, 0x005F8B, 0x007387, 0x007ACB, 0x007565, 0x006D41, 0x007559, 0x00786B, 0x007C92, 0x009686, 0x007ADC, 0x00616E, 0x0065C5, 0x00865C, 0x004E86, 0x0050DA, 0x004E21, 0x005BEE, 0x006599, 0x006DBC, 0x00731F, 0x007642, 0x007CE7, 0x00826F, 0x0091CF,
	0x009675, 0x009818, 0x00529B, 0x007DD1, 0x00502B, 0x005398, 0x006797, 0x0081E8, 0x008F2A, 0x0096A3, 0x005841, 0x006D99, 0x007D2F, 0x00985E, 0x004EE4, 0x004F8B, 0x0051B7, 0x0052B1, 0x00793C, 0x009234, 0x0096B7, 0x0096F6, 0x00970A, 0x009E97, 0x009F62, 0x0066A6, 0x006B74, 0x005217, 0x0052A3, 0x0070C8, 0x0088C2, 0x005EC9,
	0x00604B, 0x007DF4, 0x009023, 0x00932C, 0x007089, 0x008DEF, 0x009732, 0x0052B4, 0x005ECA, 0x006717, 0x00697C, 0x006D6A, 0x006F0F, 0x008001, 0x0090CE, 0x00516D, 0x009332, 0x008AD6, 0x00548C, 0x008A71, 0x008CC4, 0x0060D1, 0x0067A0, 0x006E7E, 0x008155,
	0
};

// Added to Jōyō Kanji in 2010
static const unsigned map_kanji_regular_3[] = {
	0x006328, 0x0066D6, 0x005B9B, 0x005D50, 0x00754F, 0x00840E, 0x006905, 0x005F59, 0x008328, 0x0054BD, 0x006DEB, 0x005504, 0x009B31, 0x006028, 0x005A9B, 0x008276,
	0x0065FA, 0x005CA1, 0x0081C6, 0x004FFA, 0x0082DB, 0x007259, 0x0074E6, 0x006977, 0x006F70, 0x008AE7, 0x005D16, 0x0084CB, 0x009AB8, 0x0067FF, 0x00984E, 0x00845B,
	0x0091DC, 0x00938C, 0x0097D3, 0x0073A9, 0x004F0E, 0x004E80, 0x006BC0, 0x00757F, 0x0081FC, 0x0055C5, 0x005DFE, 0x0050C5, 0x009326, 0x0060E7, 0x004E32, 0x007A9F,
	0x00718A, 0x008A63, 0x0061AC, 0x007A3D, 0x009699, 0x006841, 0x0062F3, 0x009375, 0x008237, 0x0080A1, 0x00864E, 0x00932E, 0x0052FE, 0x006897, 0x005589, 0x004E5E,
	0x0050B2, 0x0099D2, 0x009803, 0x0075D5, 0x006C99, 0x00632B, 0x0091C7, 0x00585E, 0x0057FC, 0x0067F5, 0x005239, 0x0062F6, 0x0065AC, 0x006063, 0x00646F, 0x00990C,
	0x009E7F, 0x020B9F, 0x005AC9, 0x00816B, 0x00546A, 0x008896, 0x007F9E, 0x008E74, 0x0061A7, 0x0062ED, 0x005C3B, 0x0082AF, 0x00814E, 0x009808, 0x0088FE, 0x0051C4,
	0x009192, 0x00810A, 0x00621A, 0x00714E, 0x007FA8, 0x00817A, 0x008A6E, 0x007B8B, 0x0081B3, 0x0072D9, 0x009061, 0x0066FD, 0x00723D, 0x0075E9, 0x008E2A, 0x006349,
	0x00905C, 0x006C70, 0x00553E, 0x005806, 0x006234, 0x008AB0, 0x0065E6, 0x007DBB, 0x007DFB, 0x00914E, 0x008CBC, 0x005632, 0x006357, 0x00690E, 0x00722A, 0x009DB4,
	0x008AE6, 0x006EBA, 0x005861, 0x0059AC, 0x008CED, 0x0085E4, 0x0077B3, 0x006803, 0x009813, 0x008CAA, 0x004E3C, 0x0090A3, 0x005948, 0x0068A8, 0x008B0E, 0x00934B,
	0x005302, 0x008679, 0x00637B, 0x007F75, 0x00525D, 0x007BB8, 0x006C3E, 0x006C4E, 0x00962A, 0x006591, 0x007709, 0x00819D, 0x008098, 0x008A03, 0x00961C, 0x00853D,
	0x009905, 0x0074A7, 0x008511, 0x0054FA, 0x008702, 0x008C8C, 0x009830, 0x007766, 0x0052C3, 0x006627, 0x006795, 0x00871C, 0x0051A5, 0x009EBA, 0x0051B6, 0x005F25,
	0x0095C7, 0x0055A9, 0x006E67, 0x005996, 0x00760D, 0x006C83, 0x0062C9, 0x008FA3, 0x0085CD, 0x007483, 0x006144, 0x004FB6, 0x0077AD, 0x007460, 0x005442, 0x008CC2,
	0x005F04, 0x007C60, 0x009E93, 0x008107,
	0
};

// JIS X 0208 Level 1
static const unsigned map_kanji_jis1[] = {
	0x005516, 0x005A03, 0x00963F, 0x0059F6, 0x009022, 0x008475, 0x00831C, 0x007A50, 0x006E25, 0x0065ED, 0x008466, 0x0082A6, 0x009BF5, 0x006893, 0x0065A1, 0x0059D0, 0x00867B, 0x0098F4, 0x007D62, 0x007DBE, 0x009B8E, 0x006216, 0x007C9F, 0x0088B7, 0x005EB5, 0x006309, 0x00978D, 0x00674F, 0x004F0A, 0x005937, 0x0060DF, 0x008B02,
	0x004EA5, 0x0090C1, 0x0078EF, 0x006EA2, 0x009C2F, 0x005141, 0x0080E4, 0x00852D, 0x00540B, 0x0070CF, 0x008FC2, 0x00536F, 0x009D5C, 0x007ABA, 0x004E11, 0x007893, 0x005618, 0x006B1D, 0x00851A, 0x009C3B, 0x0059E5, 0x0053A9, 0x0074DC, 0x00958F, 0x005642, 0x004E91, 0x00834F, 0x0053E1, 0x005B30, 0x0066F3, 0x006D29, 0x00745B,
	0x0076C8, 0x007A4E, 0x009834, 0x00698E, 0x0053AD, 0x005830, 0x005944, 0x0063A9, 0x007114, 0x0071D5, 0x0082D1, 0x008597, 0x009D1B, 0x0065BC, 0x007525, 0x008956, 0x009D2C, 0x009D0E, 0x00837B, 0x006876, 0x007261, 0x004F3D, 0x005609, 0x0073C2, 0x0079BE, 0x008304, 0x008766, 0x005629, 0x008FE6, 0x00971E, 0x004FC4, 0x005CE8,
	0x0081E5, 0x0086FE, 0x0099D5, 0x005EFB, 0x006062, 0x009B41, 0x006666, 0x0082A5, 0x0087F9, 0x0051F1, 0x0054B3, 0x00788D, 0x0093A7, 0x006D6C, 0x0099A8, 0x0086D9, 0x0086CE, 0x00920E, 0x005283, 0x005ED3, 0x0064B9, 0x008D6B, 0x007B20, 0x006A2B, 0x006A7F, 0x0068B6, 0x009C0D, 0x006070, 0x009C39, 0x0053F6, 0x00691B, 0x006A3A,
	0x009784, 0x00515C, 0x007AC3, 0x0084B2, 0x00565B, 0x009D28, 0x006822, 0x008305, 0x008431, 0x007CA5, 0x0082C5, 0x004F83, 0x0059E6, 0x0067D1, 0x006853, 0x006F97, 0x006F45, 0x007AFF, 0x007FF0, 0x00839E, 0x008ACC, 0x008218, 0x005DCC, 0x00764C, 0x007FEB, 0x008D0B, 0x0096C1, 0x005B09, 0x006BC5, 0x007A00, 0x005FBD, 0x005993,
	0x007947, 0x0087FB, 0x008ABC, 0x0063AC, 0x0097A0, 0x005403, 0x006854, 0x006A58, 0x007827, 0x006775, 0x009ECD, 0x004EC7, 0x006C72, 0x007078, 0x007B08, 0x006E20, 0x0092F8, 0x0079A6, 0x004EA8, 0x004FA0, 0x0050D1, 0x005147, 0x005321, 0x00537F, 0x0055AC, 0x005F4A, 0x00602F, 0x00854E, 0x009957, 0x005C2D, 0x006850, 0x007C81,
	0x006B23, 0x006B3D, 0x0079BD, 0x0082B9, 0x00887F, 0x005036, 0x0072D7, 0x007396, 0x0077E9, 0x008EAF, 0x0099C8, 0x0055B0, 0x005BD3, 0x006ADB, 0x0091E7, 0x005C51, 0x006C93, 0x008F61, 0x007AAA, 0x009688, 0x007C82, 0x006817, 0x00936C, 0x005366, 0x008888, 0x007941, 0x00572D, 0x0073EA, 0x006167, 0x006842, 0x007566, 0x007E4B,
	0x007F6B, 0x00834A, 0x00981A, 0x00621F, 0x008A23, 0x005026, 0x0055A7, 0x006372, 0x00727D, 0x00786F, 0x009E78, 0x007D43, 0x008AFA, 0x004E4E, 0x0059D1, 0x0072D0, 0x007CCA, 0x0088B4, 0x0080E1, 0x0083F0, 0x008DE8, 0x009237, 0x004F0D, 0x00543E, 0x0068A7, 0x006A8E, 0x00745A, 0x009190, 0x009BC9, 0x004F7C, 0x005016, 0x0057A2,
	0x005B8F, 0x005DF7, 0x005E9A, 0x005F18, 0x006602, 0x006643, 0x00676D, 0x006D69, 0x007CE0, 0x007D18, 0x0080B1, 0x008154, 0x00818F, 0x00783F, 0x0095A4, 0x009D3B, 0x0052AB, 0x0058D5, 0x006FE0, 0x008F5F, 0x009EB9, 0x009D60, 0x006F09, 0x007511, 0x005FFD, 0x0060DA, 0x0072DB, 0x006B64, 0x005764, 0x00660F, 0x0068B1, 0x00826E,
	0x004E9B, 0x0053C9, 0x005D6F, 0x007473, 0x0088DF, 0x005750, 0x0054C9, 0x007280, 0x007826, 0x0051B4, 0x00583A, 0x00698A, 0x0080B4, 0x007895, 0x009DFA, 0x00548B, 0x006714, 0x007A84, 0x009BAD, 0x007B39, 0x005319, 0x0085A9, 0x007690, 0x009BD6, 0x00634C, 0x009306, 0x009BAB, 0x006652, 0x006492, 0x0071E6, 0x0073CA, 0x007E82,
	0x008B83, 0x009910, 0x004ED4, 0x005C4D, 0x005B5C, 0x0065AF, 0x007345, 0x00723E, 0x0075D4, 0x00800C, 0x008494, 0x006C50, 0x009D2B, 0x007AFA, 0x005B8D, 0x0096EB, 0x0053F1, 0x006089, 0x008500, 0x007BE0, 0x005072, 0x0067F4, 0x005C61, 0x00854A, 0x007E1E, 0x007D17, 0x006753, 0x00707C, 0x00932B, 0x0060F9, 0x007DAC, 0x006D32,
	0x007E4D, 0x008490, 0x008B90, 0x008F2F, 0x00914B, 0x004EC0, 0x00620E, 0x005919, 0x005CFB, 0x007AE3, 0x00821C, 0x0099FF, 0x00696F, 0x006DF3, 0x009187, 0x0066D9, 0x006E1A, 0x0085AF, 0x0085F7, 0x006055, 0x0092E4, 0x0054E8, 0x005617, 0x0059BE, 0x005A3C, 0x005E84, 0x005EE0, 0x006377, 0x00660C, 0x0068A2, 0x006A1F, 0x006A35,
	0x006E58, 0x0083D6, 0x00848B, 0x008549, 0x0088F3, 0x0091A4, 0x009266, 0x00937E, 0x009798, 0x004E1E, 0x0064FE, 0x006756, 0x007A63, 0x0057F4, 0x0071ED, 0x008755, 0x00664B, 0x00699B, 0x0075B9, 0x0079E6, 0x005875, 0x0058EC, 0x008A0A, 0x00976D, 0x007B25, 0x008ACF, 0x0053A8, 0x009017, 0x007FE0, 0x009310, 0x00745E, 0x005D69,
	0x008DA8, 0x0096DB, 0x006919, 0x0083C5, 0x009817, 0x0096C0, 0x00647A, 0x0068F2, 0x006816, 0x008106, 0x008E5F, 0x0078A9, 0x008749, 0x005C16, 0x0064B0, 0x006834, 0x00717D, 0x007A7F, 0x007BAD, 0x00821B, 0x008CCE, 0x009583, 0x007CCE, 0x00564C, 0x005CA8, 0x0066FE, 0x00695A, 0x00758F, 0x008607, 0x009F20, 0x0053E2, 0x005B8B,
	0x00531D, 0x0060E3, 0x0063BB, 0x0069CD, 0x006F15, 0x007CDF, 0x007D9C, 0x008061, 0x0084BC, 0x009397, 0x005176, 0x0063C3, 0x008A51, 0x0067C1, 0x008235, 0x006955, 0x009640, 0x009A28, 0x005CB1, 0x00817F, 0x0082D4, 0x009EDB, 0x009BDB, 0x00918D, 0x009DF9, 0x007027, 0x005544, 0x006258, 0x007422, 0x009438, 0x008338, 0x0051E7,
	0x0086F8, 0x0053EA, 0x0053E9, 0x008FB0, 0x005DFD, 0x007AEA, 0x008FBF, 0x0072F8, 0x009C48, 0x006A3D, 0x005766, 0x006B4E, 0x006E5B, 0x007BAA, 0x00803D, 0x0086CB, 0x006A80, 0x005F1B, 0x00667A, 0x008718, 0x0099B3, 0x007B51, 0x008A3B, 0x006A17, 0x007026, 0x00732A, 0x0082E7, 0x0051CB, 0x00558B, 0x005BF5, 0x005E16, 0x0066A2,
	0x007252, 0x008776, 0x008ADC, 0x00929A, 0x0069CC, 0x00939A, 0x006802, 0x0063B4, 0x0069FB, 0x004F43, 0x0067D8, 0x008FBB, 0x008526, 0x007DB4, 0x009354, 0x00693F, 0x0058F7, 0x005B2C, 0x007D2C, 0x00540A, 0x005243, 0x00608C, 0x00633A, 0x0068AF, 0x006C40, 0x007887, 0x00798E, 0x008E44, 0x00912D, 0x0091D8, 0x009F0E, 0x0064E2,
	0x0093D1, 0x008F4D, 0x00586B, 0x007E8F, 0x00751C, 0x00985B, 0x006FB1, 0x00514E, 0x005835, 0x005C60, 0x00675C, 0x0083DF, 0x00934D, 0x007825, 0x00783A, 0x005858, 0x005957, 0x005B95, 0x005D8B, 0x0068BC, 0x006DD8, 0x006D9B, 0x0071C8, 0x007977, 0x008463, 0x008569, 0x009419, 0x00649E, 0x008404, 0x009D07, 0x006D9C, 0x0079BF,
	0x006A61, 0x006934, 0x009CF6, 0x0082EB, 0x005BC5, 0x009149, 0x00701E, 0x005678, 0x0060C7, 0x006566, 0x006C8C, 0x009041, 0x005451, 0x004E4D, 0x0051EA, 0x008599, 0x007058, 0x00637A, 0x006962, 0x0099B4, 0x007577, 0x006960, 0x006C5D, 0x008FE9, 0x008CD1, 0x005EFF, 0x0097EE, 0x006FE1, 0x0079B0, 0x007962, 0x008471, 0x00649A,
	0x004E43, 0x005EFC, 0x004E4B, 0x0057DC, 0x0056A2, 0x0081BF, 0x008997, 0x0086A4, 0x005DF4, 0x0064AD, 0x006777, 0x007436, 0x0082AD, 0x0076C3, 0x00724C, 0x006973, 0x007164, 0x0072FD, 0x009019, 0x00877F, 0x0079E4, 0x0077E7, 0x008429, 0x005265, 0x0067CF, 0x007B94, 0x007C95, 0x0066DD, 0x0083AB, 0x0099C1, 0x0051FD, 0x007872,
	0x008087, 0x007B48, 0x006AE8, 0x005E61, 0x007560, 0x006E8C, 0x009197, 0x007B4F, 0x009CE9, 0x00567A, 0x005859, 0x0086E4, 0x0096BC, 0x0053DB, 0x0091C6, 0x00633D, 0x0078D0, 0x008543, 0x00532A, 0x005E87, 0x006590, 0x007DCB, 0x008AB9, 0x006A0B, 0x007C38, 0x006787, 0x006BD8, 0x007435, 0x0067CA, 0x007A17, 0x00758B, 0x009AED,
	0x005F66, 0x0083F1, 0x005F3C, 0x007562, 0x00903C, 0x006867, 0x007D10, 0x008B2C, 0x005F6A, 0x0074E2, 0x008C79, 0x005EDF, 0x009328, 0x0092F2, 0x00849C, 0x0086ED, 0x009C2D, 0x005F6C, 0x00658C, 0x007015, 0x0057E0, 0x0051A8, 0x0065A7, 0x008299, 0x0064AB, 0x008461, 0x00856A, 0x006953, 0x00847A, 0x008557, 0x006DF5, 0x005F17,
	0x009B92, 0x00543B, 0x00626E, 0x00711A, 0x007CDE, 0x009801, 0x0050FB, 0x0078A7, 0x0077A5, 0x007B86, 0x007BC7, 0x005A29, 0x0097AD, 0x0092EA, 0x005703, 0x00752B, 0x008F14, 0x00620A, 0x0083E9, 0x005446, 0x005CEF, 0x005E96, 0x006367, 0x00670B, 0x0070F9, 0x00840C, 0x0084EC, 0x0092D2, 0x009CF3, 0x009D6C, 0x00927E, 0x005420,
	0x00982C, 0x00535C, 0x007A46, 0x0091E6, 0x006B86, 0x005E4C, 0x0054E9, 0x0069D9, 0x009BAA, 0x0067FE, 0x009C52, 0x00685D, 0x004EA6, 0x004FE3, 0x006CAB, 0x008FC4, 0x004FAD, 0x009EBF, 0x008513, 0x005DF3, 0x007B95, 0x006E4A, 0x0084D1, 0x007A14, 0x007C8D, 0x00725F, 0x009D61, 0x00690B, 0x0059EA, 0x00725D, 0x0068C9, 0x007DEC,
	0x006478, 0x005B5F, 0x008499, 0x005132, 0x006762, 0x0052FF, 0x005C24, 0x007C7E, 0x008CB0, 0x0060B6, 0x004E5F, 0x00723A, 0x008036, 0x009756, 0x0085AE, 0x009453, 0x006108, 0x004F51, 0x005BA5, 0x0063D6, 0x0067DA, 0x006D8C, 0x007337, 0x007950, 0x009091, 0x008F3F, 0x0050AD, 0x00694A, 0x007194, 0x008000, 0x0084C9, 0x009065,
	0x00617E, 0x006DC0, 0x0087BA, 0x0083B1, 0x006D1B, 0x00862D, 0x00674E, 0x0088E1, 0x00844E, 0x0063A0, 0x005289, 0x006E9C, 0x007409, 0x009F8D, 0x004EAE, 0x0051CC, 0x006881, 0x007A1C, 0x008AD2, 0x00907C, 0x006DCB, 0x0071D0, 0x007433, 0x009C57, 0x009E9F, 0x004F36, 0x005DBA, 0x00601C, 0x0073B2, 0x0082D3, 0x006190, 0x006F23,
	0x007149, 0x007C3E, 0x00806F, 0x0084EE, 0x009B6F, 0x006AD3, 0x005A41, 0x006994, 0x007262, 0x0072FC, 0x007BED, 0x00807E, 0x00874B, 0x007984, 0x00808B, 0x00502D, 0x006B6A, 0x009DF2, 0x004E99, 0x004E98, 0x009C10, 0x008A6B, 0x0085C1, 0x008568, 0x006900, 0x007897,
	0
};

// JIS X 0208 Level 2
static const unsigned map_kanji_jis2_1[] = {
	0x005F0C, 0x004E10, 0x004E15, 0x004E2A, 0x004E31, 0x004E36, 0x004E3F, 0x004E42, 0x004E56, 0x004E58, 0x004E82, 0x004E85, 0x008C6B, 0x004E8A, 0x008212, 0x005F0D, 0x004E8E, 0x004E9E, 0x004E9F, 0x004EA0, 0x004EA2, 0x004EB0, 0x004EB3, 0x004EB6, 0x004ECE, 0x004ECD, 0x004EC4, 0x004EC6, 0x004EC2, 0x004ED7, 0x004EDE, 0x004EED,
	0x004EDF, 0x004EF7, 0x004F09, 0x004F5A, 0x004F30, 0x004F5B, 0x004F5D, 0x004F57, 0x004F47, 0x004F76, 0x004F88, 0x004F8F, 0x004F98, 0x004F7B, 0x004F69, 0x004F70, 0x004F91, 0x004F6F, 0x004F86, 0x004F96, 0x005118, 0x004FD4, 0x004FDF, 0x004FCE, 0x004FD8, 0x004FDB, 0x004FD1, 0x004FDA, 0x004FD0, 0x004FE4, 0x004FE5, 0x00501A,
	0x005028, 0x005014, 0x00502A, 0x005025, 0x005005, 0x004F1C, 0x004FF6, 0x005021, 0x005029, 0x00502C, 0x004FFE, 0x004FEF, 0x005011, 0x005006, 0x005043, 0x005047, 0x006703, 0x005055, 0x005050, 0x005048, 0x00505A, 0x005056, 0x00506C, 0x005078, 0x005080, 0x00509A, 0x005085, 0x0050B4, 0x0050C9, 0x0050CA, 0x0050B3, 0x0050C2,
	0x0050D6, 0x0050DE, 0x0050E5, 0x0050ED, 0x0050E3, 0x0050EE, 0x0050F9, 0x0050F5, 0x005109, 0x005101, 0x005102, 0x005116, 0x005115, 0x005114, 0x00511A, 0x005121, 0x00513A, 0x005137, 0x00513C, 0x00513B, 0x00513F, 0x005140, 0x005152, 0x00514C, 0x005154, 0x005162, 0x007AF8, 0x005169, 0x00516A, 0x00516E, 0x005180, 0x005182,
	0x0056D8, 0x00518C, 0x005189, 0x00518F, 0x005191, 0x005193, 0x005195, 0x005196, 0x0051A4, 0x0051A6, 0x0051A2, 0x0051A9, 0x0051AA, 0x0051AB, 0x0051B3, 0x0051B1, 0x0051B2, 0x0051B0, 0x0051B5, 0x0051BD, 0x0051C5, 0x0051C9, 0x0051DB, 0x0051E0, 0x008655, 0x0051E9, 0x0051ED, 0x0051F0, 0x0051F5, 0x0051FE, 0x005204, 0x00520B,
	0x005214, 0x00520E, 0x005227, 0x00522A, 0x00522E, 0x005233, 0x00524F, 0x005244, 0x00524B, 0x00524C, 0x00525E, 0x005254, 0x00526A, 0x005274, 0x005269, 0x005273, 0x00527F, 0x00527D, 0x00528D, 0x005294, 0x005292, 0x005271, 0x005288, 0x005291, 0x008FA8, 0x008FA7, 0x0052AC, 0x0052AD, 0x0052BC, 0x0052B5, 0x0052C1, 0x0052CD,
	0x0052D7, 0x0052DE, 0x0052E3, 0x0052E6, 0x0098ED, 0x0052E0, 0x0052F3, 0x0052F5, 0x0052F8, 0x0052F9, 0x005306, 0x005308, 0x007538, 0x00530D, 0x005310, 0x00530F, 0x005315, 0x00531A, 0x005323, 0x00532F, 0x005331, 0x005333, 0x005338, 0x005340, 0x005346, 0x005345, 0x004E17, 0x005349, 0x00534D, 0x0051D6, 0x00535E, 0x005369,
	0x00536E, 0x005918, 0x00537B, 0x005377, 0x005382, 0x005396, 0x0053A0, 0x0053A6, 0x0053A5, 0x0053AE, 0x0053B0, 0x0053B6, 0x0053C3, 0x007C12, 0x0096D9, 0x0053DF, 0x0066FC, 0x0071EE, 0x0053EE, 0x0053E8, 0x0053ED, 0x0053FA, 0x005401, 0x00543D, 0x005440, 0x00542C, 0x00542D, 0x00543C, 0x00542E, 0x005436, 0x005429, 0x00541D,
	0x00544E, 0x00548F, 0x005475, 0x00548E, 0x00545F, 0x005471, 0x005477, 0x005470, 0x005492, 0x00547B, 0x005480, 0x005476, 0x005484, 0x005490, 0x005486, 0x0054C7, 0x0054A2, 0x0054B8, 0x0054A5, 0x0054AC, 0x0054C4, 0x0054C8, 0x0054A8, 0x0054AB, 0x0054C2, 0x0054A4, 0x0054BE, 0x0054BC, 0x0054D8, 0x0054E5, 0x0054E6, 0x00550F,
	0x005514, 0x0054FD, 0x0054EE, 0x0054ED, 0x0054E2, 0x005539, 0x005540, 0x005563, 0x00554C, 0x00552E, 0x00555C, 0x005545, 0x005556, 0x005557, 0x005538, 0x005533, 0x00555D, 0x005599, 0x005580, 0x0054AF, 0x00558A, 0x00559F, 0x00557B, 0x00557E, 0x005598, 0x00559E, 0x0055AE, 0x00557C, 0x005583, 0x005587, 0x0055A8, 0x0055DA,
	0x0055DF, 0x0055C4, 0x0055DC, 0x0055E4, 0x0055D4, 0x005614, 0x0055F7, 0x005616, 0x0055FE, 0x0055FD, 0x00561B, 0x0055F9, 0x00564E, 0x005650, 0x0071DF, 0x005634, 0x005636, 0x005638, 0x00566B, 0x005664, 0x00562F, 0x00566C, 0x00566A, 0x005686, 0x005680, 0x00568A, 0x0056A0, 0x005694, 0x00568F, 0x0056A5, 0x0056AE, 0x0056B6,
	0x0056B4, 0x0056C2, 0x0056BC, 0x0056C1, 0x0056C3, 0x0056C0, 0x0056C8, 0x0056CE, 0x0056D1, 0x0056D3, 0x0056D7, 0x0056EE, 0x0056F9, 0x005700, 0x0056FF, 0x005704, 0x005709, 0x005708, 0x00570B, 0x00570D, 0x005713, 0x005718, 0x005716, 0x0055C7, 0x00571C, 0x005726, 0x005737, 0x005738, 0x00574E, 0x00573B, 0x005740, 0x00574F,
	0x005769, 0x0057C0, 0x005788, 0x005761, 0x00577F, 0x005789, 0x005793, 0x0057A0, 0x0057B3, 0x0057A4, 0x0057AA, 0x0057B0, 0x0057C3, 0x0057C6, 0x0057D4, 0x0057D2, 0x0057D3, 0x00580A, 0x0057D6, 0x0057E3, 0x00580B, 0x005819, 0x00581D, 0x005872, 0x005821, 0x005862, 0x00584B, 0x005870, 0x005852, 0x00583D, 0x005879, 0x005885,
	0x0058B9, 0x00589F, 0x0058AB, 0x0058BA, 0x0058DE, 0x0058BB, 0x0058B8, 0x0058AE, 0x0058C5, 0x0058D3, 0x0058D1, 0x0058D7, 0x0058D9, 0x0058D8, 0x0058E5, 0x0058DC, 0x0058E4, 0x0058DF, 0x0058EF, 0x0058FA, 0x0058F9, 0x0058FB, 0x0058FC, 0x0058FD, 0x005902, 0x00590A, 0x005910, 0x00591B, 0x0068A6, 0x005925, 0x00592C, 0x00592D,
	0x005932, 0x005938, 0x00593E, 0x007AD2, 0x005955, 0x005950, 0x00594E, 0x00595A, 0x005958, 0x005962, 0x005960, 0x005967, 0x00596C, 0x005969, 0x005978, 0x005981, 0x00599D, 0x004F5E, 0x004FAB, 0x0059A3, 0x0059B2, 0x0059C6, 0x0059E8, 0x0059DC, 0x00598D, 0x0059D9, 0x0059DA, 0x005A25, 0x005A1F, 0x005A11, 0x005A1C, 0x005A09,
	0x005A1A, 0x005A40, 0x005A6C, 0x005A49, 0x005A35, 0x005A36, 0x005A62, 0x005A6A, 0x005A9A, 0x005ABC, 0x005ABE, 0x005ACB, 0x005AC2, 0x005ABD, 0x005AE3, 0x005AD7, 0x005AE6, 0x005AE9, 0x005AD6, 0x005AFA, 0x005AFB, 0x005B0C, 0x005B0B, 0x005B16, 0x005B32, 0x005AD0, 0x005B2A, 0x005B36, 0x005B3E, 0x005B43, 0x005B45, 0x005B40,
	0x005B51, 0x005B55, 0x005B5A, 0x005B5B, 0x005B65, 0x005B69, 0x005B70, 0x005B73, 0x005B75, 0x005B78, 0x006588, 0x005B7A, 0x005B80, 0x005B83, 0x005BA6, 0x005BB8, 0x005BC3, 0x005BC7, 0x005BC9, 0x005BD4, 0x005BD0, 0x005BE4, 0x005BE6, 0x005BE2, 0x005BDE, 0x005BE5, 0x005BEB, 0x005BF0, 0x005BF6, 0x005BF3, 0x005C05, 0x005C07,
	0x005C08, 0x005C0D, 0x005C13, 0x005C20, 0x005C22, 0x005C28, 0x005C38, 0x005C39, 0x005C41, 0x005C46, 0x005C4E, 0x005C53, 0x005C50, 0x005C4F, 0x005B71, 0x005C6C, 0x005C6E, 0x004E62, 0x005C76, 0x005C79, 0x005C8C, 0x005C91, 0x005C94, 0x00599B, 0x005CAB, 0x005CBB, 0x005CB6, 0x005CBC, 0x005CB7, 0x005CC5, 0x005CBE, 0x005CC7,
	0x005CD9, 0x005CE9, 0x005CFD, 0x005CFA, 0x005CED, 0x005D8C, 0x005CEA, 0x005D0B, 0x005D15, 0x005D17, 0x005D5C, 0x005D1F, 0x005D1B, 0x005D11, 0x005D14, 0x005D22, 0x005D1A, 0x005D19, 0x005D18, 0x005D4C, 0x005D52, 0x005D4E, 0x005D4B, 0x005D6C, 0x005D73, 0x005D76, 0x005D87, 0x005D84, 0x005D82, 0x005DA2, 0x005D9D, 0x005DAC,
	0x005DAE, 0x005DBD, 0x005D90, 0x005DB7, 0x005DBC, 0x005DC9, 0x005DCD, 0x005DD3, 0x005DD2, 0x005DD6, 0x005DDB, 0x005DEB, 0x005DF2, 0x005DF5, 0x005E0B, 0x005E1A, 0x005E19, 0x005E11, 0x005E1B, 0x005E36, 0x005E37, 0x005E44, 0x005E43, 0x005E40, 0x005E4E, 0x005E57, 0x005E54, 0x005E5F, 0x005E62, 0x005E64, 0x005E47, 0x005E75,
	0x005E76, 0x005E7A, 0x009EBC, 0x005E7F, 0x005EA0, 0x005EC1, 0x005EC2, 0x005EC8, 0x005ED0, 0x005ECF, 0x005ED6, 0x005EE3, 0x005EDD, 0x005EDA, 0x005EDB, 0x005EE2, 0x005EE1, 0x005EE8, 0x005EE9, 0x005EEC, 0x005EF1, 0x005EF3, 0x005EF0, 0x005EF4, 0x005EF8, 0x005EFE, 0x005F03, 0x005F09, 0x005F5D, 0x005F5C, 0x005F0B, 0x005F11,
	0x005F16, 0x005F29, 0x005F2D, 0x005F38, 0x005F41, 0x005F48, 0x005F4C, 0x005F4E, 0x005F2F, 0x005F51, 0x005F56, 0x005F57, 0x005F61, 0x005F6D, 0x005F73, 0x005F77, 0x005F83, 0x005F82, 0x005F7F, 0x005F8A, 0x005F88, 0x005F91, 0x005F87, 0x005F9E, 0x005F99, 0x005F98, 0x005FA0, 0x005FA8, 0x005FAD, 0x005FBC, 0x005FD6, 0x005FFB,
	0x005FE4, 0x005FF8, 0x005FF1, 0x005FDD, 0x0060B3, 0x005FFF, 0x006021, 0x006060, 0x006019, 0x006010, 0x006029, 0x00600E, 0x006031, 0x00601B, 0x006015, 0x00602B, 0x006026, 0x00600F, 0x00603A, 0x00605A, 0x006041, 0x00606A, 0x006077, 0x00605F, 0x00604A, 0x006046, 0x00604D, 0x006043, 0x006064, 0x006042, 0x00606C, 0x00606B,
	0x006059, 0x006081, 0x00608D, 0x006083, 0x00609A, 0x006084, 0x00609B, 0x006096, 0x006097, 0x006092, 0x0060A7, 0x00608B, 0x0060E1, 0x0060B8, 0x0060E0, 0x0060D3, 0x0060B4, 0x005FF0, 0x0060BD, 0x0060C6, 0x0060B5, 0x0060D8, 0x00614D, 0x006115, 0x006106, 0x0060F6, 0x0060F7, 0x006100, 0x0060F4, 0x0060FA, 0x006103, 0x006121,
	0x0060FB, 0x0060F1, 0x00610D, 0x00610E, 0x006147, 0x00613E, 0x006128, 0x006127, 0x00614A, 0x00613F, 0x00613C, 0x00612C, 0x006134, 0x00613D, 0x006142, 0x006173, 0x006177, 0x006158, 0x006159, 0x00615A, 0x00616B, 0x006174, 0x00616F, 0x006165, 0x006171, 0x00615F, 0x00615D, 0x006153, 0x006175, 0x006199, 0x006196, 0x006187,
	0x006194, 0x00619A, 0x00618A, 0x006191, 0x0061AB, 0x0061AE, 0x0061CC, 0x0061CA, 0x0061C9, 0x0061F7, 0x0061C8, 0x0061C3, 0x0061C6, 0x0061BA, 0x0061CB, 0x007F79, 0x0061CD, 0x0061E6, 0x0061E3, 0x0061F6, 0x0061FA, 0x0061F4, 0x0061FF, 0x0061FD, 0x0061FC, 0x0061FE, 0x006200, 0x006208, 0x006209, 0x00620D, 0x00620C, 0x006214,
	0x00621B, 0x00621E, 0x006221, 0x00622A, 0x00622E, 0x006230, 0x006232, 0x006233, 0x006241, 0x00624E, 0x00625E, 0x006263, 0x00625B, 0x006260, 0x006268, 0x00627C, 0x006282, 0x006289, 0x00627E, 0x006292, 0x006293, 0x006296, 0x0062D4, 0x006283, 0x006294, 0x0062D7, 0x0062D1, 0x0062BB, 0x0062CF, 0x0062FF, 0x0062C6, 0x0064D4,
	0x0062C8, 0x0062DC, 0x0062CC, 0x0062CA, 0x0062C2, 0x0062C7, 0x00629B, 0x00630C, 0x0062EE, 0x0062F1, 0x006327, 0x006302, 0x006308, 0x0062EF, 0x0062F5, 0x006350, 0x00633E, 0x00634D, 0x00641C, 0x00634F, 0x006396, 0x00638E, 0x006380, 0x0063AB, 0x006376, 0x0063A3, 0x00638F, 0x006389, 0x00639F, 0x0063B5, 0x00636B, 0x006369,
	0x0063BE, 0x0063E9, 0x0063C0, 0x0063C6, 0x0063E3, 0x0063C9, 0x0063D2, 0x0063F6, 0x0063C4, 0x006416, 0x006434, 0x006406, 0x006413, 0x006426, 0x006436, 0x00651D, 0x006417, 0x006428, 0x00640F, 0x006467, 0x006476, 0x00644E, 0x00652A, 0x006495, 0x006493, 0x0064A5, 0x0064A9, 0x006488, 0x0064BC, 0x0064DA, 0x0064D2, 0x0064C5,
	0x0064C7, 0x0064BB, 0x0064D8, 0x0064C2, 0x0064F1, 0x0064E7, 0x008209, 0x0064E0, 0x0064E1, 0x0062AC, 0x0064E3, 0x0064EF, 0x00652C, 0x0064F6, 0x0064F4, 0x0064F2, 0x0064FA, 0x006500, 0x0064FD, 0x006518, 0x00651C, 0x006505, 0x006524, 0x006523, 0x00652B, 0x006534, 0x006535, 0x006537, 0x006536, 0x006538, 0x00754B, 0x006548,
	0x006556, 0x006555, 0x00654D, 0x006558, 0x00655E, 0x00655D, 0x006572, 0x006578, 0x006582, 0x006583, 0x008B8A, 0x00659B, 0x00659F, 0x0065AB, 0x0065B7, 0x0065C3, 0x0065C6, 0x0065C1, 0x0065C4, 0x0065CC, 0x0065D2, 0x0065DB, 0x0065D9, 0x0065E0, 0x0065E1, 0x0065F1, 0x006772, 0x00660A, 0x006603, 0x0065FB, 0x006773, 0x006635,
	0x006636, 0x006634, 0x00661C, 0x00664F, 0x006644, 0x006649, 0x006641, 0x00665E, 0x00665D, 0x006664, 0x006667, 0x006668, 0x00665F, 0x006662, 0x006670, 0x006683, 0x006688, 0x00668E, 0x006689, 0x006684, 0x006698, 0x00669D, 0x0066C1, 0x0066B9, 0x0066C9, 0x0066BE, 0x0066BC, 0x0066C4, 0x0066B8, 0x0066DA, 0x0066E0, 0x00663F,
	0
};

static const unsigned map_kanji_jis2_2[] = {
	0x0066E6, 0x0066E9, 0x0066F0, 0x0066F5, 0x0066F7, 0x00670F, 0x006716, 0x00671E, 0x006726, 0x006727, 0x009738, 0x00672E, 0x00673F, 0x006736, 0x006741, 0x006738, 0x006737, 0x006746, 0x00675E, 0x006760, 0x006759, 0x006763, 0x006764, 0x006789, 0x006770, 0x0067A9, 0x00677C, 0x00676A, 0x00678C, 0x00678B, 0x0067A6, 0x0067A1,
	0x006785, 0x0067B7, 0x0067EF, 0x0067B4, 0x0067EC, 0x0067B3, 0x0067E9, 0x0067B8, 0x0067E4, 0x0067DE, 0x0067DD, 0x0067E2, 0x0067EE, 0x0067B9, 0x0067CE, 0x0067C6, 0x0067E7, 0x006A9C, 0x00681E, 0x006846, 0x006829, 0x006840, 0x00684D, 0x006832, 0x00684E, 0x0068B3, 0x00682B, 0x006859, 0x006863, 0x006877, 0x00687F, 0x00689F,
	0x00688F, 0x0068AD, 0x006894, 0x00689D, 0x00689B, 0x006883, 0x006AAE, 0x0068B9, 0x006874, 0x0068B5, 0x0068A0, 0x0068BA, 0x00690F, 0x00688D, 0x00687E, 0x006901, 0x0068CA, 0x006908, 0x0068D8, 0x006922, 0x006926, 0x0068E1, 0x00690C, 0x0068CD, 0x0068D4, 0x0068E7, 0x0068D5, 0x006936, 0x006912, 0x006904, 0x0068D7, 0x0068E3,
	0x006925, 0x0068F9, 0x0068E0, 0x0068EF, 0x006928, 0x00692A, 0x00691A, 0x006923, 0x006921, 0x0068C6, 0x006979, 0x00695C, 0x006978, 0x00696B, 0x006954, 0x00697E, 0x00696E, 0x006939, 0x006974, 0x00693D, 0x006959, 0x006930, 0x006961, 0x00695E, 0x00695D, 0x006981, 0x00696A, 0x0069B2, 0x0069AE, 0x0069D0, 0x0069BF, 0x0069C1,
	0x0069D3, 0x0069BE, 0x0069CE, 0x005BE8, 0x0069CA, 0x0069DD, 0x0069BB, 0x0069C3, 0x0069A7, 0x006A2E, 0x006991, 0x0069A0, 0x00699C, 0x006995, 0x0069B4, 0x0069DE, 0x0069E8, 0x006A02, 0x006A1B, 0x0069FF, 0x006B0A, 0x0069F9, 0x0069F2, 0x0069E7, 0x006A05, 0x0069B1, 0x006A1E, 0x0069ED, 0x006A14, 0x0069EB, 0x006A0A, 0x006A12,
	0x006AC1, 0x006A23, 0x006A13, 0x006A44, 0x006A0C, 0x006A72, 0x006A36, 0x006A78, 0x006A47, 0x006A62, 0x006A59, 0x006A66, 0x006A48, 0x006A38, 0x006A22, 0x006A90, 0x006A8D, 0x006AA0, 0x006A84, 0x006AA2, 0x006AA3, 0x006A97, 0x008617, 0x006ABB, 0x006AC3, 0x006AC2, 0x006AB8, 0x006AB3, 0x006AAC, 0x006ADE, 0x006AD1, 0x006ADF,
	0x006AAA, 0x006ADA, 0x006AEA, 0x006AFB, 0x006B05, 0x008616, 0x006AFA, 0x006B12, 0x006B16, 0x006B1F, 0x006B38, 0x006B37, 0x0076DC, 0x006B39, 0x0098EE, 0x006B47, 0x006B43, 0x006B49, 0x006B50, 0x006B59, 0x006B54, 0x006B5B, 0x006B5F, 0x006B61, 0x006B78, 0x006B79, 0x006B7F, 0x006B80, 0x006B84, 0x006B83, 0x006B8D, 0x006B98,
	0x006B95, 0x006B9E, 0x006BA4, 0x006BAA, 0x006BAB, 0x006BAF, 0x006BB2, 0x006BB1, 0x006BB3, 0x006BB7, 0x006BBC, 0x006BC6, 0x006BCB, 0x006BD3, 0x006BDF, 0x006BEC, 0x006BEB, 0x006BF3, 0x006BEF, 0x009EBE, 0x006C08, 0x006C13, 0x006C14, 0x006C1B, 0x006C24, 0x006C23, 0x006C5E, 0x006C55, 0x006C62, 0x006C6A, 0x006C82, 0x006C8D,
	0x006C9A, 0x006C81, 0x006C9B, 0x006C7E, 0x006C68, 0x006C73, 0x006C92, 0x006C90, 0x006CC4, 0x006CF1, 0x006CD3, 0x006CBD, 0x006CD7, 0x006CC5, 0x006CDD, 0x006CAE, 0x006CB1, 0x006CBE, 0x006CBA, 0x006CDB, 0x006CEF, 0x006CD9, 0x006CEA, 0x006D1F, 0x00884D, 0x006D36, 0x006D2B, 0x006D3D, 0x006D38, 0x006D19, 0x006D35, 0x006D33,
	0x006D12, 0x006D0C, 0x006D63, 0x006D93, 0x006D64, 0x006D5A, 0x006D79, 0x006D59, 0x006D8E, 0x006D95, 0x006FE4, 0x006D85, 0x006DF9, 0x006E15, 0x006E0A, 0x006DB5, 0x006DC7, 0x006DE6, 0x006DB8, 0x006DC6, 0x006DEC, 0x006DDE, 0x006DCC, 0x006DE8, 0x006DD2, 0x006DC5, 0x006DFA, 0x006DD9, 0x006DE4, 0x006DD5, 0x006DEA, 0x006DEE,
	0x006E2D, 0x006E6E, 0x006E2E, 0x006E19, 0x006E72, 0x006E5F, 0x006E3E, 0x006E23, 0x006E6B, 0x006E2B, 0x006E76, 0x006E4D, 0x006E1F, 0x006E43, 0x006E3A, 0x006E4E, 0x006E24, 0x006EFF, 0x006E1D, 0x006E38, 0x006E82, 0x006EAA, 0x006E98, 0x006EC9, 0x006EB7, 0x006ED3, 0x006EBD, 0x006EAF, 0x006EC4, 0x006EB2, 0x006ED4, 0x006ED5,
	0x006E8F, 0x006EA5, 0x006EC2, 0x006E9F, 0x006F41, 0x006F11, 0x00704C, 0x006EEC, 0x006EF8, 0x006EFE, 0x006F3F, 0x006EF2, 0x006F31, 0x006EEF, 0x006F32, 0x006ECC, 0x006F3E, 0x006F13, 0x006EF7, 0x006F86, 0x006F7A, 0x006F78, 0x006F81, 0x006F80, 0x006F6F, 0x006F5B, 0x006FF3, 0x006F6D, 0x006F82, 0x006F7C, 0x006F58, 0x006F8E,
	0x006F91, 0x006FC2, 0x006F66, 0x006FB3, 0x006FA3, 0x006FA1, 0x006FA4, 0x006FB9, 0x006FC6, 0x006FAA, 0x006FDF, 0x006FD5, 0x006FEC, 0x006FD4, 0x006FD8, 0x006FF1, 0x006FEE, 0x006FDB, 0x007009, 0x00700B, 0x006FFA, 0x007011, 0x007001, 0x00700F, 0x006FFE, 0x00701B, 0x00701A, 0x006F74, 0x00701D, 0x007018, 0x00701F, 0x007030,
	0x00703E, 0x007032, 0x007051, 0x007063, 0x007099, 0x007092, 0x0070AF, 0x0070F1, 0x0070AC, 0x0070B8, 0x0070B3, 0x0070AE, 0x0070DF, 0x0070CB, 0x0070DD, 0x0070D9, 0x007109, 0x0070FD, 0x00711C, 0x007119, 0x007165, 0x007155, 0x007188, 0x007166, 0x007162, 0x00714C, 0x007156, 0x00716C, 0x00718F, 0x0071FB, 0x007184, 0x007195,
	0x0071A8, 0x0071AC, 0x0071D7, 0x0071B9, 0x0071BE, 0x0071D2, 0x0071C9, 0x0071D4, 0x0071CE, 0x0071E0, 0x0071EC, 0x0071E7, 0x0071F5, 0x0071FC, 0x0071F9, 0x0071FF, 0x00720D, 0x007210, 0x00721B, 0x007228, 0x00722D, 0x00722C, 0x007230, 0x007232, 0x00723B, 0x00723C, 0x00723F, 0x007240, 0x007246, 0x00724B, 0x007258, 0x007274,
	0x00727E, 0x007282, 0x007281, 0x007287, 0x007292, 0x007296, 0x0072A2, 0x0072A7, 0x0072B9, 0x0072B2, 0x0072C3, 0x0072C6, 0x0072C4, 0x0072CE, 0x0072D2, 0x0072E2, 0x0072E0, 0x0072E1, 0x0072F9, 0x0072F7, 0x00500F, 0x007317, 0x00730A, 0x00731C, 0x007316, 0x00731D, 0x007334, 0x00732F, 0x007329, 0x007325, 0x00733E, 0x00734E,
	0x00734F, 0x009ED8, 0x007357, 0x00736A, 0x007368, 0x007370, 0x007378, 0x007375, 0x00737B, 0x00737A, 0x0073C8, 0x0073B3, 0x0073CE, 0x0073BB, 0x0073C0, 0x0073E5, 0x0073EE, 0x0073DE, 0x0074A2, 0x007405, 0x00746F, 0x007425, 0x0073F8, 0x007432, 0x00743A, 0x007455, 0x00743F, 0x00745F, 0x007459, 0x007441, 0x00745C, 0x007469,
	0x007470, 0x007463, 0x00746A, 0x007476, 0x00747E, 0x00748B, 0x00749E, 0x0074CA, 0x0074CF, 0x0074D4, 0x0073F1, 0x0074E0, 0x0074E3, 0x0074E7, 0x0074E9, 0x0074EE, 0x0074F2, 0x0074F0, 0x0074F1, 0x0074F8, 0x0074F7, 0x007504, 0x007503, 0x007505, 0x00750C, 0x00750E, 0x00750D, 0x007515, 0x007513, 0x00751E, 0x007526, 0x00752C,
	0x00753C, 0x007544, 0x00754D, 0x00754A, 0x007549, 0x00755B, 0x007546, 0x00755A, 0x007569, 0x007564, 0x007567, 0x00756B, 0x00756D, 0x007578, 0x007576, 0x007586, 0x007587, 0x007574, 0x00758A, 0x007589, 0x007582, 0x007594, 0x00759A, 0x00759D, 0x0075A5, 0x0075A3, 0x0075C2, 0x0075B3, 0x0075C3, 0x0075B5, 0x0075BD, 0x0075B8,
	0x0075BC, 0x0075B1, 0x0075CD, 0x0075CA, 0x0075D2, 0x0075D9, 0x0075E3, 0x0075DE, 0x0075FE, 0x0075FF, 0x0075FC, 0x007601, 0x0075F0, 0x0075FA, 0x0075F2, 0x0075F3, 0x00760B, 0x007609, 0x00761F, 0x007627, 0x007620, 0x007621, 0x007622, 0x007624, 0x007634, 0x007630, 0x00763B, 0x007647, 0x007648, 0x007646, 0x00765C, 0x007658,
	0x007661, 0x007662, 0x007668, 0x007669, 0x00766A, 0x007667, 0x00766C, 0x007670, 0x007672, 0x007676, 0x007678, 0x00767C, 0x007680, 0x007683, 0x007688, 0x00768B, 0x00768E, 0x007696, 0x007693, 0x007699, 0x00769A, 0x0076B0, 0x0076B4, 0x0076B8, 0x0076B9, 0x0076BA, 0x0076C2, 0x0076CD, 0x0076D6, 0x0076D2, 0x0076DE, 0x0076E1,
	0x0076E5, 0x0076E7, 0x0076EA, 0x00862F, 0x0076FB, 0x007708, 0x007707, 0x007704, 0x007729, 0x007724, 0x00771E, 0x007725, 0x007726, 0x00771B, 0x007737, 0x007738, 0x007747, 0x00775A, 0x007768, 0x00776B, 0x00775B, 0x007765, 0x00777F, 0x00777E, 0x007779, 0x00778E, 0x00778B, 0x007791, 0x0077A0, 0x00779E, 0x0077B0, 0x0077B6,
	0x0077B9, 0x0077BF, 0x0077BC, 0x0077BD, 0x0077BB, 0x0077C7, 0x0077CD, 0x0077D7, 0x0077DA, 0x0077DC, 0x0077E3, 0x0077EE, 0x0077FC, 0x00780C, 0x007812, 0x007926, 0x007820, 0x00792A, 0x007845, 0x00788E, 0x007874, 0x007886, 0x00787C, 0x00789A, 0x00788C, 0x0078A3, 0x0078B5, 0x0078AA, 0x0078AF, 0x0078D1, 0x0078C6, 0x0078CB,
	0x0078D4, 0x0078BE, 0x0078BC, 0x0078C5, 0x0078CA, 0x0078EC, 0x0078E7, 0x0078DA, 0x0078FD, 0x0078F4, 0x007907, 0x007912, 0x007911, 0x007919, 0x00792C, 0x00792B, 0x007940, 0x007960, 0x007957, 0x00795F, 0x00795A, 0x007955, 0x007953, 0x00797A, 0x00797F, 0x00798A, 0x00799D, 0x0079A7, 0x009F4B, 0x0079AA, 0x0079AE, 0x0079B3,
	0x0079B9, 0x0079BA, 0x0079C9, 0x0079D5, 0x0079E7, 0x0079EC, 0x0079E1, 0x0079E3, 0x007A08, 0x007A0D, 0x007A18, 0x007A19, 0x007A20, 0x007A1F, 0x007980, 0x007A31, 0x007A3B, 0x007A3E, 0x007A37, 0x007A43, 0x007A57, 0x007A49, 0x007A61, 0x007A62, 0x007A69, 0x009F9D, 0x007A70, 0x007A79, 0x007A7D, 0x007A88, 0x007A97, 0x007A95,
	0x007A98, 0x007A96, 0x007AA9, 0x007AC8, 0x007AB0, 0x007AB6, 0x007AC5, 0x007AC4, 0x007ABF, 0x009083, 0x007AC7, 0x007ACA, 0x007ACD, 0x007ACF, 0x007AD5, 0x007AD3, 0x007AD9, 0x007ADA, 0x007ADD, 0x007AE1, 0x007AE2, 0x007AE6, 0x007AED, 0x007AF0, 0x007B02, 0x007B0F, 0x007B0A, 0x007B06, 0x007B33, 0x007B18, 0x007B19, 0x007B1E,
	0x007B35, 0x007B28, 0x007B36, 0x007B50, 0x007B7A, 0x007B04, 0x007B4D, 0x007B0B, 0x007B4C, 0x007B45, 0x007B75, 0x007B65, 0x007B74, 0x007B67, 0x007B70, 0x007B71, 0x007B6C, 0x007B6E, 0x007B9D, 0x007B98, 0x007B9F, 0x007B8D, 0x007B9C, 0x007B9A, 0x007B92, 0x007B8F, 0x007B5D, 0x007B99, 0x007BCB, 0x007BC1, 0x007BCC, 0x007BCF,
	0x007BB4, 0x007BC6, 0x007BDD, 0x007BE9, 0x007C11, 0x007C14, 0x007BE6, 0x007BE5, 0x007C00, 0x007C07, 0x007C13, 0x007BF3, 0x007BF7, 0x007C17, 0x007C0D, 0x007BF6, 0x007C23, 0x007C27, 0x007C2A, 0x007C1F, 0x007C37, 0x007C2B, 0x007C3D, 0x007C4C, 0x007C43, 0x007C54, 0x007C4F, 0x007C40, 0x007C50, 0x007C58, 0x007C5F, 0x007C64,
	0x007C56, 0x007C65, 0x007C6C, 0x007C75, 0x007C83, 0x007C90, 0x007CA4, 0x007CAD, 0x007CA2, 0x007CAB, 0x007CA1, 0x007CA8, 0x007CB3, 0x007CB2, 0x007CB1, 0x007CAE, 0x007CB9, 0x007CBD, 0x007CC0, 0x007CC5, 0x007CC2, 0x007CD8, 0x007CD2, 0x007CDC, 0x007CE2, 0x009B3B, 0x007CEF, 0x007CF2, 0x007CF4, 0x007CF6, 0x007CFA, 0x007D06,
	0x007D02, 0x007D1C, 0x007D15, 0x007D0A, 0x007D45, 0x007D4B, 0x007D2E, 0x007D32, 0x007D3F, 0x007D35, 0x007D46, 0x007D73, 0x007D56, 0x007D4E, 0x007D72, 0x007D68, 0x007D6E, 0x007D4F, 0x007D63, 0x007D93, 0x007D89, 0x007D5B, 0x007D8F, 0x007D7D, 0x007D9B, 0x007DBA, 0x007DAE, 0x007DA3, 0x007DB5, 0x007DC7, 0x007DBD, 0x007DAB,
	0x007E3D, 0x007DA2, 0x007DAF, 0x007DDC, 0x007DB8, 0x007D9F, 0x007DB0, 0x007DD8, 0x007DDD, 0x007DE4, 0x007DDE, 0x007DF2, 0x007DE1, 0x007E05, 0x007E0A, 0x007E23, 0x007E21, 0x007E12, 0x007E31, 0x007E1F, 0x007E09, 0x007E0B, 0x007E22, 0x007E46, 0x007E66, 0x007E3B, 0x007E35, 0x007E39, 0x007E43, 0x007E37, 0x007E32, 0x007E3A,
	0x007E67, 0x007E5D, 0x007E56, 0x007E5E, 0x007E59, 0x007E5A, 0x007E79, 0x007E6A, 0x007E69, 0x007E7C, 0x007E7B, 0x007E83, 0x007DD5, 0x007E7D, 0x008FAE, 0x007E7F, 0x007E88, 0x007E89, 0x007E8C, 0x007E92, 0x007E90, 0x007E93, 0x007E94, 0x007E96, 0x007E8E, 0x007E9B, 0x007E9C, 0x007F38, 0x007F3A, 0x007F45, 0x007F4C, 0x007F4D,
	0
};

static const unsigned map_kanji_jis2_3[] = {
	0x007F4E, 0x007F50, 0x007F51, 0x007F55, 0x007F54, 0x007F58, 0x007F5F, 0x007F60, 0x007F68, 0x007F69, 0x007F67, 0x007F78, 0x007F82, 0x007F86, 0x007F83, 0x007F88, 0x007F87, 0x007F8C, 0x007F94, 0x007F9D, 0x007F9A, 0x007FA3, 0x007FAF, 0x007FB2, 0x007FB9, 0x007FAE, 0x007FB6, 0x007FB8, 0x008B71, 0x007FC5, 0x007FC6, 0x007FCA,
	0x007FD5, 0x007FD4, 0x007FE1, 0x007FE6, 0x007FE9, 0x007FF3, 0x007FF9, 0x0098DC, 0x008006, 0x008004, 0x00800B, 0x008012, 0x008018, 0x008019, 0x00801C, 0x008021, 0x008028, 0x00803F, 0x00803B, 0x00804A, 0x008046, 0x008052, 0x008058, 0x00805A, 0x00805F, 0x008062, 0x008068, 0x008073, 0x008072, 0x008070, 0x008076, 0x008079,
	0x00807D, 0x00807F, 0x008084, 0x008086, 0x008085, 0x00809B, 0x008093, 0x00809A, 0x0080AD, 0x005190, 0x0080AC, 0x0080DB, 0x0080E5, 0x0080D9, 0x0080DD, 0x0080C4, 0x0080DA, 0x0080D6, 0x008109, 0x0080EF, 0x0080F1, 0x00811B, 0x008129, 0x008123, 0x00812F, 0x00814B, 0x00968B, 0x008146, 0x00813E, 0x008153, 0x008151, 0x0080FC,
	0x008171, 0x00816E, 0x008165, 0x008166, 0x008174, 0x008183, 0x008188, 0x00818A, 0x008180, 0x008182, 0x0081A0, 0x008195, 0x0081A4, 0x0081A3, 0x00815F, 0x008193, 0x0081A9, 0x0081B0, 0x0081B5, 0x0081BE, 0x0081B8, 0x0081BD, 0x0081C0, 0x0081C2, 0x0081BA, 0x0081C9, 0x0081CD, 0x0081D1, 0x0081D9, 0x0081D8, 0x0081C8, 0x0081DA,
	0x0081DF, 0x0081E0, 0x0081E7, 0x0081FA, 0x0081FB, 0x0081FE, 0x008201, 0x008202, 0x008205, 0x008207, 0x00820A, 0x00820D, 0x008210, 0x008216, 0x008229, 0x00822B, 0x008238, 0x008233, 0x008240, 0x008259, 0x008258, 0x00825D, 0x00825A, 0x00825F, 0x008264, 0x008262, 0x008268, 0x00826A, 0x00826B, 0x00822E, 0x008271, 0x008277,
	0x008278, 0x00827E, 0x00828D, 0x008292, 0x0082AB, 0x00829F, 0x0082BB, 0x0082AC, 0x0082E1, 0x0082E3, 0x0082DF, 0x0082D2, 0x0082F4, 0x0082F3, 0x0082FA, 0x008393, 0x008303, 0x0082FB, 0x0082F9, 0x0082DE, 0x008306, 0x0082DC, 0x008309, 0x0082D9, 0x008335, 0x008334, 0x008316, 0x008332, 0x008331, 0x008340, 0x008339, 0x008350,
	0x008345, 0x00832F, 0x00832B, 0x008317, 0x008318, 0x008385, 0x00839A, 0x0083AA, 0x00839F, 0x0083A2, 0x008396, 0x008323, 0x00838E, 0x008387, 0x00838A, 0x00837C, 0x0083B5, 0x008373, 0x008375, 0x0083A0, 0x008389, 0x0083A8, 0x0083F4, 0x008413, 0x0083EB, 0x0083CE, 0x0083FD, 0x008403, 0x0083D8, 0x00840B, 0x0083C1, 0x0083F7,
	0x008407, 0x0083E0, 0x0083F2, 0x00840D, 0x008422, 0x008420, 0x0083BD, 0x008438, 0x008506, 0x0083FB, 0x00846D, 0x00842A, 0x00843C, 0x00855A, 0x008484, 0x008477, 0x00846B, 0x0084AD, 0x00846E, 0x008482, 0x008469, 0x008446, 0x00842C, 0x00846F, 0x008479, 0x008435, 0x0084CA, 0x008462, 0x0084B9, 0x0084BF, 0x00849F, 0x0084D9,
	0x0084CD, 0x0084BB, 0x0084DA, 0x0084D0, 0x0084C1, 0x0084C6, 0x0084D6, 0x0084A1, 0x008521, 0x0084FF, 0x0084F4, 0x008517, 0x008518, 0x00852C, 0x00851F, 0x008515, 0x008514, 0x0084FC, 0x008540, 0x008563, 0x008558, 0x008548, 0x008541, 0x008602, 0x00854B, 0x008555, 0x008580, 0x0085A4, 0x008588, 0x008591, 0x00858A, 0x0085A8,
	0x00856D, 0x008594, 0x00859B, 0x0085EA, 0x008587, 0x00859C, 0x008577, 0x00857E, 0x008590, 0x0085C9, 0x0085BA, 0x0085CF, 0x0085B9, 0x0085D0, 0x0085D5, 0x0085DD, 0x0085E5, 0x0085DC, 0x0085F9, 0x00860A, 0x008613, 0x00860B, 0x0085FE, 0x0085FA, 0x008606, 0x008622, 0x00861A, 0x008630, 0x00863F, 0x00864D, 0x004E55, 0x008654,
	0x00865F, 0x008667, 0x008671, 0x008693, 0x0086A3, 0x0086A9, 0x0086AA, 0x00868B, 0x00868C, 0x0086B6, 0x0086AF, 0x0086C4, 0x0086C6, 0x0086B0, 0x0086C9, 0x008823, 0x0086AB, 0x0086D4, 0x0086DE, 0x0086E9, 0x0086EC, 0x0086DF, 0x0086DB, 0x0086EF, 0x008712, 0x008706, 0x008708, 0x008700, 0x008703, 0x0086FB, 0x008711, 0x008709,
	0x00870D, 0x0086F9, 0x00870A, 0x008734, 0x00873F, 0x008737, 0x00873B, 0x008725, 0x008729, 0x00871A, 0x008760, 0x00875F, 0x008778, 0x00874C, 0x00874E, 0x008774, 0x008757, 0x008768, 0x00876E, 0x008759, 0x008753, 0x008763, 0x00876A, 0x008805, 0x0087A2, 0x00879F, 0x008782, 0x0087AF, 0x0087CB, 0x0087BD, 0x0087C0, 0x0087D0,
	0x0096D6, 0x0087AB, 0x0087C4, 0x0087B3, 0x0087C7, 0x0087C6, 0x0087BB, 0x0087EF, 0x0087F2, 0x0087E0, 0x00880F, 0x00880D, 0x0087FE, 0x0087F6, 0x0087F7, 0x00880E, 0x0087D2, 0x008811, 0x008816, 0x008815, 0x008822, 0x008821, 0x008831, 0x008836, 0x008839, 0x008827, 0x00883B, 0x008844, 0x008842, 0x008852, 0x008859, 0x00885E,
	0x008862, 0x00886B, 0x008881, 0x00887E, 0x00889E, 0x008875, 0x00887D, 0x0088B5, 0x008872, 0x008882, 0x008897, 0x008892, 0x0088AE, 0x008899, 0x0088A2, 0x00888D, 0x0088A4, 0x0088B0, 0x0088BF, 0x0088B1, 0x0088C3, 0x0088C4, 0x0088D4, 0x0088D8, 0x0088D9, 0x0088DD, 0x0088F9, 0x008902, 0x0088FC, 0x0088F4, 0x0088E8, 0x0088F2,
	0x008904, 0x00890C, 0x00890A, 0x008913, 0x008943, 0x00891E, 0x008925, 0x00892A, 0x00892B, 0x008941, 0x008944, 0x00893B, 0x008936, 0x008938, 0x00894C, 0x00891D, 0x008960, 0x00895E, 0x008966, 0x008964, 0x00896D, 0x00896A, 0x00896F, 0x008974, 0x008977, 0x00897E, 0x008983, 0x008988, 0x00898A, 0x008993, 0x008998, 0x0089A1,
	0x0089A9, 0x0089A6, 0x0089AC, 0x0089AF, 0x0089B2, 0x0089BA, 0x0089BD, 0x0089BF, 0x0089C0, 0x0089DA, 0x0089DC, 0x0089DD, 0x0089E7, 0x0089F4, 0x0089F8, 0x008A16, 0x008A10, 0x008A0C, 0x008A1B, 0x008A1D, 0x008A25, 0x008A36, 0x008A41, 0x008A5B, 0x008A52, 0x008A46, 0x008A48, 0x008A7C, 0x008A6D, 0x008A6C, 0x008A62, 0x008A85,
	0x008A82, 0x008A84, 0x008AA8, 0x008AA1, 0x008A91, 0x008AA5, 0x008AA6, 0x008A9A, 0x008AA3, 0x008AC4, 0x008ACD, 0x008AC2, 0x008ADA, 0x008AEB, 0x008AF3, 0x008AE4, 0x008AF1, 0x008B14, 0x008AE0, 0x008AE2, 0x008AF7, 0x008ADE, 0x008ADB, 0x008B0C, 0x008B07, 0x008B1A, 0x008AE1, 0x008B16, 0x008B10, 0x008B17, 0x008B20, 0x008B33,
	0x0097AB, 0x008B26, 0x008B2B, 0x008B3E, 0x008B28, 0x008B41, 0x008B4C, 0x008B4F, 0x008B4E, 0x008B49, 0x008B56, 0x008B5B, 0x008B5A, 0x008B6B, 0x008B5F, 0x008B6C, 0x008B6F, 0x008B74, 0x008B7D, 0x008B80, 0x008B8C, 0x008B8E, 0x008B92, 0x008B93, 0x008B96, 0x008B99, 0x008B9A, 0x008C3A, 0x008C41, 0x008C3F, 0x008C48, 0x008C4C,
	0x008C4E, 0x008C50, 0x008C55, 0x008C62, 0x008C6C, 0x008C78, 0x008C7A, 0x008C82, 0x008C89, 0x008C85, 0x008C8A, 0x008C8D, 0x008C8E, 0x008C94, 0x008C7C, 0x008C98, 0x00621D, 0x008CAD, 0x008CBD, 0x008CB2, 0x008CB3, 0x008CAE, 0x008CB6, 0x008CC8, 0x008CC1, 0x008CE4, 0x008CE3, 0x008CDA, 0x008CFD, 0x008CFA, 0x008CFB, 0x008D04,
	0x008D05, 0x008D0A, 0x008D07, 0x008D0F, 0x008D0D, 0x008D10, 0x009F4E, 0x008D13, 0x008CCD, 0x008D14, 0x008D16, 0x008D67, 0x008D6D, 0x008D71, 0x008D73, 0x008D81, 0x008D99, 0x008DC2, 0x008DBE, 0x008DBA, 0x008DCF, 0x008DDA, 0x008DD6, 0x008DCC, 0x008DDB, 0x008DCB, 0x008DEA, 0x008DEB, 0x008DDF, 0x008DE3, 0x008DFC, 0x008E08,
	0x008E09, 0x008DFF, 0x008E1D, 0x008E1E, 0x008E10, 0x008E1F, 0x008E42, 0x008E35, 0x008E30, 0x008E34, 0x008E4A, 0x008E47, 0x008E49, 0x008E4C, 0x008E50, 0x008E48, 0x008E59, 0x008E64, 0x008E60, 0x008E63, 0x008E55, 0x008E76, 0x008E72, 0x008E7C, 0x008E81, 0x008E87, 0x008E85, 0x008E84, 0x008E8B, 0x008E8A, 0x008E93, 0x008E91,
	0x008E94, 0x008E99, 0x008EAA, 0x008EA1, 0x008EAC, 0x008EB0, 0x008EC6, 0x008EB1, 0x008EBE, 0x008EC5, 0x008EC8, 0x008ECB, 0x008EDB, 0x008EE3, 0x008EFC, 0x008EFB, 0x008EEB, 0x008EFE, 0x008F0A, 0x008F05, 0x008F15, 0x008F12, 0x008F19, 0x008F13, 0x008F1C, 0x008F1F, 0x008F1B, 0x008F0C, 0x008F26, 0x008F33, 0x008F3B, 0x008F39,
	0x008F45, 0x008F42, 0x008F3E, 0x008F4C, 0x008F49, 0x008F46, 0x008F4E, 0x008F57, 0x008F5C, 0x008F62, 0x008F63, 0x008F64, 0x008F9C, 0x008F9F, 0x008FAD, 0x008FAF, 0x008FB7, 0x008FDA, 0x008FE5, 0x008FE2, 0x008FEA, 0x008FEF, 0x009087, 0x008FF4, 0x009005, 0x008FF9, 0x008FFA, 0x009011, 0x009015, 0x009021, 0x00900D, 0x00901E,
	0x009016, 0x00900B, 0x009027, 0x009036, 0x009035, 0x009039, 0x008FF8, 0x00904F, 0x009050, 0x009051, 0x009052, 0x00900E, 0x009049, 0x00903E, 0x009056, 0x009058, 0x00905E, 0x009068, 0x00906F, 0x009076, 0x0096A8, 0x009072, 0x009082, 0x00907D, 0x009081, 0x009080, 0x00908A, 0x009089, 0x00908F, 0x0090A8, 0x0090AF, 0x0090B1,
	0x0090B5, 0x0090E2, 0x0090E4, 0x006248, 0x0090DB, 0x009102, 0x009112, 0x009119, 0x009132, 0x009130, 0x00914A, 0x009156, 0x009158, 0x009163, 0x009165, 0x009169, 0x009173, 0x009172, 0x00918B, 0x009189, 0x009182, 0x0091A2, 0x0091AB, 0x0091AF, 0x0091AA, 0x0091B5, 0x0091B4, 0x0091BA, 0x0091C0, 0x0091C1, 0x0091C9, 0x0091CB,
	0x0091D0, 0x0091D6, 0x0091DF, 0x0091E1, 0x0091DB, 0x0091FC, 0x0091F5, 0x0091F6, 0x00921E, 0x0091FF, 0x009214, 0x00922C, 0x009215, 0x009211, 0x00925E, 0x009257, 0x009245, 0x009249, 0x009264, 0x009248, 0x009295, 0x00923F, 0x00924B, 0x009250, 0x00929C, 0x009296, 0x009293, 0x00929B, 0x00925A, 0x0092CF, 0x0092B9, 0x0092B7,
	0x0092E9, 0x00930F, 0x0092FA, 0x009344, 0x009319, 0x009322, 0x00931A, 0x009323, 0x00933A, 0x009335, 0x00933B, 0x00935C, 0x009360, 0x00937C, 0x00936E, 0x009356, 0x0093B0, 0x0093AC, 0x0093AD, 0x009394, 0x0093B9, 0x0093D6, 0x0093D7, 0x0093E8, 0x0093E5, 0x0093D8, 0x0093C3, 0x0093DD, 0x0093D0, 0x0093C8, 0x0093E4, 0x00941A,
	0x009414, 0x009413, 0x009403, 0x009407, 0x009410, 0x009436, 0x00942B, 0x009435, 0x009421, 0x00943A, 0x009441, 0x009452, 0x009444, 0x00945B, 0x009460, 0x009462, 0x00945E, 0x00946A, 0x009229, 0x009470, 0x009475, 0x009477, 0x00947D, 0x00945A, 0x00947C, 0x00947E, 0x009481, 0x00947F, 0x009582, 0x009587, 0x00958A, 0x009594,
	0x009596, 0x009598, 0x009599, 0x0095A0, 0x0095A8, 0x0095A7, 0x0095AD, 0x0095BC, 0x0095BB, 0x0095B9, 0x0095BE, 0x0095CA, 0x006FF6, 0x0095C3, 0x0095CD, 0x0095CC, 0x0095D5, 0x0095D4, 0x0095D6, 0x0095DC, 0x0095E1, 0x0095E5, 0x0095E2, 0x009621, 0x009628, 0x00962E, 0x00962F, 0x009642, 0x00964C, 0x00964F, 0x00964B, 0x009677,
	0x00965C, 0x00965E, 0x00965D, 0x00965F, 0x009666, 0x009672, 0x00966C, 0x00968D, 0x009698, 0x009695, 0x009697, 0x0096AA, 0x0096A7, 0x0096B1, 0x0096B2, 0x0096B0, 0x0096B4, 0x0096B6, 0x0096B8, 0x0096B9, 0x0096CE, 0x0096CB, 0x0096C9, 0x0096CD, 0x00894D, 0x0096DC, 0x00970D, 0x0096D5, 0x0096F9, 0x009704, 0x009706, 0x009708,
	0x009713, 0x00970E, 0x009711, 0x00970F, 0x009716, 0x009719, 0x009724, 0x00972A, 0x009730, 0x009739, 0x00973D, 0x00973E, 0x009744, 0x009746, 0x009748, 0x009742, 0x009749, 0x00975C, 0x009760, 0x009764, 0x009766, 0x009768, 0x0052D2, 0x00976B, 0x009771, 0x009779, 0x009785, 0x00977C, 0x009781, 0x00977A, 0x009786, 0x00978B,
	0x00978F, 0x009790, 0x00979C, 0x0097A8, 0x0097A6, 0x0097A3, 0x0097B3, 0x0097B4, 0x0097C3, 0x0097C6, 0x0097C8, 0x0097CB, 0x0097DC, 0x0097ED, 0x009F4F, 0x0097F2, 0x007ADF, 0x0097F6, 0x0097F5, 0x00980F, 0x00980C, 0x009838, 0x009824, 0x009821, 0x009837, 0x00983D, 0x009846, 0x00984F, 0x00984B, 0x00986B, 0x00986F, 0x009870,
	0
};

static const unsigned map_kanji_jis2_4[] = {
	0x009871, 0x009874, 0x009873, 0x0098AA, 0x0098AF, 0x0098B1, 0x0098B6, 0x0098C4, 0x0098C3, 0x0098C6, 0x0098E9, 0x0098EB, 0x009903, 0x009909, 0x009912, 0x009914, 0x009918, 0x009921, 0x00991D, 0x00991E, 0x009924, 0x009920, 0x00992C, 0x00992E, 0x00993D, 0x00993E, 0x009942, 0x009949, 0x009945, 0x009950, 0x00994B, 0x009951,
	0x009952, 0x00994C, 0x009955, 0x009997, 0x009998, 0x0099A5, 0x0099AD, 0x0099AE, 0x0099BC, 0x0099DF, 0x0099DB, 0x0099DD, 0x0099D8, 0x0099D1, 0x0099ED, 0x0099EE, 0x0099F1, 0x0099F2, 0x0099FB, 0x0099F8, 0x009A01, 0x009A0F, 0x009A05, 0x0099E2, 0x009A19, 0x009A2B, 0x009A37, 0x009A45, 0x009A42, 0x009A40, 0x009A43, 0x009A3E,
	0x009A55, 0x009A4D, 0x009A5B, 0x009A57, 0x009A5F, 0x009A62, 0x009A65, 0x009A64, 0x009A69, 0x009A6B, 0x009A6A, 0x009AAD, 0x009AB0, 0x009ABC, 0x009AC0, 0x009ACF, 0x009AD1, 0x009AD3, 0x009AD4, 0x009ADE, 0x009ADF, 0x009AE2, 0x009AE3, 0x009AE6, 0x009AEF, 0x009AEB, 0x009AEE, 0x009AF4, 0x009AF1, 0x009AF7, 0x009AFB, 0x009B06,
	0x009B18, 0x009B1A, 0x009B1F, 0x009B22, 0x009B23, 0x009B25, 0x009B27, 0x009B28, 0x009B29, 0x009B2A, 0x009B2E, 0x009B2F, 0x009B32, 0x009B44, 0x009B43, 0x009B4F, 0x009B4D, 0x009B4E, 0x009B51, 0x009B58, 0x009B74, 0x009B93, 0x009B83, 0x009B91, 0x009B96, 0x009B97, 0x009B9F, 0x009BA0, 0x009BA8, 0x009BB4, 0x009BC0, 0x009BCA,
	0x009BB9, 0x009BC6, 0x009BCF, 0x009BD1, 0x009BD2, 0x009BE3, 0x009BE2, 0x009BE4, 0x009BD4, 0x009BE1, 0x009C3A, 0x009BF2, 0x009BF1, 0x009BF0, 0x009C15, 0x009C14, 0x009C09, 0x009C13, 0x009C0C, 0x009C06, 0x009C08, 0x009C12, 0x009C0A, 0x009C04, 0x009C2E, 0x009C1B, 0x009C25, 0x009C24, 0x009C21, 0x009C30, 0x009C47, 0x009C32,
	0x009C46, 0x009C3E, 0x009C5A, 0x009C60, 0x009C67, 0x009C76, 0x009C78, 0x009CE7, 0x009CEC, 0x009CF0, 0x009D09, 0x009D08, 0x009CEB, 0x009D03, 0x009D06, 0x009D2A, 0x009D26, 0x009DAF, 0x009D23, 0x009D1F, 0x009D44, 0x009D15, 0x009D12, 0x009D41, 0x009D3F, 0x009D3E, 0x009D46, 0x009D48, 0x009D5D, 0x009D5E, 0x009D64, 0x009D51,
	0x009D50, 0x009D59, 0x009D72, 0x009D89, 0x009D87, 0x009DAB, 0x009D6F, 0x009D7A, 0x009D9A, 0x009DA4, 0x009DA9, 0x009DB2, 0x009DC4, 0x009DC1, 0x009DBB, 0x009DB8, 0x009DBA, 0x009DC6, 0x009DCF, 0x009DC2, 0x009DD9, 0x009DD3, 0x009DF8, 0x009DE6, 0x009DED, 0x009DEF, 0x009DFD, 0x009E1A, 0x009E1B, 0x009E1E, 0x009E75, 0x009E79,
	0x009E7D, 0x009E81, 0x009E88, 0x009E8B, 0x009E8C, 0x009E92, 0x009E95, 0x009E91, 0x009E9D, 0x009EA5, 0x009EA9, 0x009EB8, 0x009EAA, 0x009EAD, 0x009761, 0x009ECC, 0x009ECE, 0x009ECF, 0x009ED0, 0x009ED4, 0x009EDC, 0x009EDE, 0x009EDD, 0x009EE0, 0x009EE5, 0x009EE8, 0x009EEF, 0x009EF4, 0x009EF6, 0x009EF7, 0x009EF9, 0x009EFB,
	0x009EFC, 0x009EFD, 0x009F07, 0x009F08, 0x0076B7, 0x009F15, 0x009F21, 0x009F2C, 0x009F3E, 0x009F4A, 0x009F52, 0x009F54, 0x009F63, 0x009F5F, 0x009F60, 0x009F61, 0x009F66, 0x009F67, 0x009F6C, 0x009F6A, 0x009F77, 0x009F72, 0x009F76, 0x009F95, 0x009F9C, 0x009FA0, 0x00582F, 0x0069C7, 0x009059, 0x007464, 0x0051DC, 0x007199,
	0
};

// IBM Extension, including Compatibility Ideograph
static const unsigned map_kanji_ibm[] = {
	0x007E8A, 0x00891C, 0x009348, 0x009288, 0x0084DC, 0x004FC9, 0x0070BB, 0x006631, 0x0068C8, 0x0092F9, 0x0066FB, 0x005F45, 0x004E28, 0x004EE1, 0x004EFC, 0x004F00, 0x004F03, 0x004F39, 0x004F56, 0x004F92, 0x004F8A, 0x004F9A, 0x004F94, 0x004FCD, 0x005040, 0x005022, 0x004FFF, 0x00501E, 0x005046, 0x005070, 0x005042, 0x005094,
	0x0050F4, 0x0050D8, 0x00514A, 0x005164, 0x00519D, 0x0051BE, 0x0051EC, 0x005215, 0x00529C, 0x0052A6, 0x0052C0, 0x0052DB, 0x005300, 0x005307, 0x005324, 0x005372, 0x005393, 0x0053B2, 0x0053DD, 0x00FA0E, 0x00549C, 0x00548A, 0x0054A9, 0x0054FF, 0x005586, 0x005759, 0x005765, 0x0057AC, 0x0057C8, 0x0057C7, 0x00FA0F, 0x00FA10,
	0x00589E, 0x0058B2, 0x00590B, 0x005953, 0x00595B, 0x00595D, 0x005963, 0x0059A4, 0x0059BA, 0x005B56, 0x005BC0, 0x00752F, 0x005BD8, 0x005BEC, 0x005C1E, 0x005CA6, 0x005CBA, 0x005CF5, 0x005D27, 0x005D53, 0x00FA11, 0x005D42, 0x005D6D, 0x005DB8, 0x005DB9, 0x005DD0, 0x005F21, 0x005F34, 0x005F67, 0x005FB7, 0x005FDE, 0x00605D,
	0x006085, 0x00608A, 0x0060DE, 0x0060D5, 0x006120, 0x0060F2, 0x006111, 0x006137, 0x006130, 0x006198, 0x006213, 0x0062A6, 0x0063F5, 0x006460, 0x00649D, 0x0064CE, 0x00654E, 0x006600, 0x006615, 0x00663B, 0x006609, 0x00662E, 0x00661E, 0x006624, 0x006665, 0x006657, 0x006659, 0x00FA12, 0x006673, 0x006699, 0x0066A0, 0x0066B2,
	0x0066BF, 0x0066FA, 0x00670E, 0x00F929, 0x006766, 0x0067BB, 0x006852, 0x0067C0, 0x006801, 0x006844, 0x0068CF, 0x00FA13, 0x006968, 0x00FA14, 0x006998, 0x0069E2, 0x006A30, 0x006A6B, 0x006A46, 0x006A73, 0x006A7E, 0x006AE2, 0x006AE4, 0x006BD6, 0x006C3F, 0x006C5C, 0x006C86, 0x006C6F, 0x006CDA, 0x006D04, 0x006D87, 0x006D6F,
	0x006D96, 0x006DAC, 0x006DCF, 0x006DF8, 0x006DF2, 0x006DFC, 0x006E39, 0x006E5C, 0x006E27, 0x006E3C, 0x006EBF, 0x006F88, 0x006FB5, 0x006FF5, 0x007005, 0x007007, 0x007028, 0x007085, 0x0070AB, 0x00710F, 0x007104, 0x00715C, 0x007146, 0x007147, 0x00FA15, 0x0071C1, 0x0071FE, 0x0072B1, 0x0072BE, 0x007324, 0x00FA16, 0x007377,
	0x0073BD, 0x0073C9, 0x0073D6, 0x0073E3, 0x0073D2, 0x007407, 0x0073F5, 0x007426, 0x00742A, 0x007429, 0x00742E, 0x007462, 0x007489, 0x00749F, 0x007501, 0x00756F, 0x007682, 0x00769C, 0x00769E, 0x00769B, 0x0076A6, 0x00FA17, 0x007746, 0x0052AF, 0x007821, 0x00784E, 0x007864, 0x00787A, 0x007930, 0x00FA18, 0x00FA19, 0x00FA1A,
	0x007994, 0x00FA1B, 0x00799B, 0x007AD1, 0x007AE7, 0x00FA1C, 0x007AEB, 0x007B9E, 0x00FA1D, 0x007D48, 0x007D5C, 0x007DB7, 0x007DA0, 0x007DD6, 0x007E52, 0x007F47, 0x007FA1, 0x00FA1E, 0x008301, 0x008362, 0x00837F, 0x0083C7, 0x0083F6, 0x008448, 0x0084B4, 0x008553, 0x008559, 0x00856B, 0x00FA1F, 0x0085B0, 0x00FA20, 0x00FA21,
	0x008807, 0x0088F5, 0x008A12, 0x008A37, 0x008A79, 0x008AA7, 0x008ABE, 0x008ADF, 0x00FA22, 0x008AF6, 0x008B53, 0x008B7F, 0x008CF0, 0x008CF4, 0x008D12, 0x008D76, 0x00FA23, 0x008ECF, 0x00FA24, 0x00FA25, 0x009067, 0x0090DE, 0x00FA26, 0x009115, 0x009127, 0x0091DA, 0x0091D7, 0x0091DE, 0x0091ED, 0x0091EE, 0x0091E4, 0x0091E5,
	0x009206, 0x009210, 0x00920A, 0x00923A, 0x009240, 0x00923C, 0x00924E, 0x009259, 0x009251, 0x009239, 0x009267, 0x0092A7, 0x009277, 0x009278, 0x0092E7, 0x0092D7, 0x0092D9, 0x0092D0, 0x00FA27, 0x0092D5, 0x0092E0, 0x0092D3, 0x009325, 0x009321, 0x0092FB, 0x00FA28, 0x00931E, 0x0092FF, 0x00931D, 0x009302, 0x009370, 0x009357,
	0x0093A4, 0x0093C6, 0x0093DE, 0x0093F8, 0x009431, 0x009445, 0x009448, 0x009592, 0x00F9DC, 0x00FA29, 0x00969D, 0x0096AF, 0x009733, 0x00973B, 0x009743, 0x00974D, 0x00974F, 0x009751, 0x009755, 0x009857, 0x009865, 0x00FA2A, 0x00FA2B, 0x009927, 0x00FA2C, 0x00999E, 0x009A4E, 0x009AD9, 0x009ADC, 0x009B75, 0x009B72, 0x009B8F,
	0x009BB1, 0x009BBB, 0x009C00, 0x009D70, 0x009D6B, 0x00FA2D, 0x009E19, 0x009ED1,
	0
};

// JIS X 0213 Level 3
// Including Compatibility Ideograph for KS X 1001
// Including Compatibility Ideograph in Unicode 3.2
static const unsigned map_kanji_jis3_1[] = {
	0x004FF1, 0x02000B, 0x003402, 0x004E2F, 0x004E30, 0x004E8D, 0x004EFD, 0x004EFF, 0x004F0B, 0x004F60, 0x004F48, 0x004F49, 0x004F5F, 0x004F6A, 0x004F6C, 0x004F7E, 0x004F97, 0x00FA30, 0x004FE0, 0x005001, 0x005002, 0x00500E, 0x005018, 0x005027, 0x00502E, 0x00503B, 0x005041, 0x0050CC, 0x0050F2, 0x0050D0, 0x0050E6, 0x00FA31,
	0x005106, 0x005103, 0x00510B, 0x00511E, 0x005135, 0x00FA32, 0x005155, 0x005157, 0x0034B5, 0x0051C3, 0x0051CA, 0x0051DE, 0x0051E2, 0x0051EE, 0x005201, 0x0034DB, 0x005213, 0x005249, 0x005257, 0x005261, 0x005293, 0x0052C8, 0x00FA33, 0x0052CC, 0x0052D0, 0x0052D6, 0x00FA34, 0x0052F0, 0x0052FB, 0x00531C, 0x00FA35, 0x005361,
	0x005363, 0x00537D, 0x00539D, 0x005412, 0x005427, 0x00544D, 0x00546B, 0x005474, 0x00547F, 0x005488, 0x005496, 0x0054A1, 0x0054C6, 0x00550E, 0x00552B, 0x005535, 0x005550, 0x00555E, 0x005581, 0x00558E, 0x00FA36, 0x0055AD, 0x0055CE, 0x00FA37, 0x005608, 0x00560E, 0x00563B, 0x005649, 0x005676, 0x005666, 0x00FA38, 0x00566F,
	0x005671, 0x005672, 0x005699, 0x00569E, 0x0056A9, 0x0056AC, 0x0056B3, 0x0056C9, 0x0056CA, 0x00570A, 0x02123D, 0x005721, 0x00572F, 0x005733, 0x005734, 0x005770, 0x005777, 0x00577C, 0x00579C, 0x02131B, 0x0057B8, 0x0057CF, 0x0057E4, 0x0057ED, 0x0057F5, 0x0057F6, 0x0057FF, 0x005809, 0x005864, 0x00FA39, 0x00587C, 0x005889,
	0x00FA3A, 0x0058A9, 0x02146E, 0x0058D2, 0x0058CE, 0x0058D4, 0x0058DA, 0x0058E0, 0x0058E9, 0x00590C, 0x008641, 0x00596D, 0x00598B, 0x005992, 0x0059C3, 0x0059D2, 0x0059DD, 0x005A13, 0x005A23, 0x005A67, 0x005A6D, 0x005A77, 0x005A7E, 0x005A84, 0x005A9E, 0x005AA7, 0x005AC4, 0x0218BD, 0x005B19, 0x005B25, 0x005B41, 0x005B7D,
	0x005B93, 0x005C12, 0x005C23, 0x005C2B, 0x00378D, 0x005C62, 0x00FA3B, 0x00FA3C, 0x0216B4, 0x005C7A, 0x005C8F, 0x005C9F, 0x005CA3, 0x005CAA, 0x005CCB, 0x005CD0, 0x005CD2, 0x005CF4, 0x021E34, 0x0037E2, 0x005D0D, 0x005D46, 0x005D47, 0x005D4A, 0x005D81, 0x005DA0, 0x005DA4, 0x005DA7, 0x005DCB, 0x00541E, 0x005653, 0x005DE2,
	0x005E14, 0x005E18, 0x005E58, 0x005E5E, 0x005EBE, 0x00F928, 0x005ECB, 0x005EF9, 0x005F00, 0x005F02, 0x005F07, 0x005F1D, 0x005F23, 0x005F36, 0x005F3D, 0x005F40, 0x005F54, 0x005F58, 0x005F64, 0x005F7D, 0x005F89, 0x005F9C, 0x005FA7, 0x005FAF, 0x005FB5, 0x005FC9, 0x005FE1, 0x005FE9, 0x00600D, 0x006014, 0x006018, 0x006033,
	0x006035, 0x006047, 0x00FA3D, 0x00609D, 0x00609E, 0x0060CB, 0x0060D4, 0x0060DD, 0x0060F8, 0x00611C, 0x00612B, 0x00FA3E, 0x00618D, 0x00FA3F, 0x0061BC, 0x0061B9, 0x00FA40, 0x006222, 0x00623E, 0x006243, 0x006256, 0x00625A, 0x00626F, 0x006285, 0x0062C4, 0x0062D6, 0x0062FC, 0x00630A, 0x006318, 0x006339, 0x006343, 0x006365,
	0x00637C, 0x0063E5, 0x0063ED, 0x006410, 0x006414, 0x006422, 0x006479, 0x006451, 0x00646D, 0x0064BE, 0x0064BF, 0x0064C4, 0x0064CA, 0x0064D0, 0x0064F7, 0x0064FB, 0x006522, 0x006529, 0x00FA41, 0x006567, 0x00659D, 0x00FA42, 0x00663A, 0x006622, 0x00662B, 0x006630, 0x006633, 0x006648, 0x00664C, 0x0231C4, 0x00665A, 0x006661,
	0x006677, 0x006678, 0x00668D, 0x00FA43, 0x0066BB, 0x0066C6, 0x0066C8, 0x003B22, 0x0066DB, 0x0066E8, 0x006713, 0x006733, 0x006747, 0x006748, 0x00677B, 0x006781, 0x006793, 0x006798, 0x00679B, 0x0067F9, 0x0067D7, 0x0067FC, 0x00681D, 0x00682C, 0x006831, 0x00685B, 0x006872, 0x006875, 0x00FA44, 0x0068A3, 0x0068A5, 0x0068B2,
	0x0068D0, 0x0068E8, 0x0068ED, 0x0068F0, 0x0068F1, 0x0068FC, 0x00690A, 0x006949, 0x0235C4, 0x006935, 0x006942, 0x006957, 0x006963, 0x006964, 0x006980, 0x0069A5, 0x0069AD, 0x0069CF, 0x003BB6, 0x003BC3, 0x0069E9, 0x0069EA, 0x0069F5, 0x0069F6, 0x006A0F, 0x006A15, 0x02373F, 0x006A3B, 0x006A3E, 0x006A45, 0x006A50, 0x006A56,
	0x006A5B, 0x023763, 0x006A89, 0x006A94, 0x006A9D, 0x006A9E, 0x006AA5, 0x006AE7, 0x003C0F, 0x00F91D, 0x006B1B, 0x006B1E, 0x006B2C, 0x006B35, 0x006B46, 0x006B56, 0x006B60, 0x006B65, 0x006B67, 0x006B77, 0x006B82, 0x006BA9, 0x006BAD, 0x00F970, 0x006BCF, 0x006BD7, 0x006BFF, 0x006C05, 0x006C10, 0x006C33, 0x006C59, 0x006CAA,
	0x006C74, 0x006C76, 0x006C85, 0x006C98, 0x006C9C, 0x006CFB, 0x006CC6, 0x006CD4, 0x006CE0, 0x006CEB, 0x006CEE, 0x023CFE, 0x006D0E, 0x006D2E, 0x006D31, 0x006D39, 0x006D3F, 0x006D58, 0x006D65, 0x00FA45, 0x006D82, 0x006D89, 0x006D94, 0x006DAA, 0x006DBF, 0x006DC4, 0x006DD6, 0x006DDA, 0x006DDB, 0x006DDD, 0x00FA46, 0x006E34,
	0x006E44, 0x006E5E, 0x006EAB, 0x006EB1, 0x006EC1, 0x006EC7, 0x006ECE, 0x006F10, 0x006F1A, 0x00FA47, 0x006F2A, 0x006F2F, 0x006F33, 0x006F51, 0x006F59, 0x006F5E, 0x006F61, 0x006F62, 0x006F7E, 0x006F8C, 0x006F8D, 0x006F94, 0x006FA0, 0x006FA7, 0x006FB6, 0x006FBC, 0x006FC7, 0x006FCA, 0x006FF9, 0x006FF0, 0x007006, 0x00704A,
	0x00705D, 0x00705E, 0x00704E, 0x007064, 0x007075, 0x0070A4, 0x0070B7, 0x0070D4, 0x0070D8, 0x0070E4, 0x00712B, 0x00711E, 0x007120, 0x00712E, 0x007130, 0x007151, 0x00FA48, 0x007152, 0x007160, 0x007168, 0x007185, 0x007187, 0x007192, 0x0071BA, 0x0071C4, 0x007200, 0x007215, 0x007255, 0x007256, 0x003E3F, 0x00728D, 0x00729B,
	0x0072C0, 0x0072FB, 0x0247F1, 0x007327, 0x007328, 0x007350, 0x007366, 0x00737C, 0x007395, 0x00739F, 0x0073A0, 0x0073A2, 0x0073A6, 0x0073AB, 0x0073CF, 0x0073D9, 0x0073E9, 0x00740A, 0x00741A, 0x00741B, 0x00FA4A, 0x007428, 0x00742B, 0x00742C, 0x00742F, 0x007430, 0x007444, 0x007446, 0x007447, 0x00744B, 0x007457, 0x00746B,
	0x00746D, 0x007486, 0x007487, 0x007498, 0x00749C, 0x0074A3, 0x007490, 0x0074A6, 0x0074A8, 0x0074A9, 0x0074B5, 0x0074BF, 0x0074C8, 0x0074C9, 0x0074DA, 0x0074FF, 0x007517, 0x007579, 0x007592, 0x003F72, 0x0075CE, 0x0075E4, 0x007600, 0x007602, 0x007608, 0x007615, 0x007616, 0x007619, 0x00761E, 0x00762D, 0x007635, 0x007643,
	0x00764B, 0x007664, 0x007665, 0x00766D, 0x00766F, 0x007671, 0x007681, 0x00769D, 0x0076AA, 0x0076B6, 0x0076C5, 0x0076CC, 0x0076CE, 0x0076D4, 0x0076E6, 0x0076F1, 0x0076FC, 0x00770A, 0x007719, 0x007734, 0x007736, 0x00774D, 0x00774E, 0x00775C, 0x00775F, 0x007762, 0x00777A, 0x007780, 0x007794, 0x0077AA, 0x0077E0, 0x00782D,
	0x02548E, 0x007843, 0x00784F, 0x007851, 0x007868, 0x00786E, 0x00FA4B, 0x0078B0, 0x02550E, 0x0078AD, 0x0078E4, 0x0078F2, 0x007900, 0x0078F7, 0x00791C, 0x00792E, 0x007931, 0x007934, 0x00FA4C, 0x00FA4D, 0x007945, 0x007946, 0x00FA4E, 0x00FA4F, 0x00FA50, 0x00795C, 0x00FA51, 0x007979, 0x00FA52, 0x00FA53, 0x007998, 0x0079B1,
	0x0079B8, 0x0079C8, 0x0079CA, 0x025771, 0x0079D4, 0x0079DE, 0x0079EB, 0x0079ED, 0x007A03, 0x00FA54, 0x007A39, 0x007A5D, 0x007A6D, 0x00FA55, 0x007A85, 0x007AA0, 0x0259C4, 0x007AB3, 0x007ABB, 0x007ACE, 0x007AFD, 0x007B12, 0x007B2D, 0x007B3B, 0x007B47, 0x007B4E, 0x007B60, 0x007B6D, 0x007B6F, 0x007B72, 0x00FA56, 0x007BD7,
	0x007BD9, 0x007C01, 0x007C31, 0x007C1E, 0x007C20, 0x007C33, 0x007C36, 0x004264, 0x025DA1, 0x007C59, 0x007C6D, 0x007C79, 0x007C8F, 0x007C94, 0x007CA0, 0x007CBC, 0x007CD5, 0x007CD9, 0x007CDD, 0x007D07, 0x007D08, 0x007D13, 0x007D1D, 0x007D23, 0x007D31, 0x007D41, 0x007D53, 0x007D7A, 0x007D83, 0x007D8B, 0x007DA6, 0x007DC2,
	0x007DCC, 0x007DE3, 0x00FA57, 0x007E28, 0x007E08, 0x007E11, 0x007E15, 0x00FA59, 0x007E47, 0x007E61, 0x007E8D, 0x00FA5A, 0x007F91, 0x007F97, 0x007FBF, 0x007FCE, 0x007FDB, 0x007FDF, 0x007FEC, 0x007FEE, 0x007FFA, 0x00FA5B, 0x008014, 0x008026, 0x008035, 0x008037, 0x00803C, 0x0080CA, 0x0080D7, 0x0080E0, 0x0080F3, 0x008118,
	0x00814A, 0x008160, 0x008167, 0x008168, 0x00816D, 0x0081BB, 0x0081CA, 0x0081CF, 0x0081D7, 0x00FA5C, 0x004453, 0x00445B, 0x008260, 0x008274, 0x026AFF, 0x00828E, 0x0082A1, 0x0082A3, 0x0082A4, 0x0082A9, 0x0082AE, 0x0082B7, 0x0082BE, 0x0082BF, 0x0082C6, 0x0082D5, 0x0082FD, 0x0082FE, 0x008300, 0x008322, 0x00832D, 0x00833A,
	0x008343, 0x008347, 0x008351, 0x008355, 0x00837D, 0x008386, 0x008392, 0x008398, 0x0083A7, 0x0083A9, 0x0083BF, 0x0083C0, 0x0083CF, 0x0083D1, 0x0083E1, 0x0083EA, 0x008401, 0x008406, 0x00840A, 0x00FA5F, 0x00845F, 0x008470, 0x008473, 0x008485, 0x00849E, 0x0084AF, 0x0084BA, 0x0084C0, 0x0084C2, 0x026E40, 0x008532, 0x00851E,
	0x008523, 0x00852F, 0x008564, 0x0085AD, 0x00857A, 0x00858C, 0x00858F, 0x0085A2, 0x0085CB, 0x0085CE, 0x0085ED, 0x008612, 0x0085FF, 0x008604, 0x008605, 0x008610, 0x0270F4, 0x008618, 0x008629, 0x008638, 0x008657, 0x00865B, 0x00F936, 0x008662, 0x00459D, 0x00866C, 0x008675, 0x008698, 0x0086B8, 0x0086FA, 0x0086FC, 0x0086FD,
	0x00870B, 0x008771, 0x008787, 0x008788, 0x0087AC, 0x0087AD, 0x0087B5, 0x0045EA, 0x0087D6, 0x0087EC, 0x008806, 0x00880A, 0x008810, 0x008814, 0x00881F, 0x008898, 0x0088AA, 0x0088CA, 0x0088CE, 0x027684, 0x00FA60, 0x008918, 0x008919, 0x00891A, 0x008927, 0x008930, 0x008932, 0x008939, 0x008940, 0x008994, 0x00FA61, 0x0089D4,
	0x0089E5, 0x0089F6, 0x008A15, 0x008A22, 0x008A47, 0x008A4E, 0x008A5D, 0x008A61, 0x008A75, 0x008AD0, 0x008AF4, 0x00FA62, 0x00FA63, 0x008B46, 0x008B54, 0x008B59, 0x008B69, 0x008B9D, 0x008C49, 0x008C68, 0x00FA64, 0x008CE1, 0x008CF8, 0x008CFE, 0x00FA65, 0x008D1B, 0x008DAF, 0x008DCE, 0x008DD1, 0x008DD7, 0x008E20, 0x008E23,
	0x008E3D, 0x008E70, 0x008E7B, 0x028277, 0x008EC0, 0x004844, 0x008EFA, 0x008F1E, 0x008F2D, 0x008F36, 0x008F54, 0x0283CD, 0x008FA6, 0x008FB5, 0x008FE4, 0x008FE8, 0x008FEE, 0x009008, 0x00902D, 0x00FA67, 0x009088, 0x009095, 0x009097, 0x009099, 0x00909B, 0x0090A2, 0x0090B3, 0x0090BE, 0x0090C4, 0x0090C5, 0x0090C7, 0x0090D7,
	0x0090DD, 0x0090EF, 0x0090F4, 0x009114, 0x009116, 0x009122, 0x009123, 0x00912F, 0x009131, 0x009134, 0x00913D, 0x009148, 0x00915B, 0x009183, 0x00919E, 0x0091AC, 0x0091B1, 0x0091BC, 0x0091FB, 0x0091F1, 0x009207, 0x009238, 0x009243, 0x00924F, 0x0092C2, 0x0092CB, 0x0092CC, 0x009304, 0x00931F, 0x009349, 0x00934A, 0x009364,
	0x009365, 0x00936A, 0x00939B, 0x0093A3, 0x0093BA, 0x0093DF, 0x009404, 0x0093FD, 0x009433, 0x00944A, 0x009463, 0x00946B, 0x009471, 0x009472, 0x00958E, 0x00959F, 0x0095A6, 0x0095A9, 0x0095AC, 0x0095B6, 0x0095BD, 0x0095CB, 0x0095D0, 0x0095D3, 0x0049B0, 0x0095DA, 0x0095DE, 0x009658, 0x009684, 0x0096A4, 0x0096A5, 0x0096D2,
	0x0096DE, 0x00FA68, 0x0096E9, 0x0096EF, 0x00974E, 0x00975A, 0x00976E, 0x009773, 0x009795, 0x0097AE, 0x0097BA, 0x0097C1, 0x0097C9, 0x0097DE, 0x0097DB, 0x0097F4, 0x00FA69, 0x00980A, 0x00981E, 0x00982B, 0x00FA6A, 0x009852, 0x009853, 0x009856, 0x009859, 0x00985A, 0x00F9D0, 0x00986C, 0x0098BA, 0x0098C8, 0x0098E7, 0x009958,
	0x009A02, 0x009A03, 0x009A24, 0x009A2D, 0x009A2E, 0x009A38, 0x009A4A, 0x009A52, 0x009AB6, 0x009AC1, 0x009AC3, 0x009ACE, 0x009AD6, 0x009AF9, 0x009B02, 0x009B08, 0x009B20, 0x004C17, 0x009B2D, 0x009B5E, 0x009B79, 0x009B66, 0x009B84, 0x009B8A, 0x009B9E, 0x009BA7, 0x009BC1, 0x009BCE, 0x009BE5, 0x009BF8, 0x009BFD, 0x009C23,
	0
};

static const unsigned map_kanji_jis3_2[] = {
	0x009C41, 0x009C4F, 0x009C50, 0x009C53, 0x009C63, 0x009C65, 0x009C77, 0x009D1D,
	0x009D1E, 0x009D43, 0x009D47, 0x009D52, 0x009D63, 0x009D7C, 0x009D8A, 0x009D96,
	0x009DC0, 0x009DAC, 0x009DBC, 0x009DD7, 0x02A190, 0x009DE7, 0x009E07, 0x009E15,
	0x009E7C, 0x009E9E, 0x009EA4, 0x009EAC, 0x009EAF, 0x009EB4, 0x009EB5, 0x009EC3,
	0x009F10, 0x009F39, 0x009F57, 0x009F90, 0x009F94, 0x009F97, 0x009FA2, 0x0059F8,
	0x005C5B, 0x005E77, 0x007626, 0x007E6B,
	0
};

// JIS X 0213 Level 4
// Including Compatibility Ideograph for KS X 1001
// Including Compatibility Ideograph in Unicode 3.2
static const unsigned map_kanji_jis4_1[] = {
	0x020089, 0x004E02, 0x004E0F, 0x004E12, 0x004E29, 0x004E2B, 0x004E2E, 0x004E40, 0x004E47, 0x004E48, 0x0200A2, 0x004E51, 0x003406, 0x0200A4, 0x004E5A, 0x004E69, 0x004E9D, 0x00342C, 0x00342E, 0x004EB9, 0x004EBB, 0x0201A2, 0x004EBC, 0x004EC3, 0x004EC8, 0x004ED0, 0x004EEB, 0x004EDA, 0x004EF1, 0x004EF5, 0x004F16, 0x004F64,
	0x004F37, 0x004F3E, 0x004F54, 0x004F58, 0x020213, 0x004F77, 0x004F78, 0x004F7A, 0x004F7D, 0x004F82, 0x004F85, 0x004FE6, 0x004FB2, 0x004FBE, 0x004FC5, 0x004FCB, 0x004FCF, 0x004FD2, 0x00346A, 0x004FF2, 0x005000, 0x005010, 0x005013, 0x00501C, 0x003468, 0x00504E, 0x005053, 0x005057, 0x005063, 0x005066, 0x00506A, 0x0050A3,
	0x005088, 0x005092, 0x005093, 0x005095, 0x005096, 0x00509C, 0x0050AA, 0x02032B, 0x0050B1, 0x0050BA, 0x0050BB, 0x0050C4, 0x0050C7, 0x0050F3, 0x020381, 0x0050CE, 0x020371, 0x0050D4, 0x0050D9, 0x0050E1, 0x0050E9, 0x003492, 0x005108, 0x0203F9, 0x005117, 0x00511B, 0x02044A, 0x005160, 0x020509, 0x005173, 0x005183, 0x00518B,
	0x0034BC, 0x005198, 0x0051A3, 0x0051AD, 0x0034C7, 0x0051BC, 0x0205D6, 0x020628, 0x0051F3, 0x0051F4, 0x005202, 0x005212, 0x005216, 0x02074F, 0x005255, 0x00525C, 0x00526C, 0x005277, 0x005284, 0x005282, 0x020807, 0x005298, 0x02083A, 0x0052A4, 0x0052BA, 0x0052BB, 0x0052CA, 0x00351F, 0x0052D1, 0x0208B9, 0x0052F7, 0x00530A,
	0x00530B, 0x005335, 0x00533E, 0x005342, 0x02097C, 0x02099D, 0x005367, 0x00536C, 0x00537A, 0x0053A4, 0x0053B4, 0x020AD3, 0x0053B7, 0x0053C0, 0x020B1D, 0x00355D, 0x00355E, 0x0053D5, 0x0053DA, 0x003563, 0x0053F4, 0x0053F5, 0x005455, 0x005424, 0x005428, 0x00356E, 0x005443, 0x005462, 0x005466, 0x00546C, 0x00548D, 0x005495,
	0x0054A0, 0x0054A6, 0x0054AD, 0x0054AE, 0x0054B7, 0x0054BA, 0x0054BF, 0x0054C3, 0x020D45, 0x0054EC, 0x0054EF, 0x0054F1, 0x0054F3, 0x005500, 0x005501, 0x005509, 0x00553C, 0x005541, 0x0035A6, 0x005547, 0x00554A, 0x0035A8, 0x005560, 0x005561, 0x005564, 0x020DE1, 0x00557D, 0x005582, 0x005588, 0x005591, 0x0035C5, 0x0055D2,
	0x020E95, 0x020E6D, 0x0055BF, 0x0055C9, 0x0055CC, 0x0055D1, 0x0055DD, 0x0035DA, 0x0055E2, 0x020E64, 0x0055E9, 0x005628, 0x020F5F, 0x005607, 0x005610, 0x005630, 0x005637, 0x0035F4, 0x00563D, 0x00563F, 0x005640, 0x005647, 0x00565E, 0x005660, 0x00566D, 0x003605, 0x005688, 0x00568C, 0x005695, 0x00569A, 0x00569D, 0x0056A8,
	0x0056AD, 0x0056B2, 0x0056C5, 0x0056CD, 0x0056DF, 0x0056E8, 0x0056F6, 0x0056F7, 0x021201, 0x005715, 0x005723, 0x021255, 0x005729, 0x02127B, 0x005745, 0x005746, 0x00574C, 0x00574D, 0x021274, 0x005768, 0x00576F, 0x005773, 0x005774, 0x005775, 0x00577B, 0x0212E4, 0x0212D7, 0x00579A, 0x00579D, 0x00579E, 0x0057A8, 0x0057D7,
	0x0212FD, 0x0057CC, 0x021336, 0x021344, 0x0057DE, 0x0057E6, 0x0057F0, 0x00364A, 0x0057F8, 0x0057FB, 0x0057FD, 0x005804, 0x00581E, 0x005820, 0x005827, 0x005832, 0x005839, 0x0213C4, 0x005849, 0x00584C, 0x005867, 0x00588A, 0x00588B, 0x00588D, 0x00588F, 0x005890, 0x005894, 0x00589D, 0x0058AA, 0x0058B1, 0x02146D, 0x0058C3,
	0x0058CD, 0x0058E2, 0x0058F3, 0x0058F4, 0x005905, 0x005906, 0x00590D, 0x005914, 0x005924, 0x0215D7, 0x003691, 0x00593D, 0x003699, 0x005946, 0x003696, 0x026C29, 0x00595F, 0x021647, 0x005975, 0x005976, 0x00597C, 0x00599F, 0x0059AE, 0x0059BC, 0x0059C8, 0x0059CD, 0x0059DE, 0x0059E3, 0x0059E4, 0x0059E7, 0x0059EE, 0x021706,
	0x021742, 0x0036CF, 0x005A0C, 0x005A0D, 0x005A17, 0x005A27, 0x005A2D, 0x005A55, 0x005A65, 0x005A7A, 0x005A8B, 0x005A9C, 0x005A9F, 0x005AA0, 0x005AA2, 0x005AB1, 0x005AB3, 0x005AB5, 0x005ABA, 0x005ABF, 0x005ADA, 0x005ADC, 0x005AE0, 0x005AE5, 0x005AF0, 0x005AEE, 0x005AF5, 0x005B00, 0x005B08, 0x005B17, 0x005B34, 0x005B2D,
	0x005B4C, 0x005B52, 0x005B68, 0x005B6F, 0x005B7C, 0x005B7F, 0x005B81, 0x005B84, 0x0219C3, 0x005B96, 0x005BAC, 0x003761, 0x003762, 0x005BCE, 0x005BD6, 0x00376C, 0x00376B, 0x005BF1, 0x005BFD, 0x003775, 0x005C03, 0x005C29, 0x005C30, 0x021C56, 0x005C5F, 0x005C63, 0x005C67, 0x005C68, 0x005C69, 0x005C70, 0x021D2D, 0x021D45,
	0x005C7C, 0x021D78, 0x021D62, 0x005C88, 0x005C8A, 0x0037C1, 0x021DA1, 0x021D9C, 0x005CA0, 0x005CA2, 0x005CA7, 0x021D92, 0x005CAD, 0x005CB5, 0x021DB7, 0x005CC9, 0x021DE0, 0x021E33, 0x005D06, 0x005D10, 0x005D2B, 0x005D1D, 0x005D20, 0x005D24, 0x005D26, 0x005D31, 0x005D39, 0x0037E8, 0x005D61, 0x005D6A, 0x0037F4, 0x005D70,
	0x021F1E, 0x0037FD, 0x005D88, 0x003800, 0x005D92, 0x005D94, 0x005D97, 0x005D99, 0x005DB0, 0x005DB2, 0x005DB4, 0x021F76, 0x005DD1, 0x005DD7, 0x005DD8, 0x005DE0, 0x021FFA, 0x005DE4, 0x005DE9, 0x00382F, 0x005E00, 0x003836, 0x005E12, 0x005E15, 0x003840, 0x005E1F, 0x005E2E, 0x005E3E, 0x005E49, 0x00385C, 0x005E56, 0x003861,
	0x005E6B, 0x005E6C, 0x005E6D, 0x005E6E, 0x02217B, 0x005EA5, 0x005EAA, 0x005EAC, 0x005EB9, 0x005EBF, 0x005EC6, 0x005ED2, 0x005ED9, 0x02231E, 0x005EFD, 0x005F08, 0x005F0E, 0x005F1C, 0x0223AD, 0x005F1E, 0x005F47, 0x005F63, 0x005F72, 0x005F7E, 0x005F8F, 0x005FA2, 0x005FA4, 0x005FB8, 0x005FC4, 0x0038FA, 0x005FC7, 0x005FCB,
	0x005FD2, 0x005FD3, 0x005FD4, 0x005FE2, 0x005FEE, 0x005FEF, 0x005FF3, 0x005FFC, 0x003917, 0x006017, 0x006022, 0x006024, 0x00391A, 0x00604C, 0x00607F, 0x006095, 0x0060A8, 0x0226F3, 0x0060B0, 0x0060B1, 0x0060BE, 0x0060C8, 0x0060D9, 0x0060DB, 0x0060EE, 0x0060F5, 0x006110, 0x006112, 0x006113, 0x006119, 0x00611E, 0x00613A,
	0x00396F, 0x006141, 0x006146, 0x006160, 0x00617C, 0x02285B, 0x006192, 0x006193, 0x006197, 0x0061A5, 0x0061A8, 0x0061AD, 0x0228AB, 0x0061D5, 0x0061DD, 0x0061DF, 0x0061F5, 0x02298F, 0x006215, 0x006223, 0x006229, 0x006246, 0x00624C, 0x006251, 0x006252, 0x006261, 0x006264, 0x00627B, 0x00626D, 0x006273, 0x006299, 0x0062D5,
	0x022AB8, 0x0062FD, 0x006303, 0x00630D, 0x006310, 0x022B4F, 0x022B50, 0x006332, 0x006335, 0x00633B, 0x00633C, 0x006341, 0x006344, 0x00634E, 0x022B46, 0x006359, 0x022C1D, 0x022BA6, 0x00636C, 0x006384, 0x006399, 0x022C24, 0x006394, 0x0063BD, 0x0063F7, 0x0063D4, 0x0063D5, 0x0063DC, 0x0063E0, 0x0063EB, 0x0063EC, 0x0063F2,
	0x006409, 0x00641E, 0x006425, 0x006429, 0x00642F, 0x00645A, 0x00645B, 0x00645D, 0x006473, 0x00647D, 0x006487, 0x006491, 0x00649F, 0x0064CB, 0x0064CC, 0x0064D5, 0x0064D7, 0x022DE1, 0x0064E4, 0x0064E5, 0x0064FF, 0x006504, 0x003A6E, 0x00650F, 0x006514, 0x006516, 0x003A73, 0x00651E, 0x006532, 0x006544, 0x006554, 0x00656B,
	0x00657A, 0x006581, 0x006584, 0x006585, 0x00658A, 0x0065B2, 0x0065B5, 0x0065B8, 0x0065BF, 0x0065C2, 0x0065C9, 0x0065D4, 0x003AD6, 0x0065F2, 0x0065F9, 0x0065FC, 0x006604, 0x006608, 0x006621, 0x00662A, 0x006645, 0x006651, 0x00664E, 0x003AEA, 0x0231C3, 0x00665B, 0x006663, 0x0231F5, 0x0231B6, 0x00666A, 0x00666B, 0x00666C,
	0x00666D, 0x00667B, 0x006680, 0x006690, 0x006692, 0x003B0E, 0x0066AD, 0x0066B1, 0x0066B5, 0x003B1A, 0x003B1C, 0x0066EC, 0x003AD7, 0x006701, 0x006705, 0x006712, 0x023372, 0x006719, 0x0233D3, 0x0233D2, 0x00674C, 0x00674D, 0x006754, 0x00675D, 0x0233D0, 0x0233E4, 0x0233D5, 0x006774, 0x006776, 0x0233DA, 0x006792, 0x0233DF,
	0x008363, 0x006810, 0x0067B0, 0x0067B2, 0x0067C3, 0x0067C8, 0x0067D2, 0x0067D9, 0x0067DB, 0x0067F0, 0x0067F7, 0x02344A, 0x023451, 0x02344B, 0x006818, 0x00681F, 0x00682D, 0x023465, 0x006833, 0x00683B, 0x00683E, 0x006845, 0x006849, 0x00684C, 0x006855, 0x006857, 0x003B77, 0x00686B, 0x00686E, 0x00687A, 0x00687C, 0x006882,
	0x006890, 0x006896, 0x003B6D, 0x006898, 0x006899, 0x00689A, 0x00689C, 0x0068AA, 0x0068AB, 0x0068B4, 0x0068BB, 0x0068FB, 0x0234E4, 0x02355A, 0x0068C3, 0x0068C5, 0x0068CC, 0x0068D6, 0x0068D9, 0x0068E4, 0x0068E5, 0x0068EC, 0x0068F7, 0x006903, 0x006907, 0x003B87, 0x003B88, 0x023594, 0x00693B, 0x003B8D, 0x006946, 0x006969,
	0x00696C, 0x006972, 0x00697A, 0x00697F, 0x006992, 0x003BA4, 0x006996, 0x0069A6, 0x0069B0, 0x0069B7, 0x0069BA, 0x0069BC, 0x0069C0, 0x0069D1, 0x0069D6, 0x023639, 0x023647, 0x023638, 0x02363A, 0x0069E3, 0x0069EE, 0x0069EF, 0x0069F3, 0x003BCD, 0x0069F4, 0x0069FE, 0x006A11, 0x006A1A, 0x006A1D, 0x02371C, 0x006A32, 0x006A33,
	0x006A34, 0x006A3F, 0x006A49, 0x006A7A, 0x006A4E, 0x006A52, 0x006A64, 0x02370C, 0x006A83, 0x006A8B, 0x003BF0, 0x006A91, 0x006A9F, 0x006AA1, 0x023764, 0x006AAB, 0x006ABD, 0x006AC6, 0x006AD4, 0x006AD0, 0x006ADC, 0x006ADD, 0x0237FF, 0x0237E7, 0x006AEC, 0x006AF1, 0x006AF2, 0x006AF3, 0x006AFD, 0x023824, 0x006B0B, 0x006B0F,
	0x006B10, 0x006B11, 0x02383D, 0x006B17, 0x003C26, 0x006B2F, 0x006B4A, 0x006B58, 0x006B6C, 0x006B75, 0x006B7A, 0x006B81, 0x006B9B, 0x006BAE, 0x023A98, 0x006BBD, 0x006BBE, 0x006BC7, 0x006BC8, 0x006BC9, 0x006BDA, 0x006BE6, 0x006BE7, 0x006BEE, 0x006BF1, 0x006C02, 0x006C0A, 0x006C0E, 0x006C35, 0x006C36, 0x006C3A, 0x023C7F,
	0x006C4D, 0x006C5B, 0x006C6D, 0x006C84, 0x006C89, 0x003CC3, 0x006C94, 0x006C95, 0x006C97, 0x006CAD, 0x006CC2, 0x006CD0, 0x003CD2, 0x006CD6, 0x006CDC, 0x006CE9, 0x006CEC, 0x006CED, 0x023D00, 0x006D00, 0x006D0A, 0x006D24, 0x006D26, 0x006D27, 0x006C67, 0x006D2F, 0x006D3C, 0x006D5B, 0x006D5E, 0x006D60, 0x006D70, 0x006D80,
	0x006D81, 0x006D8A, 0x006D8D, 0x006D91, 0x006D98, 0x023D40, 0x006E17, 0x023DFA, 0x023DF9, 0x023DD3, 0x006DAB, 0x006DAE, 0x006DB4, 0x006DC2, 0x006D34, 0x006DC8, 0x006DCE, 0x006DD0, 0x006DDF, 0x006DE9, 0x006DF6, 0x006E36, 0x006E1E, 0x006E22, 0x003D11, 0x006E32, 0x006E48, 0x006E49, 0x006E4B, 0x006E4C, 0x006E4F, 0x006E51,
	0x006E53, 0x006E54, 0x006E57, 0x006E63, 0x003D1E, 0x006E93, 0x006EA7, 0x006EB4, 0x006EC3, 0x006ECA, 0x006ED9, 0x006F35, 0x006EEB, 0x006EF9, 0x006EFB, 0x006F0A, 0x006F0C, 0x006F18, 0x006F25, 0x006F36, 0x006F3C, 0x023F7E, 0x006F52, 0x006F57, 0x006F5A, 0x006F60, 0x006F68, 0x006F98, 0x006F7D, 0x006F90, 0x006F96, 0x006FBE,
	0x006F9F, 0x006FA5, 0x006FAF, 0x003D64, 0x006FC8, 0x006FC9, 0x006FDA, 0x006FDE, 0x006FE9, 0x024096, 0x006FFC, 0x007000, 0x00700A, 0x007023, 0x024103, 0x007039, 0x00703A, 0x00703C, 0x007043, 0x007047, 0x00704B, 0x003D9A, 0x007054, 0x007065, 0x007069, 0x00706C, 0x00706E, 0x007076, 0x00707E, 0x007081, 0x007086, 0x007095,
	0x007097, 0x0241C6, 0x00709F, 0x0070B1, 0x0241FE, 0x0070EC, 0x0070CA, 0x0070D1, 0x0070D3, 0x0070DC, 0x007103, 0x007106, 0x007107, 0x007108, 0x00710C, 0x003DC0, 0x00712F, 0x007131, 0x007150, 0x00714A, 0x007153, 0x00715E, 0x003DD4, 0x007196, 0x007180, 0x00719B, 0x0071A0, 0x0071A2, 0x0071AE, 0x0071AF, 0x0071B3, 0x0243BC,
	0x0071CB, 0x0071D3, 0x0071D9, 0x0071DC, 0x007207, 0x003E05, 0x00FA49, 0x00722B, 0x007234, 0x007238, 0x007239, 0x004E2C, 0x007242, 0x007253, 0x007257, 0x007263, 0x024629, 0x00726E, 0x00726F, 0x007278, 0x00727F, 0x00728E, 0x0246A5, 0x0072AD, 0x0072AE, 0x0072B0, 0x0072C1, 0x003E60, 0x0072CC, 0x003E66, 0x003E68, 0x0072F3,
	0
};

static const unsigned map_kanji_jis4_2[] = {
	0x0072FA, 0x007307, 0x007312, 0x007318, 0x007319, 0x003E83, 0x007339, 0x00732C, 0x007331, 0x007333, 0x00733D, 0x007352, 0x003E94, 0x00736B, 0x00736C, 0x024896, 0x00736E, 0x00736F, 0x007371, 0x007381, 0x007385, 0x00738A, 0x007394, 0x007398, 0x00739C, 0x00739E, 0x0073A5, 0x0073A8, 0x0073B5, 0x0073B7, 0x0073B9, 0x0073BC,
	0x0073BF, 0x0073C5, 0x0073CB, 0x0073E1, 0x0073E7, 0x0073F9, 0x007413, 0x0073FA, 0x007401, 0x007424, 0x007431, 0x007439, 0x007453, 0x007440, 0x007443, 0x00744D, 0x007452, 0x00745D, 0x007471, 0x007481, 0x007485, 0x007488, 0x024A4D, 0x007492, 0x007497, 0x007499, 0x0074A0, 0x0074A1, 0x0074A5, 0x0074AA, 0x0074AB, 0x0074B9,
	0x0074BB, 0x0074BA, 0x0074D6, 0x0074D8, 0x0074DE, 0x0074EF, 0x0074EB, 0x024B56, 0x0074FA, 0x024B6F, 0x007520, 0x007524, 0x00752A, 0x003F57, 0x024C16, 0x00753D, 0x00753E, 0x007540, 0x007548, 0x00754E, 0x007550, 0x007552, 0x00756C, 0x007572, 0x007571, 0x00757A, 0x00757D, 0x00757E, 0x007581, 0x024D14, 0x00758C, 0x003F75,
	0x0075A2, 0x003F77, 0x0075B0, 0x0075B7, 0x0075BF, 0x0075C0, 0x0075C6, 0x0075CF, 0x0075D3, 0x0075DD, 0x0075DF, 0x0075E0, 0x0075E7, 0x0075EC, 0x0075EE, 0x0075F1, 0x0075F9, 0x007603, 0x007618, 0x007607, 0x00760F, 0x003FAE, 0x024E0E, 0x007613, 0x00761B, 0x00761C, 0x024E37, 0x007625, 0x007628, 0x00763C, 0x007633, 0x024E6A,
	0x003FC9, 0x007641, 0x024E8B, 0x007649, 0x007655, 0x003FD7, 0x00766E, 0x007695, 0x0076A1, 0x0076A0, 0x0076A7, 0x0076A8, 0x0076AF, 0x02504A, 0x0076C9, 0x025055, 0x0076E8, 0x0076EC, 0x025122, 0x007717, 0x00771A, 0x00772D, 0x007735, 0x0251A9, 0x004039, 0x0251E5, 0x0251CD, 0x007758, 0x007760, 0x00776A, 0x02521E, 0x007772,
	0x00777C, 0x00777D, 0x02524C, 0x004058, 0x00779A, 0x00779F, 0x0077A2, 0x0077A4, 0x0077A9, 0x0077DE, 0x0077DF, 0x0077E4, 0x0077E6, 0x0077EA, 0x0077EC, 0x004093, 0x0077F0, 0x0077F4, 0x0077FB, 0x02542E, 0x007805, 0x007806, 0x007809, 0x00780D, 0x007819, 0x00782C, 0x007847, 0x00786A, 0x0254D9, 0x00788A, 0x007894, 0x0078A4,
	0x00789D, 0x00789E, 0x00789F, 0x0078BB, 0x0078C8, 0x0078CC, 0x0078CE, 0x0078D5, 0x0078E0, 0x0078E1, 0x0078E6, 0x0078F9, 0x0078FA, 0x0078FB, 0x0078FE, 0x0255A7, 0x007910, 0x00791B, 0x007925, 0x00793B, 0x00794A, 0x007958, 0x00795B, 0x004105, 0x007967, 0x007972, 0x007995, 0x007996, 0x0079A1, 0x0079A9, 0x0079B4, 0x0079BB,
	0x0079C2, 0x0079C7, 0x0079CC, 0x0079CD, 0x0079D6, 0x004148, 0x0257A9, 0x0257B4, 0x00414F, 0x007A0A, 0x007A11, 0x007A15, 0x007A1B, 0x007A1E, 0x004163, 0x007A2D, 0x007A38, 0x007A47, 0x007A4C, 0x007A56, 0x007A59, 0x007A5C, 0x007A5F, 0x007A60, 0x007A67, 0x007A6A, 0x007A75, 0x007A78, 0x007A82, 0x007A8A, 0x007A90, 0x007AA3,
	0x007AAC, 0x0259D4, 0x0041B4, 0x007AB9, 0x007ABC, 0x007ABE, 0x0041BF, 0x007ACC, 0x007AE8, 0x007AF4, 0x025AE4, 0x025AE3, 0x007B07, 0x025AF1, 0x007B3D, 0x007B27, 0x007B2A, 0x007B2E, 0x007B2F, 0x007B31, 0x0041E6, 0x0041F3, 0x007B7F, 0x007B41, 0x0041EE, 0x007B55, 0x007B79, 0x007B64, 0x007B66, 0x007B69, 0x007B73, 0x025BB2,
	0x004207, 0x007B90, 0x007B91, 0x007B9B, 0x00420E, 0x007BAF, 0x007BB5, 0x007BBC, 0x007BC5, 0x007BCA, 0x025C4B, 0x025C64, 0x007BD4, 0x007BD6, 0x007BDA, 0x007BEA, 0x007BF0, 0x007C03, 0x007C0B, 0x007C0E, 0x007C0F, 0x007C26, 0x007C45, 0x007C4A, 0x007C51, 0x007C57, 0x007C5E, 0x007C61, 0x007C69, 0x007C6E, 0x007C6F, 0x007C70,
	0x025E2E, 0x025E56, 0x025E65, 0x007CA6, 0x025E62, 0x007CB6, 0x007CB7, 0x007CBF, 0x025ED8, 0x007CC4, 0x025EC2, 0x007CC8, 0x007CCD, 0x025EE8, 0x007CD7, 0x025F23, 0x007CE6, 0x007CEB, 0x025F5C, 0x007CF5, 0x007D03, 0x007D09, 0x0042C6, 0x007D12, 0x007D1E, 0x025FE0, 0x025FD4, 0x007D3D, 0x007D3E, 0x007D40, 0x007D47, 0x02600C,
	0x025FFB, 0x0042D6, 0x007D59, 0x007D5A, 0x007D6A, 0x007D70, 0x0042DD, 0x007D7F, 0x026017, 0x007D86, 0x007D88, 0x007D8C, 0x007D97, 0x026060, 0x007D9D, 0x007DA7, 0x007DAA, 0x007DB6, 0x007DC0, 0x007DD7, 0x007DD9, 0x007DE6, 0x007DF1, 0x007DF9, 0x004302, 0x0260ED, 0x00FA58, 0x007E10, 0x007E17, 0x007E1D, 0x007E20, 0x007E27,
	0x007E2C, 0x007E45, 0x007E73, 0x007E75, 0x007E7E, 0x007E86, 0x007E87, 0x00432B, 0x007E91, 0x007E98, 0x007E9A, 0x004343, 0x007F3C, 0x007F3B, 0x007F3E, 0x007F43, 0x007F44, 0x007F4F, 0x0034C1, 0x026270, 0x007F52, 0x026286, 0x007F61, 0x007F63, 0x007F64, 0x007F6D, 0x007F7D, 0x007F7E, 0x02634C, 0x007F90, 0x00517B, 0x023D0E,
	0x007F96, 0x007F9C, 0x007FAD, 0x026402, 0x007FC3, 0x007FCF, 0x007FE3, 0x007FE5, 0x007FEF, 0x007FF2, 0x008002, 0x00800A, 0x008008, 0x00800E, 0x008011, 0x008016, 0x008024, 0x00802C, 0x008030, 0x008043, 0x008066, 0x008071, 0x008075, 0x00807B, 0x008099, 0x00809C, 0x0080A4, 0x0080A7, 0x0080B8, 0x02667E, 0x0080C5, 0x0080D5,
	0x0080D8, 0x0080E6, 0x0266B0, 0x00810D, 0x0080F5, 0x0080FB, 0x0043EE, 0x008135, 0x008116, 0x00811E, 0x0043F0, 0x008124, 0x008127, 0x00812C, 0x02671D, 0x00813D, 0x004408, 0x008169, 0x004417, 0x008181, 0x00441C, 0x008184, 0x008185, 0x004422, 0x008198, 0x0081B2, 0x0081C1, 0x0081C3, 0x0081D6, 0x0081DB, 0x0268DD, 0x0081E4,
	0x0268EA, 0x0081EC, 0x026951, 0x0081FD, 0x0081FF, 0x02696F, 0x008204, 0x0269DD, 0x008219, 0x008221, 0x008222, 0x026A1E, 0x008232, 0x008234, 0x00823C, 0x008246, 0x008249, 0x008245, 0x026A58, 0x00824B, 0x004476, 0x00824F, 0x00447A, 0x008257, 0x026A8C, 0x00825C, 0x008263, 0x026AB7, 0x00FA5D, 0x00FA5E, 0x008279, 0x004491,
	0x00827D, 0x00827F, 0x008283, 0x00828A, 0x008293, 0x0082A7, 0x0082A8, 0x0082B2, 0x0082B4, 0x0082BA, 0x0082BC, 0x0082E2, 0x0082E8, 0x0082F7, 0x008307, 0x008308, 0x00830C, 0x008354, 0x00831B, 0x00831D, 0x008330, 0x00833C, 0x008344, 0x008357, 0x0044BE, 0x0044D4, 0x0044B3, 0x00838D, 0x008394, 0x008395, 0x00839B, 0x00839D,
	0x0083C9, 0x0083D0, 0x0083D4, 0x0083DD, 0x0083E5, 0x0083F9, 0x00840F, 0x008411, 0x008415, 0x026C73, 0x008417, 0x008439, 0x00844A, 0x00844F, 0x008451, 0x008452, 0x008459, 0x00845A, 0x00845C, 0x026CDD, 0x008465, 0x008476, 0x008478, 0x00847C, 0x008481, 0x00450D, 0x008497, 0x0084A6, 0x0084BE, 0x004508, 0x0084CE, 0x0084CF,
	0x0084D3, 0x026E65, 0x0084E7, 0x0084EA, 0x0084EF, 0x0084F0, 0x0084F1, 0x0084FA, 0x0084FD, 0x00850C, 0x00851B, 0x008524, 0x008525, 0x00852B, 0x008534, 0x00854F, 0x00856F, 0x004525, 0x004543, 0x00853E, 0x008551, 0x00855E, 0x008561, 0x008562, 0x026F94, 0x00857B, 0x00857D, 0x00857F, 0x008581, 0x008586, 0x008593, 0x00859D,
	0x00859F, 0x026FF8, 0x026FF6, 0x026FF7, 0x0085B7, 0x0085BC, 0x0085C7, 0x0085CA, 0x0085D8, 0x0085D9, 0x0085DF, 0x0085E1, 0x0085E6, 0x0085F6, 0x008600, 0x008611, 0x00861E, 0x008621, 0x008624, 0x008627, 0x02710D, 0x008639, 0x00863C, 0x027139, 0x008640, 0x008653, 0x008656, 0x00866F, 0x008677, 0x00867A, 0x008687, 0x008689,
	0x00868D, 0x008691, 0x00869C, 0x00869D, 0x0086A8, 0x0086B1, 0x0086B3, 0x0086C1, 0x0086C3, 0x0086D1, 0x0086D5, 0x0086D7, 0x0086E3, 0x0086E6, 0x0045B8, 0x008705, 0x008707, 0x00870E, 0x008710, 0x008713, 0x008719, 0x00871F, 0x008721, 0x008723, 0x008731, 0x00873A, 0x00873E, 0x008740, 0x008743, 0x008751, 0x008758, 0x008764,
	0x008765, 0x008772, 0x00877C, 0x0273DB, 0x0273DA, 0x0087A7, 0x008789, 0x00878B, 0x008793, 0x0087A0, 0x0273FE, 0x0045E5, 0x0087BE, 0x027410, 0x0087C1, 0x0087CE, 0x0087F5, 0x0087DF, 0x027449, 0x0087E3, 0x0087E5, 0x0087E6, 0x0087EA, 0x0087EB, 0x0087ED, 0x008801, 0x008803, 0x00880B, 0x008813, 0x008828, 0x00882E, 0x008832,
	0x00883C, 0x00460F, 0x00884A, 0x008858, 0x00885F, 0x008864, 0x027615, 0x027614, 0x008869, 0x027631, 0x00886F, 0x0088A0, 0x0088BC, 0x0088BD, 0x0088BE, 0x0088C0, 0x0088D2, 0x027693, 0x0088D1, 0x0088D3, 0x0088DB, 0x0088F0, 0x0088F1, 0x004641, 0x008901, 0x02770E, 0x008937, 0x027723, 0x008942, 0x008945, 0x008949, 0x027752,
	0x004665, 0x008962, 0x008980, 0x008989, 0x008990, 0x00899F, 0x0089B0, 0x0089B7, 0x0089D6, 0x0089D8, 0x0089EB, 0x0046A1, 0x0089F1, 0x0089F3, 0x0089FD, 0x0089FF, 0x0046AF, 0x008A11, 0x008A14, 0x027985, 0x008A21, 0x008A35, 0x008A3E, 0x008A45, 0x008A4D, 0x008A58, 0x008AAE, 0x008A90, 0x008AB7, 0x008AD7, 0x008AFC, 0x027A84,
	0x008B0A, 0x008B05, 0x008B0D, 0x008B1C, 0x008B1F, 0x008B2D, 0x008B43, 0x00470C, 0x008B51, 0x008B5E, 0x008B76, 0x008B81, 0x008B8B, 0x008B94, 0x008B95, 0x008B9C, 0x008B9E, 0x008C39, 0x027BB3, 0x008C3D, 0x027BBE, 0x027BC7, 0x008C45, 0x008C47, 0x008C4F, 0x008C54, 0x008C57, 0x008C69, 0x008C6D, 0x008C73, 0x027CB8, 0x008C93,
	0x008C92, 0x008C99, 0x004764, 0x008C9B, 0x008CA4, 0x008CD6, 0x008CD5, 0x008CD9, 0x027DA0, 0x008CF1, 0x027E10, 0x008D09, 0x008D0E, 0x008D6C, 0x008D84, 0x008D95, 0x008DA6, 0x027FB7, 0x008DC6, 0x008DC8, 0x008DD9, 0x008DEC, 0x008E0C, 0x0047FD, 0x008DFD, 0x008E06, 0x02808A, 0x008E14, 0x008E16, 0x008E21, 0x008E22, 0x008E27,
	0x0280BB, 0x004816, 0x008E36, 0x008E39, 0x008E4B, 0x008E54, 0x008E62, 0x008E6C, 0x008E6D, 0x008E6F, 0x008E98, 0x008E9E, 0x008EAE, 0x008EB3, 0x008EB5, 0x008EB6, 0x008EBB, 0x028282, 0x008ED1, 0x008ED4, 0x00484E, 0x008EF9, 0x0282F3, 0x008F00, 0x008F08, 0x008F17, 0x008F2B, 0x008F40, 0x008F4A, 0x008F58, 0x02840C, 0x008FA4,
	0x008FB4, 0x00FA66, 0x008FB6, 0x028455, 0x008FC1, 0x008FC6, 0x008FCA, 0x008FCD, 0x008FD3, 0x008FD5, 0x008FE0, 0x008FF1, 0x008FF5, 0x008FFB, 0x009002, 0x00900C, 0x009037, 0x02856B, 0x009043, 0x009044, 0x00905D, 0x0285C8, 0x0285C9, 0x009085, 0x00908C, 0x009090, 0x00961D, 0x0090A1, 0x0048B5, 0x0090B0, 0x0090B6, 0x0090C3,
	0x0090C8, 0x0286D7, 0x0090DC, 0x0090DF, 0x0286FA, 0x0090F6, 0x0090F2, 0x009100, 0x0090EB, 0x0090FE, 0x0090FF, 0x009104, 0x009106, 0x009118, 0x00911C, 0x00911E, 0x009137, 0x009139, 0x00913A, 0x009146, 0x009147, 0x009157, 0x009159, 0x009161, 0x009164, 0x009174, 0x009179, 0x009185, 0x00918E, 0x0091A8, 0x0091AE, 0x0091B3,
	0x0091B6, 0x0091C3, 0x0091C4, 0x028949, 0x028946, 0x0091EC, 0x009201, 0x009216, 0x009217, 0x02896B, 0x009233, 0x009242, 0x009247, 0x00924A, 0x009256, 0x009260, 0x009261, 0x009265, 0x009268, 0x028987, 0x028988, 0x00927C, 0x00927D, 0x00927F, 0x009289, 0x00928D, 0x009297, 0x009299, 0x00929F, 0x0092AB, 0x0289BA, 0x0289BB,
	0x0092B2, 0x0092BF, 0x0092C0, 0x0092C6, 0x0092CE, 0x0092E5, 0x009311, 0x028A1E, 0x028A29, 0x0092F7, 0x00930D, 0x009315, 0x009327, 0x009329, 0x028A71, 0x028A43, 0x009347, 0x009351, 0x00935A, 0x00936B, 0x009371, 0x009373, 0x0093A1, 0x028A99, 0x028ACD, 0x009388, 0x00938B, 0x00938F, 0x00939E, 0x0093F5, 0x028AE4, 0x028ADD,
	0x0093F1, 0x0093C1, 0x0093C7, 0x0093DC, 0x0093E2, 0x0093E7, 0x009409, 0x00940F, 0x009416, 0x009417, 0x0093FB, 0x009432, 0x009434, 0x00943B, 0x028BC1, 0x028BEF, 0x00946D, 0x00946F, 0x009578, 0x009579, 0x009586, 0x00958C, 0x00958D, 0x028D10, 0x0095AB, 0x0095B4, 0x028D71, 0x0095C8, 0x028DFB, 0x028E1F, 0x00962C, 0x009633,
	0
};

static const unsigned map_kanji_jis4_3[] = {
	0x009634, 0x028E36, 0x00963C, 0x009641, 0x009661, 0x028E89, 0x009682, 0x028EEB, 0x00969A, 0x028F32, 0x0049E7, 0x0096A9, 0x0096B3, 0x0096BA, 0x0096BD, 0x0049FA, 0x028FF8, 0x0096D8, 0x0096DA, 0x0096DD, 0x004A04, 0x009714, 0x009723, 0x004A29, 0x009736, 0x009741, 0x009747, 0x009757, 0x00975B, 0x00976A, 0x0292A0, 0x0292B1,
	0x009796, 0x00979A, 0x00979E, 0x0097A2, 0x0097B1, 0x0097B2, 0x0097BE, 0x0097CC, 0x0097D1, 0x0097D4, 0x0097D8, 0x0097D9, 0x0097E1, 0x0097F1, 0x009804, 0x00980D, 0x00980E, 0x009814, 0x009816, 0x004ABC, 0x029490, 0x009823, 0x009832, 0x009833, 0x009825, 0x009847, 0x009866, 0x0098AB, 0x0098AD, 0x0098B0, 0x0295CF, 0x0098B7,
	0x0098B8, 0x0098BB, 0x0098BC, 0x0098BF, 0x0098C2, 0x0098C7, 0x0098CB, 0x0098E0, 0x02967F, 0x0098E1, 0x0098E3, 0x0098E5, 0x0098EA, 0x0098F0, 0x0098F1, 0x0098F3, 0x009908, 0x004B3B, 0x0296F0, 0x009916, 0x009917, 0x029719, 0x00991A, 0x00991B, 0x00991C, 0x029750, 0x009931, 0x009932, 0x009933, 0x00993A, 0x00993B, 0x00993C,
	0x009940, 0x009941, 0x009946, 0x00994D, 0x00994E, 0x00995C, 0x00995F, 0x009960, 0x0099A3, 0x0099A6, 0x0099B9, 0x0099BD, 0x0099BF, 0x0099C3, 0x0099C9, 0x0099D4, 0x0099D9, 0x0099DE, 0x0298C6, 0x0099F0, 0x0099F9, 0x0099FC, 0x009A0A, 0x009A11, 0x009A16, 0x009A1A, 0x009A20, 0x009A31, 0x009A36, 0x009A44, 0x009A4C, 0x009A58,
	0x004BC2, 0x009AAF, 0x004BCA, 0x009AB7, 0x004BD2, 0x009AB9, 0x029A72, 0x009AC6, 0x009AD0, 0x009AD2, 0x009AD5, 0x004BE8, 0x009AE0, 0x009AE5, 0x009AE9, 0x009B03, 0x009B0C, 0x009B10, 0x009B12, 0x009B16, 0x009B1C, 0x009B2B, 0x009B33, 0x009B3D, 0x004C20, 0x009B4B, 0x009B63, 0x009B65, 0x009B6B, 0x009B6C, 0x009B73, 0x009B76,
	0x009B77, 0x009BA6, 0x009BAC, 0x029DDB, 0x029E3D, 0x009BB2, 0x009BB8, 0x009BBE, 0x009BC7, 0x009BF3, 0x009BD8, 0x009BDD, 0x009BE7, 0x009BEA, 0x009BEB, 0x009BEF, 0x009BEE, 0x029E15, 0x009BFA, 0x029E8A, 0x009BF7, 0x029E49, 0x009C16, 0x009C18, 0x009C19, 0x009C1A, 0x009C1D, 0x009C22, 0x009C27, 0x009C29, 0x009C2A, 0x029EC4,
	0x009C31, 0x009C36, 0x009C37, 0x009C45, 0x009C5C, 0x029EE9, 0x009C49, 0x009C4A, 0x029EDB, 0x009C54, 0x009C58, 0x009C5B, 0x009C5D, 0x009C5F, 0x009C69, 0x009C6A, 0x009C6B, 0x009C6D, 0x009C6E, 0x009C70, 0x009C72, 0x009C75, 0x009C7A, 0x009CE6, 0x009CF2, 0x009D0B, 0x009D02, 0x029FCE, 0x009D11, 0x009D17, 0x009D18, 0x02A02F,
	0x004CC4, 0x02A01A, 0x009D32, 0x004CD1, 0x009D42, 0x009D4A, 0x009D5F, 0x009D62, 0x02A0F9, 0x009D69, 0x02A082, 0x009D73, 0x009D76, 0x009D77, 0x009D7E, 0x009D84, 0x009D8D, 0x009D99, 0x009DA1, 0x009DBF, 0x009DB5, 0x009DB9, 0x009DBD, 0x009DC3, 0x009DC7, 0x009DC9, 0x009DD6, 0x009DDA, 0x009DDF, 0x009DE0, 0x009DE3, 0x009DF4,
	0x004D07, 0x009E0A, 0x009E02, 0x009E0D, 0x009E1C, 0x009E1D, 0x009E7B, 0x022218, 0x009E80, 0x009E85, 0x009E9B, 0x009EA8, 0x02A38C, 0x009EBD, 0x02A437, 0x009EDF, 0x009EE7, 0x009EEE, 0x009EFF, 0x009F02, 0x004D77, 0x009F03, 0x009F17, 0x009F19, 0x009F2F, 0x009F37, 0x009F3A, 0x009F3D, 0x009F41, 0x009F45, 0x009F46, 0x009F53,
	0x009F55, 0x009F58, 0x02A5F1, 0x009F5D, 0x02A602, 0x009F69, 0x02A61A, 0x009F6D, 0x009F70, 0x009F75, 0x02A6B2,
	0
};

// JIS X 0212
static const unsigned map_kanji_hojo_1[] = {
	0x004E02, 0x004E04, 0x004E05, 0x004E0C, 0x004E12, 0x004E1F, 0x004E23, 0x004E24, 0x004E28, 0x004E2B, 0x004E2E, 0x004E2F, 0x004E30, 0x004E35, 0x004E40, 0x004E41, 0x004E44, 0x004E47, 0x004E51, 0x004E5A, 0x004E5C, 0x004E63, 0x004E68, 0x004E69, 0x004E74, 0x004E75, 0x004E79, 0x004E7F, 0x004E8D, 0x004E96, 0x004E97, 0x004E9D,
	0x004EAF, 0x004EB9, 0x004EC3, 0x004ED0, 0x004EDA, 0x004EDB, 0x004EE0, 0x004EE1, 0x004EE2, 0x004EE8, 0x004EEF, 0x004EF1, 0x004EF3, 0x004EF5, 0x004EFD, 0x004EFE, 0x004EFF, 0x004F00, 0x004F02, 0x004F03, 0x004F08, 0x004F0B, 0x004F0C, 0x004F12, 0x004F15, 0x004F16, 0x004F17, 0x004F19, 0x004F2E, 0x004F31, 0x004F60, 0x004F33,
	0x004F35, 0x004F37, 0x004F39, 0x004F3B, 0x004F3E, 0x004F40, 0x004F42, 0x004F48, 0x004F49, 0x004F4B, 0x004F4C, 0x004F52, 0x004F54, 0x004F56, 0x004F58, 0x004F5F, 0x004F63, 0x004F6A, 0x004F6C, 0x004F6E, 0x004F71, 0x004F77, 0x004F78, 0x004F79, 0x004F7A, 0x004F7D, 0x004F7E, 0x004F81, 0x004F82, 0x004F84, 0x004F85, 0x004F89,
	0x004F8A, 0x004F8C, 0x004F8E, 0x004F90, 0x004F92, 0x004F93, 0x004F94, 0x004F97, 0x004F99, 0x004F9A, 0x004F9E, 0x004F9F, 0x004FB2, 0x004FB7, 0x004FB9, 0x004FBB, 0x004FBC, 0x004FBD, 0x004FBE, 0x004FC0, 0x004FC1, 0x004FC5, 0x004FC6, 0x004FC8, 0x004FC9, 0x004FCB, 0x004FCC, 0x004FCD, 0x004FCF, 0x004FD2, 0x004FDC, 0x004FE0,
	0x004FE2, 0x004FF0, 0x004FF2, 0x004FFC, 0x004FFD, 0x004FFF, 0x005000, 0x005001, 0x005004, 0x005007, 0x00500A, 0x00500C, 0x00500E, 0x005010, 0x005013, 0x005017, 0x005018, 0x00501B, 0x00501C, 0x00501D, 0x00501E, 0x005022, 0x005027, 0x00502E, 0x005030, 0x005032, 0x005033, 0x005035, 0x005040, 0x005041, 0x005042, 0x005045,
	0x005046, 0x00504A, 0x00504C, 0x00504E, 0x005051, 0x005052, 0x005053, 0x005057, 0x005059, 0x00505F, 0x005060, 0x005062, 0x005063, 0x005066, 0x005067, 0x00506A, 0x00506D, 0x005070, 0x005071, 0x00503B, 0x005081, 0x005083, 0x005084, 0x005086, 0x00508A, 0x00508E, 0x00508F, 0x005090, 0x005092, 0x005093, 0x005094, 0x005096,
	0x00509B, 0x00509C, 0x00509E, 0x00509F, 0x0050A0, 0x0050A1, 0x0050A2, 0x0050AA, 0x0050AF, 0x0050B0, 0x0050B9, 0x0050BA, 0x0050BD, 0x0050C0, 0x0050C3, 0x0050C4, 0x0050C7, 0x0050CC, 0x0050CE, 0x0050D0, 0x0050D3, 0x0050D4, 0x0050D8, 0x0050DC, 0x0050DD, 0x0050DF, 0x0050E2, 0x0050E4, 0x0050E6, 0x0050E8, 0x0050E9, 0x0050EF,
	0x0050F1, 0x0050F6, 0x0050FA, 0x0050FE, 0x005103, 0x005106, 0x005107, 0x005108, 0x00510B, 0x00510C, 0x00510D, 0x00510E, 0x0050F2, 0x005110, 0x005117, 0x005119, 0x00511B, 0x00511C, 0x00511D, 0x00511E, 0x005123, 0x005127, 0x005128, 0x00512C, 0x00512D, 0x00512F, 0x005131, 0x005133, 0x005134, 0x005135, 0x005138, 0x005139,
	0x005142, 0x00514A, 0x00514F, 0x005153, 0x005155, 0x005157, 0x005158, 0x00515F, 0x005164, 0x005166, 0x00517E, 0x005183, 0x005184, 0x00518B, 0x00518E, 0x005198, 0x00519D, 0x0051A1, 0x0051A3, 0x0051AD, 0x0051B8, 0x0051BA, 0x0051BC, 0x0051BE, 0x0051BF, 0x0051C2, 0x0051C8, 0x0051CF, 0x0051D1, 0x0051D2, 0x0051D3, 0x0051D5,
	0x0051D8, 0x0051DE, 0x0051E2, 0x0051E5, 0x0051EE, 0x0051F2, 0x0051F3, 0x0051F4, 0x0051F7, 0x005201, 0x005202, 0x005205, 0x005212, 0x005213, 0x005215, 0x005216, 0x005218, 0x005222, 0x005228, 0x005231, 0x005232, 0x005235, 0x00523C, 0x005245, 0x005249, 0x005255, 0x005257, 0x005258, 0x00525A, 0x00525C, 0x00525F, 0x005260,
	0x005261, 0x005266, 0x00526E, 0x005277, 0x005278, 0x005279, 0x005280, 0x005282, 0x005285, 0x00528A, 0x00528C, 0x005293, 0x005295, 0x005296, 0x005297, 0x005298, 0x00529A, 0x00529C, 0x0052A4, 0x0052A5, 0x0052A6, 0x0052A7, 0x0052AF, 0x0052B0, 0x0052B6, 0x0052B7, 0x0052B8, 0x0052BA, 0x0052BB, 0x0052BD, 0x0052C0, 0x0052C4,
	0x0052C6, 0x0052C8, 0x0052CC, 0x0052CF, 0x0052D1, 0x0052D4, 0x0052D6, 0x0052DB, 0x0052DC, 0x0052E1, 0x0052E5, 0x0052E8, 0x0052E9, 0x0052EA, 0x0052EC, 0x0052F0, 0x0052F1, 0x0052F4, 0x0052F6, 0x0052F7, 0x005300, 0x005303, 0x00530A, 0x00530B, 0x00530C, 0x005311, 0x005313, 0x005318, 0x00531B, 0x00531C, 0x00531E, 0x00531F,
	0x005325, 0x005327, 0x005328, 0x005329, 0x00532B, 0x00532C, 0x00532D, 0x005330, 0x005332, 0x005335, 0x00533C, 0x00533D, 0x00533E, 0x005342, 0x00534C, 0x00534B, 0x005359, 0x00535B, 0x005361, 0x005363, 0x005365, 0x00536C, 0x00536D, 0x005372, 0x005379, 0x00537E, 0x005383, 0x005387, 0x005388, 0x00538E, 0x005393, 0x005394,
	0x005399, 0x00539D, 0x0053A1, 0x0053A4, 0x0053AA, 0x0053AB, 0x0053AF, 0x0053B2, 0x0053B4, 0x0053B5, 0x0053B7, 0x0053B8, 0x0053BA, 0x0053BD, 0x0053C0, 0x0053C5, 0x0053CF, 0x0053D2, 0x0053D3, 0x0053D5, 0x0053DA, 0x0053DD, 0x0053DE, 0x0053E0, 0x0053E6, 0x0053E7, 0x0053F5, 0x005402, 0x005413, 0x00541A, 0x005421, 0x005427,
	0x005428, 0x00542A, 0x00542F, 0x005431, 0x005434, 0x005435, 0x005443, 0x005444, 0x005447, 0x00544D, 0x00544F, 0x00545E, 0x005462, 0x005464, 0x005466, 0x005467, 0x005469, 0x00546B, 0x00546D, 0x00546E, 0x005474, 0x00547F, 0x005481, 0x005483, 0x005485, 0x005488, 0x005489, 0x00548D, 0x005491, 0x005495, 0x005496, 0x00549C,
	0x00549F, 0x0054A1, 0x0054A6, 0x0054A7, 0x0054A9, 0x0054AA, 0x0054AD, 0x0054AE, 0x0054B1, 0x0054B7, 0x0054B9, 0x0054BA, 0x0054BB, 0x0054BF, 0x0054C6, 0x0054CA, 0x0054CD, 0x0054CE, 0x0054E0, 0x0054EA, 0x0054EC, 0x0054EF, 0x0054F6, 0x0054FC, 0x0054FE, 0x0054FF, 0x005500, 0x005501, 0x005505, 0x005508, 0x005509, 0x00550C,
	0x00550D, 0x00550E, 0x005515, 0x00552A, 0x00552B, 0x005532, 0x005535, 0x005536, 0x00553B, 0x00553C, 0x00553D, 0x005541, 0x005547, 0x005549, 0x00554A, 0x00554D, 0x005550, 0x005551, 0x005558, 0x00555A, 0x00555B, 0x00555E, 0x005560, 0x005561, 0x005564, 0x005566, 0x00557F, 0x005581, 0x005582, 0x005586, 0x005588, 0x00558E,
	0x00558F, 0x005591, 0x005592, 0x005593, 0x005594, 0x005597, 0x0055A3, 0x0055A4, 0x0055AD, 0x0055B2, 0x0055BF, 0x0055C1, 0x0055C3, 0x0055C6, 0x0055C9, 0x0055CB, 0x0055CC, 0x0055CE, 0x0055D1, 0x0055D2, 0x0055D3, 0x0055D7, 0x0055D8, 0x0055DB, 0x0055DE, 0x0055E2, 0x0055E9, 0x0055F6, 0x0055FF, 0x005605, 0x005608, 0x00560A,
	0x00560D, 0x00560E, 0x00560F, 0x005610, 0x005611, 0x005612, 0x005619, 0x00562C, 0x005630, 0x005633, 0x005635, 0x005637, 0x005639, 0x00563B, 0x00563C, 0x00563D, 0x00563F, 0x005640, 0x005641, 0x005643, 0x005644, 0x005646, 0x005649, 0x00564B, 0x00564D, 0x00564F, 0x005654, 0x00565E, 0x005660, 0x005661, 0x005662, 0x005663,
	0x005666, 0x005669, 0x00566D, 0x00566F, 0x005671, 0x005672, 0x005675, 0x005684, 0x005685, 0x005688, 0x00568B, 0x00568C, 0x005695, 0x005699, 0x00569A, 0x00569D, 0x00569E, 0x00569F, 0x0056A6, 0x0056A7, 0x0056A8, 0x0056A9, 0x0056AB, 0x0056AC, 0x0056AD, 0x0056B1, 0x0056B3, 0x0056B7, 0x0056BE, 0x0056C5, 0x0056C9, 0x0056CA,
	0x0056CB, 0x0056CF, 0x0056D0, 0x0056CC, 0x0056CD, 0x0056D9, 0x0056DC, 0x0056DD, 0x0056DF, 0x0056E1, 0x0056E4, 0x0056E5, 0x0056E6, 0x0056E7, 0x0056E8, 0x0056F1, 0x0056EB, 0x0056ED, 0x0056F6, 0x0056F7, 0x005701, 0x005702, 0x005707, 0x00570A, 0x00570C, 0x005711, 0x005715, 0x00571A, 0x00571B, 0x00571D, 0x005720, 0x005722,
	0x005723, 0x005724, 0x005725, 0x005729, 0x00572A, 0x00572C, 0x00572E, 0x00572F, 0x005733, 0x005734, 0x00573D, 0x00573E, 0x00573F, 0x005745, 0x005746, 0x00574C, 0x00574D, 0x005752, 0x005762, 0x005765, 0x005767, 0x005768, 0x00576B, 0x00576D, 0x00576E, 0x00576F, 0x005770, 0x005771, 0x005773, 0x005774, 0x005775, 0x005777,
	0x005779, 0x00577A, 0x00577B, 0x00577C, 0x00577E, 0x005781, 0x005783, 0x00578C, 0x005794, 0x005797, 0x005799, 0x00579A, 0x00579C, 0x00579D, 0x00579E, 0x00579F, 0x0057A1, 0x005795, 0x0057A7, 0x0057A8, 0x0057A9, 0x0057AC, 0x0057B8, 0x0057BD, 0x0057C7, 0x0057C8, 0x0057CC, 0x0057CF, 0x0057D5, 0x0057DD, 0x0057DE, 0x0057E4,
	0x0057E6, 0x0057E7, 0x0057E9, 0x0057ED, 0x0057F0, 0x0057F5, 0x0057F6, 0x0057F8, 0x0057FD, 0x0057FE, 0x0057FF, 0x005803, 0x005804, 0x005808, 0x005809, 0x0057E1, 0x00580C, 0x00580D, 0x00581B, 0x00581E, 0x00581F, 0x005820, 0x005826, 0x005827, 0x00582D, 0x005832, 0x005839, 0x00583F, 0x005849, 0x00584C, 0x00584D, 0x00584F,
	0x005850, 0x005855, 0x00585F, 0x005861, 0x005864, 0x005867, 0x005868, 0x005878, 0x00587C, 0x00587F, 0x005880, 0x005881, 0x005887, 0x005888, 0x005889, 0x00588A, 0x00588C, 0x00588D, 0x00588F, 0x005890, 0x005894, 0x005896, 0x00589D, 0x0058A0, 0x0058A1, 0x0058A2, 0x0058A6, 0x0058A9, 0x0058B1, 0x0058B2, 0x0058C4, 0x0058BC,
	0x0058C2, 0x0058C8, 0x0058CD, 0x0058CE, 0x0058D0, 0x0058D2, 0x0058D4, 0x0058D6, 0x0058DA, 0x0058DD, 0x0058E1, 0x0058E2, 0x0058E9, 0x0058F3, 0x005905, 0x005906, 0x00590B, 0x00590C, 0x005912, 0x005913, 0x005914, 0x008641, 0x00591D, 0x005921, 0x005923, 0x005924, 0x005928, 0x00592F, 0x005930, 0x005933, 0x005935, 0x005936,
	0x00593F, 0x005943, 0x005946, 0x005952, 0x005953, 0x005959, 0x00595B, 0x00595D, 0x00595E, 0x00595F, 0x005961, 0x005963, 0x00596B, 0x00596D, 0x00596F, 0x005972, 0x005975, 0x005976, 0x005979, 0x00597B, 0x00597C, 0x00598B, 0x00598C, 0x00598E, 0x005992, 0x005995, 0x005997, 0x00599F, 0x0059A4, 0x0059A7, 0x0059AD, 0x0059AE,
	0x0059AF, 0x0059B0, 0x0059B3, 0x0059B7, 0x0059BA, 0x0059BC, 0x0059C1, 0x0059C3, 0x0059C4, 0x0059C8, 0x0059CA, 0x0059CD, 0x0059D2, 0x0059DD, 0x0059DE, 0x0059DF, 0x0059E3, 0x0059E4, 0x0059E7, 0x0059EE, 0x0059EF, 0x0059F1, 0x0059F2, 0x0059F4, 0x0059F7, 0x005A00, 0x005A04, 0x005A0C, 0x005A0D, 0x005A0E, 0x005A12, 0x005A13,
	0x005A1E, 0x005A23, 0x005A24, 0x005A27, 0x005A28, 0x005A2A, 0x005A2D, 0x005A30, 0x005A44, 0x005A45, 0x005A47, 0x005A48, 0x005A4C, 0x005A50, 0x005A55, 0x005A5E, 0x005A63, 0x005A65, 0x005A67, 0x005A6D, 0x005A77, 0x005A7A, 0x005A7B, 0x005A7E, 0x005A8B, 0x005A90, 0x005A93, 0x005A96, 0x005A99, 0x005A9C, 0x005A9E, 0x005A9F,
	0x005AA0, 0x005AA2, 0x005AA7, 0x005AAC, 0x005AB1, 0x005AB2, 0x005AB3, 0x005AB5, 0x005AB8, 0x005ABA, 0x005ABB, 0x005ABF, 0x005AC4, 0x005AC6, 0x005AC8, 0x005ACF, 0x005ADA, 0x005ADC, 0x005AE0, 0x005AE5, 0x005AEA, 0x005AEE, 0x005AF5, 0x005AF6, 0x005AFD, 0x005B00, 0x005B01, 0x005B08, 0x005B17, 0x005B34, 0x005B19, 0x005B1B,
	0x005B1D, 0x005B21, 0x005B25, 0x005B2D, 0x005B38, 0x005B41, 0x005B4B, 0x005B4C, 0x005B52, 0x005B56, 0x005B5E, 0x005B68, 0x005B6E, 0x005B6F, 0x005B7C, 0x005B7D, 0x005B7E, 0x005B7F, 0x005B81, 0x005B84, 0x005B86, 0x005B8A, 0x005B8E, 0x005B90, 0x005B91, 0x005B93, 0x005B94, 0x005B96, 0x005BA8, 0x005BA9, 0x005BAC, 0x005BAD,
	0x005BAF, 0x005BB1, 0x005BB2, 0x005BB7, 0x005BBA, 0x005BBC, 0x005BC0, 0x005BC1, 0x005BCD, 0x005BCF, 0x005BD6, 0x005BD7, 0x005BD8, 0x005BD9, 0x005BDA, 0x005BE0, 0x005BEF, 0x005BF1, 0x005BF4, 0x005BFD, 0x005C0C, 0x005C17, 0x005C1E, 0x005C1F, 0x005C23, 0x005C26, 0x005C29, 0x005C2B, 0x005C2C, 0x005C2E, 0x005C30, 0x005C32,
	0
};

static const unsigned map_kanji_hojo_2[] = {
	0x005C35, 0x005C36, 0x005C59, 0x005C5A, 0x005C5C, 0x005C62, 0x005C63, 0x005C67, 0x005C68, 0x005C69, 0x005C6D, 0x005C70, 0x005C74, 0x005C75, 0x005C7A, 0x005C7B, 0x005C7C, 0x005C7D, 0x005C87, 0x005C88, 0x005C8A, 0x005C8F, 0x005C92, 0x005C9D, 0x005C9F, 0x005CA0, 0x005CA2, 0x005CA3, 0x005CA6, 0x005CAA, 0x005CB2, 0x005CB4,
	0x005CB5, 0x005CBA, 0x005CC9, 0x005CCB, 0x005CD2, 0x005CDD, 0x005CD7, 0x005CEE, 0x005CF1, 0x005CF2, 0x005CF4, 0x005D01, 0x005D06, 0x005D0D, 0x005D12, 0x005D2B, 0x005D23, 0x005D24, 0x005D26, 0x005D27, 0x005D31, 0x005D34, 0x005D39, 0x005D3D, 0x005D3F, 0x005D42, 0x005D43, 0x005D46, 0x005D48, 0x005D55, 0x005D51, 0x005D59,
	0x005D4A, 0x005D5F, 0x005D60, 0x005D61, 0x005D62, 0x005D64, 0x005D6A, 0x005D6D, 0x005D70, 0x005D79, 0x005D7A, 0x005D7E, 0x005D7F, 0x005D81, 0x005D83, 0x005D88, 0x005D8A, 0x005D92, 0x005D93, 0x005D94, 0x005D95, 0x005D99, 0x005D9B, 0x005D9F, 0x005DA0, 0x005DA7, 0x005DAB, 0x005DB0, 0x005DB4, 0x005DB8, 0x005DB9, 0x005DC3,
	0x005DC7, 0x005DCB, 0x005DD0, 0x005DCE, 0x005DD8, 0x005DD9, 0x005DE0, 0x005DE4, 0x005DE9, 0x005DF8, 0x005DF9, 0x005E00, 0x005E07, 0x005E0D, 0x005E12, 0x005E14, 0x005E15, 0x005E18, 0x005E1F, 0x005E20, 0x005E2E, 0x005E28, 0x005E32, 0x005E35, 0x005E3E, 0x005E4B, 0x005E50, 0x005E49, 0x005E51, 0x005E56, 0x005E58, 0x005E5B,
	0x005E5C, 0x005E5E, 0x005E68, 0x005E6A, 0x005E6B, 0x005E6C, 0x005E6D, 0x005E6E, 0x005E70, 0x005E80, 0x005E8B, 0x005E8E, 0x005EA2, 0x005EA4, 0x005EA5, 0x005EA8, 0x005EAA, 0x005EAC, 0x005EB1, 0x005EB3, 0x005EBD, 0x005EBE, 0x005EBF, 0x005EC6, 0x005ECC, 0x005ECB, 0x005ECE, 0x005ED1, 0x005ED2, 0x005ED4, 0x005ED5, 0x005EDC,
	0x005EDE, 0x005EE5, 0x005EEB, 0x005F02, 0x005F06, 0x005F07, 0x005F08, 0x005F0E, 0x005F19, 0x005F1C, 0x005F1D, 0x005F21, 0x005F22, 0x005F23, 0x005F24, 0x005F28, 0x005F2B, 0x005F2C, 0x005F2E, 0x005F30, 0x005F34, 0x005F36, 0x005F3B, 0x005F3D, 0x005F3F, 0x005F40, 0x005F44, 0x005F45, 0x005F47, 0x005F4D, 0x005F50, 0x005F54,
	0x005F58, 0x005F5B, 0x005F60, 0x005F63, 0x005F64, 0x005F67, 0x005F6F, 0x005F72, 0x005F74, 0x005F75, 0x005F78, 0x005F7A, 0x005F7D, 0x005F7E, 0x005F89, 0x005F8D, 0x005F8F, 0x005F96, 0x005F9C, 0x005F9D, 0x005FA2, 0x005FA7, 0x005FAB, 0x005FA4, 0x005FAC, 0x005FAF, 0x005FB0, 0x005FB1, 0x005FB8, 0x005FC4, 0x005FC7, 0x005FC8,
	0x005FC9, 0x005FCB, 0x005FD0, 0x005FD1, 0x005FD2, 0x005FD3, 0x005FD4, 0x005FDE, 0x005FE1, 0x005FE2, 0x005FE8, 0x005FE9, 0x005FEA, 0x005FEC, 0x005FED, 0x005FEE, 0x005FEF, 0x005FF2, 0x005FF3, 0x005FF6, 0x005FFA, 0x005FFC, 0x006007, 0x00600A, 0x00600D, 0x006013, 0x006014, 0x006017, 0x006018, 0x00601A, 0x00601F, 0x006024,
	0x00602D, 0x006033, 0x006035, 0x006040, 0x006047, 0x006048, 0x006049, 0x00604C, 0x006051, 0x006054, 0x006056, 0x006057, 0x00605D, 0x006061, 0x006067, 0x006071, 0x00607E, 0x00607F, 0x006082, 0x006086, 0x006088, 0x00608A, 0x00608E, 0x006091, 0x006093, 0x006095, 0x006098, 0x00609D, 0x00609E, 0x0060A2, 0x0060A4, 0x0060A5,
	0x0060A8, 0x0060B0, 0x0060B1, 0x0060B7, 0x0060BB, 0x0060BE, 0x0060C2, 0x0060C4, 0x0060C8, 0x0060C9, 0x0060CA, 0x0060CB, 0x0060CE, 0x0060CF, 0x0060D4, 0x0060D5, 0x0060D9, 0x0060DB, 0x0060DD, 0x0060DE, 0x0060E2, 0x0060E5, 0x0060F2, 0x0060F5, 0x0060F8, 0x0060FC, 0x0060FD, 0x006102, 0x006107, 0x00610A, 0x00610C, 0x006110,
	0x006111, 0x006112, 0x006113, 0x006114, 0x006116, 0x006117, 0x006119, 0x00611C, 0x00611E, 0x006122, 0x00612A, 0x00612B, 0x006130, 0x006131, 0x006135, 0x006136, 0x006137, 0x006139, 0x006141, 0x006145, 0x006146, 0x006149, 0x00615E, 0x006160, 0x00616C, 0x006172, 0x006178, 0x00617B, 0x00617C, 0x00617F, 0x006180, 0x006181,
	0x006183, 0x006184, 0x00618B, 0x00618D, 0x006192, 0x006193, 0x006197, 0x006198, 0x00619C, 0x00619D, 0x00619F, 0x0061A0, 0x0061A5, 0x0061A8, 0x0061AA, 0x0061AD, 0x0061B8, 0x0061B9, 0x0061BC, 0x0061C0, 0x0061C1, 0x0061C2, 0x0061CE, 0x0061CF, 0x0061D5, 0x0061DC, 0x0061DD, 0x0061DE, 0x0061DF, 0x0061E1, 0x0061E2, 0x0061E7,
	0x0061E9, 0x0061E5, 0x0061EC, 0x0061ED, 0x0061EF, 0x006201, 0x006203, 0x006204, 0x006207, 0x006213, 0x006215, 0x00621C, 0x006220, 0x006222, 0x006223, 0x006227, 0x006229, 0x00622B, 0x006239, 0x00623D, 0x006242, 0x006243, 0x006244, 0x006246, 0x00624C, 0x006250, 0x006251, 0x006252, 0x006254, 0x006256, 0x00625A, 0x00625C,
	0x006264, 0x00626D, 0x00626F, 0x006273, 0x00627A, 0x00627D, 0x00628D, 0x00628E, 0x00628F, 0x006290, 0x0062A6, 0x0062A8, 0x0062B3, 0x0062B6, 0x0062B7, 0x0062BA, 0x0062BE, 0x0062BF, 0x0062C4, 0x0062CE, 0x0062D5, 0x0062D6, 0x0062DA, 0x0062EA, 0x0062F2, 0x0062F4, 0x0062FC, 0x0062FD, 0x006303, 0x006304, 0x00630A, 0x00630B,
	0x00630D, 0x006310, 0x006313, 0x006316, 0x006318, 0x006329, 0x00632A, 0x00632D, 0x006335, 0x006336, 0x006339, 0x00633C, 0x006341, 0x006342, 0x006343, 0x006344, 0x006346, 0x00634A, 0x00634B, 0x00634E, 0x006352, 0x006353, 0x006354, 0x006358, 0x00635B, 0x006365, 0x006366, 0x00636C, 0x00636D, 0x006371, 0x006374, 0x006375,
	0x006378, 0x00637C, 0x00637D, 0x00637F, 0x006382, 0x006384, 0x006387, 0x00638A, 0x006390, 0x006394, 0x006395, 0x006399, 0x00639A, 0x00639E, 0x0063A4, 0x0063A6, 0x0063AD, 0x0063AE, 0x0063AF, 0x0063BD, 0x0063C1, 0x0063C5, 0x0063C8, 0x0063CE, 0x0063D1, 0x0063D3, 0x0063D4, 0x0063D5, 0x0063DC, 0x0063E0, 0x0063E5, 0x0063EA,
	0x0063EC, 0x0063F2, 0x0063F3, 0x0063F5, 0x0063F8, 0x0063F9, 0x006409, 0x00640A, 0x006410, 0x006412, 0x006414, 0x006418, 0x00641E, 0x006420, 0x006422, 0x006424, 0x006425, 0x006429, 0x00642A, 0x00642F, 0x006430, 0x006435, 0x00643D, 0x00643F, 0x00644B, 0x00644F, 0x006451, 0x006452, 0x006453, 0x006454, 0x00645A, 0x00645B,
	0x00645C, 0x00645D, 0x00645F, 0x006460, 0x006461, 0x006463, 0x00646D, 0x006473, 0x006474, 0x00647B, 0x00647D, 0x006485, 0x006487, 0x00648F, 0x006490, 0x006491, 0x006498, 0x006499, 0x00649B, 0x00649D, 0x00649F, 0x0064A1, 0x0064A3, 0x0064A6, 0x0064A8, 0x0064AC, 0x0064B3, 0x0064BD, 0x0064BE, 0x0064BF, 0x0064C4, 0x0064C9,
	0x0064CA, 0x0064CB, 0x0064CC, 0x0064CE, 0x0064D0, 0x0064D1, 0x0064D5, 0x0064D7, 0x0064E4, 0x0064E5, 0x0064E9, 0x0064EA, 0x0064ED, 0x0064F0, 0x0064F5, 0x0064F7, 0x0064FB, 0x0064FF, 0x006501, 0x006504, 0x006508, 0x006509, 0x00650A, 0x00650F, 0x006513, 0x006514, 0x006516, 0x006519, 0x00651B, 0x00651E, 0x00651F, 0x006522,
	0x006526, 0x006529, 0x00652E, 0x006531, 0x00653A, 0x00653C, 0x00653D, 0x006543, 0x006547, 0x006549, 0x006550, 0x006552, 0x006554, 0x00655F, 0x006560, 0x006567, 0x00656B, 0x00657A, 0x00657D, 0x006581, 0x006585, 0x00658A, 0x006592, 0x006595, 0x006598, 0x00659D, 0x0065A0, 0x0065A3, 0x0065A6, 0x0065AE, 0x0065B2, 0x0065B3,
	0x0065B4, 0x0065BF, 0x0065C2, 0x0065C8, 0x0065C9, 0x0065CE, 0x0065D0, 0x0065D4, 0x0065D6, 0x0065D8, 0x0065DF, 0x0065F0, 0x0065F2, 0x0065F4, 0x0065F5, 0x0065F9, 0x0065FE, 0x0065FF, 0x006600, 0x006604, 0x006608, 0x006609, 0x00660D, 0x006611, 0x006612, 0x006615, 0x006616, 0x00661D, 0x00661E, 0x006621, 0x006622, 0x006623,
	0x006624, 0x006626, 0x006629, 0x00662A, 0x00662B, 0x00662C, 0x00662E, 0x006630, 0x006631, 0x006633, 0x006639, 0x006637, 0x006640, 0x006645, 0x006646, 0x00664A, 0x00664C, 0x006651, 0x00664E, 0x006657, 0x006658, 0x006659, 0x00665B, 0x00665C, 0x006660, 0x006661, 0x0066FB, 0x00666A, 0x00666B, 0x00666C, 0x00667E, 0x006673,
	0x006675, 0x00667F, 0x006677, 0x006678, 0x006679, 0x00667B, 0x006680, 0x00667C, 0x00668B, 0x00668C, 0x00668D, 0x006690, 0x006692, 0x006699, 0x00669A, 0x00669B, 0x00669C, 0x00669F, 0x0066A0, 0x0066A4, 0x0066AD, 0x0066B1, 0x0066B2, 0x0066B5, 0x0066BB, 0x0066BF, 0x0066C0, 0x0066C2, 0x0066C3, 0x0066C8, 0x0066CC, 0x0066CE,
	0x0066CF, 0x0066D4, 0x0066DB, 0x0066DF, 0x0066E8, 0x0066EB, 0x0066EC, 0x0066EE, 0x0066FA, 0x006705, 0x006707, 0x00670E, 0x006713, 0x006719, 0x00671C, 0x006720, 0x006722, 0x006733, 0x00673E, 0x006745, 0x006747, 0x006748, 0x00674C, 0x006754, 0x006755, 0x00675D, 0x006766, 0x00676C, 0x00676E, 0x006774, 0x006776, 0x00677B,
	0x006781, 0x006784, 0x00678E, 0x00678F, 0x006791, 0x006793, 0x006796, 0x006798, 0x006799, 0x00679B, 0x0067B0, 0x0067B1, 0x0067B2, 0x0067B5, 0x0067BB, 0x0067BC, 0x0067BD, 0x0067F9, 0x0067C0, 0x0067C2, 0x0067C3, 0x0067C5, 0x0067C8, 0x0067C9, 0x0067D2, 0x0067D7, 0x0067D9, 0x0067DC, 0x0067E1, 0x0067E6, 0x0067F0, 0x0067F2,
	0x0067F6, 0x0067F7, 0x006852, 0x006814, 0x006819, 0x00681D, 0x00681F, 0x006828, 0x006827, 0x00682C, 0x00682D, 0x00682F, 0x006830, 0x006831, 0x006833, 0x00683B, 0x00683F, 0x006844, 0x006845, 0x00684A, 0x00684C, 0x006855, 0x006857, 0x006858, 0x00685B, 0x00686B, 0x00686E, 0x00686F, 0x006870, 0x006871, 0x006872, 0x006875,
	0x006879, 0x00687A, 0x00687B, 0x00687C, 0x006882, 0x006884, 0x006886, 0x006888, 0x006896, 0x006898, 0x00689A, 0x00689C, 0x0068A1, 0x0068A3, 0x0068A5, 0x0068A9, 0x0068AA, 0x0068AE, 0x0068B2, 0x0068BB, 0x0068C5, 0x0068C8, 0x0068CC, 0x0068CF, 0x0068D0, 0x0068D1, 0x0068D3, 0x0068D6, 0x0068D9, 0x0068DC, 0x0068DD, 0x0068E5,
	0x0068E8, 0x0068EA, 0x0068EB, 0x0068EC, 0x0068ED, 0x0068F0, 0x0068F1, 0x0068F5, 0x0068F6, 0x0068FB, 0x0068FC, 0x0068FD, 0x006906, 0x006909, 0x00690A, 0x006910, 0x006911, 0x006913, 0x006916, 0x006917, 0x006931, 0x006933, 0x006935, 0x006938, 0x00693B, 0x006942, 0x006945, 0x006949, 0x00694E, 0x006957, 0x00695B, 0x006963,
	0x006964, 0x006965, 0x006966, 0x006968, 0x006969, 0x00696C, 0x006970, 0x006971, 0x006972, 0x00697A, 0x00697B, 0x00697F, 0x006980, 0x00698D, 0x006992, 0x006996, 0x006998, 0x0069A1, 0x0069A5, 0x0069A6, 0x0069A8, 0x0069AB, 0x0069AD, 0x0069AF, 0x0069B7, 0x0069B8, 0x0069BA, 0x0069BC, 0x0069C5, 0x0069C8, 0x0069D1, 0x0069D6,
	0x0069D7, 0x0069E2, 0x0069E5, 0x0069EE, 0x0069EF, 0x0069F1, 0x0069F3, 0x0069F5, 0x0069FE, 0x006A00, 0x006A01, 0x006A03, 0x006A0F, 0x006A11, 0x006A15, 0x006A1A, 0x006A1D, 0x006A20, 0x006A24, 0x006A28, 0x006A30, 0x006A32, 0x006A34, 0x006A37, 0x006A3B, 0x006A3E, 0x006A3F, 0x006A45, 0x006A46, 0x006A49, 0x006A4A, 0x006A4E,
	0x006A50, 0x006A51, 0x006A52, 0x006A55, 0x006A56, 0x006A5B, 0x006A64, 0x006A67, 0x006A6A, 0x006A71, 0x006A73, 0x006A7E, 0x006A81, 0x006A83, 0x006A86, 0x006A87, 0x006A89, 0x006A8B, 0x006A91, 0x006A9B, 0x006A9D, 0x006A9E, 0x006A9F, 0x006AA5, 0x006AAB, 0x006AAF, 0x006AB0, 0x006AB1, 0x006AB4, 0x006ABD, 0x006ABE, 0x006ABF,
	0x006AC6, 0x006AC9, 0x006AC8, 0x006ACC, 0x006AD0, 0x006AD4, 0x006AD5, 0x006AD6, 0x006ADC, 0x006ADD, 0x006AE4, 0x006AE7, 0x006AEC, 0x006AF0, 0x006AF1, 0x006AF2, 0x006AFC, 0x006AFD, 0x006B02, 0x006B03, 0x006B06, 0x006B07, 0x006B09, 0x006B0F, 0x006B10, 0x006B11, 0x006B17, 0x006B1B, 0x006B1E, 0x006B24, 0x006B28, 0x006B2B,
	0
};

static const unsigned map_kanji_hojo_3[] = {
	0x006B2C, 0x006B2F, 0x006B35, 0x006B36, 0x006B3B, 0x006B3F, 0x006B46, 0x006B4A, 0x006B4D, 0x006B52, 0x006B56, 0x006B58, 0x006B5D, 0x006B60, 0x006B67, 0x006B6B, 0x006B6E, 0x006B70, 0x006B75, 0x006B7D, 0x006B7E, 0x006B82, 0x006B85, 0x006B97, 0x006B9B, 0x006B9F, 0x006BA0, 0x006BA3, 0x006BA8, 0x006BA9, 0x006BAC, 0x006BAD,
	0x006BAE, 0x006BB0, 0x006BB8, 0x006BB9, 0x006BBD, 0x006BBE, 0x006BC3, 0x006BC4, 0x006BC9, 0x006BCC, 0x006BD6, 0x006BDA, 0x006BE1, 0x006BE3, 0x006BE6, 0x006BE7, 0x006BEE, 0x006BF1, 0x006BF7, 0x006BF9, 0x006BFF, 0x006C02, 0x006C04, 0x006C05, 0x006C09, 0x006C0D, 0x006C0E, 0x006C10, 0x006C12, 0x006C19, 0x006C26, 0x006C27,
	0x006C28, 0x006C2C, 0x006C2E, 0x006C33, 0x006C35, 0x006C36, 0x006C3A, 0x006C3B, 0x006C3F, 0x006C4A, 0x006C4B, 0x006C4D, 0x006C4F, 0x006C52, 0x006C54, 0x006C59, 0x006C5B, 0x006C5C, 0x006C6B, 0x006C6D, 0x006C6F, 0x006C74, 0x006C76, 0x006C78, 0x006C79, 0x006C7B, 0x006C85, 0x006C86, 0x006C87, 0x006C89, 0x006C94, 0x006C95,
	0x006C97, 0x006C98, 0x006C9C, 0x006C9F, 0x006CB0, 0x006CB2, 0x006CB4, 0x006CC2, 0x006CC6, 0x006CCD, 0x006CCF, 0x006CD0, 0x006CD1, 0x006CD2, 0x006CD4, 0x006CD6, 0x006CDA, 0x006CDC, 0x006CE0, 0x006CE7, 0x006CE9, 0x006CEB, 0x006CEC, 0x006CEE, 0x006CF2, 0x006CF4, 0x006D04, 0x006D07, 0x006D0A, 0x006D0E, 0x006D0F, 0x006D11,
	0x006D13, 0x006D1A, 0x006D26, 0x006D27, 0x006D28, 0x006C67, 0x006D2E, 0x006D2F, 0x006D31, 0x006D39, 0x006D3C, 0x006D3F, 0x006D57, 0x006D5E, 0x006D5F, 0x006D61, 0x006D65, 0x006D67, 0x006D6F, 0x006D70, 0x006D7C, 0x006D82, 0x006D87, 0x006D91, 0x006D92, 0x006D94, 0x006D96, 0x006D97, 0x006D98, 0x006DAA, 0x006DAC, 0x006DB4,
	0x006DB7, 0x006DB9, 0x006DBD, 0x006DBF, 0x006DC4, 0x006DC8, 0x006DCA, 0x006DCE, 0x006DCF, 0x006DD6, 0x006DDB, 0x006DDD, 0x006DDF, 0x006DE0, 0x006DE2, 0x006DE5, 0x006DE9, 0x006DEF, 0x006DF0, 0x006DF4, 0x006DF6, 0x006DFC, 0x006E00, 0x006E04, 0x006E1E, 0x006E22, 0x006E27, 0x006E32, 0x006E36, 0x006E39, 0x006E3B, 0x006E3C,
	0x006E44, 0x006E45, 0x006E48, 0x006E49, 0x006E4B, 0x006E4F, 0x006E51, 0x006E52, 0x006E53, 0x006E54, 0x006E57, 0x006E5C, 0x006E5D, 0x006E5E, 0x006E62, 0x006E63, 0x006E68, 0x006E73, 0x006E7B, 0x006E7D, 0x006E8D, 0x006E93, 0x006E99, 0x006EA0, 0x006EA7, 0x006EAD, 0x006EAE, 0x006EB1, 0x006EB3, 0x006EBB, 0x006EBF, 0x006EC0,
	0x006EC1, 0x006EC3, 0x006EC7, 0x006EC8, 0x006ECA, 0x006ECD, 0x006ECE, 0x006ECF, 0x006EEB, 0x006EED, 0x006EEE, 0x006EF9, 0x006EFB, 0x006EFD, 0x006F04, 0x006F08, 0x006F0A, 0x006F0C, 0x006F0D, 0x006F16, 0x006F18, 0x006F1A, 0x006F1B, 0x006F26, 0x006F29, 0x006F2A, 0x006F2F, 0x006F30, 0x006F33, 0x006F36, 0x006F3B, 0x006F3C,
	0x006F2D, 0x006F4F, 0x006F51, 0x006F52, 0x006F53, 0x006F57, 0x006F59, 0x006F5A, 0x006F5D, 0x006F5E, 0x006F61, 0x006F62, 0x006F68, 0x006F6C, 0x006F7D, 0x006F7E, 0x006F83, 0x006F87, 0x006F88, 0x006F8B, 0x006F8C, 0x006F8D, 0x006F90, 0x006F92, 0x006F93, 0x006F94, 0x006F96, 0x006F9A, 0x006F9F, 0x006FA0, 0x006FA5, 0x006FA6,
	0x006FA7, 0x006FA8, 0x006FAE, 0x006FAF, 0x006FB0, 0x006FB5, 0x006FB6, 0x006FBC, 0x006FC5, 0x006FC7, 0x006FC8, 0x006FCA, 0x006FDA, 0x006FDE, 0x006FE8, 0x006FE9, 0x006FF0, 0x006FF5, 0x006FF9, 0x006FFC, 0x006FFD, 0x007000, 0x007005, 0x007006, 0x007007, 0x00700D, 0x007017, 0x007020, 0x007023, 0x00702F, 0x007034, 0x007037,
	0x007039, 0x00703C, 0x007043, 0x007044, 0x007048, 0x007049, 0x00704A, 0x00704B, 0x007054, 0x007055, 0x00705D, 0x00705E, 0x00704E, 0x007064, 0x007065, 0x00706C, 0x00706E, 0x007075, 0x007076, 0x00707E, 0x007081, 0x007085, 0x007086, 0x007094, 0x007095, 0x007096, 0x007097, 0x007098, 0x00709B, 0x0070A4, 0x0070AB, 0x0070B0,
	0x0070B1, 0x0070B4, 0x0070B7, 0x0070CA, 0x0070D1, 0x0070D3, 0x0070D4, 0x0070D5, 0x0070D6, 0x0070D8, 0x0070DC, 0x0070E4, 0x0070FA, 0x007103, 0x007104, 0x007105, 0x007106, 0x007107, 0x00710B, 0x00710C, 0x00710F, 0x00711E, 0x007120, 0x00712B, 0x00712D, 0x00712F, 0x007130, 0x007131, 0x007138, 0x007141, 0x007145, 0x007146,
	0x007147, 0x00714A, 0x00714B, 0x007150, 0x007152, 0x007157, 0x00715A, 0x00715C, 0x00715E, 0x007160, 0x007168, 0x007179, 0x007180, 0x007185, 0x007187, 0x00718C, 0x007192, 0x00719A, 0x00719B, 0x0071A0, 0x0071A2, 0x0071AF, 0x0071B0, 0x0071B2, 0x0071B3, 0x0071BA, 0x0071BF, 0x0071C0, 0x0071C1, 0x0071C4, 0x0071CB, 0x0071CC,
	0x0071D3, 0x0071D6, 0x0071D9, 0x0071DA, 0x0071DC, 0x0071F8, 0x0071FE, 0x007200, 0x007207, 0x007208, 0x007209, 0x007213, 0x007217, 0x00721A, 0x00721D, 0x00721F, 0x007224, 0x00722B, 0x00722F, 0x007234, 0x007238, 0x007239, 0x007241, 0x007242, 0x007243, 0x007245, 0x00724E, 0x00724F, 0x007250, 0x007253, 0x007255, 0x007256,
	0x00725A, 0x00725C, 0x00725E, 0x007260, 0x007263, 0x007268, 0x00726B, 0x00726E, 0x00726F, 0x007271, 0x007277, 0x007278, 0x00727B, 0x00727C, 0x00727F, 0x007284, 0x007289, 0x00728D, 0x00728E, 0x007293, 0x00729B, 0x0072A8, 0x0072AD, 0x0072AE, 0x0072B1, 0x0072B4, 0x0072BE, 0x0072C1, 0x0072C7, 0x0072C9, 0x0072CC, 0x0072D5,
	0x0072D6, 0x0072D8, 0x0072DF, 0x0072E5, 0x0072F3, 0x0072F4, 0x0072FA, 0x0072FB, 0x0072FE, 0x007302, 0x007304, 0x007305, 0x007307, 0x00730B, 0x00730D, 0x007312, 0x007313, 0x007318, 0x007319, 0x00731E, 0x007322, 0x007324, 0x007327, 0x007328, 0x00732C, 0x007331, 0x007332, 0x007335, 0x00733A, 0x00733B, 0x00733D, 0x007343,
	0x00734D, 0x007350, 0x007352, 0x007356, 0x007358, 0x00735D, 0x00735E, 0x00735F, 0x007360, 0x007366, 0x007367, 0x007369, 0x00736B, 0x00736C, 0x00736E, 0x00736F, 0x007371, 0x007377, 0x007379, 0x00737C, 0x007380, 0x007381, 0x007383, 0x007385, 0x007386, 0x00738E, 0x007390, 0x007393, 0x007395, 0x007397, 0x007398, 0x00739C,
	0x00739E, 0x00739F, 0x0073A0, 0x0073A2, 0x0073A5, 0x0073A6, 0x0073AA, 0x0073AB, 0x0073AD, 0x0073B5, 0x0073B7, 0x0073B9, 0x0073BC, 0x0073BD, 0x0073BF, 0x0073C5, 0x0073C6, 0x0073C9, 0x0073CB, 0x0073CC, 0x0073CF, 0x0073D2, 0x0073D3, 0x0073D6, 0x0073D9, 0x0073DD, 0x0073E1, 0x0073E3, 0x0073E6, 0x0073E7, 0x0073E9, 0x0073F4,
	0x0073F5, 0x0073F7, 0x0073F9, 0x0073FA, 0x0073FB, 0x0073FD, 0x0073FF, 0x007400, 0x007401, 0x007404, 0x007407, 0x00740A, 0x007411, 0x00741A, 0x00741B, 0x007424, 0x007426, 0x007428, 0x007429, 0x00742A, 0x00742B, 0x00742C, 0x00742D, 0x00742E, 0x00742F, 0x007430, 0x007431, 0x007439, 0x007440, 0x007443, 0x007444, 0x007446,
	0x007447, 0x00744B, 0x00744D, 0x007451, 0x007452, 0x007457, 0x00745D, 0x007462, 0x007466, 0x007467, 0x007468, 0x00746B, 0x00746D, 0x00746E, 0x007471, 0x007472, 0x007480, 0x007481, 0x007485, 0x007486, 0x007487, 0x007489, 0x00748F, 0x007490, 0x007491, 0x007492, 0x007498, 0x007499, 0x00749A, 0x00749C, 0x00749F, 0x0074A0,
	0x0074A1, 0x0074A3, 0x0074A6, 0x0074A8, 0x0074A9, 0x0074AA, 0x0074AB, 0x0074AE, 0x0074AF, 0x0074B1, 0x0074B2, 0x0074B5, 0x0074B9, 0x0074BB, 0x0074BF, 0x0074C8, 0x0074C9, 0x0074CC, 0x0074D0, 0x0074D3, 0x0074D8, 0x0074DA, 0x0074DB, 0x0074DE, 0x0074DF, 0x0074E4, 0x0074E8, 0x0074EA, 0x0074EB, 0x0074EF, 0x0074F4, 0x0074FA,
	0x0074FB, 0x0074FC, 0x0074FF, 0x007506, 0x007512, 0x007516, 0x007517, 0x007520, 0x007521, 0x007524, 0x007527, 0x007529, 0x00752A, 0x00752F, 0x007536, 0x007539, 0x00753D, 0x00753E, 0x00753F, 0x007540, 0x007543, 0x007547, 0x007548, 0x00754E, 0x007550, 0x007552, 0x007557, 0x00755E, 0x00755F, 0x007561, 0x00756F, 0x007571,
	0x007579, 0x00757A, 0x00757B, 0x00757C, 0x00757D, 0x00757E, 0x007581, 0x007585, 0x007590, 0x007592, 0x007593, 0x007595, 0x007599, 0x00759C, 0x0075A2, 0x0075A4, 0x0075B4, 0x0075BA, 0x0075BF, 0x0075C0, 0x0075C1, 0x0075C4, 0x0075C6, 0x0075CC, 0x0075CE, 0x0075CF, 0x0075D7, 0x0075DC, 0x0075DF, 0x0075E0, 0x0075E1, 0x0075E4,
	0x0075E7, 0x0075EC, 0x0075EE, 0x0075EF, 0x0075F1, 0x0075F9, 0x007600, 0x007602, 0x007603, 0x007604, 0x007607, 0x007608, 0x00760A, 0x00760C, 0x00760F, 0x007612, 0x007613, 0x007615, 0x007616, 0x007619, 0x00761B, 0x00761C, 0x00761D, 0x00761E, 0x007623, 0x007625, 0x007626, 0x007629, 0x00762D, 0x007632, 0x007633, 0x007635,
	0x007638, 0x007639, 0x00763A, 0x00763C, 0x00764A, 0x007640, 0x007641, 0x007643, 0x007644, 0x007645, 0x007649, 0x00764B, 0x007655, 0x007659, 0x00765F, 0x007664, 0x007665, 0x00766D, 0x00766E, 0x00766F, 0x007671, 0x007674, 0x007681, 0x007685, 0x00768C, 0x00768D, 0x007695, 0x00769B, 0x00769C, 0x00769D, 0x00769F, 0x0076A0,
	0x0076A2, 0x0076A3, 0x0076A4, 0x0076A5, 0x0076A6, 0x0076A7, 0x0076A8, 0x0076AA, 0x0076AD, 0x0076BD, 0x0076C1, 0x0076C5, 0x0076C9, 0x0076CB, 0x0076CC, 0x0076CE, 0x0076D4, 0x0076D9, 0x0076E0, 0x0076E6, 0x0076E8, 0x0076EC, 0x0076F0, 0x0076F1, 0x0076F6, 0x0076F9, 0x0076FC, 0x007700, 0x007706, 0x00770A, 0x00770E, 0x007712,
	0x007714, 0x007715, 0x007717, 0x007719, 0x00771A, 0x00771C, 0x007722, 0x007728, 0x00772D, 0x00772E, 0x00772F, 0x007734, 0x007735, 0x007736, 0x007739, 0x00773D, 0x00773E, 0x007742, 0x007745, 0x007746, 0x00774A, 0x00774D, 0x00774E, 0x00774F, 0x007752, 0x007756, 0x007757, 0x00775C, 0x00775E, 0x00775F, 0x007760, 0x007762,
	0x007764, 0x007767, 0x00776A, 0x00776C, 0x007770, 0x007772, 0x007773, 0x007774, 0x00777A, 0x00777D, 0x007780, 0x007784, 0x00778C, 0x00778D, 0x007794, 0x007795, 0x007796, 0x00779A, 0x00779F, 0x0077A2, 0x0077A7, 0x0077AA, 0x0077AE, 0x0077AF, 0x0077B1, 0x0077B5, 0x0077BE, 0x0077C3, 0x0077C9, 0x0077D1, 0x0077D2, 0x0077D5,
	0x0077D9, 0x0077DE, 0x0077DF, 0x0077E0, 0x0077E4, 0x0077E6, 0x0077EA, 0x0077EC, 0x0077F0, 0x0077F1, 0x0077F4, 0x0077F8, 0x0077FB, 0x007805, 0x007806, 0x007809, 0x00780D, 0x00780E, 0x007811, 0x00781D, 0x007821, 0x007822, 0x007823, 0x00782D, 0x00782E, 0x007830, 0x007835, 0x007837, 0x007843, 0x007844, 0x007847, 0x007848,
	0x00784C, 0x00784E, 0x007852, 0x00785C, 0x00785E, 0x007860, 0x007861, 0x007863, 0x007864, 0x007868, 0x00786A, 0x00786E, 0x00787A, 0x00787E, 0x00788A, 0x00788F, 0x007894, 0x007898, 0x0078A1, 0x00789D, 0x00789E, 0x00789F, 0x0078A4, 0x0078A8, 0x0078AC, 0x0078AD, 0x0078B0, 0x0078B1, 0x0078B2, 0x0078B3, 0x0078BB, 0x0078BD,
	0x0078BF, 0x0078C7, 0x0078C8, 0x0078C9, 0x0078CC, 0x0078CE, 0x0078D2, 0x0078D3, 0x0078D5, 0x0078D6, 0x0078E4, 0x0078DB, 0x0078DF, 0x0078E0, 0x0078E1, 0x0078E6, 0x0078EA, 0x0078F2, 0x0078F3, 0x007900, 0x0078F6, 0x0078F7, 0x0078FA, 0x0078FB, 0x0078FF, 0x007906, 0x00790C, 0x007910, 0x00791A, 0x00791C, 0x00791E, 0x00791F,
	0x007920, 0x007925, 0x007927, 0x007929, 0x00792D, 0x007931, 0x007934, 0x007935, 0x00793B, 0x00793D, 0x00793F, 0x007944, 0x007945, 0x007946, 0x00794A, 0x00794B, 0x00794F, 0x007951, 0x007954, 0x007958, 0x00795B, 0x00795C, 0x007967, 0x007969, 0x00796B, 0x007972, 0x007979, 0x00797B, 0x00797C, 0x00797E, 0x00798B, 0x00798C,
	0
};

static const unsigned map_kanji_hojo_4[] = {
	0x007991, 0x007993, 0x007994, 0x007995, 0x007996, 0x007998, 0x00799B, 0x00799C, 0x0079A1, 0x0079A8, 0x0079A9, 0x0079AB, 0x0079AF, 0x0079B1, 0x0079B4, 0x0079B8, 0x0079BB, 0x0079C2, 0x0079C4, 0x0079C7, 0x0079C8, 0x0079CA, 0x0079CF, 0x0079D4, 0x0079D6, 0x0079DA, 0x0079DD, 0x0079DE, 0x0079E0, 0x0079E2, 0x0079E5, 0x0079EA,
	0x0079EB, 0x0079ED, 0x0079F1, 0x0079F8, 0x0079FC, 0x007A02, 0x007A03, 0x007A07, 0x007A09, 0x007A0A, 0x007A0C, 0x007A11, 0x007A15, 0x007A1B, 0x007A1E, 0x007A21, 0x007A27, 0x007A2B, 0x007A2D, 0x007A2F, 0x007A30, 0x007A34, 0x007A35, 0x007A38, 0x007A39, 0x007A3A, 0x007A44, 0x007A45, 0x007A47, 0x007A48, 0x007A4C, 0x007A55,
	0x007A56, 0x007A59, 0x007A5C, 0x007A5D, 0x007A5F, 0x007A60, 0x007A65, 0x007A67, 0x007A6A, 0x007A6D, 0x007A75, 0x007A78, 0x007A7E, 0x007A80, 0x007A82, 0x007A85, 0x007A86, 0x007A8A, 0x007A8B, 0x007A90, 0x007A91, 0x007A94, 0x007A9E, 0x007AA0, 0x007AA3, 0x007AAC, 0x007AB3, 0x007AB5, 0x007AB9, 0x007ABB, 0x007ABC, 0x007AC6,
	0x007AC9, 0x007ACC, 0x007ACE, 0x007AD1, 0x007ADB, 0x007AE8, 0x007AE9, 0x007AEB, 0x007AEC, 0x007AF1, 0x007AF4, 0x007AFB, 0x007AFD, 0x007AFE, 0x007B07, 0x007B14, 0x007B1F, 0x007B23, 0x007B27, 0x007B29, 0x007B2A, 0x007B2B, 0x007B2D, 0x007B2E, 0x007B2F, 0x007B30, 0x007B31, 0x007B34, 0x007B3D, 0x007B3F, 0x007B40, 0x007B41,
	0x007B47, 0x007B4E, 0x007B55, 0x007B60, 0x007B64, 0x007B66, 0x007B69, 0x007B6A, 0x007B6D, 0x007B6F, 0x007B72, 0x007B73, 0x007B77, 0x007B84, 0x007B89, 0x007B8E, 0x007B90, 0x007B91, 0x007B96, 0x007B9B, 0x007B9E, 0x007BA0, 0x007BA5, 0x007BAC, 0x007BAF, 0x007BB0, 0x007BB2, 0x007BB5, 0x007BB6, 0x007BBA, 0x007BBB, 0x007BBC,
	0x007BBD, 0x007BC2, 0x007BC5, 0x007BC8, 0x007BCA, 0x007BD4, 0x007BD6, 0x007BD7, 0x007BD9, 0x007BDA, 0x007BDB, 0x007BE8, 0x007BEA, 0x007BF2, 0x007BF4, 0x007BF5, 0x007BF8, 0x007BF9, 0x007BFA, 0x007BFC, 0x007BFE, 0x007C01, 0x007C02, 0x007C03, 0x007C04, 0x007C06, 0x007C09, 0x007C0B, 0x007C0C, 0x007C0E, 0x007C0F, 0x007C19,
	0x007C1B, 0x007C20, 0x007C25, 0x007C26, 0x007C28, 0x007C2C, 0x007C31, 0x007C33, 0x007C34, 0x007C36, 0x007C39, 0x007C3A, 0x007C46, 0x007C4A, 0x007C55, 0x007C51, 0x007C52, 0x007C53, 0x007C59, 0x007C5A, 0x007C5B, 0x007C5C, 0x007C5D, 0x007C5E, 0x007C61, 0x007C63, 0x007C67, 0x007C69, 0x007C6D, 0x007C6E, 0x007C70, 0x007C72,
	0x007C79, 0x007C7C, 0x007C7D, 0x007C86, 0x007C87, 0x007C8F, 0x007C94, 0x007C9E, 0x007CA0, 0x007CA6, 0x007CB0, 0x007CB6, 0x007CB7, 0x007CBA, 0x007CBB, 0x007CBC, 0x007CBF, 0x007CC4, 0x007CC7, 0x007CC8, 0x007CC9, 0x007CCD, 0x007CCF, 0x007CD3, 0x007CD4, 0x007CD5, 0x007CD7, 0x007CD9, 0x007CDA, 0x007CDD, 0x007CE6, 0x007CE9,
	0x007CEB, 0x007CF5, 0x007D03, 0x007D07, 0x007D08, 0x007D09, 0x007D0F, 0x007D11, 0x007D12, 0x007D13, 0x007D16, 0x007D1D, 0x007D1E, 0x007D23, 0x007D26, 0x007D2A, 0x007D2D, 0x007D31, 0x007D3C, 0x007D3D, 0x007D3E, 0x007D40, 0x007D41, 0x007D47, 0x007D48, 0x007D4D, 0x007D51, 0x007D53, 0x007D57, 0x007D59, 0x007D5A, 0x007D5C,
	0x007D5D, 0x007D65, 0x007D67, 0x007D6A, 0x007D70, 0x007D78, 0x007D7A, 0x007D7B, 0x007D7F, 0x007D81, 0x007D82, 0x007D83, 0x007D85, 0x007D86, 0x007D88, 0x007D8B, 0x007D8C, 0x007D8D, 0x007D91, 0x007D96, 0x007D97, 0x007D9D, 0x007D9E, 0x007DA6, 0x007DA7, 0x007DAA, 0x007DB3, 0x007DB6, 0x007DB7, 0x007DB9, 0x007DC2, 0x007DC3,
	0x007DC4, 0x007DC5, 0x007DC6, 0x007DCC, 0x007DCD, 0x007DCE, 0x007DD7, 0x007DD9, 0x007E00, 0x007DE2, 0x007DE5, 0x007DE6, 0x007DEA, 0x007DEB, 0x007DED, 0x007DF1, 0x007DF5, 0x007DF6, 0x007DF9, 0x007DFA, 0x007E08, 0x007E10, 0x007E11, 0x007E15, 0x007E17, 0x007E1C, 0x007E1D, 0x007E20, 0x007E27, 0x007E28, 0x007E2C, 0x007E2D,
	0x007E2F, 0x007E33, 0x007E36, 0x007E3F, 0x007E44, 0x007E45, 0x007E47, 0x007E4E, 0x007E50, 0x007E52, 0x007E58, 0x007E5F, 0x007E61, 0x007E62, 0x007E65, 0x007E6B, 0x007E6E, 0x007E6F, 0x007E73, 0x007E78, 0x007E7E, 0x007E81, 0x007E86, 0x007E87, 0x007E8A, 0x007E8D, 0x007E91, 0x007E95, 0x007E98, 0x007E9A, 0x007E9D, 0x007E9E,
	0x007F3C, 0x007F3B, 0x007F3D, 0x007F3E, 0x007F3F, 0x007F43, 0x007F44, 0x007F47, 0x007F4F, 0x007F52, 0x007F53, 0x007F5B, 0x007F5C, 0x007F5D, 0x007F61, 0x007F63, 0x007F64, 0x007F65, 0x007F66, 0x007F6D, 0x007F71, 0x007F7D, 0x007F7E, 0x007F7F, 0x007F80, 0x007F8B, 0x007F8D, 0x007F8F, 0x007F90, 0x007F91, 0x007F96, 0x007F97,
	0x007F9C, 0x007FA1, 0x007FA2, 0x007FA6, 0x007FAA, 0x007FAD, 0x007FB4, 0x007FBC, 0x007FBF, 0x007FC0, 0x007FC3, 0x007FC8, 0x007FCE, 0x007FCF, 0x007FDB, 0x007FDF, 0x007FE3, 0x007FE5, 0x007FE8, 0x007FEC, 0x007FEE, 0x007FEF, 0x007FF2, 0x007FFA, 0x007FFD, 0x007FFE, 0x007FFF, 0x008007, 0x008008, 0x00800A, 0x00800D, 0x00800E,
	0x00800F, 0x008011, 0x008013, 0x008014, 0x008016, 0x00801D, 0x00801E, 0x00801F, 0x008020, 0x008024, 0x008026, 0x00802C, 0x00802E, 0x008030, 0x008034, 0x008035, 0x008037, 0x008039, 0x00803A, 0x00803C, 0x00803E, 0x008040, 0x008044, 0x008060, 0x008064, 0x008066, 0x00806D, 0x008071, 0x008075, 0x008081, 0x008088, 0x00808E,
	0x00809C, 0x00809E, 0x0080A6, 0x0080A7, 0x0080AB, 0x0080B8, 0x0080B9, 0x0080C8, 0x0080CD, 0x0080CF, 0x0080D2, 0x0080D4, 0x0080D5, 0x0080D7, 0x0080D8, 0x0080E0, 0x0080ED, 0x0080EE, 0x0080F0, 0x0080F2, 0x0080F3, 0x0080F6, 0x0080F9, 0x0080FA, 0x0080FE, 0x008103, 0x00810B, 0x008116, 0x008117, 0x008118, 0x00811C, 0x00811E,
	0x008120, 0x008124, 0x008127, 0x00812C, 0x008130, 0x008135, 0x00813A, 0x00813C, 0x008145, 0x008147, 0x00814A, 0x00814C, 0x008152, 0x008157, 0x008160, 0x008161, 0x008167, 0x008168, 0x008169, 0x00816D, 0x00816F, 0x008177, 0x008181, 0x008190, 0x008184, 0x008185, 0x008186, 0x00818B, 0x00818E, 0x008196, 0x008198, 0x00819B,
	0x00819E, 0x0081A2, 0x0081AE, 0x0081B2, 0x0081B4, 0x0081BB, 0x0081CB, 0x0081C3, 0x0081C5, 0x0081CA, 0x0081CE, 0x0081CF, 0x0081D5, 0x0081D7, 0x0081DB, 0x0081DD, 0x0081DE, 0x0081E1, 0x0081E4, 0x0081EB, 0x0081EC, 0x0081F0, 0x0081F1, 0x0081F2, 0x0081F5, 0x0081F6, 0x0081F8, 0x0081F9, 0x0081FD, 0x0081FF, 0x008200, 0x008203,
	0x00820F, 0x008213, 0x008214, 0x008219, 0x00821A, 0x00821D, 0x008221, 0x008222, 0x008228, 0x008232, 0x008234, 0x00823A, 0x008243, 0x008244, 0x008245, 0x008246, 0x00824B, 0x00824E, 0x00824F, 0x008251, 0x008256, 0x00825C, 0x008260, 0x008263, 0x008267, 0x00826D, 0x008274, 0x00827B, 0x00827D, 0x00827F, 0x008280, 0x008281,
	0x008283, 0x008284, 0x008287, 0x008289, 0x00828A, 0x00828E, 0x008291, 0x008294, 0x008296, 0x008298, 0x00829A, 0x00829B, 0x0082A0, 0x0082A1, 0x0082A3, 0x0082A4, 0x0082A7, 0x0082A8, 0x0082A9, 0x0082AA, 0x0082AE, 0x0082B0, 0x0082B2, 0x0082B4, 0x0082B7, 0x0082BA, 0x0082BC, 0x0082BE, 0x0082BF, 0x0082C6, 0x0082D0, 0x0082D5,
	0x0082DA, 0x0082E0, 0x0082E2, 0x0082E4, 0x0082E8, 0x0082EA, 0x0082ED, 0x0082EF, 0x0082F6, 0x0082F7, 0x0082FD, 0x0082FE, 0x008300, 0x008301, 0x008307, 0x008308, 0x00830A, 0x00830B, 0x008354, 0x00831B, 0x00831D, 0x00831E, 0x00831F, 0x008321, 0x008322, 0x00832C, 0x00832D, 0x00832E, 0x008330, 0x008333, 0x008337, 0x00833A,
	0x00833C, 0x00833D, 0x008342, 0x008343, 0x008344, 0x008347, 0x00834D, 0x00834E, 0x008351, 0x008355, 0x008356, 0x008357, 0x008370, 0x008378, 0x00837D, 0x00837F, 0x008380, 0x008382, 0x008384, 0x008386, 0x00838D, 0x008392, 0x008394, 0x008395, 0x008398, 0x008399, 0x00839B, 0x00839C, 0x00839D, 0x0083A6, 0x0083A7, 0x0083A9,
	0x0083AC, 0x0083BE, 0x0083BF, 0x0083C0, 0x0083C7, 0x0083C9, 0x0083CF, 0x0083D0, 0x0083D1, 0x0083D4, 0x0083DD, 0x008353, 0x0083E8, 0x0083EA, 0x0083F6, 0x0083F8, 0x0083F9, 0x0083FC, 0x008401, 0x008406, 0x00840A, 0x00840F, 0x008411, 0x008415, 0x008419, 0x0083AD, 0x00842F, 0x008439, 0x008445, 0x008447, 0x008448, 0x00844A,
	0x00844D, 0x00844F, 0x008451, 0x008452, 0x008456, 0x008458, 0x008459, 0x00845A, 0x00845C, 0x008460, 0x008464, 0x008465, 0x008467, 0x00846A, 0x008470, 0x008473, 0x008474, 0x008476, 0x008478, 0x00847C, 0x00847D, 0x008481, 0x008485, 0x008492, 0x008493, 0x008495, 0x00849E, 0x0084A6, 0x0084A8, 0x0084A9, 0x0084AA, 0x0084AF,
	0x0084B1, 0x0084B4, 0x0084BA, 0x0084BD, 0x0084BE, 0x0084C0, 0x0084C2, 0x0084C7, 0x0084C8, 0x0084CC, 0x0084CF, 0x0084D3, 0x0084DC, 0x0084E7, 0x0084EA, 0x0084EF, 0x0084F0, 0x0084F1, 0x0084F2, 0x0084F7, 0x008532, 0x0084FA, 0x0084FB, 0x0084FD, 0x008502, 0x008503, 0x008507, 0x00850C, 0x00850E, 0x008510, 0x00851C, 0x00851E,
	0x008522, 0x008523, 0x008524, 0x008525, 0x008527, 0x00852A, 0x00852B, 0x00852F, 0x008533, 0x008534, 0x008536, 0x00853F, 0x008546, 0x00854F, 0x008550, 0x008551, 0x008552, 0x008553, 0x008556, 0x008559, 0x00855C, 0x00855D, 0x00855E, 0x00855F, 0x008560, 0x008561, 0x008562, 0x008564, 0x00856B, 0x00856F, 0x008579, 0x00857A,
	0x00857B, 0x00857D, 0x00857F, 0x008581, 0x008585, 0x008586, 0x008589, 0x00858B, 0x00858C, 0x00858F, 0x008593, 0x008598, 0x00859D, 0x00859F, 0x0085A0, 0x0085A2, 0x0085A5, 0x0085A7, 0x0085B4, 0x0085B6, 0x0085B7, 0x0085B8, 0x0085BC, 0x0085BD, 0x0085BE, 0x0085BF, 0x0085C2, 0x0085C7, 0x0085CA, 0x0085CB, 0x0085CE, 0x0085AD,
	0x0085D8, 0x0085DA, 0x0085DF, 0x0085E0, 0x0085E6, 0x0085E8, 0x0085ED, 0x0085F3, 0x0085F6, 0x0085FC, 0x0085FF, 0x008600, 0x008604, 0x008605, 0x00860D, 0x00860E, 0x008610, 0x008611, 0x008612, 0x008618, 0x008619, 0x00861B, 0x00861E, 0x008621, 0x008627, 0x008629, 0x008636, 0x008638, 0x00863A, 0x00863C, 0x00863D, 0x008640,
	0x008642, 0x008646, 0x008652, 0x008653, 0x008656, 0x008657, 0x008658, 0x008659, 0x00865D, 0x008660, 0x008661, 0x008662, 0x008663, 0x008664, 0x008669, 0x00866C, 0x00866F, 0x008675, 0x008676, 0x008677, 0x00867A, 0x00868D, 0x008691, 0x008696, 0x008698, 0x00869A, 0x00869C, 0x0086A1, 0x0086A6, 0x0086A7, 0x0086A8, 0x0086AD,
	0x0086B1, 0x0086B3, 0x0086B4, 0x0086B5, 0x0086B7, 0x0086B8, 0x0086B9, 0x0086BF, 0x0086C0, 0x0086C1, 0x0086C3, 0x0086C5, 0x0086D1, 0x0086D2, 0x0086D5, 0x0086D7, 0x0086DA, 0x0086DC, 0x0086E0, 0x0086E3, 0x0086E5, 0x0086E7, 0x008688, 0x0086FA, 0x0086FC, 0x0086FD, 0x008704, 0x008705, 0x008707, 0x00870B, 0x00870E, 0x00870F,
	0x008710, 0x008713, 0x008714, 0x008719, 0x00871E, 0x00871F, 0x008721, 0x008723, 0x008728, 0x00872E, 0x00872F, 0x008731, 0x008732, 0x008739, 0x00873A, 0x00873C, 0x00873D, 0x00873E, 0x008740, 0x008743, 0x008745, 0x00874D, 0x008758, 0x00875D, 0x008761, 0x008764, 0x008765, 0x00876F, 0x008771, 0x008772, 0x00877B, 0x008783,
	0x008784, 0x008785, 0x008786, 0x008787, 0x008788, 0x008789, 0x00878B, 0x00878C, 0x008790, 0x008793, 0x008795, 0x008797, 0x008798, 0x008799, 0x00879E, 0x0087A0, 0x0087A3, 0x0087A7, 0x0087AC, 0x0087AD, 0x0087AE, 0x0087B1, 0x0087B5, 0x0087BE, 0x0087BF, 0x0087C1, 0x0087C8, 0x0087C9, 0x0087CA, 0x0087CE, 0x0087D5, 0x0087D6,
	0
};

static const unsigned map_kanji_hojo_5[] = {
	0x0087D9, 0x0087DA, 0x0087DC, 0x0087DF, 0x0087E2, 0x0087E3, 0x0087E4, 0x0087EA, 0x0087EB, 0x0087ED, 0x0087F1, 0x0087F3, 0x0087F8, 0x0087FA, 0x0087FF, 0x008801, 0x008803, 0x008806, 0x008809, 0x00880A, 0x00880B, 0x008810, 0x008819, 0x008812, 0x008813, 0x008814, 0x008818, 0x00881A, 0x00881B, 0x00881C, 0x00881E, 0x00881F,
	0x008828, 0x00882D, 0x00882E, 0x008830, 0x008832, 0x008835, 0x00883A, 0x00883C, 0x008841, 0x008843, 0x008845, 0x008848, 0x008849, 0x00884A, 0x00884B, 0x00884E, 0x008851, 0x008855, 0x008856, 0x008858, 0x00885A, 0x00885C, 0x00885F, 0x008860, 0x008864, 0x008869, 0x008871, 0x008879, 0x00887B, 0x008880, 0x008898, 0x00889A,
	0x00889B, 0x00889C, 0x00889F, 0x0088A0, 0x0088A8, 0x0088AA, 0x0088BA, 0x0088BD, 0x0088BE, 0x0088C0, 0x0088CA, 0x0088CB, 0x0088CC, 0x0088CD, 0x0088CE, 0x0088D1, 0x0088D2, 0x0088D3, 0x0088DB, 0x0088DE, 0x0088E7, 0x0088EF, 0x0088F0, 0x0088F1, 0x0088F5, 0x0088F7, 0x008901, 0x008906, 0x00890D, 0x00890E, 0x00890F, 0x008915,
	0x008916, 0x008918, 0x008919, 0x00891A, 0x00891C, 0x008920, 0x008926, 0x008927, 0x008928, 0x008930, 0x008931, 0x008932, 0x008935, 0x008939, 0x00893A, 0x00893E, 0x008940, 0x008942, 0x008945, 0x008946, 0x008949, 0x00894F, 0x008952, 0x008957, 0x00895A, 0x00895B, 0x00895C, 0x008961, 0x008962, 0x008963, 0x00896B, 0x00896E,
	0x008970, 0x008973, 0x008975, 0x00897A, 0x00897B, 0x00897C, 0x00897D, 0x008989, 0x00898D, 0x008990, 0x008994, 0x008995, 0x00899B, 0x00899C, 0x00899F, 0x0089A0, 0x0089A5, 0x0089B0, 0x0089B4, 0x0089B5, 0x0089B6, 0x0089B7, 0x0089BC, 0x0089D4, 0x0089D5, 0x0089D6, 0x0089D7, 0x0089D8, 0x0089E5, 0x0089E9, 0x0089EB, 0x0089ED,
	0x0089F1, 0x0089F3, 0x0089F6, 0x0089F9, 0x0089FD, 0x0089FF, 0x008A04, 0x008A05, 0x008A07, 0x008A0F, 0x008A11, 0x008A12, 0x008A14, 0x008A15, 0x008A1E, 0x008A20, 0x008A22, 0x008A24, 0x008A26, 0x008A2B, 0x008A2C, 0x008A2F, 0x008A35, 0x008A37, 0x008A3D, 0x008A3E, 0x008A40, 0x008A43, 0x008A45, 0x008A47, 0x008A49, 0x008A4D,
	0x008A4E, 0x008A53, 0x008A56, 0x008A57, 0x008A58, 0x008A5C, 0x008A5D, 0x008A61, 0x008A65, 0x008A67, 0x008A75, 0x008A76, 0x008A77, 0x008A79, 0x008A7A, 0x008A7B, 0x008A7E, 0x008A7F, 0x008A80, 0x008A83, 0x008A86, 0x008A8B, 0x008A8F, 0x008A90, 0x008A92, 0x008A96, 0x008A97, 0x008A99, 0x008A9F, 0x008AA7, 0x008AA9, 0x008AAE,
	0x008AAF, 0x008AB3, 0x008AB6, 0x008AB7, 0x008ABB, 0x008ABE, 0x008AC3, 0x008AC6, 0x008AC8, 0x008AC9, 0x008ACA, 0x008AD1, 0x008AD3, 0x008AD4, 0x008AD5, 0x008AD7, 0x008ADD, 0x008ADF, 0x008AEC, 0x008AF0, 0x008AF4, 0x008AF5, 0x008AF6, 0x008AFC, 0x008AFF, 0x008B05, 0x008B06, 0x008B0B, 0x008B11, 0x008B1C, 0x008B1E, 0x008B1F,
	0x008B0A, 0x008B2D, 0x008B30, 0x008B37, 0x008B3C, 0x008B42, 0x008B43, 0x008B44, 0x008B45, 0x008B46, 0x008B48, 0x008B52, 0x008B53, 0x008B54, 0x008B59, 0x008B4D, 0x008B5E, 0x008B63, 0x008B6D, 0x008B76, 0x008B78, 0x008B79, 0x008B7C, 0x008B7E, 0x008B81, 0x008B84, 0x008B85, 0x008B8B, 0x008B8D, 0x008B8F, 0x008B94, 0x008B95,
	0x008B9C, 0x008B9E, 0x008B9F, 0x008C38, 0x008C39, 0x008C3D, 0x008C3E, 0x008C45, 0x008C47, 0x008C49, 0x008C4B, 0x008C4F, 0x008C51, 0x008C53, 0x008C54, 0x008C57, 0x008C58, 0x008C5B, 0x008C5D, 0x008C59, 0x008C63, 0x008C64, 0x008C66, 0x008C68, 0x008C69, 0x008C6D, 0x008C73, 0x008C75, 0x008C76, 0x008C7B, 0x008C7E, 0x008C86,
	0x008C87, 0x008C8B, 0x008C90, 0x008C92, 0x008C93, 0x008C99, 0x008C9B, 0x008C9C, 0x008CA4, 0x008CB9, 0x008CBA, 0x008CC5, 0x008CC6, 0x008CC9, 0x008CCB, 0x008CCF, 0x008CD6, 0x008CD5, 0x008CD9, 0x008CDD, 0x008CE1, 0x008CE8, 0x008CEC, 0x008CEF, 0x008CF0, 0x008CF2, 0x008CF5, 0x008CF7, 0x008CF8, 0x008CFE, 0x008CFF, 0x008D01,
	0x008D03, 0x008D09, 0x008D12, 0x008D17, 0x008D1B, 0x008D65, 0x008D69, 0x008D6C, 0x008D6E, 0x008D7F, 0x008D82, 0x008D84, 0x008D88, 0x008D8D, 0x008D90, 0x008D91, 0x008D95, 0x008D9E, 0x008D9F, 0x008DA0, 0x008DA6, 0x008DAB, 0x008DAC, 0x008DAF, 0x008DB2, 0x008DB5, 0x008DB7, 0x008DB9, 0x008DBB, 0x008DC0, 0x008DC5, 0x008DC6,
	0x008DC7, 0x008DC8, 0x008DCA, 0x008DCE, 0x008DD1, 0x008DD4, 0x008DD5, 0x008DD7, 0x008DD9, 0x008DE4, 0x008DE5, 0x008DE7, 0x008DEC, 0x008DF0, 0x008DBC, 0x008DF1, 0x008DF2, 0x008DF4, 0x008DFD, 0x008E01, 0x008E04, 0x008E05, 0x008E06, 0x008E0B, 0x008E11, 0x008E14, 0x008E16, 0x008E20, 0x008E21, 0x008E22, 0x008E23, 0x008E26,
	0x008E27, 0x008E31, 0x008E33, 0x008E36, 0x008E37, 0x008E38, 0x008E39, 0x008E3D, 0x008E40, 0x008E41, 0x008E4B, 0x008E4D, 0x008E4E, 0x008E4F, 0x008E54, 0x008E5B, 0x008E5C, 0x008E5D, 0x008E5E, 0x008E61, 0x008E62, 0x008E69, 0x008E6C, 0x008E6D, 0x008E6F, 0x008E70, 0x008E71, 0x008E79, 0x008E7A, 0x008E7B, 0x008E82, 0x008E83,
	0x008E89, 0x008E90, 0x008E92, 0x008E95, 0x008E9A, 0x008E9B, 0x008E9D, 0x008E9E, 0x008EA2, 0x008EA7, 0x008EA9, 0x008EAD, 0x008EAE, 0x008EB3, 0x008EB5, 0x008EBA, 0x008EBB, 0x008EC0, 0x008EC1, 0x008EC3, 0x008EC4, 0x008EC7, 0x008ECF, 0x008ED1, 0x008ED4, 0x008EDC, 0x008EE8, 0x008EEE, 0x008EF0, 0x008EF1, 0x008EF7, 0x008EF9,
	0x008EFA, 0x008EED, 0x008F00, 0x008F02, 0x008F07, 0x008F08, 0x008F0F, 0x008F10, 0x008F16, 0x008F17, 0x008F18, 0x008F1E, 0x008F20, 0x008F21, 0x008F23, 0x008F25, 0x008F27, 0x008F28, 0x008F2C, 0x008F2D, 0x008F2E, 0x008F34, 0x008F35, 0x008F36, 0x008F37, 0x008F3A, 0x008F40, 0x008F41, 0x008F43, 0x008F47, 0x008F4F, 0x008F51,
	0x008F52, 0x008F53, 0x008F54, 0x008F55, 0x008F58, 0x008F5D, 0x008F5E, 0x008F65, 0x008F9D, 0x008FA0, 0x008FA1, 0x008FA4, 0x008FA5, 0x008FA6, 0x008FB5, 0x008FB6, 0x008FB8, 0x008FBE, 0x008FC0, 0x008FC1, 0x008FC6, 0x008FCA, 0x008FCB, 0x008FCD, 0x008FD0, 0x008FD2, 0x008FD3, 0x008FD5, 0x008FE0, 0x008FE3, 0x008FE4, 0x008FE8,
	0x008FEE, 0x008FF1, 0x008FF5, 0x008FF6, 0x008FFB, 0x008FFE, 0x009002, 0x009004, 0x009008, 0x00900C, 0x009018, 0x00901B, 0x009028, 0x009029, 0x00902F, 0x00902A, 0x00902C, 0x00902D, 0x009033, 0x009034, 0x009037, 0x00903F, 0x009043, 0x009044, 0x00904C, 0x00905B, 0x00905D, 0x009062, 0x009066, 0x009067, 0x00906C, 0x009070,
	0x009074, 0x009079, 0x009085, 0x009088, 0x00908B, 0x00908C, 0x00908E, 0x009090, 0x009095, 0x009097, 0x009098, 0x009099, 0x00909B, 0x0090A0, 0x0090A1, 0x0090A2, 0x0090A5, 0x0090B0, 0x0090B2, 0x0090B3, 0x0090B4, 0x0090B6, 0x0090BD, 0x0090CC, 0x0090BE, 0x0090C3, 0x0090C4, 0x0090C5, 0x0090C7, 0x0090C8, 0x0090D5, 0x0090D7,
	0x0090D8, 0x0090D9, 0x0090DC, 0x0090DD, 0x0090DF, 0x0090E5, 0x0090D2, 0x0090F6, 0x0090EB, 0x0090EF, 0x0090F0, 0x0090F4, 0x0090FE, 0x0090FF, 0x009100, 0x009104, 0x009105, 0x009106, 0x009108, 0x00910D, 0x009110, 0x009114, 0x009116, 0x009117, 0x009118, 0x00911A, 0x00911C, 0x00911E, 0x009120, 0x009125, 0x009122, 0x009123,
	0x009127, 0x009129, 0x00912E, 0x00912F, 0x009131, 0x009134, 0x009136, 0x009137, 0x009139, 0x00913A, 0x00913C, 0x00913D, 0x009143, 0x009147, 0x009148, 0x00914F, 0x009153, 0x009157, 0x009159, 0x00915A, 0x00915B, 0x009161, 0x009164, 0x009167, 0x00916D, 0x009174, 0x009179, 0x00917A, 0x00917B, 0x009181, 0x009183, 0x009185,
	0x009186, 0x00918A, 0x00918E, 0x009191, 0x009193, 0x009194, 0x009195, 0x009198, 0x00919E, 0x0091A1, 0x0091A6, 0x0091A8, 0x0091AC, 0x0091AD, 0x0091AE, 0x0091B0, 0x0091B1, 0x0091B2, 0x0091B3, 0x0091B6, 0x0091BB, 0x0091BC, 0x0091BD, 0x0091BF, 0x0091C2, 0x0091C3, 0x0091C5, 0x0091D3, 0x0091D4, 0x0091D7, 0x0091D9, 0x0091DA,
	0x0091DE, 0x0091E4, 0x0091E5, 0x0091E9, 0x0091EA, 0x0091EC, 0x0091ED, 0x0091EE, 0x0091EF, 0x0091F0, 0x0091F1, 0x0091F7, 0x0091F9, 0x0091FB, 0x0091FD, 0x009200, 0x009201, 0x009204, 0x009205, 0x009206, 0x009207, 0x009209, 0x00920A, 0x00920C, 0x009210, 0x009212, 0x009213, 0x009216, 0x009218, 0x00921C, 0x00921D, 0x009223,
	0x009224, 0x009225, 0x009226, 0x009228, 0x00922E, 0x00922F, 0x009230, 0x009233, 0x009235, 0x009236, 0x009238, 0x009239, 0x00923A, 0x00923C, 0x00923E, 0x009240, 0x009242, 0x009243, 0x009246, 0x009247, 0x00924A, 0x00924D, 0x00924E, 0x00924F, 0x009251, 0x009258, 0x009259, 0x00925C, 0x00925D, 0x009260, 0x009261, 0x009265,
	0x009267, 0x009268, 0x009269, 0x00926E, 0x00926F, 0x009270, 0x009275, 0x009276, 0x009277, 0x009278, 0x009279, 0x00927B, 0x00927C, 0x00927D, 0x00927F, 0x009288, 0x009289, 0x00928A, 0x00928D, 0x00928E, 0x009292, 0x009297, 0x009299, 0x00929F, 0x0092A0, 0x0092A4, 0x0092A5, 0x0092A7, 0x0092A8, 0x0092AB, 0x0092AF, 0x0092B2,
	0x0092B6, 0x0092B8, 0x0092BA, 0x0092BB, 0x0092BC, 0x0092BD, 0x0092BF, 0x0092C0, 0x0092C1, 0x0092C2, 0x0092C3, 0x0092C5, 0x0092C6, 0x0092C7, 0x0092C8, 0x0092CB, 0x0092CC, 0x0092CD, 0x0092CE, 0x0092D0, 0x0092D3, 0x0092D5, 0x0092D7, 0x0092D8, 0x0092D9, 0x0092DC, 0x0092DD, 0x0092DF, 0x0092E0, 0x0092E1, 0x0092E3, 0x0092E5,
	0x0092E7, 0x0092E8, 0x0092EC, 0x0092EE, 0x0092F0, 0x0092F9, 0x0092FB, 0x0092FF, 0x009300, 0x009302, 0x009308, 0x00930D, 0x009311, 0x009314, 0x009315, 0x00931C, 0x00931D, 0x00931E, 0x00931F, 0x009321, 0x009324, 0x009325, 0x009327, 0x009329, 0x00932A, 0x009333, 0x009334, 0x009336, 0x009337, 0x009347, 0x009348, 0x009349,
	0x009350, 0x009351, 0x009352, 0x009355, 0x009357, 0x009358, 0x00935A, 0x00935E, 0x009364, 0x009365, 0x009367, 0x009369, 0x00936A, 0x00936D, 0x00936F, 0x009370, 0x009371, 0x009373, 0x009374, 0x009376, 0x00937A, 0x00937D, 0x00937F, 0x009380, 0x009381, 0x009382, 0x009388, 0x00938A, 0x00938B, 0x00938D, 0x00938F, 0x009392,
	0x009395, 0x009398, 0x00939B, 0x00939E, 0x0093A1, 0x0093A3, 0x0093A4, 0x0093A6, 0x0093A8, 0x0093AB, 0x0093B4, 0x0093B5, 0x0093B6, 0x0093BA, 0x0093A9, 0x0093C1, 0x0093C4, 0x0093C5, 0x0093C6, 0x0093C7, 0x0093C9, 0x0093CA, 0x0093CB, 0x0093CC, 0x0093CD, 0x0093D3, 0x0093D9, 0x0093DC, 0x0093DE, 0x0093DF, 0x0093E2, 0x0093E6,
	0x0093E7, 0x0093F9, 0x0093F7, 0x0093F8, 0x0093FA, 0x0093FB, 0x0093FD, 0x009401, 0x009402, 0x009404, 0x009408, 0x009409, 0x00940D, 0x00940E, 0x00940F, 0x009415, 0x009416, 0x009417, 0x00941F, 0x00942E, 0x00942F, 0x009431, 0x009432, 0x009433, 0x009434, 0x00943B, 0x00943F, 0x00943D, 0x009443, 0x009445, 0x009448, 0x00944A,
	0x00944C, 0x009455, 0x009459, 0x00945C, 0x00945F, 0x009461, 0x009463, 0x009468, 0x00946B, 0x00946D, 0x00946E, 0x00946F, 0x009471, 0x009472, 0x009484, 0x009483, 0x009578, 0x009579, 0x00957E, 0x009584, 0x009588, 0x00958C, 0x00958D, 0x00958E, 0x00959D, 0x00959E, 0x00959F, 0x0095A1, 0x0095A6, 0x0095A9, 0x0095AB, 0x0095AC,
	0x0095B4, 0x0095B6, 0x0095BA, 0x0095BD, 0x0095BF, 0x0095C6, 0x0095C8, 0x0095C9, 0x0095CB, 0x0095D0, 0x0095D1, 0x0095D2, 0x0095D3, 0x0095D9, 0x0095DA, 0x0095DD, 0x0095DE, 0x0095DF, 0x0095E0, 0x0095E4, 0x0095E6, 0x00961D, 0x00961E, 0x009622, 0x009624, 0x009625, 0x009626, 0x00962C, 0x009631, 0x009633, 0x009637, 0x009638,
	0
};

static const unsigned map_kanji_hojo_6[] = {
	0x009639, 0x00963A, 0x00963C, 0x00963D, 0x009641, 0x009652, 0x009654, 0x009656, 0x009657, 0x009658, 0x009661, 0x00966E, 0x009674, 0x00967B, 0x00967C, 0x00967E, 0x00967F, 0x009681, 0x009682, 0x009683, 0x009684, 0x009689, 0x009691, 0x009696, 0x00969A, 0x00969D, 0x00969F, 0x0096A4, 0x0096A5, 0x0096A6, 0x0096A9, 0x0096AE,
	0x0096AF, 0x0096B3, 0x0096BA, 0x0096CA, 0x0096D2, 0x005DB2, 0x0096D8, 0x0096DA, 0x0096DD, 0x0096DE, 0x0096DF, 0x0096E9, 0x0096EF, 0x0096F1, 0x0096FA, 0x009702, 0x009703, 0x009705, 0x009709, 0x00971A, 0x00971B, 0x00971D, 0x009721, 0x009722, 0x009723, 0x009728, 0x009731, 0x009733, 0x009741, 0x009743, 0x00974A, 0x00974E,
	0x00974F, 0x009755, 0x009757, 0x009758, 0x00975A, 0x00975B, 0x009763, 0x009767, 0x00976A, 0x00976E, 0x009773, 0x009776, 0x009777, 0x009778, 0x00977B, 0x00977D, 0x00977F, 0x009780, 0x009789, 0x009795, 0x009796, 0x009797, 0x009799, 0x00979A, 0x00979E, 0x00979F, 0x0097A2, 0x0097AC, 0x0097AE, 0x0097B1, 0x0097B2, 0x0097B5,
	0x0097B6, 0x0097B8, 0x0097B9, 0x0097BA, 0x0097BC, 0x0097BE, 0x0097BF, 0x0097C1, 0x0097C4, 0x0097C5, 0x0097C7, 0x0097C9, 0x0097CA, 0x0097CC, 0x0097CD, 0x0097CE, 0x0097D0, 0x0097D1, 0x0097D4, 0x0097D7, 0x0097D8, 0x0097D9, 0x0097DD, 0x0097DE, 0x0097E0, 0x0097DB, 0x0097E1, 0x0097E4, 0x0097EF, 0x0097F1, 0x0097F4, 0x0097F7,
	0x0097F8, 0x0097FA, 0x009807, 0x00980A, 0x009819, 0x00980D, 0x00980E, 0x009814, 0x009816, 0x00981C, 0x00981E, 0x009820, 0x009823, 0x009826, 0x00982B, 0x00982E, 0x00982F, 0x009830, 0x009832, 0x009833, 0x009835, 0x009825, 0x00983E, 0x009844, 0x009847, 0x00984A, 0x009851, 0x009852, 0x009853, 0x009856, 0x009857, 0x009859,
	0x00985A, 0x009862, 0x009863, 0x009865, 0x009866, 0x00986A, 0x00986C, 0x0098AB, 0x0098AD, 0x0098AE, 0x0098B0, 0x0098B4, 0x0098B7, 0x0098B8, 0x0098BA, 0x0098BB, 0x0098BF, 0x0098C2, 0x0098C5, 0x0098C8, 0x0098CC, 0x0098E1, 0x0098E3, 0x0098E5, 0x0098E6, 0x0098E7, 0x0098EA, 0x0098F3, 0x0098F6, 0x009902, 0x009907, 0x009908,
	0x009911, 0x009915, 0x009916, 0x009917, 0x00991A, 0x00991B, 0x00991C, 0x00991F, 0x009922, 0x009926, 0x009927, 0x00992B, 0x009931, 0x009932, 0x009933, 0x009934, 0x009935, 0x009939, 0x00993A, 0x00993B, 0x00993C, 0x009940, 0x009941, 0x009946, 0x009947, 0x009948, 0x00994D, 0x00994E, 0x009954, 0x009958, 0x009959, 0x00995B,
	0x00995C, 0x00995E, 0x00995F, 0x009960, 0x00999B, 0x00999D, 0x00999F, 0x0099A6, 0x0099B0, 0x0099B1, 0x0099B2, 0x0099B5, 0x0099B9, 0x0099BA, 0x0099BD, 0x0099BF, 0x0099C3, 0x0099C9, 0x0099D3, 0x0099D4, 0x0099D9, 0x0099DA, 0x0099DC, 0x0099DE, 0x0099E7, 0x0099EA, 0x0099EB, 0x0099EC, 0x0099F0, 0x0099F4, 0x0099F5, 0x0099F9,
	0x0099FD, 0x0099FE, 0x009A02, 0x009A03, 0x009A04, 0x009A0B, 0x009A0C, 0x009A10, 0x009A11, 0x009A16, 0x009A1E, 0x009A20, 0x009A22, 0x009A23, 0x009A24, 0x009A27, 0x009A2D, 0x009A2E, 0x009A33, 0x009A35, 0x009A36, 0x009A38, 0x009A47, 0x009A41, 0x009A44, 0x009A4A, 0x009A4B, 0x009A4C, 0x009A4E, 0x009A51, 0x009A54, 0x009A56,
	0x009A5D, 0x009AAA, 0x009AAC, 0x009AAE, 0x009AAF, 0x009AB2, 0x009AB4, 0x009AB5, 0x009AB6, 0x009AB9, 0x009ABB, 0x009ABE, 0x009ABF, 0x009AC1, 0x009AC3, 0x009AC6, 0x009AC8, 0x009ACE, 0x009AD0, 0x009AD2, 0x009AD5, 0x009AD6, 0x009AD7, 0x009ADB, 0x009ADC, 0x009AE0, 0x009AE4, 0x009AE5, 0x009AE7, 0x009AE9, 0x009AEC, 0x009AF2,
	0x009AF3, 0x009AF5, 0x009AF9, 0x009AFA, 0x009AFD, 0x009AFF, 0x009B00, 0x009B01, 0x009B02, 0x009B03, 0x009B04, 0x009B05, 0x009B08, 0x009B09, 0x009B0B, 0x009B0C, 0x009B0D, 0x009B0E, 0x009B10, 0x009B12, 0x009B16, 0x009B19, 0x009B1B, 0x009B1C, 0x009B20, 0x009B26, 0x009B2B, 0x009B2D, 0x009B33, 0x009B34, 0x009B35, 0x009B37,
	0x009B39, 0x009B3A, 0x009B3D, 0x009B48, 0x009B4B, 0x009B4C, 0x009B55, 0x009B56, 0x009B57, 0x009B5B, 0x009B5E, 0x009B61, 0x009B63, 0x009B65, 0x009B66, 0x009B68, 0x009B6A, 0x009B6B, 0x009B6C, 0x009B6D, 0x009B6E, 0x009B73, 0x009B75, 0x009B77, 0x009B78, 0x009B79, 0x009B7F, 0x009B80, 0x009B84, 0x009B85, 0x009B86, 0x009B87,
	0x009B89, 0x009B8A, 0x009B8B, 0x009B8D, 0x009B8F, 0x009B90, 0x009B94, 0x009B9A, 0x009B9D, 0x009B9E, 0x009BA6, 0x009BA7, 0x009BA9, 0x009BAC, 0x009BB0, 0x009BB1, 0x009BB2, 0x009BB7, 0x009BB8, 0x009BBB, 0x009BBC, 0x009BBE, 0x009BBF, 0x009BC1, 0x009BC7, 0x009BC8, 0x009BCE, 0x009BD0, 0x009BD7, 0x009BD8, 0x009BDD, 0x009BDF,
	0x009BE5, 0x009BE7, 0x009BEA, 0x009BEB, 0x009BEF, 0x009BF3, 0x009BF7, 0x009BF8, 0x009BF9, 0x009BFA, 0x009BFD, 0x009BFF, 0x009C00, 0x009C02, 0x009C0B, 0x009C0F, 0x009C11, 0x009C16, 0x009C18, 0x009C19, 0x009C1A, 0x009C1C, 0x009C1E, 0x009C22, 0x009C23, 0x009C26, 0x009C27, 0x009C28, 0x009C29, 0x009C2A, 0x009C31, 0x009C35,
	0x009C36, 0x009C37, 0x009C3D, 0x009C41, 0x009C43, 0x009C44, 0x009C45, 0x009C49, 0x009C4A, 0x009C4E, 0x009C4F, 0x009C50, 0x009C53, 0x009C54, 0x009C56, 0x009C58, 0x009C5B, 0x009C5D, 0x009C5E, 0x009C5F, 0x009C63, 0x009C69, 0x009C6A, 0x009C5C, 0x009C6B, 0x009C68, 0x009C6E, 0x009C70, 0x009C72, 0x009C75, 0x009C77, 0x009C7B,
	0x009CE6, 0x009CF2, 0x009CF7, 0x009CF9, 0x009D0B, 0x009D02, 0x009D11, 0x009D17, 0x009D18, 0x009D1C, 0x009D1D, 0x009D1E, 0x009D2F, 0x009D30, 0x009D32, 0x009D33, 0x009D34, 0x009D3A, 0x009D3C, 0x009D45, 0x009D3D, 0x009D42, 0x009D43, 0x009D47, 0x009D4A, 0x009D53, 0x009D54, 0x009D5F, 0x009D63, 0x009D62, 0x009D65, 0x009D69,
	0x009D6A, 0x009D6B, 0x009D70, 0x009D76, 0x009D77, 0x009D7B, 0x009D7C, 0x009D7E, 0x009D83, 0x009D84, 0x009D86, 0x009D8A, 0x009D8D, 0x009D8E, 0x009D92, 0x009D93, 0x009D95, 0x009D96, 0x009D97, 0x009D98, 0x009DA1, 0x009DAA, 0x009DAC, 0x009DAE, 0x009DB1, 0x009DB5, 0x009DB9, 0x009DBC, 0x009DBF, 0x009DC3, 0x009DC7, 0x009DC9,
	0x009DCA, 0x009DD4, 0x009DD5, 0x009DD6, 0x009DD7, 0x009DDA, 0x009DDE, 0x009DDF, 0x009DE0, 0x009DE5, 0x009DE7, 0x009DE9, 0x009DEB, 0x009DEE, 0x009DF0, 0x009DF3, 0x009DF4, 0x009DFE, 0x009E0A, 0x009E02, 0x009E07, 0x009E0E, 0x009E10, 0x009E11, 0x009E12, 0x009E15, 0x009E16, 0x009E19, 0x009E1C, 0x009E1D, 0x009E7A, 0x009E7B,
	0x009E7C, 0x009E80, 0x009E82, 0x009E83, 0x009E84, 0x009E85, 0x009E87, 0x009E8E, 0x009E8F, 0x009E96, 0x009E98, 0x009E9B, 0x009E9E, 0x009EA4, 0x009EA8, 0x009EAC, 0x009EAE, 0x009EAF, 0x009EB0, 0x009EB3, 0x009EB4, 0x009EB5, 0x009EC6, 0x009EC8, 0x009ECB, 0x009ED5, 0x009EDF, 0x009EE4, 0x009EE7, 0x009EEC, 0x009EED, 0x009EEE,
	0x009EF0, 0x009EF1, 0x009EF2, 0x009EF5, 0x009EF8, 0x009EFF, 0x009F02, 0x009F03, 0x009F09, 0x009F0F, 0x009F10, 0x009F11, 0x009F12, 0x009F14, 0x009F16, 0x009F17, 0x009F19, 0x009F1A, 0x009F1B, 0x009F1F, 0x009F22, 0x009F26, 0x009F2A, 0x009F2B, 0x009F2F, 0x009F31, 0x009F32, 0x009F34, 0x009F37, 0x009F39, 0x009F3A, 0x009F3C,
	0x009F3D, 0x009F3F, 0x009F41, 0x009F43, 0x009F44, 0x009F45, 0x009F46, 0x009F47, 0x009F53, 0x009F55, 0x009F56, 0x009F57, 0x009F58, 0x009F5A, 0x009F5D, 0x009F5E, 0x009F68, 0x009F69, 0x009F6D, 0x009F6E, 0x009F6F, 0x009F70, 0x009F71, 0x009F73, 0x009F75, 0x009F7A, 0x009F7D, 0x009F8F, 0x009F90, 0x009F91, 0x009F92, 0x009F94,
	0x009F96, 0x009F97, 0x009F9E, 0x009FA1, 0x009FA2, 0x009FA3, 0x009FA5,
	0
};

// GBK/1
static const unsigned map_symbol_gbk1[] = {
	0x0002C9, 0x002236, 0x002211, 0x00220F, 0x002237, 0x002299, 0x00224C, 0x00226E, 0x00226F, 0x002264, 0x002265, 0x002488, 0x002489, 0x00248A, 0x00248B, 0x00248C,
	0x00248D, 0x00248E, 0x00248F, 0x002490, 0x002491, 0x002492, 0x002493, 0x002494, 0x002495, 0x002496, 0x002497, 0x002498, 0x002499, 0x00249A, 0x00249B, 0x002474,
	0x002475, 0x002476, 0x002477, 0x002478, 0x002479, 0x00247A, 0x00247B, 0x00247C, 0x00247D, 0x00247E, 0x00247F, 0x002480, 0x002481, 0x002482, 0x002483, 0x002484,
	0x002485, 0x002486, 0x002487, 0x003220, 0x003221, 0x003222, 0x003223, 0x003224, 0x003225, 0x003226, 0x003227, 0x003228, 0x003229, 0x003105, 0x003106, 0x003107,
	0x003108, 0x003109, 0x00310A, 0x00310B, 0x00310C, 0x00310D, 0x00310E, 0x00310F, 0x003110, 0x003111, 0x003112, 0x003113, 0x003114, 0x003115, 0x003116, 0x003117,
	0x003118, 0x003119, 0x00311A, 0x00311B, 0x00311C, 0x00311D, 0x00311E, 0x00311F, 0x003120, 0x003121, 0x003122, 0x003123, 0x003124, 0x003125, 0x003126, 0x003127,
	0x003128, 0x003129, 0x002504, 0x002505, 0x002506, 0x002507, 0x002508, 0x002509, 0x00250A, 0x00250B, 0x00250D, 0x00250E, 0x002511, 0x002512, 0x002515, 0x002516,
	0x002519, 0x00251A, 0x00251E, 0x00251F, 0x002521, 0x002522, 0x002526, 0x002527, 0x002529, 0x00252A, 0x00252D, 0x00252E, 0x002531, 0x002532, 0x002535, 0x002536,
	0x002539, 0x00253A, 0x00253D, 0x00253E, 0x002540, 0x002541, 0x002543, 0x002544, 0x002545, 0x002546, 0x002547, 0x002548, 0x002549, 0x00254A,
	0
};

// GBK/2
static const unsigned map_kanji_gbk2_1[] = {
	0x007691, 0x00853C, 0x007231, 0x0080AE, 0x007FF1, 0x008884, 0x00575D, 0x007F62, 0x006446, 0x008D25, 0x009881, 0x00529E, 0x007ECA, 0x007ED1, 0x009551, 0x008C24, 0x009971, 0x0062A5, 0x009C8D, 0x008F88, 0x008D1D, 0x0094A1, 0x0072C8, 0x005907, 0x0060EB, 0x007EF7, 0x00752D, 0x006CF5, 0x008E66, 0x006BD5, 0x006BD9, 0x005E01,
	0x0095ED, 0x008FB9, 0x007F16, 0x008D2C, 0x0053D8, 0x008FA9, 0x008FAB, 0x006807, 0x009CD6, 0x00522B, 0x00762A, 0x006FD2, 0x006EE8, 0x005BBE, 0x006448, 0x00997C, 0x0062E8, 0x0094B5, 0x0094C2, 0x009A73, 0x008865, 0x008D22, 0x008E29, 0x0060ED, 0x00707F, 0x0082CD, 0x008231, 0x004ED3, 0x006CA7, 0x005395, 0x004FA7, 0x006D4B,
	0x005C42, 0x0067E5, 0x0078B4, 0x008BE7, 0x006400, 0x0063BA, 0x00998B, 0x008C17, 0x007F20, 0x0094F2, 0x004EA7, 0x009610, 0x0098A4, 0x00573A, 0x005C1D, 0x00957F, 0x00507F, 0x0080A0, 0x007545, 0x00949E, 0x008F66, 0x005F7B, 0x005C18, 0x009648, 0x00886C, 0x0060E9, 0x008BDA, 0x009A8B, 0x008FDF, 0x009A70, 0x009F7F, 0x0070BD,
	0x005BA0, 0x007EF8, 0x007785, 0x009504, 0x0096CF, 0x007840, 0x0050A8, 0x005904, 0x004F20, 0x0075AE, 0x0095EF, 0x00521B, 0x009524, 0x007EAF, 0x007EF0, 0x008BCD, 0x008D50, 0x00806A, 0x004E1B, 0x008E7F, 0x007BE1, 0x007A9C, 0x009519, 0x005E26, 0x008D37, 0x005355, 0x0090F8, 0x0063B8, 0x008BDE, 0x005F39, 0x006321, 0x008361,
	0x006363, 0x005C9B, 0x005BFC, 0x009093, 0x00654C, 0x006DA4, 0x009012, 0x007F14, 0x0098A0, 0x0057AB, 0x007535, 0x0060E6, 0x007889, 0x0053FC, 0x009493, 0x008C03, 0x008C0D, 0x0076EF, 0x009489, 0x009876, 0x00952D, 0x008BA2, 0x004E22, 0x004E1C, 0x0052A8, 0x00680B, 0x0051BB, 0x00728A, 0x008BFB, 0x008D4C, 0x009540, 0x00953B,
	0x007F0E, 0x005151, 0x00961F, 0x005BF9, 0x00987F, 0x00949D, 0x00593A, 0x00579B, 0x008EB2, 0x006735, 0x008DFA, 0x005241, 0x009E45, 0x00989D, 0x008BB9, 0x006076, 0x00997F, 0x005C14, 0x009975, 0x008D30, 0x0053D1, 0x007F5A, 0x009600, 0x0073D0, 0x0077FE, 0x009492, 0x0070E6, 0x008D29, 0x00996D, 0x008BBF, 0x007EBA, 0x0098DE,
	0x008BFD, 0x005E9F, 0x008D39, 0x007EB7, 0x00575F, 0x00594B, 0x006124, 0x007CAA, 0x0067AB, 0x00950B, 0x0098CE, 0x0075AF, 0x0051AF, 0x007F1D, 0x008BBD, 0x0051E4, 0x008F90, 0x00629A, 0x008F85, 0x008D4B, 0x008D1F, 0x008BA3, 0x005987, 0x007F1A, 0x008BE5, 0x009499, 0x006E89, 0x0079C6, 0x008D63, 0x005188, 0x00521A, 0x0094A2,
	0x007EB2, 0x005C97, 0x009550, 0x006401, 0x009E3D, 0x009601, 0x0094EC, 0x007ED9, 0x0057C2, 0x009F9A, 0x005BAB, 0x008D21, 0x0094A9, 0x008D2D, 0x00591F, 0x0086CA, 0x00987E, 0x005250, 0x0089C2, 0x009986, 0x0060EF, 0x008D2F, 0x0089C4, 0x005F52, 0x009F9F, 0x0095FA, 0x008F68, 0x008BE1, 0x008D35, 0x00523D, 0x008F8A, 0x006EDA,
	0x009505, 0x008FC7, 0x009A87, 0x0097E9, 0x00710A, 0x006C49, 0x00568E, 0x009602, 0x009E64, 0x008D3A, 0x008F70, 0x009E3F, 0x007EA2, 0x0058F6, 0x00552C, 0x0062A4, 0x006237, 0x0054D7, 0x00534E, 0x008BDD, 0x006000, 0x006B22, 0x0073AF, 0x008FD8, 0x007F13, 0x006362, 0x005524, 0x0075EA, 0x007115, 0x006DA3, 0x008C0E, 0x006325,
	0x008F89, 0x006BC1, 0x008D3F, 0x0079FD, 0x0070E9, 0x006C47, 0x008BB3, 0x008BF2, 0x007ED8, 0x008364, 0x006D51, 0x0083B7, 0x008D27, 0x007978, 0x0051FB, 0x0079EF, 0x009965, 0x008BA5, 0x009E21, 0x0059EC, 0x007EE9, 0x007F09, 0x008F91, 0x007EA7, 0x006324, 0x0084DF, 0x005242, 0x006D4E, 0x008BA1, 0x008BB0, 0x009645, 0x007EE7,
	0x007EAA, 0x005939, 0x00835A, 0x00988A, 0x008D3E, 0x0094BE, 0x009A7E, 0x006B7C, 0x0076D1, 0x00575A, 0x007B3A, 0x0095F4, 0x008270, 0x007F04, 0x008327, 0x0068C0, 0x007877, 0x0062E3, 0x006361, 0x007B80, 0x004FED, 0x0069DB, 0x009274, 0x008D31, 0x0089C1, 0x00952E, 0x008230, 0x005251, 0x00996F, 0x006E10, 0x006E85, 0x006DA7,
	0x006D46, 0x006868, 0x005956, 0x008BB2, 0x009171, 0x006D47, 0x009A84, 0x005A07, 0x006405, 0x0094F0, 0x0077EB, 0x004FA5, 0x00997A, 0x007F34, 0x007EDE, 0x008F7F, 0x008F83, 0x009636, 0x008282, 0x006D01, 0x007ED3, 0x008BEB, 0x007D27, 0x009526, 0x004EC5, 0x008C28, 0x008FDB, 0x0052B2, 0x008346, 0x009CB8, 0x007ECF, 0x009888,
	0x00955C, 0x0075C9, 0x007ADE, 0x0051C0, 0x007EA0, 0x009A79, 0x004E3E, 0x00952F, 0x005267, 0x009E43, 0x007EE2, 0x0089C9, 0x008BC0, 0x007EDD, 0x0094A7, 0x00519B, 0x009A8F, 0x0051EF, 0x009897, 0x008BFE, 0x005543, 0x0057A6, 0x006073, 0x0062A0, 0x005E93, 0x0088E4, 0x0057AE, 0x00630E, 0x005757, 0x004FA9, 0x005BBD, 0x0077FF,
	0x0065F7, 0x004E8F, 0x005CBF, 0x007AA5, 0x009988, 0x006E83, 0x006269, 0x009614, 0x008D56, 0x0084DD, 0x00680F, 0x0062E6, 0x007BEE, 0x009611, 0x005170, 0x006F9C, 0x008C30, 0x0063FD, 0x0089C8, 0x0061D2, 0x007F06, 0x0070C2, 0x006EE5, 0x00635E, 0x0052B3, 0x006D9D, 0x004E50, 0x00956D, 0x005792, 0x007C7B, 0x007BF1, 0x009CA4,
	0x004E3D, 0x005389, 0x00783E, 0x005386, 0x006CA5, 0x004FE9, 0x008054, 0x0083B2, 0x008FDE, 0x009570, 0x006D9F, 0x00655B, 0x008138, 0x0094FE, 0x0070BC, 0x007EC3, 0x008F86, 0x008C05, 0x007597, 0x008FBD, 0x006482, 0x009563, 0x00730E, 0x004E34, 0x0090BB, 0x009CDE, 0x008D41, 0x009F84, 0x0094C3, 0x009886, 0x00998F, 0x009F99,
	0x00804B, 0x005499, 0x007B3C, 0x005784, 0x0062E2, 0x009647, 0x006402, 0x007BD3, 0x005362, 0x009885, 0x005E90, 0x0063B3, 0x005364, 0x00864F, 0x009C81, 0x008D42, 0x005F55, 0x009646, 0x009A74, 0x005415, 0x0094DD, 0x004FA3, 0x007F15, 0x008651, 0x006C2F, 0x006EE4, 0x007EFF, 0x005CE6, 0x00631B, 0x005B6A, 0x006EE6, 0x0062A1,
	0x008F6E, 0x004F26, 0x004ED1, 0x006CA6, 0x007EB6, 0x008BBA, 0x00841D, 0x007F57, 0x00903B, 0x009523, 0x007BA9, 0x009AA1, 0x009A86, 0x007EDC, 0x005988, 0x00739B, 0x007801, 0x008682, 0x009A6C, 0x009A82, 0x005417, 0x004E70, 0x005356, 0x008FC8, 0x007792, 0x009992, 0x006EE1, 0x008C29, 0x00951A, 0x0094C6, 0x008D38, 0x009176,
	0x009541, 0x0095E8, 0x0095F7, 0x004EEC, 0x009530, 0x00919A, 0x008C1C, 0x0089C5, 0x005E42, 0x007EF5, 0x007F05, 0x005E99, 0x00706D, 0x0060AF, 0x0095FD, 0x009E23, 0x0094ED, 0x008C2C, 0x008C0B, 0x004EA9, 0x005450, 0x0094A0, 0x007EB3, 0x006C16, 0x0096BE, 0x006320, 0x008111, 0x00607C, 0x0095F9, 0x009981, 0x0062DF, 0x00817B,
	0x0064B5, 0x00917F, 0x009E1F, 0x008042, 0x00556E, 0x00954A, 0x00954D, 0x0067E0, 0x0072DE, 0x0062E7, 0x006CDE, 0x0094AE, 0x007EBD, 0x008113, 0x006D53, 0x00519C, 0x00759F, 0x008BFA, 0x009E25, 0x006CA4, 0x00556A, 0x008DB4, 0x0076D8, 0x004E53, 0x005E9E, 0x00802A, 0x005478, 0x008D54, 0x0055B7, 0x009E4F, 0x009A97, 0x0098D8,
	0x009891, 0x008D2B, 0x004E52, 0x008BC4, 0x006CFC, 0x009887, 0x0094FA, 0x008C31, 0x006C8F, 0x008110, 0x009F50, 0x009A91, 0x005C82, 0x008BAB, 0x007275, 0x006266, 0x00948E, 0x0094C5, 0x007B7E, 0x008C26, 0x0094B1, 0x0094B3, 0x008C34, 0x005811, 0x0067AA, 0x00545B, 0x005899, 0x008537, 0x005F3A, 0x0062A2, 0x009539, 0x006865,
	0x004E54, 0x004FA8, 0x007FD8, 0x007A8D, 0x0094A6, 0x004EB2, 0x008F7B, 0x006C22, 0x00503E, 0x006C30, 0x009877, 0x008BF7, 0x005E86, 0x00743C, 0x007A77, 0x008D8B, 0x009A71, 0x009F8B, 0x0098A7, 0x006743, 0x00919B, 0x00529D, 0x009E4A, 0x008BA9, 0x009976, 0x006270, 0x007ED5, 0x0070ED, 0x0097E7, 0x008BA4, 0x007EAB, 0x007ED2,
	0x008F6F, 0x009510, 0x0095F0, 0x006DA6, 0x008428, 0x009CC3, 0x008D5B, 0x0053C1, 0x004F1E, 0x004E27, 0x009A9A, 0x00626B, 0x006DA9, 0x006740, 0x007EB1, 0x005565, 0x007B5B, 0x005220, 0x0095EA, 0x009655, 0x008D61, 0x007F2E, 0x005892, 0x004F24, 0x008D4F, 0x0070E7, 0x007ECD, 0x008D4A, 0x006444, 0x006151, 0x008BBE, 0x007EC5,
	0x005BA1, 0x005A76, 0x0080BE, 0x007EF3, 0x0080DC, 0x005E08, 0x0072EE, 0x008BD7, 0x0065F6, 0x008680, 0x005B9E, 0x008BC6, 0x009A76, 0x0052BF, 0x0091CA, 0x009970, 0x0089C6, 0x008BD5, 0x00517D, 0x008F93, 0x004E66, 0x008D4E, 0x00672F, 0x006811, 0x007AD6, 0x005E05, 0x008C01, 0x00987A, 0x008BF4, 0x007855, 0x0070C1, 0x004E1D,
	0x009972, 0x008038, 0x006002, 0x009882, 0x008BBC, 0x008BF5, 0x0064DE, 0x0082CF, 0x008BC9, 0x008083, 0x00867D, 0x007EE5, 0x005C81, 0x005B59, 0x00635F, 0x007F29, 0x007410, 0x009501, 0x00736D, 0x00631E, 0x00915E, 0x006001, 0x00644A, 0x008D2A, 0x00762B, 0x006EE9, 0x00575B, 0x008C2D, 0x008C08, 0x0053F9, 0x006C64, 0x0070EB,
	0x007EE6, 0x008BA8, 0x00817E, 0x008A8A, 0x009511, 0x009898, 0x005C49, 0x008D34, 0x0094C1, 0x005385, 0x0070C3, 0x00916E, 0x0094DC, 0x006345, 0x007EDF, 0x005077, 0x005934, 0x0079C3, 0x0056FE, 0x0056E2, 0x009893, 0x008715, 0x009E35, 0x009A6E, 0x009A7C, 0x00692D, 0x00987D, 0x0070F7, 0x0097E6, 0x008FDD, 0x0056F4, 0x004E3A,
	0x006F4D, 0x007EF4, 0x0082C7, 0x004F1F, 0x004F2A, 0x007EAC, 0x008C13, 0x00536B, 0x0095FB, 0x007EB9, 0x007A33, 0x0095EE, 0x0055E1, 0x00631D, 0x008717, 0x006DA1, 0x007A9D, 0x00545C, 0x0094A8, 0x004E4C, 0x006C61, 0x008BEC, 0x00829C, 0x00575E, 0x0096FE, 0x0052A1, 0x008BEF, 0x0077FD, 0x009521, 0x00727A, 0x0070EF, 0x0088AD,
	0x004E60, 0x0094E3, 0x00620F, 0x007EC6, 0x00867E, 0x008F96, 0x009528, 0x009C9C, 0x007EA4, 0x008D24, 0x008854, 0x0095F2, 0x00663E, 0x009669, 0x0073B0, 0x0053BF, 0x009985, 0x005BAA, 0x007EBF, 0x0053A2, 0x009576, 0x004E61, 0x008BE6, 0x009879, 0x008427, 0x0056A3, 0x009500, 0x006653, 0x005578, 0x00534F, 0x0080C1, 0x008C10,
	0x008C22, 0x00950C, 0x005174, 0x009508, 0x007EE3, 0x00987B, 0x008BB8, 0x007EEA, 0x007EED, 0x008F69, 0x0060AC, 0x009009, 0x007663, 0x007EDA, 0x0052CB, 0x008BE2, 0x005BFB, 0x009A6F, 0x008BAD, 0x008BAF, 0x00900A, 0x00538B, 0x009E26, 0x009E2D, 0x0054D1, 0x004E9A, 0x008BB6, 0x009609, 0x0076D0, 0x004E25, 0x00989C, 0x00960E,
	0x008273, 0x00538C, 0x00781A, 0x008C1A, 0x009A8C, 0x009E2F, 0x006768, 0x00626C, 0x0075A1, 0x006837, 0x006447, 0x005C27, 0x008C23, 0x00836F, 0x007237, 0x009875, 0x004E1A, 0x0094F1, 0x009890, 0x009057, 0x004EEA, 0x008681, 0x00827A, 0x004EBF, 0x005FC6, 0x004E49, 0x008BE3, 0x008BAE, 0x008C0A, 0x008BD1, 0x007ECE, 0x00836B,
	0x0094F6, 0x00996E, 0x009690, 0x006A31, 0x005A74, 0x009E70, 0x005E94, 0x007F28, 0x0083B9, 0x008424, 0x008425, 0x008367, 0x008747, 0x008D62, 0x009896, 0x0054DF, 0x0062E5, 0x0075C8, 0x004F18, 0x005FE7, 0x0090AE, 0x0094C0, 0x008BF1, 0x006986, 0x008206, 0x004FDE, 0x009C7C, 0x006E14, 0x005A31, 0x005C7F, 0x008BED, 0x0055BB,
	0x0072F1, 0x009884, 0x009A6D, 0x009E33, 0x008F95, 0x005458, 0x005706, 0x007F18, 0x008FDC, 0x007EA6, 0x008DC3, 0x0094A5, 0x009605, 0x0090E7, 0x009668, 0x008574, 0x00915D, 0x006655, 0x007838, 0x006742, 0x008F7D, 0x006512, 0x006682, 0x008D5E, 0x008D43, 0x00810F, 0x0051FF, 0x0067A3, 0x008D23, 0x0062E9, 0x005219, 0x006CFD,
	0x008D3C, 0x008D60, 0x0055B3, 0x008F67, 0x0094E1, 0x0095F8, 0x006805, 0x008BC8, 0x00658B, 0x00503A, 0x0076CF, 0x0065A9, 0x008F97, 0x005D2D, 0x006808, 0x006218, 0x007EFD, 0x005F20, 0x006DA8, 0x005E10, 0x008D26, 0x0080C0, 0x008D75, 0x0086F0, 0x008F99, 0x009517, 0x008FD9, 0x008D1E, 0x009488, 0x004FA6, 0x008BCA, 0x009547,
	0
};

static const unsigned map_kanji_gbk2_2[] = {
	0x009635, 0x006323, 0x007741, 0x0072F0, 0x005E27, 0x0090D1, 0x008BC1, 0x007EC7, 0x00804C, 0x006267, 0x00503C, 0x007EB8, 0x00631A, 0x0063B7, 0x005E1C, 0x008D28, 0x00949F, 0x007EC8, 0x0080BF, 0x008BCC, 0x008F74, 0x0076B1, 0x009AA4, 0x008BF8, 0x008BDB, 0x0070DB, 0x008D2E, 0x0094F8, 0x009A7B, 0x004E13, 0x007816, 0x008F6C,
	0x008D5A, 0x006869, 0x005986, 0x009525, 0x008D58, 0x005760, 0x007F00, 0x008C06, 0x006D4A, 0x005179, 0x008D44, 0x006E0D, 0x007EFC, 0x00603B, 0x007EB5, 0x0090B9, 0x0063CD, 0x008BC5, 0x007EC4, 0x0094BB, 0x005B6C, 0x004E93, 0x008288, 0x00556C, 0x00538D, 0x0053A3, 0x009765, 0x008D5D, 0x005326, 0x00532E, 0x008D5C, 0x00522D,
	0x00523F, 0x005240, 0x005281, 0x005290, 0x004EC9, 0x004F1B, 0x004F22, 0x004F25, 0x004F27, 0x004F2B, 0x004F67, 0x004F32, 0x004F74, 0x004FAA, 0x004FAC, 0x004FE8, 0x004FEA, 0x00507E, 0x00507B, 0x0050A5, 0x0050A7, 0x0050A9, 0x0050EC, 0x006C3D, 0x004F65, 0x006C46, 0x007C74, 0x009EC9, 0x005181, 0x0051EB, 0x005156, 0x00886E,
	0x004EB5, 0x008114, 0x0051C7, 0x008BA0, 0x008BA6, 0x008BA7, 0x008BAA, 0x008BB4, 0x008BB5, 0x008BB7, 0x008BC2, 0x008BC3, 0x008BCB, 0x008BCF, 0x008BCE, 0x008BD2, 0x008BD3, 0x008BD4, 0x008BD6, 0x008BD8, 0x008BD9, 0x008BDC, 0x008BDF, 0x008BE0, 0x008BE4, 0x008BE8, 0x008BE9, 0x008BEE, 0x008BF0, 0x008BF3, 0x008BF6, 0x008BF9,
	0x008BFC, 0x008BFF, 0x008C00, 0x008C02, 0x008C04, 0x008C07, 0x008C0C, 0x008C0F, 0x008C11, 0x008C12, 0x008C14, 0x008C15, 0x008C16, 0x008C19, 0x008C1B, 0x008C18, 0x008C1D, 0x008C1F, 0x008C20, 0x008C21, 0x008C25, 0x008C27, 0x008C2A, 0x008C2B, 0x008C2E, 0x008C2F, 0x008C32, 0x008C33, 0x008C35, 0x008C36, 0x009649, 0x009667,
	0x00909D, 0x0090AC, 0x0090BA, 0x0090CF, 0x0090D0, 0x0090D3, 0x0090E6, 0x00520D, 0x005942, 0x0052A2, 0x0051FC, 0x005DEF, 0x005739, 0x00575C, 0x005785, 0x005786, 0x005776, 0x0057AD, 0x0057B2, 0x0057B4, 0x0057D8, 0x0057DA, 0x0057D9, 0x0057EF, 0x005844, 0x005865, 0x00586C, 0x00589A, 0x00828F, 0x008297, 0x0082C8, 0x0082CA,
	0x0082CB, 0x0082CC, 0x0082C1, 0x0082C4, 0x0082CE, 0x00830F, 0x0082D8, 0x008311, 0x00831A, 0x008314, 0x008315, 0x00835B, 0x00835C, 0x00835E, 0x00835F, 0x008360, 0x008366, 0x008365, 0x008368, 0x008369, 0x00836C, 0x00836A, 0x00836D, 0x00836E, 0x0083B0, 0x0083B3, 0x0083B4, 0x0083B6, 0x0083B8, 0x0083BA, 0x0083BC, 0x005807,
	0x008418, 0x00841C, 0x008426, 0x008487, 0x008488, 0x008489, 0x00848C, 0x00848E, 0x0084E6, 0x0084E0, 0x0084E5, 0x0084E3, 0x008538, 0x008539, 0x00853A, 0x00853B, 0x008572, 0x0085D3, 0x00593C, 0x005941, 0x005C25, 0x005C34, 0x00626A, 0x00629F, 0x006322, 0x0063BC, 0x0063FF, 0x0063DE, 0x006445, 0x006441, 0x00640B, 0x00641B,
	0x00640C, 0x006421, 0x00645E, 0x006484, 0x006496, 0x0064B7, 0x0064B8, 0x0064BA, 0x0064C0, 0x006525, 0x007519, 0x00535F, 0x0053FD, 0x0053FB, 0x005416, 0x005406, 0x00544B, 0x005452, 0x005453, 0x005454, 0x005456, 0x005457, 0x005459, 0x005423, 0x005432, 0x005482, 0x005494, 0x00549A, 0x00549B, 0x00549D, 0x0054D0, 0x0054B4,
	0x0054D2, 0x0054D3, 0x0054D4, 0x005472, 0x0054A3, 0x0054D5, 0x0054CC, 0x0054D9, 0x0054DA, 0x0054DC, 0x0054DD, 0x0054CF, 0x0054DE, 0x00551B, 0x0054E7, 0x005520, 0x005522, 0x005523, 0x005511, 0x005527, 0x005567, 0x0055B5, 0x00556D, 0x005555, 0x00553F, 0x005537, 0x005575, 0x005576, 0x005577, 0x005530, 0x0055B1, 0x0055B9,
	0x0055D6, 0x0055BD, 0x0055BE, 0x0055EA, 0x00561F, 0x0055EB, 0x0055EC, 0x0055E6, 0x0055EF, 0x0055E5, 0x0055F2, 0x0055F3, 0x0055CD, 0x0055E8, 0x0055F5, 0x008F94, 0x00561E, 0x00560C, 0x005601, 0x005624, 0x005623, 0x005600, 0x005627, 0x00562D, 0x005658, 0x005657, 0x005659, 0x00565C, 0x00567B, 0x00567C, 0x005693, 0x0056AF,
	0x0056D4, 0x0056F5, 0x005E0F, 0x005E31, 0x005E3B, 0x005E3C, 0x005C8D, 0x005C96, 0x005C98, 0x005C99, 0x005C9A, 0x005C9C, 0x005CBD, 0x005CC1, 0x005CC4, 0x005CE4, 0x005CE5, 0x005D02, 0x005D03, 0x005D2E, 0x005D1E, 0x005D58, 0x005D3E, 0x005D5B, 0x005D5D, 0x005D6B, 0x005D74, 0x005DC5, 0x005F95, 0x0072B7, 0x0072B8, 0x0072CD,
	0x0072E8, 0x0072EF, 0x0072F2, 0x007301, 0x007303, 0x007321, 0x007315, 0x007338, 0x00737E, 0x009963, 0x009967, 0x009968, 0x009969, 0x00996A, 0x00996B, 0x00996C, 0x009974, 0x009977, 0x00997D, 0x009980, 0x009984, 0x009987, 0x00998A, 0x00998D, 0x009990, 0x009991, 0x009993, 0x009994, 0x009995, 0x005E91, 0x008D53, 0x005EEA,
	0x005FCF, 0x006003, 0x006004, 0x005FFE, 0x006005, 0x006006, 0x00603F, 0x006078, 0x006079, 0x00607B, 0x00607A, 0x00607D, 0x0060AD, 0x0060EC, 0x006126, 0x006123, 0x0061B7, 0x0061D4, 0x0095E9, 0x0095EB, 0x0095F1, 0x0095F3, 0x0095F5, 0x0095F6, 0x0095FC, 0x0095FE, 0x009603, 0x009604, 0x009606, 0x009608, 0x00960A, 0x00960B,
	0x00960C, 0x00960D, 0x00960F, 0x009612, 0x009615, 0x009616, 0x009617, 0x009619, 0x00961A, 0x006CA3, 0x006C69, 0x006CA9, 0x006CF7, 0x006CF8, 0x006CFA, 0x006CFE, 0x006D43, 0x006D48, 0x006D4D, 0x006D4F, 0x006D52, 0x006D54, 0x006D9E, 0x006DA0, 0x006E0E, 0x006E11, 0x006E16, 0x006E0C, 0x006E86, 0x006EDF, 0x006EE0, 0x006EE2,
	0x006ED7, 0x006F46, 0x006F47, 0x006F24, 0x006F4B, 0x006F89, 0x006F72, 0x006FD1, 0x007035, 0x00704F, 0x009A9E, 0x008FF3, 0x009026, 0x005C66, 0x005F2A, 0x0059A9, 0x0059AA, 0x0059AB, 0x00599E, 0x0059D7, 0x005A05, 0x005A06, 0x005A08, 0x0059D8, 0x0059F9, 0x005A32, 0x005A34, 0x005A4A, 0x005A75, 0x0080EC, 0x005AAA, 0x005AEB,
	0x005AD2, 0x005AD4, 0x005AF1, 0x005AD8, 0x005B37, 0x005C15, 0x005C1C, 0x005B53, 0x005B62, 0x009A75, 0x009A77, 0x009A78, 0x009A7A, 0x009A7F, 0x009A7D, 0x009A80, 0x009A81, 0x009A85, 0x009A88, 0x009A8A, 0x009A90, 0x009A92, 0x009A93, 0x009A96, 0x009A98, 0x009A9B, 0x009A9C, 0x009A9D, 0x009A9F, 0x009AA0, 0x009AA2, 0x009AA3,
	0x009AA5, 0x009AA7, 0x007E9F, 0x007EA1, 0x007EA3, 0x007EA5, 0x007EA8, 0x007EA9, 0x007EAD, 0x007EB0, 0x007EBE, 0x007EC0, 0x007EC1, 0x007EC2, 0x007EC9, 0x007ECB, 0x007ECC, 0x007ED0, 0x007ED4, 0x007ED7, 0x007EDB, 0x007EE0, 0x007EE1, 0x007EE8, 0x007EEB, 0x007EEE, 0x007EEF, 0x007EF1, 0x007EF2, 0x007F0D, 0x007EF6, 0x007EFA,
	0x007EFB, 0x007EFE, 0x007F01, 0x007F02, 0x007F03, 0x007F07, 0x007F08, 0x007F0B, 0x007F0C, 0x007F0F, 0x007F11, 0x007F12, 0x007F17, 0x007F19, 0x007F1C, 0x007F1B, 0x007F1F, 0x007F21, 0x007F22, 0x007F23, 0x007F24, 0x007F25, 0x007F26, 0x007F27, 0x007F2A, 0x007F2B, 0x007F2C, 0x007F2D, 0x007F2F, 0x007F30, 0x007F31, 0x007F32,
	0x007F33, 0x007F35, 0x007391, 0x0073AE, 0x0073D1, 0x00987C, 0x0073BA, 0x0073F2, 0x00740F, 0x007477, 0x00748E, 0x0074D2, 0x0097EA, 0x0097EB, 0x0097EC, 0x006769, 0x0067A5, 0x0067A7, 0x0067A8, 0x00679E, 0x0067AD, 0x006809, 0x00680A, 0x00680C, 0x006800, 0x00680E, 0x0067FD, 0x006860, 0x006861, 0x006862, 0x006864, 0x006866,
	0x0068C2, 0x00691F, 0x006920, 0x006924, 0x006984, 0x006987, 0x006988, 0x006989, 0x0069D4, 0x0069DF, 0x0069E0, 0x006A2F, 0x006A18, 0x006A65, 0x006A79, 0x006A7C, 0x006AA9, 0x006B87, 0x006B92, 0x006B93, 0x006B9A, 0x006BA1, 0x008F6B, 0x008F6D, 0x008F71, 0x008F72, 0x008F73, 0x008F75, 0x008F76, 0x008F78, 0x008F77, 0x008F79,
	0x008F7A, 0x008F7C, 0x008F7E, 0x008F81, 0x008F82, 0x008F84, 0x008F87, 0x008F8B, 0x008F8D, 0x008F8E, 0x008F8F, 0x008F98, 0x008F9A, 0x008ECE, 0x00620B, 0x006217, 0x006225, 0x006224, 0x00622C, 0x00750F, 0x0065EE, 0x0065EF, 0x006619, 0x006654, 0x006656, 0x0066A7, 0x008D32, 0x008D33, 0x008D36, 0x008D3B, 0x008D3D, 0x008D40,
	0x008D45, 0x008D46, 0x008D48, 0x008D49, 0x008D47, 0x008D4D, 0x008D55, 0x008D59, 0x0089C7, 0x0089CA, 0x0089CB, 0x0089CC, 0x0089CE, 0x0089CF, 0x0089D0, 0x0089D1, 0x00729F, 0x007266, 0x00728B, 0x00728F, 0x0063B0, 0x006BEA, 0x006BFD, 0x006BF5, 0x006C07, 0x006C06, 0x006C15, 0x006C18, 0x006C1A, 0x006C21, 0x006C29, 0x006C2A,
	0x006C32, 0x00724D, 0x00809F, 0x0080BC, 0x00670A, 0x0080BD, 0x0080B7, 0x0080E7, 0x0080E8, 0x0080E9, 0x0080EA, 0x0080C2, 0x006710, 0x0080EB, 0x00810E, 0x008112, 0x008136, 0x008132, 0x008148, 0x008159, 0x00815A, 0x00817C, 0x00817D, 0x008191, 0x0081AA, 0x0081CC, 0x0081A6, 0x0098D1, 0x0098D2, 0x0098D3, 0x0098D5, 0x0098D9,
	0x0098DA, 0x006BC2, 0x009F51, 0x006593, 0x007080, 0x00709C, 0x00709D, 0x0070C0, 0x0070E8, 0x007110, 0x007113, 0x007116, 0x007173, 0x007172, 0x007178, 0x00717A, 0x007198, 0x0071B5, 0x007118, 0x00796F, 0x00799A, 0x00603C, 0x0060AB, 0x0061D1, 0x006206, 0x008080, 0x006CF6, 0x0077F6, 0x007800, 0x007817, 0x007818, 0x00781C,
	0x007839, 0x00783B, 0x00781F, 0x00783C, 0x007829, 0x00786D, 0x007856, 0x007857, 0x007850, 0x00789B, 0x00789C, 0x0078B9, 0x0078A5, 0x0078D9, 0x007905, 0x007913, 0x007924, 0x009F9B, 0x00770D, 0x007750, 0x007751, 0x007743, 0x007583, 0x007F74, 0x007F81, 0x009485, 0x009486, 0x009487, 0x00948B, 0x00948A, 0x00948C, 0x00948D,
	0x00948F, 0x009490, 0x009494, 0x009497, 0x009495, 0x00949A, 0x00949B, 0x00949C, 0x0094A3, 0x0094A4, 0x0094AB, 0x0094AA, 0x0094AD, 0x0094AC, 0x0094AF, 0x0094B0, 0x0094B2, 0x0094B4, 0x0094B6, 0x0094B7, 0x0094B8, 0x0094B9, 0x0094BA, 0x0094BC, 0x0094BD, 0x0094BF, 0x0094C4, 0x0094C8, 0x0094C9, 0x0094CA, 0x0094CB, 0x0094CC,
	0x0094CD, 0x0094CE, 0x0094D0, 0x0094D1, 0x0094D2, 0x0094D5, 0x0094D6, 0x0094D7, 0x0094D9, 0x0094D8, 0x0094DB, 0x0094DE, 0x0094DF, 0x0094E0, 0x0094E2, 0x0094E4, 0x0094E5, 0x0094E7, 0x0094E8, 0x0094EA, 0x0094E9, 0x0094EB, 0x0094EE, 0x0094EF, 0x0094F3, 0x0094F4, 0x0094F5, 0x0094F7, 0x0094F9, 0x0094FC, 0x0094FD, 0x0094FF,
	0x009503, 0x009502, 0x009506, 0x009507, 0x009509, 0x00950A, 0x00950D, 0x00950E, 0x00950F, 0x009512, 0x009513, 0x009514, 0x009515, 0x009516, 0x009518, 0x00951B, 0x00951D, 0x00951E, 0x00951F, 0x009522, 0x00952A, 0x00952B, 0x009529, 0x00952C, 0x009531, 0x009532, 0x009534, 0x009536, 0x009537, 0x009538, 0x00953C, 0x00953E,
	0x00953F, 0x009542, 0x009535, 0x009544, 0x009545, 0x009546, 0x009549, 0x00954C, 0x00954E, 0x00954F, 0x009552, 0x009553, 0x009554, 0x009556, 0x009557, 0x009558, 0x009559, 0x00955B, 0x00955E, 0x00955F, 0x00955D, 0x009561, 0x009562, 0x009564, 0x009565, 0x009566, 0x009567, 0x009568, 0x009569, 0x00956A, 0x00956B, 0x00956C,
	0x00956F, 0x009571, 0x009572, 0x009573, 0x00953A, 0x007A06, 0x007A51, 0x009E20, 0x009E22, 0x009E28, 0x009E29, 0x009E2A, 0x009E2B, 0x009E2C, 0x009E32, 0x009E31, 0x009E36, 0x009E38, 0x009E37, 0x009E39, 0x009E3A, 0x009E3E, 0x009E41, 0x009E42, 0x009E44, 0x009E46, 0x009E47, 0x009E48, 0x009E49, 0x009E4B, 0x009E4C, 0x009E4E,
	0x009E51, 0x009E55, 0x009E57, 0x009E5A, 0x009E5B, 0x009E5C, 0x009E5E, 0x009E63, 0x009E66, 0x009E67, 0x009E68, 0x009E69, 0x009E6A, 0x009E6B, 0x009E6C, 0x009E71, 0x009E6D, 0x009E73, 0x007596, 0x0075A0, 0x0075AC, 0x0075D6, 0x0075E8, 0x0075E6, 0x0075EB, 0x007610, 0x007605, 0x007617, 0x00763F, 0x00763E, 0x00764D, 0x00765E,
	0
};

static const unsigned map_kanji_gbk2_3[] = {
	0x007654, 0x00766B, 0x007AA6, 0x007AA8, 0x007AAD, 0x0088C6, 0x0088C9, 0x0088E2, 0x0088E3, 0x0088E5, 0x008921, 0x00891B, 0x008934, 0x0076B2, 0x008022, 0x008025, 0x008027, 0x008029, 0x008031, 0x00804D, 0x008069, 0x009878, 0x009880, 0x009883, 0x009889, 0x00988C, 0x00988D, 0x00988F, 0x009894, 0x00989A, 0x00989B, 0x00989E,
	0x00989F, 0x0098A1, 0x0098A2, 0x0098A5, 0x0098A6, 0x00866E, 0x00867F, 0x00867C, 0x0086AC, 0x0086BA, 0x0086CF, 0x0086F1, 0x0086F2, 0x0086F3, 0x0086D0, 0x0086F4, 0x0086D8, 0x008748, 0x008722, 0x00877D, 0x00877E, 0x008770, 0x0087A8, 0x0087D3, 0x0087D1, 0x0087DB, 0x0087EE, 0x007F42, 0x007B03, 0x007B15, 0x007B38, 0x007B24,
	0x007B3E, 0x007B58, 0x007B5A, 0x007B7B, 0x007B62, 0x007BA6, 0x007BA7, 0x007BA8, 0x007B85, 0x007BA2, 0x007BAB, 0x007BD1, 0x007C16, 0x007C41, 0x008223, 0x00822D, 0x00822F, 0x00823B, 0x00823E, 0x008885, 0x007F9F, 0x007FA7, 0x007FB0, 0x007C91, 0x007C9D, 0x007C9C, 0x007CC1, 0x007CCC, 0x007CE8, 0x0066A8, 0x007D77, 0x008D94,
	0x008DB1, 0x009150, 0x009170, 0x00916F, 0x00917D, 0x00917E, 0x00918C, 0x0091A3, 0x009E7E, 0x008DB8, 0x008DBF, 0x008DC4, 0x008DDE, 0x008DF7, 0x008DF8, 0x008DF9, 0x008DFB, 0x008E2C, 0x008E2E, 0x008E2F, 0x008E3A, 0x008E51, 0x008E52, 0x008E8F, 0x008E9C, 0x0089DE, 0x0089EF, 0x009753, 0x0096F3, 0x009701, 0x00972D, 0x009F80,
	0x009F83, 0x009F85, 0x009F86, 0x009F87, 0x009F88, 0x009F89, 0x009F8A, 0x009F8C, 0x009EFE, 0x009F0B, 0x009F0D, 0x0096E0, 0x0092AE, 0x00933E, 0x00943E, 0x009C7F, 0x009C82, 0x009C85, 0x009C86, 0x009C87, 0x009C88, 0x007A23, 0x009C8B, 0x009C8E, 0x009C90, 0x009C91, 0x009C92, 0x009C94, 0x009C95, 0x009C9A, 0x009C9B, 0x009C9E,
	0x009C9F, 0x009CA0, 0x009CA1, 0x009CA2, 0x009CA3, 0x009CA5, 0x009CA6, 0x009CA7, 0x009CA8, 0x009CA9, 0x009CAB, 0x009CAD, 0x009CAE, 0x009CB0, 0x009CB1, 0x009CB2, 0x009CB3, 0x009CB4, 0x009CB5, 0x009CB6, 0x009CB7, 0x009CBA, 0x009CBB, 0x009CBC, 0x009CBD, 0x009CC4, 0x009CC5, 0x009CC6, 0x009CC7, 0x009CCA, 0x009CCB, 0x009CCC,
	0x009CCD, 0x009CCE, 0x009CCF, 0x009CD0, 0x009CD3, 0x009CD4, 0x009CD5, 0x009CD7, 0x009CD8, 0x009CD9, 0x009CDC, 0x009CDD, 0x009CDF, 0x009CE2, 0x009791, 0x009792, 0x009794, 0x0097AF, 0x009AB1, 0x009E58, 0x009ABA, 0x009AC5, 0x009AC2, 0x009ACB, 0x009ACC, 0x009B47, 0x009B49, 0x0098E8, 0x00990D, 0x009AE1, 0x009B0F, 0x009B13,
	0x009EE2, 0x009EE9, 0x009EEA,
	0
};

// GBK/3
static const unsigned map_kanji_gbk3_1[] = {
	0x004E06, 0x004E20, 0x004E33, 0x004E37, 0x004E46, 0x004E4A, 0x004E5B, 0x004E64, 0x004E65, 0x004E67, 0x004E6A, 0x004E6B, 0x004E6C, 0x004E6D, 0x004E6E, 0x004E6F, 0x004E72, 0x004E76, 0x004E77, 0x004E78, 0x004E7A, 0x004E7B, 0x004E7C, 0x004E7D, 0x004E81, 0x004E83, 0x004E84, 0x004E87, 0x004E90, 0x004EA3, 0x004EAA, 0x004EB1,
	0x004EB4, 0x004EB7, 0x004EB8, 0x004EBD, 0x004EBE, 0x004ECC, 0x004ED2, 0x004EDC, 0x004EE6, 0x004EE7, 0x004EE9, 0x004EF4, 0x004EF8, 0x004EF9, 0x004EFA, 0x004F04, 0x004F05, 0x004F06, 0x004F07, 0x004F13, 0x004F14, 0x004F21, 0x004F23, 0x004F28, 0x004F29, 0x004F2C, 0x004F2D, 0x004F3F, 0x004F41, 0x004F44, 0x004F45, 0x004F4A,
	0x004F61, 0x004F62, 0x004F66, 0x004F68, 0x004F6B, 0x004F6D, 0x004F72, 0x004F80, 0x004F87, 0x004F95, 0x004F9C, 0x004FA2, 0x004FA4, 0x004FB0, 0x004FB1, 0x004FB3, 0x004FB4, 0x004FB8, 0x004FBA, 0x004FC7, 0x004FD3, 0x004FD5, 0x004FD6, 0x004FD9, 0x004FE7, 0x004FEB, 0x004FEC, 0x004FF4, 0x004FF7, 0x004FF9, 0x004FFB, 0x005003,
	0x005008, 0x005015, 0x005020, 0x00502F, 0x005031, 0x005034, 0x005037, 0x005038, 0x00503D, 0x00503F, 0x005044, 0x00504B, 0x00504D, 0x005054, 0x005058, 0x00505B, 0x00505D, 0x00505E, 0x005061, 0x005064, 0x005068, 0x005069, 0x00506B, 0x00506E, 0x00506F, 0x005073, 0x005079, 0x00507A, 0x00507C, 0x005082, 0x005087, 0x005089,
	0x00508B, 0x00508C, 0x005097, 0x00509D, 0x0050A4, 0x0050A6, 0x0050AB, 0x0050AE, 0x0050B6, 0x0050B8, 0x0050BC, 0x0050BF, 0x0050C1, 0x0050C6, 0x0050C8, 0x0050CB, 0x0050D2, 0x0050D7, 0x0050DB, 0x0050E0, 0x0050EA, 0x0050EB, 0x0050F0, 0x0050F7, 0x0050F8, 0x0050FC, 0x0050FD, 0x0050FF, 0x005105, 0x00510A, 0x00510F, 0x005111,
	0x005113, 0x005120, 0x005122, 0x005124, 0x005125, 0x005126, 0x005129, 0x00512B, 0x00512E, 0x005130, 0x005136, 0x00513D, 0x00513E, 0x005159, 0x00515B, 0x00515D, 0x00515E, 0x005161, 0x005163, 0x005167, 0x00516F, 0x005172, 0x00517A, 0x00517F, 0x005187, 0x005194, 0x00519A, 0x00519E, 0x00519F, 0x0051A7, 0x0051AE, 0x0051B9,
	0x0051C1, 0x0051CE, 0x0051D0, 0x0051D4, 0x0051D7, 0x0051D9, 0x0051DA, 0x0051DF, 0x0051E3, 0x0051E8, 0x005209, 0x00520C, 0x00520F, 0x005210, 0x00521C, 0x00521E, 0x00521F, 0x005221, 0x005223, 0x005226, 0x00522C, 0x00522F, 0x005234, 0x00523E, 0x005246, 0x005248, 0x00524E, 0x005252, 0x005253, 0x005259, 0x005262, 0x005268,
	0x00526B, 0x00526D, 0x005276, 0x00527A, 0x00527B, 0x00527C, 0x00527E, 0x005286, 0x00528B, 0x00528E, 0x00528F, 0x005299, 0x0052AE, 0x0052C2, 0x0052CE, 0x0052D3, 0x0052DA, 0x0052EB, 0x0052ED, 0x0052EE, 0x0052EF, 0x0052FC, 0x0052FD, 0x005304, 0x005309, 0x00530E, 0x005312, 0x005314, 0x005322, 0x005334, 0x005336, 0x005337,
	0x005344, 0x005350, 0x00535D, 0x005368, 0x00536A, 0x005376, 0x00537C, 0x005380, 0x005381, 0x00538A, 0x00538F, 0x005390, 0x005391, 0x005392, 0x005397, 0x00539B, 0x00539C, 0x00539E, 0x0053A7, 0x0053AC, 0x0053B1, 0x0053B9, 0x0053BC, 0x0053BE, 0x0053C4, 0x0053C6, 0x0053C7, 0x0053D0, 0x0053DC, 0x0053FE, 0x0053FF, 0x005400,
	0x005405, 0x005407, 0x005414, 0x005418, 0x005419, 0x00541C, 0x005422, 0x005425, 0x005430, 0x005433, 0x005437, 0x00543A, 0x00543F, 0x005441, 0x005445, 0x00544C, 0x00545A, 0x00545D, 0x005460, 0x005461, 0x005463, 0x005465, 0x00546F, 0x005479, 0x00547A, 0x00547E, 0x005487, 0x005493, 0x005497, 0x005498, 0x00549E, 0x0054B0,
	0x0054B5, 0x0054B6, 0x0054C5, 0x0054CB, 0x0054D6, 0x0054DB, 0x0054E3, 0x0054E4, 0x0054EB, 0x0054F0, 0x0054F4, 0x0054F5, 0x0054F7, 0x0054F8, 0x0054F9, 0x0054FB, 0x005502, 0x005503, 0x00550A, 0x00550B, 0x005512, 0x005513, 0x005517, 0x005518, 0x005519, 0x00551A, 0x00551C, 0x00551D, 0x00551E, 0x00551F, 0x005521, 0x005525,
	0x005526, 0x005528, 0x005529, 0x00552D, 0x005534, 0x00553A, 0x005542, 0x005548, 0x00554B, 0x00554E, 0x005552, 0x005554, 0x005559, 0x00555F, 0x005562, 0x005568, 0x005569, 0x00556B, 0x00556F, 0x005570, 0x005571, 0x005572, 0x005573, 0x005574, 0x005579, 0x00557A, 0x005585, 0x00558C, 0x00558D, 0x005590, 0x005595, 0x005596,
	0x00559B, 0x0055A0, 0x0055A1, 0x0055A2, 0x0055A5, 0x0055A6, 0x0055AF, 0x0055B4, 0x0055B8, 0x0055BA, 0x0055BC, 0x0055C0, 0x0055C2, 0x0055C8, 0x0055CA, 0x0055CF, 0x0055D0, 0x0055D5, 0x0055D9, 0x0055E0, 0x0055E7, 0x0055ED, 0x0055EE, 0x0055F0, 0x0055F1, 0x0055F4, 0x0055F8, 0x0055FA, 0x0055FB, 0x0055FC, 0x005602, 0x005603,
	0x005604, 0x00560B, 0x005613, 0x005615, 0x00561A, 0x00561C, 0x00561D, 0x005620, 0x005621, 0x005622, 0x005625, 0x005626, 0x00562A, 0x00562B, 0x00562E, 0x00563A, 0x00563E, 0x005645, 0x005648, 0x00564A, 0x005651, 0x005652, 0x005655, 0x005656, 0x00565A, 0x00565D, 0x00565F, 0x005665, 0x005667, 0x00566E, 0x005670, 0x005673,
	0x005677, 0x005679, 0x00567D, 0x00567E, 0x00567F, 0x005681, 0x005682, 0x005683, 0x005689, 0x00568D, 0x005690, 0x005691, 0x005692, 0x005696, 0x005697, 0x005698, 0x00569B, 0x00569C, 0x0056A1, 0x0056A4, 0x0056AA, 0x0056B0, 0x0056B5, 0x0056B8, 0x0056B9, 0x0056BA, 0x0056BB, 0x0056BD, 0x0056BF, 0x0056C4, 0x0056C6, 0x0056C7,
	0x0056D2, 0x0056D5, 0x0056D6, 0x0056E9, 0x0056EA, 0x0056EC, 0x0056EF, 0x0056F8, 0x0056FB, 0x0056FC, 0x005705, 0x00570E, 0x005710, 0x005714, 0x005717, 0x005719, 0x00571E, 0x00572B, 0x005731, 0x005732, 0x005735, 0x005736, 0x00573C, 0x005741, 0x005743, 0x005744, 0x005748, 0x005749, 0x00574B, 0x005753, 0x005754, 0x005755,
	0x005756, 0x005758, 0x005763, 0x00576C, 0x005772, 0x005778, 0x00577D, 0x005780, 0x005787, 0x00578A, 0x00578D, 0x00578E, 0x00578F, 0x005790, 0x005791, 0x005796, 0x005798, 0x0057A5, 0x0057AF, 0x0057B1, 0x0057B5, 0x0057B6, 0x0057B7, 0x0057B9, 0x0057BA, 0x0057BB, 0x0057BC, 0x0057BE, 0x0057BF, 0x0057C1, 0x0057C4, 0x0057C5,
	0x0057C9, 0x0057CA, 0x0057CD, 0x0057D0, 0x0057D1, 0x0057DB, 0x0057E2, 0x0057E5, 0x0057E8, 0x0057EA, 0x0057EB, 0x0057EC, 0x0057EE, 0x0057F1, 0x0057F2, 0x0057F3, 0x005801, 0x00580E, 0x00580F, 0x005810, 0x005812, 0x005813, 0x005814, 0x005816, 0x005817, 0x005818, 0x00581A, 0x00581C, 0x005822, 0x005823, 0x005825, 0x005828,
	0x005829, 0x00582B, 0x00582C, 0x00582E, 0x005833, 0x005836, 0x005837, 0x005838, 0x00583B, 0x00583C, 0x00583E, 0x005842, 0x005843, 0x005845, 0x005846, 0x005847, 0x005848, 0x00584E, 0x005853, 0x005856, 0x00585B, 0x00585C, 0x00585D, 0x005860, 0x005863, 0x005866, 0x00586A, 0x00586D, 0x00586E, 0x00586F, 0x005871, 0x005873,
	0x005874, 0x005876, 0x005877, 0x00587A, 0x00587B, 0x00587D, 0x005882, 0x005884, 0x005886, 0x00588E, 0x005891, 0x005895, 0x005898, 0x00589B, 0x0058A3, 0x0058A4, 0x0058A5, 0x0058A7, 0x0058AC, 0x0058AD, 0x0058AF, 0x0058B0, 0x0058B4, 0x0058B5, 0x0058B6, 0x0058B7, 0x0058BD, 0x0058BF, 0x0058C0, 0x0058C6, 0x0058C9, 0x0058CB,
	0x0058CF, 0x0058DB, 0x0058E3, 0x0058E6, 0x0058E7, 0x0058E8, 0x0058EA, 0x0058ED, 0x0058F5, 0x0058F8, 0x0058FE, 0x0058FF, 0x005900, 0x005901, 0x005903, 0x005908, 0x00590E, 0x005911, 0x005917, 0x00591E, 0x005920, 0x005926, 0x00593B, 0x005940, 0x005945, 0x00594A, 0x00594C, 0x00594D, 0x00595C, 0x005964, 0x005966, 0x005970,
	0x005971, 0x005977, 0x00597A, 0x00597E, 0x00597F, 0x005980, 0x005985, 0x005989, 0x00598F, 0x005990, 0x005991, 0x005994, 0x005998, 0x00599A, 0x00599C, 0x0059A0, 0x0059A1, 0x0059A2, 0x0059A6, 0x0059B1, 0x0059B4, 0x0059B5, 0x0059B6, 0x0059B8, 0x0059BD, 0x0059BF, 0x0059C0, 0x0059C2, 0x0059C5, 0x0059C7, 0x0059CC, 0x0059CE,
	0x0059CF, 0x0059D5, 0x0059D6, 0x0059DB, 0x0059E0, 0x0059E1, 0x0059E2, 0x0059E9, 0x0059ED, 0x0059F0, 0x0059F3, 0x0059F5, 0x0059FA, 0x0059FC, 0x0059FD, 0x0059FE, 0x005A02, 0x005A0A, 0x005A0B, 0x005A0F, 0x005A10, 0x005A14, 0x005A15, 0x005A16, 0x005A19, 0x005A1B, 0x005A1D, 0x005A21, 0x005A22, 0x005A26, 0x005A2B, 0x005A2C,
	0x005A2E, 0x005A33, 0x005A37, 0x005A38, 0x005A39, 0x005A3A, 0x005A3B, 0x005A3D, 0x005A3E, 0x005A3F, 0x005A42, 0x005A43, 0x005A4B, 0x005A4D, 0x005A4E, 0x005A4F, 0x005A51, 0x005A52, 0x005A53, 0x005A54, 0x005A56, 0x005A57, 0x005A58, 0x005A59, 0x005A5B, 0x005A5C, 0x005A5D, 0x005A5F, 0x005A60, 0x005A61, 0x005A64, 0x005A68,
	0x005A69, 0x005A6B, 0x005A6E, 0x005A6F, 0x005A70, 0x005A71, 0x005A72, 0x005A73, 0x005A78, 0x005A79, 0x005A7C, 0x005A7D, 0x005A80, 0x005A81, 0x005A82, 0x005A83, 0x005A85, 0x005A86, 0x005A87, 0x005A88, 0x005A89, 0x005A8A, 0x005A8C, 0x005A8D, 0x005A8E, 0x005A8F, 0x005A91, 0x005A94, 0x005A95, 0x005A97, 0x005A98, 0x005A9D,
	0x005AA1, 0x005AA3, 0x005AA4, 0x005AA5, 0x005AA6, 0x005AA8, 0x005AA9, 0x005AAB, 0x005AAD, 0x005AAE, 0x005AAF, 0x005AB0, 0x005AB4, 0x005AB6, 0x005AB7, 0x005AB9, 0x005AC0, 0x005AC3, 0x005AC5, 0x005AC7, 0x005ACA, 0x005ACD, 0x005ACE, 0x005AD1, 0x005AD3, 0x005AD5, 0x005AD9, 0x005ADB, 0x005ADD, 0x005ADE, 0x005ADF, 0x005AE2,
	0x005AE4, 0x005AE7, 0x005AE8, 0x005AEC, 0x005AED, 0x005AEF, 0x005AF2, 0x005AF3, 0x005AF4, 0x005AF7, 0x005AF8, 0x005AF9, 0x005AFC, 0x005AFE, 0x005AFF, 0x005B02, 0x005B03, 0x005B04, 0x005B05, 0x005B06, 0x005B07, 0x005B0A, 0x005B0D, 0x005B0E, 0x005B0F, 0x005B10, 0x005B11, 0x005B12, 0x005B13, 0x005B14, 0x005B15, 0x005B18,
	0x005B1A, 0x005B1C, 0x005B1E, 0x005B1F, 0x005B20, 0x005B23, 0x005B24, 0x005B26, 0x005B27, 0x005B28, 0x005B29, 0x005B2B, 0x005B2E, 0x005B2F, 0x005B31, 0x005B33, 0x005B35, 0x005B39, 0x005B3A, 0x005B3B, 0x005B3C, 0x005B3D, 0x005B3F, 0x005B42, 0x005B44, 0x005B46, 0x005B47, 0x005B48, 0x005B49, 0x005B4A, 0x005B4D, 0x005B4E,
	0x005B4F, 0x005B60, 0x005B61, 0x005B67, 0x005B6D, 0x005B72, 0x005B74, 0x005B76, 0x005B77, 0x005B79, 0x005B7B, 0x005B82, 0x005B92, 0x005BA7, 0x005BBB, 0x005BC8, 0x005BCA, 0x005BCB, 0x005BD1, 0x005BD5, 0x005BDC, 0x005BE3, 0x005BEA, 0x005BED, 0x005BF2, 0x005BF7, 0x005C00, 0x005C10, 0x005C19, 0x005C1B, 0x005C21, 0x005C2A,
	0x005C2F, 0x005C33, 0x005C37, 0x005C43, 0x005C44, 0x005C47, 0x005C4C, 0x005C52, 0x005C54, 0x005C56, 0x005C57, 0x005C58, 0x005C5D, 0x005C6A, 0x005C6B, 0x005C72, 0x005C73, 0x005C77, 0x005C78, 0x005C7E, 0x005C80, 0x005C83, 0x005C84, 0x005C85, 0x005C86, 0x005C89, 0x005C8B, 0x005C8E, 0x005C93, 0x005C95, 0x005C9E, 0x005CA4,
	0x005CA5, 0x005CAE, 0x005CAF, 0x005CB0, 0x005CB9, 0x005CC0, 0x005CC2, 0x005CC3, 0x005CC6, 0x005CC8, 0x005CCA, 0x005CCC, 0x005CCD, 0x005CCE, 0x005CCF, 0x005CD1, 0x005CD3, 0x005CD4, 0x005CD5, 0x005CD6, 0x005CD8, 0x005CDA, 0x005CDB, 0x005CDC, 0x005CDE, 0x005CDF, 0x005CE2, 0x005CE3, 0x005CE7, 0x005CEB, 0x005CEC, 0x005CF3,
	0x005CF7, 0x005CF8, 0x005CF9, 0x005CFC, 0x005CFE, 0x005CFF, 0x005D00, 0x005D04, 0x005D05, 0x005D08, 0x005D09, 0x005D0A, 0x005D0C, 0x005D0F, 0x005D13, 0x005D1C, 0x005D21, 0x005D25, 0x005D28, 0x005D2A, 0x005D2C, 0x005D2F, 0x005D30, 0x005D32, 0x005D33, 0x005D35, 0x005D36, 0x005D37, 0x005D38, 0x005D3A, 0x005D3B, 0x005D3C,
	0
};

static const unsigned map_kanji_gbk3_2[] = {
	0x005D40, 0x005D41, 0x005D44, 0x005D45, 0x005D49, 0x005D4D, 0x005D4F, 0x005D54, 0x005D56, 0x005D57, 0x005D5A, 0x005D5E, 0x005D63, 0x005D65, 0x005D66, 0x005D67, 0x005D68, 0x005D6E, 0x005D71, 0x005D72, 0x005D75, 0x005D77, 0x005D78, 0x005D7B, 0x005D7C, 0x005D7D, 0x005D80, 0x005D85, 0x005D86, 0x005D89, 0x005D8D, 0x005D8E,
	0x005D8F, 0x005D91, 0x005D96, 0x005D98, 0x005D9A, 0x005D9C, 0x005D9E, 0x005DA1, 0x005DA3, 0x005DA5, 0x005DA6, 0x005DA8, 0x005DA9, 0x005DAA, 0x005DAD, 0x005DAF, 0x005DB1, 0x005DB3, 0x005DB5, 0x005DB6, 0x005DBB, 0x005DBE, 0x005DBF, 0x005DC0, 0x005DC1, 0x005DC2, 0x005DC4, 0x005DC6, 0x005DC8, 0x005DCA, 0x005DCF, 0x005DD4,
	0x005DD5, 0x005DDA, 0x005DDC, 0x005DDF, 0x005DEA, 0x005DEC, 0x005DED, 0x005DF0, 0x005DF6, 0x005DFA, 0x005DFC, 0x005DFF, 0x005E04, 0x005E09, 0x005E0A, 0x005E0E, 0x005E13, 0x005E17, 0x005E1E, 0x005E21, 0x005E22, 0x005E23, 0x005E24, 0x005E29, 0x005E2A, 0x005E2C, 0x005E34, 0x005E39, 0x005E3A, 0x005E3F, 0x005E41, 0x005E46,
	0x005E48, 0x005E4A, 0x005E4D, 0x005E4F, 0x005E52, 0x005E53, 0x005E59, 0x005E5A, 0x005E5D, 0x005E60, 0x005E65, 0x005E66, 0x005E67, 0x005E69, 0x005E6F, 0x005E71, 0x005E82, 0x005E85, 0x005E88, 0x005E89, 0x005E8C, 0x005E8D, 0x005E92, 0x005E98, 0x005E9B, 0x005E9D, 0x005EA1, 0x005EA3, 0x005EA9, 0x005EAE, 0x005EAF, 0x005EB0,
	0x005EB2, 0x005EB4, 0x005EBA, 0x005EBB, 0x005EBC, 0x005EC0, 0x005EC4, 0x005EC5, 0x005EC7, 0x005ECD, 0x005ED7, 0x005ED8, 0x005EE4, 0x005EE6, 0x005EE7, 0x005EED, 0x005EEE, 0x005EEF, 0x005EF2, 0x005EF5, 0x005F05, 0x005F12, 0x005F1A, 0x005F32, 0x005F33, 0x005F42, 0x005F43, 0x005F46, 0x005F49, 0x005F4B, 0x005F4F, 0x005F5A,
	0x005F5E, 0x005F5F, 0x005F65, 0x005F68, 0x005F6E, 0x005F76, 0x005F86, 0x005F8E, 0x005F94, 0x005F9A, 0x005F9B, 0x005F9F, 0x005FA3, 0x005FA5, 0x005FA6, 0x005FB2, 0x005FB6, 0x005FBA, 0x005FBB, 0x005FBE, 0x005FBF, 0x005FC0, 0x005FC1, 0x005FC2, 0x005FCA, 0x005FCE, 0x005FD5, 0x005FDA, 0x005FDB, 0x005FDF, 0x005FE3, 0x005FE5,
	0x005FE6, 0x005FF4, 0x005FF7, 0x005FF9, 0x006008, 0x006009, 0x00600B, 0x00600C, 0x006011, 0x00601E, 0x006023, 0x00602C, 0x00602E, 0x006030, 0x006032, 0x006034, 0x006036, 0x006037, 0x006038, 0x006039, 0x00603D, 0x00603E, 0x006044, 0x006045, 0x00604E, 0x00604F, 0x006053, 0x006058, 0x00605B, 0x00605C, 0x00605E, 0x006066,
	0x00606E, 0x006072, 0x006074, 0x006080, 0x006087, 0x00608F, 0x006090, 0x006099, 0x00609C, 0x0060A1, 0x0060AE, 0x0060B9, 0x0060BA, 0x0060BF, 0x0060C0, 0x0060C1, 0x0060C3, 0x0060CC, 0x0060CD, 0x0060D0, 0x0060D2, 0x0060D6, 0x0060D7, 0x0060E4, 0x0060EA, 0x0060FE, 0x0060FF, 0x006104, 0x006105, 0x00610B, 0x006118, 0x00611D,
	0x006125, 0x006129, 0x00612D, 0x00612E, 0x00612F, 0x006132, 0x006133, 0x006138, 0x00613B, 0x006140, 0x006143, 0x00614F, 0x006150, 0x006152, 0x006154, 0x006156, 0x006157, 0x00615B, 0x00615C, 0x006161, 0x006164, 0x006166, 0x006169, 0x00616A, 0x00616D, 0x006179, 0x00617A, 0x00617D, 0x006185, 0x006186, 0x006188, 0x006189,
	0x00618C, 0x00618F, 0x006195, 0x00619B, 0x00619E, 0x0061A1, 0x0061A2, 0x0061A3, 0x0061A6, 0x0061AF, 0x0061B0, 0x0061B1, 0x0061B3, 0x0061B4, 0x0061B5, 0x0061BB, 0x0061BD, 0x0061BF, 0x0061C4, 0x0061C5, 0x0061D3, 0x0061D6, 0x0061D7, 0x0061D8, 0x0061D9, 0x0061DA, 0x0061DB, 0x0061E0, 0x0061E4, 0x0061E8, 0x0061EA, 0x0061EB,
	0x0061EE, 0x0061F0, 0x0061F1, 0x0061F3, 0x0061F9, 0x0061FB, 0x006202, 0x006205, 0x006219, 0x006228, 0x00622D, 0x006231, 0x006235, 0x006236, 0x00623A, 0x00623C, 0x006245, 0x00624A, 0x00624F, 0x006257, 0x006259, 0x00625D, 0x00625F, 0x006262, 0x006265, 0x006272, 0x006274, 0x006275, 0x006277, 0x006278, 0x006281, 0x006286,
	0x006287, 0x006288, 0x00628B, 0x00628C, 0x00629D, 0x0062A3, 0x0062A7, 0x0062A9, 0x0062AA, 0x0062AD, 0x0062AE, 0x0062AF, 0x0062B0, 0x0062B2, 0x0062B4, 0x0062B8, 0x0062C0, 0x0062C1, 0x0062C3, 0x0062CB, 0x0062DE, 0x0062E4, 0x0062EB, 0x0062F0, 0x0062F8, 0x0062F9, 0x0062FA, 0x0062FB, 0x006300, 0x006305, 0x006306, 0x00630F,
	0x006312, 0x006314, 0x006315, 0x006317, 0x00631C, 0x006326, 0x00632C, 0x00632E, 0x006330, 0x006331, 0x006333, 0x006334, 0x006337, 0x006338, 0x006340, 0x006347, 0x006348, 0x006351, 0x006356, 0x00635A, 0x00635D, 0x006360, 0x006364, 0x00636A, 0x00636F, 0x006370, 0x006373, 0x006379, 0x00637E, 0x006381, 0x006385, 0x006386,
	0x00638B, 0x00638D, 0x006391, 0x006393, 0x006397, 0x00639C, 0x00639D, 0x0063B1, 0x0063B6, 0x0063B9, 0x0063BF, 0x0063C2, 0x0063C7, 0x0063CA, 0x0063CB, 0x0063CC, 0x0063D7, 0x0063D8, 0x0063D9, 0x0063DD, 0x0063DF, 0x0063E2, 0x0063E4, 0x0063E6, 0x0063E7, 0x0063E8, 0x0063EF, 0x0063F0, 0x0063F1, 0x0063FB, 0x0063FC, 0x0063FE,
	0x006403, 0x006404, 0x006407, 0x006408, 0x00640E, 0x006411, 0x006415, 0x006419, 0x00641A, 0x00641D, 0x00641F, 0x006423, 0x006427, 0x00642B, 0x00642E, 0x006431, 0x006432, 0x006433, 0x006437, 0x006438, 0x006439, 0x00643B, 0x00643C, 0x006440, 0x006443, 0x006449, 0x00644C, 0x00644D, 0x006450, 0x006455, 0x006456, 0x006457,
	0x006459, 0x006462, 0x006464, 0x006465, 0x006466, 0x006468, 0x00646A, 0x00646B, 0x00646C, 0x00646E, 0x006470, 0x006471, 0x006472, 0x006475, 0x006477, 0x00647C, 0x00647E, 0x00647F, 0x006480, 0x006481, 0x006486, 0x006489, 0x00648A, 0x00648B, 0x00648C, 0x00648D, 0x00648E, 0x006494, 0x006497, 0x00649C, 0x0064A0, 0x0064A2,
	0x0064A7, 0x0064AA, 0x0064AF, 0x0064B1, 0x0064B4, 0x0064B6, 0x0064C3, 0x0064C6, 0x0064C8, 0x0064CF, 0x0064D3, 0x0064D6, 0x0064D9, 0x0064DB, 0x0064DC, 0x0064DD, 0x0064DF, 0x0064E8, 0x0064EB, 0x0064EE, 0x0064F3, 0x0064F8, 0x0064F9, 0x0064FC, 0x006502, 0x006503, 0x006506, 0x006507, 0x00650B, 0x00650C, 0x00650D, 0x00650E,
	0x006510, 0x006511, 0x006515, 0x006517, 0x00651A, 0x006520, 0x006521, 0x006527, 0x006528, 0x00652D, 0x006530, 0x006533, 0x006540, 0x006541, 0x006542, 0x006546, 0x00654A, 0x00654B, 0x006553, 0x00655A, 0x00655C, 0x006561, 0x006564, 0x006565, 0x006568, 0x006569, 0x00656A, 0x00656D, 0x00656E, 0x00656F, 0x006571, 0x006573,
	0x006576, 0x006579, 0x00657B, 0x00657C, 0x00657E, 0x00657F, 0x006580, 0x006586, 0x00658D, 0x00658F, 0x006594, 0x006596, 0x00659A, 0x00659E, 0x0065A2, 0x0065A8, 0x0065AA, 0x0065B1, 0x0065B6, 0x0065BA, 0x0065BB, 0x0065BE, 0x0065C0, 0x0065C7, 0x0065CA, 0x0065CD, 0x0065D1, 0x0065D3, 0x0065D5, 0x0065DA, 0x0065DC, 0x0065DD,
	0x0065DE, 0x0065E3, 0x0065E4, 0x0065EA, 0x0065EB, 0x0065F3, 0x0065F8, 0x0065FD, 0x006601, 0x006605, 0x00660B, 0x006610, 0x006617, 0x006618, 0x00661A, 0x00661B, 0x006632, 0x006638, 0x00663D, 0x006647, 0x00664D, 0x006650, 0x006671, 0x006672, 0x00667D, 0x006685, 0x006686, 0x00668A, 0x00668F, 0x006693, 0x006694, 0x006695,
	0x00669E, 0x0066A1, 0x0066A3, 0x0066A5, 0x0066A9, 0x0066AA, 0x0066AC, 0x0066AF, 0x0066B0, 0x0066B3, 0x0066B6, 0x0066B7, 0x0066BA, 0x0066BD, 0x0066C5, 0x0066CA, 0x0066CB, 0x0066CD, 0x0066D0, 0x0066D1, 0x0066D2, 0x0066D3, 0x0066D5, 0x0066D7, 0x0066D8, 0x0066DE, 0x0066E1, 0x0066E2, 0x0066E3, 0x0066E4, 0x0066E5, 0x0066E7,
	0x0066EA, 0x0066ED, 0x0066EF, 0x0066F1, 0x0066F6, 0x006702, 0x006704, 0x006706, 0x00670C, 0x006711, 0x006718, 0x00671A, 0x006721, 0x006723, 0x006724, 0x006725, 0x006729, 0x006730, 0x006732, 0x006739, 0x00673B, 0x00673C, 0x006744, 0x00674A, 0x00674B, 0x006752, 0x006757, 0x006758, 0x00675A, 0x00675B, 0x006767, 0x00676B,
	0x006778, 0x006779, 0x00677A, 0x00677D, 0x006780, 0x006782, 0x006783, 0x006786, 0x006788, 0x00678A, 0x00678D, 0x006794, 0x00679F, 0x0067A4, 0x0067AC, 0x0067AE, 0x0067BA, 0x0067BE, 0x0067BF, 0x0067C7, 0x0067CB, 0x0067CC, 0x0067CD, 0x0067D5, 0x0067D6, 0x0067DF, 0x0067E3, 0x0067E8, 0x0067EA, 0x0067EB, 0x0067ED, 0x0067F8,
	0x0067FA, 0x006806, 0x00680D, 0x006812, 0x006815, 0x00681A, 0x00681B, 0x00681C, 0x006820, 0x006823, 0x006824, 0x006825, 0x006826, 0x00682E, 0x006835, 0x006836, 0x00683A, 0x006847, 0x00684B, 0x00684F, 0x006856, 0x00685A, 0x00685E, 0x00686A, 0x00686C, 0x00686D, 0x006873, 0x006878, 0x00687D, 0x006880, 0x006887, 0x006889,
	0x00688A, 0x00688B, 0x00688C, 0x00688E, 0x006891, 0x006892, 0x006895, 0x00689E, 0x0068A4, 0x0068AC, 0x0068B7, 0x0068B8, 0x0068BD, 0x0068BE, 0x0068BF, 0x0068C1, 0x0068C7, 0x0068CE, 0x0068DB, 0x0068DE, 0x0068E2, 0x0068E6, 0x0068E9, 0x0068F3, 0x0068F4, 0x0068F8, 0x0068FE, 0x0068FF, 0x006902, 0x006914, 0x006915, 0x006918,
	0x00691D, 0x00691E, 0x006927, 0x006929, 0x00692B, 0x00692C, 0x00692E, 0x00692F, 0x006932, 0x006937, 0x00693A, 0x00693C, 0x00693E, 0x006940, 0x006941, 0x006943, 0x006944, 0x006947, 0x006948, 0x00694B, 0x00694C, 0x00694D, 0x00694F, 0x006950, 0x006951, 0x006952, 0x006956, 0x006958, 0x00695F, 0x006967, 0x006976, 0x006983,
	0x006985, 0x00698B, 0x00698C, 0x00698F, 0x006990, 0x006993, 0x006997, 0x006999, 0x00699A, 0x00699D, 0x00699E, 0x00699F, 0x0069A2, 0x0069A3, 0x0069A4, 0x0069A9, 0x0069AA, 0x0069AC, 0x0069B3, 0x0069B5, 0x0069B6, 0x0069B9, 0x0069BD, 0x0069C2, 0x0069C4, 0x0069C6, 0x0069C9, 0x0069D2, 0x0069D5, 0x0069DA, 0x0069DC, 0x0069E1,
	0x0069E4, 0x0069E6, 0x0069EC, 0x0069F0, 0x0069F7, 0x0069F8, 0x0069FA, 0x0069FC, 0x006A04, 0x006A06, 0x006A07, 0x006A08, 0x006A09, 0x006A0D, 0x006A0E, 0x006A10, 0x006A16, 0x006A1C, 0x006A25, 0x006A26, 0x006A27, 0x006A2C, 0x006A2D, 0x006A3C, 0x006A40, 0x006A41, 0x006A42, 0x006A43, 0x006A4C, 0x006A4D, 0x006A4F, 0x006A53,
	0x006A54, 0x006A57, 0x006A5A, 0x006A5C, 0x006A5D, 0x006A5E, 0x006A60, 0x006A63, 0x006A68, 0x006A69, 0x006A6C, 0x006A6D, 0x006A6E, 0x006A6F, 0x006A70, 0x006A74, 0x006A75, 0x006A76, 0x006A77, 0x006A7B, 0x006A7D, 0x006A82, 0x006A85, 0x006A88, 0x006A8A, 0x006A8C, 0x006A8F, 0x006A92, 0x006A93, 0x006A95, 0x006A96, 0x006A98,
	0x006A99, 0x006A9A, 0x006AA4, 0x006AA6, 0x006AA7, 0x006AA8, 0x006AAD, 0x006AB2, 0x006AB5, 0x006AB6, 0x006AB7, 0x006AB9, 0x006ABA, 0x006ABC, 0x006AC0, 0x006AC4, 0x006AC5, 0x006AC7, 0x006ACA, 0x006ACB, 0x006ACD, 0x006ACE, 0x006ACF, 0x006AD2, 0x006AD7, 0x006AD8, 0x006AD9, 0x006AE0, 0x006AE1, 0x006AE3, 0x006AE5, 0x006AE6,
	0x006AE9, 0x006AEB, 0x006AED, 0x006AEE, 0x006AEF, 0x006AF4, 0x006AF5, 0x006AF6, 0x006AF7, 0x006AF8, 0x006AF9, 0x006AFE, 0x006AFF, 0x006B00, 0x006B01, 0x006B08, 0x006B0C, 0x006B0D, 0x006B0E, 0x006B13, 0x006B14, 0x006B15, 0x006B18, 0x006B19, 0x006B1A, 0x006B1C, 0x006B25, 0x006B26, 0x006B29, 0x006B2A, 0x006B2D, 0x006B2E,
	0x006B30, 0x006B31, 0x006B33, 0x006B34, 0x006B3C, 0x006B40, 0x006B41, 0x006B42, 0x006B44, 0x006B45, 0x006B48, 0x006B4B, 0x006B4F, 0x006B51, 0x006B55, 0x006B57, 0x006B5A, 0x006B5C, 0x006B5E, 0x006B68, 0x006B6D, 0x006B71, 0x006B72, 0x006B76, 0x006B88, 0x006B8C, 0x006B8E, 0x006B8F, 0x006B90, 0x006B91, 0x006B94, 0x006B99,
	0
};

static const unsigned map_kanji_gbk3_3[] = {
	0x006B9C, 0x006B9D, 0x006BA5, 0x006BA6, 0x006BA7, 0x006BB6, 0x006BCA, 0x006BD0, 0x006BD1, 0x006BDC, 0x006BDD, 0x006BDE, 0x006BE0, 0x006BE2, 0x006BE4, 0x006BE5, 0x006BE8, 0x006BE9, 0x006BED, 0x006BF0, 0x006BF2, 0x006BF4, 0x006BF6, 0x006BF8, 0x006BFA, 0x006BFB, 0x006BFC, 0x006BFE, 0x006C00, 0x006C01, 0x006C03, 0x006C0B,
	0x006C0C, 0x006C1C, 0x006C1D, 0x006C1E, 0x006C20, 0x006C25, 0x006C2B, 0x006C2D, 0x006C31, 0x006C39, 0x006C3C, 0x006C43, 0x006C44, 0x006C45, 0x006C48, 0x006C4C, 0x006C51, 0x006C53, 0x006C56, 0x006C58, 0x006C63, 0x006C65, 0x006C66, 0x006C6C, 0x006C6E, 0x006C71, 0x006C75, 0x006C77, 0x006C7C, 0x006C7F, 0x006C80, 0x006C8A,
	0x006C8B, 0x006C8E, 0x006C91, 0x006C9D, 0x006C9E, 0x006CA0, 0x006CA8, 0x006CAC, 0x006CAF, 0x006CB5, 0x006CB6, 0x006CB7, 0x006CC0, 0x006CC3, 0x006CC7, 0x006CC8, 0x006CCB, 0x006CCE, 0x006CD8, 0x006CDF, 0x006CE4, 0x006CE6, 0x006CF9, 0x006CFF, 0x006D02, 0x006D03, 0x006D05, 0x006D06, 0x006D08, 0x006D09, 0x006D0D, 0x006D10,
	0x006D14, 0x006D15, 0x006D16, 0x006D18, 0x006D1C, 0x006D1D, 0x006D20, 0x006D21, 0x006D22, 0x006D23, 0x006D2C, 0x006D2D, 0x006D30, 0x006D37, 0x006D3A, 0x006D40, 0x006D42, 0x006D49, 0x006D4C, 0x006D50, 0x006D55, 0x006D56, 0x006D5D, 0x006D62, 0x006D68, 0x006D6B, 0x006D6D, 0x006D71, 0x006D72, 0x006D73, 0x006D75, 0x006D76,
	0x006D7A, 0x006D7B, 0x006D7D, 0x006D7E, 0x006D7F, 0x006D83, 0x006D84, 0x006D86, 0x006D8B, 0x006D8F, 0x006D90, 0x006D9A, 0x006DA2, 0x006DA5, 0x006DAD, 0x006DB0, 0x006DB1, 0x006DB3, 0x006DB6, 0x006DBA, 0x006DBB, 0x006DBE, 0x006DC1, 0x006DC3, 0x006DC9, 0x006DCD, 0x006DD3, 0x006DD4, 0x006DD7, 0x006DDC, 0x006DE3, 0x006DE7,
	0x006DED, 0x006DFD, 0x006DFE, 0x006DFF, 0x006E01, 0x006E02, 0x006E03, 0x006E06, 0x006E0F, 0x006E12, 0x006E18, 0x006E1C, 0x006E28, 0x006E2A, 0x006E30, 0x006E31, 0x006E33, 0x006E35, 0x006E37, 0x006E3D, 0x006E3F, 0x006E40, 0x006E41, 0x006E42, 0x006E46, 0x006E47, 0x006E50, 0x006E55, 0x006E59, 0x006E5A, 0x006E60, 0x006E61,
	0x006E64, 0x006E65, 0x006E66, 0x006E69, 0x006E6A, 0x006E6C, 0x006E6D, 0x006E70, 0x006E71, 0x006E74, 0x006E75, 0x006E77, 0x006E78, 0x006E79, 0x006E7A, 0x006E7C, 0x006E81, 0x006E84, 0x006E87, 0x006E88, 0x006E8A, 0x006E8B, 0x006E8E, 0x006E91, 0x006E92, 0x006E94, 0x006E95, 0x006E97, 0x006E9A, 0x006E9B, 0x006E9E, 0x006EA1,
	0x006EA3, 0x006EA4, 0x006EA6, 0x006EA8, 0x006EA9, 0x006EAC, 0x006EB0, 0x006EB5, 0x006EB8, 0x006EB9, 0x006EBC, 0x006EBE, 0x006EC6, 0x006ED0, 0x006ED2, 0x006ED6, 0x006ED8, 0x006EDB, 0x006EDC, 0x006EE3, 0x006EE7, 0x006EEA, 0x006EF0, 0x006EF1, 0x006EF3, 0x006EF5, 0x006EF6, 0x006EFA, 0x006EFC, 0x006F00, 0x006F03, 0x006F05,
	0x006F07, 0x006F0B, 0x006F0E, 0x006F12, 0x006F17, 0x006F19, 0x006F1C, 0x006F1D, 0x006F1E, 0x006F1F, 0x006F21, 0x006F27, 0x006F28, 0x006F2E, 0x006F34, 0x006F37, 0x006F39, 0x006F3A, 0x006F3D, 0x006F40, 0x006F42, 0x006F43, 0x006F44, 0x006F48, 0x006F49, 0x006F4A, 0x006F4C, 0x006F4E, 0x006F50, 0x006F55, 0x006F56, 0x006F63,
	0x006F65, 0x006F67, 0x006F69, 0x006F6A, 0x006F6B, 0x006F71, 0x006F73, 0x006F75, 0x006F76, 0x006F77, 0x006F79, 0x006F7B, 0x006F7F, 0x006F85, 0x006F8A, 0x006F8F, 0x006F95, 0x006F99, 0x006F9B, 0x006F9D, 0x006F9E, 0x006FA2, 0x006FA9, 0x006FAB, 0x006FAC, 0x006FAD, 0x006FB2, 0x006FB4, 0x006FB7, 0x006FB8, 0x006FBA, 0x006FBB,
	0x006FBD, 0x006FBF, 0x006FC4, 0x006FCB, 0x006FCC, 0x006FCD, 0x006FCE, 0x006FCF, 0x006FD0, 0x006FD3, 0x006FD6, 0x006FD7, 0x006FD9, 0x006FDC, 0x006FDD, 0x006FE2, 0x006FE3, 0x006FE5, 0x006FE6, 0x006FE7, 0x006FEA, 0x006FED, 0x006FF2, 0x006FF4, 0x006FF7, 0x006FF8, 0x006FFB, 0x006FFF, 0x007002, 0x007003, 0x007004, 0x007008,
	0x00700C, 0x00700E, 0x007010, 0x007012, 0x007013, 0x007014, 0x007016, 0x007019, 0x00701C, 0x007021, 0x007022, 0x007024, 0x007025, 0x007029, 0x00702A, 0x00702B, 0x00702D, 0x00702E, 0x007031, 0x007033, 0x007036, 0x007038, 0x00703B, 0x00703D, 0x00703F, 0x007040, 0x007041, 0x007042, 0x007045, 0x007046, 0x00704D, 0x007050,
	0x007052, 0x007053, 0x007056, 0x007057, 0x007059, 0x00705A, 0x00705B, 0x00705C, 0x00705F, 0x007060, 0x007061, 0x007062, 0x007066, 0x007067, 0x007068, 0x00706A, 0x007071, 0x007072, 0x007073, 0x007074, 0x007077, 0x007079, 0x00707A, 0x00707B, 0x007082, 0x007083, 0x007084, 0x007087, 0x007088, 0x00708B, 0x00708C, 0x00708D,
	0x00708F, 0x007090, 0x007091, 0x007093, 0x00709A, 0x00709E, 0x0070A0, 0x0070A1, 0x0070A2, 0x0070A3, 0x0070A5, 0x0070A6, 0x0070A7, 0x0070A8, 0x0070A9, 0x0070AA, 0x0070B2, 0x0070B5, 0x0070B6, 0x0070BE, 0x0070BF, 0x0070C4, 0x0070C5, 0x0070C6, 0x0070C7, 0x0070C9, 0x0070CC, 0x0070CD, 0x0070CE, 0x0070D0, 0x0070D2, 0x0070D7,
	0x0070DA, 0x0070DE, 0x0070E0, 0x0070E1, 0x0070E2, 0x0070E3, 0x0070E5, 0x0070EA, 0x0070EE, 0x0070F0, 0x0070F2, 0x0070F3, 0x0070F4, 0x0070F5, 0x0070F6, 0x0070F8, 0x0070FB, 0x0070FC, 0x0070FE, 0x0070FF, 0x007100, 0x007101, 0x007102, 0x00710D, 0x00710E, 0x007111, 0x007112, 0x007117, 0x00711B, 0x00711D, 0x00711F, 0x007122,
	0x007123, 0x007124, 0x007125, 0x007127, 0x007128, 0x007129, 0x00712A, 0x00712C, 0x007132, 0x007133, 0x007134, 0x007135, 0x007137, 0x007139, 0x00713A, 0x00713B, 0x00713D, 0x00713E, 0x00713F, 0x007140, 0x007142, 0x007143, 0x007144, 0x007148, 0x00714D, 0x00714F, 0x007154, 0x007158, 0x00715B, 0x00715D, 0x00715F, 0x007161,
	0x007163, 0x00716A, 0x00716B, 0x00716D, 0x00716F, 0x007170, 0x007171, 0x007174, 0x007175, 0x007176, 0x007177, 0x00717B, 0x00717C, 0x00717E, 0x00717F, 0x007181, 0x007182, 0x007183, 0x007186, 0x007189, 0x00718B, 0x00718D, 0x00718E, 0x007190, 0x007191, 0x007193, 0x007197, 0x00719C, 0x00719D, 0x00719E, 0x0071A1, 0x0071A3,
	0x0071A4, 0x0071A5, 0x0071A6, 0x0071A7, 0x0071A9, 0x0071AA, 0x0071AB, 0x0071AD, 0x0071B4, 0x0071B6, 0x0071B7, 0x0071B8, 0x0071BB, 0x0071BC, 0x0071BD, 0x0071C2, 0x0071C5, 0x0071C6, 0x0071C7, 0x0071CA, 0x0071CD, 0x0071CF, 0x0071D1, 0x0071D8, 0x0071DB, 0x0071DD, 0x0071DE, 0x0071E1, 0x0071E2, 0x0071E3, 0x0071E4, 0x0071E8,
	0x0071E9, 0x0071EA, 0x0071EB, 0x0071EF, 0x0071F0, 0x0071F1, 0x0071F2, 0x0071F3, 0x0071F4, 0x0071F6, 0x0071F7, 0x0071FA, 0x0071FD, 0x007201, 0x007202, 0x007203, 0x007204, 0x007205, 0x00720A, 0x00720B, 0x00720C, 0x00720E, 0x00720F, 0x007211, 0x007212, 0x007214, 0x007216, 0x007218, 0x007219, 0x00721C, 0x00721E, 0x007220,
	0x007221, 0x007222, 0x007223, 0x007225, 0x007226, 0x007227, 0x007229, 0x00722E, 0x007233, 0x007244, 0x007249, 0x00724A, 0x007251, 0x007254, 0x007264, 0x007265, 0x00726A, 0x00726C, 0x00726D, 0x007270, 0x007273, 0x007276, 0x007283, 0x007285, 0x007286, 0x007288, 0x00728C, 0x007290, 0x007291, 0x007294, 0x007295, 0x007297,
	0x007298, 0x007299, 0x00729A, 0x00729C, 0x00729D, 0x00729E, 0x0072A1, 0x0072A3, 0x0072A4, 0x0072A5, 0x0072A6, 0x0072A9, 0x0072AA, 0x0072AB, 0x0072B3, 0x0072B5, 0x0072BA, 0x0072BB, 0x0072BC, 0x0072BD, 0x0072BF, 0x0072C5, 0x0072CA, 0x0072CB, 0x0072CF, 0x0072D1, 0x0072D3, 0x0072D4, 0x0072DA,
	0
};

// GBK/4
static const unsigned map_kanji_gbk4_1[] = {
	0x0072DC, 0x0072DD, 0x0072E3, 0x0072E4, 0x0072E6, 0x0072E7, 0x0072EA, 0x0072EB, 0x0072F5, 0x0072F6, 0x0072FF, 0x007300, 0x007306, 0x007308, 0x007309, 0x00730C, 0x00730F, 0x007310, 0x007311, 0x007314, 0x00731A, 0x007320, 0x007323, 0x007326, 0x00732D, 0x007330, 0x00733C, 0x007340, 0x007341, 0x007342, 0x007346, 0x007347,
	0x007348, 0x007349, 0x00734A, 0x00734B, 0x00734C, 0x007351, 0x007353, 0x007354, 0x007355, 0x007359, 0x00735A, 0x00735B, 0x00735C, 0x007361, 0x007362, 0x007364, 0x007365, 0x007373, 0x007374, 0x007376, 0x00737D, 0x00737F, 0x007382, 0x007388, 0x00738C, 0x00738D, 0x00738F, 0x007392, 0x007399, 0x00739A, 0x00739D, 0x0073A1,
	0x0073A3, 0x0073A4, 0x0073A7, 0x0073AC, 0x0073B1, 0x0073B4, 0x0073B6, 0x0073B8, 0x0073BE, 0x0073C1, 0x0073C3, 0x0073C4, 0x0073C7, 0x0073D4, 0x0073D5, 0x0073D7, 0x0073D8, 0x0073DA, 0x0073DB, 0x0073DC, 0x0073DF, 0x0073E2, 0x0073E4, 0x0073E8, 0x0073EB, 0x0073EC, 0x0073EF, 0x0073F0, 0x0073F3, 0x0073F6, 0x0073FC, 0x007402,
	0x007408, 0x00740B, 0x00740C, 0x00740D, 0x00740E, 0x007412, 0x007414, 0x007415, 0x007416, 0x007417, 0x007418, 0x007419, 0x00741C, 0x00741D, 0x00741E, 0x00741F, 0x007420, 0x007421, 0x007423, 0x007427, 0x007437, 0x007438, 0x00743B, 0x00743D, 0x00743E, 0x007442, 0x007445, 0x007448, 0x007449, 0x00744A, 0x00744C, 0x00744E,
	0x00744F, 0x007450, 0x007454, 0x007456, 0x007458, 0x007461, 0x007465, 0x00746C, 0x007474, 0x007475, 0x007478, 0x007479, 0x00747A, 0x00747B, 0x00747C, 0x00747D, 0x00747F, 0x007482, 0x007484, 0x00748A, 0x00748C, 0x00748D, 0x007493, 0x007494, 0x007495, 0x007496, 0x00749B, 0x00749D, 0x0074A4, 0x0074AC, 0x0074AD, 0x0074B3,
	0x0074B4, 0x0074B6, 0x0074B7, 0x0074B8, 0x0074BC, 0x0074BE, 0x0074C0, 0x0074C1, 0x0074C2, 0x0074C3, 0x0074C4, 0x0074C5, 0x0074C6, 0x0074C7, 0x0074CB, 0x0074CD, 0x0074CE, 0x0074D1, 0x0074D5, 0x0074D7, 0x0074D9, 0x0074DD, 0x0074E1, 0x0074E5, 0x0074EC, 0x0074ED, 0x0074F3, 0x0074F5, 0x0074F9, 0x0074FD, 0x0074FE, 0x007500,
	0x007502, 0x007507, 0x007508, 0x007509, 0x00750A, 0x00750B, 0x007510, 0x007514, 0x00751B, 0x00751D, 0x007522, 0x00752E, 0x007534, 0x007541, 0x007542, 0x007553, 0x007555, 0x007556, 0x007558, 0x007563, 0x007568, 0x00756E, 0x007575, 0x007580, 0x007584, 0x007588, 0x00758D, 0x007598, 0x00759B, 0x00759E, 0x0075A6, 0x0075A7,
	0x0075A8, 0x0075A9, 0x0075AA, 0x0075AD, 0x0075B6, 0x0075BB, 0x0075CB, 0x0075D0, 0x0075D1, 0x0075DA, 0x0075E5, 0x0075ED, 0x0075F5, 0x0075F6, 0x0075F7, 0x0075F8, 0x0075FB, 0x0075FD, 0x007606, 0x00760E, 0x007611, 0x007614, 0x00761A, 0x00762C, 0x00762E, 0x00762F, 0x007631, 0x007636, 0x007637, 0x00763D, 0x00764E, 0x00764F,
	0x007650, 0x007651, 0x007653, 0x007657, 0x00765A, 0x00765B, 0x00765D, 0x007660, 0x007666, 0x007673, 0x007675, 0x007677, 0x007679, 0x00767F, 0x007689, 0x00768A, 0x00768F, 0x007692, 0x007694, 0x007697, 0x007698, 0x0076A9, 0x0076AB, 0x0076AC, 0x0076B3, 0x0076B5, 0x0076BB, 0x0076BC, 0x0076BE, 0x0076C0, 0x0076C4, 0x0076C7,
	0x0076D3, 0x0076D5, 0x0076DA, 0x0076DD, 0x0076E2, 0x0076E9, 0x0076EB, 0x0076ED, 0x0076F3, 0x0076F5, 0x0076F7, 0x0076FA, 0x0076FD, 0x0076FF, 0x007702, 0x007703, 0x007705, 0x00770F, 0x007710, 0x007711, 0x007713, 0x007716, 0x007718, 0x00771D, 0x007721, 0x007723, 0x007727, 0x00772A, 0x00772B, 0x00772C, 0x007730, 0x007731,
	0x007732, 0x007733, 0x00773B, 0x00773F, 0x007744, 0x007748, 0x007749, 0x00774B, 0x00774C, 0x007753, 0x007754, 0x007755, 0x007759, 0x00775D, 0x007769, 0x00776D, 0x00776E, 0x00776F, 0x007771, 0x007775, 0x007776, 0x007777, 0x007778, 0x00777B, 0x007781, 0x007782, 0x007783, 0x007786, 0x007787, 0x007788, 0x007789, 0x00778A,
	0x00778F, 0x007790, 0x007793, 0x007797, 0x007798, 0x007799, 0x00779B, 0x00779C, 0x00779D, 0x0077A1, 0x0077A3, 0x0077A6, 0x0077A8, 0x0077AB, 0x0077B2, 0x0077B4, 0x0077B7, 0x0077B8, 0x0077BA, 0x0077C0, 0x0077C1, 0x0077C2, 0x0077C4, 0x0077C5, 0x0077C6, 0x0077C8, 0x0077CA, 0x0077CB, 0x0077CC, 0x0077CE, 0x0077CF, 0x0077D0,
	0x0077D3, 0x0077D4, 0x0077D6, 0x0077D8, 0x0077DD, 0x0077E1, 0x0077E8, 0x0077F2, 0x0077F5, 0x0077F7, 0x0077F9, 0x0077FA, 0x007803, 0x007804, 0x007807, 0x007808, 0x00780A, 0x00780B, 0x00780F, 0x007810, 0x007813, 0x00781B, 0x00781E, 0x007824, 0x007828, 0x00782A, 0x00782B, 0x00782F, 0x007831, 0x007833, 0x007836, 0x00783D,
	0x007841, 0x007842, 0x007846, 0x007849, 0x00784A, 0x00784B, 0x00784D, 0x007853, 0x007854, 0x007858, 0x007859, 0x00785A, 0x00785B, 0x00785F, 0x007862, 0x007865, 0x007866, 0x007867, 0x007869, 0x007870, 0x007871, 0x007873, 0x007875, 0x007876, 0x007878, 0x007879, 0x00787B, 0x00787D, 0x00787F, 0x007880, 0x007882, 0x007883,
	0x007884, 0x007885, 0x007888, 0x00788B, 0x007890, 0x007892, 0x007896, 0x007899, 0x0078A0, 0x0078A2, 0x0078A6, 0x0078AB, 0x0078AE, 0x0078B6, 0x0078B7, 0x0078B8, 0x0078C0, 0x0078C2, 0x0078C3, 0x0078C4, 0x0078CD, 0x0078CF, 0x0078D7, 0x0078D8, 0x0078DC, 0x0078DD, 0x0078DE, 0x0078E2, 0x0078E3, 0x0078E5, 0x0078E9, 0x0078EB,
	0x0078ED, 0x0078EE, 0x0078F0, 0x0078F1, 0x0078F5, 0x0078F8, 0x0078FC, 0x007902, 0x007903, 0x007904, 0x007908, 0x007909, 0x00790A, 0x00790B, 0x00790D, 0x00790F, 0x007914, 0x007915, 0x007916, 0x007917, 0x007918, 0x00791D, 0x007921, 0x007922, 0x007923, 0x007928, 0x00792F, 0x007932, 0x007933, 0x007936, 0x007937, 0x007938,
	0x007939, 0x007942, 0x007943, 0x00794C, 0x00794D, 0x00794E, 0x007952, 0x007959, 0x007961, 0x007963, 0x007964, 0x007966, 0x00796A, 0x00796C, 0x00796E, 0x007970, 0x007971, 0x007973, 0x007974, 0x007975, 0x007976, 0x00797D, 0x007982, 0x007983, 0x007986, 0x007987, 0x007988, 0x007989, 0x007990, 0x007992, 0x007997, 0x007999,
	0x00799E, 0x00799F, 0x0079A0, 0x0079A2, 0x0079A3, 0x0079A4, 0x0079A5, 0x0079AC, 0x0079AD, 0x0079B2, 0x0079B5, 0x0079B6, 0x0079B7, 0x0079BC, 0x0079C5, 0x0079CE, 0x0079D0, 0x0079D3, 0x0079D7, 0x0079D9, 0x0079DB, 0x0079DC, 0x0079E8, 0x0079EE, 0x0079F2, 0x0079F3, 0x0079F4, 0x0079F5, 0x0079F6, 0x0079F7, 0x0079F9, 0x0079FA,
	0x0079FE, 0x0079FF, 0x007A01, 0x007A04, 0x007A05, 0x007A0F, 0x007A10, 0x007A12, 0x007A13, 0x007A16, 0x007A1D, 0x007A22, 0x007A24, 0x007A25, 0x007A26, 0x007A28, 0x007A29, 0x007A2A, 0x007A2C, 0x007A36, 0x007A41, 0x007A4A, 0x007A4B, 0x007A52, 0x007A53, 0x007A54, 0x007A58, 0x007A5A, 0x007A5B, 0x007A5E, 0x007A64, 0x007A66,
	0x007A68, 0x007A6C, 0x007A6E, 0x007A6F, 0x007A71, 0x007A72, 0x007A73, 0x007A7B, 0x007A7C, 0x007A87, 0x007A89, 0x007A8C, 0x007A8E, 0x007A8F, 0x007A99, 0x007A9A, 0x007A9B, 0x007AA1, 0x007AA2, 0x007AA4, 0x007AA7, 0x007AAB, 0x007AB1, 0x007AB2, 0x007AB4, 0x007AB7, 0x007AB8, 0x007ABD, 0x007AC0, 0x007AC1, 0x007AC2, 0x007AD0,
	0x007AD4, 0x007AD7, 0x007AD8, 0x007AE4, 0x007AEE, 0x007AF2, 0x007AF3, 0x007AF5, 0x007AF7, 0x007AFC, 0x007B00, 0x007B01, 0x007B05, 0x007B09, 0x007B0C, 0x007B0D, 0x007B0E, 0x007B10, 0x007B13, 0x007B16, 0x007B17, 0x007B1A, 0x007B1C, 0x007B1D, 0x007B21, 0x007B22, 0x007B32, 0x007B37, 0x007B42, 0x007B43, 0x007B44, 0x007B4A,
	0x007B53, 0x007B57, 0x007B59, 0x007B5C, 0x007B5E, 0x007B5F, 0x007B61, 0x007B63, 0x007B68, 0x007B6B, 0x007B76, 0x007B78, 0x007B7C, 0x007B7D, 0x007B81, 0x007B82, 0x007B83, 0x007B88, 0x007B8A, 0x007B8C, 0x007B93, 0x007BA3, 0x007BA4, 0x007BAE, 0x007BB3, 0x007BB7, 0x007BB9, 0x007BBE, 0x007BBF, 0x007BC3, 0x007BCD, 0x007BCE,
	0x007BD0, 0x007BD2, 0x007BD5, 0x007BD8, 0x007BDC, 0x007BDE, 0x007BDF, 0x007BE2, 0x007BE3, 0x007BE7, 0x007BEB, 0x007BEC, 0x007BEF, 0x007BFB, 0x007BFD, 0x007BFF, 0x007C05, 0x007C08, 0x007C0A, 0x007C10, 0x007C15, 0x007C18, 0x007C1A, 0x007C1C, 0x007C1D, 0x007C22, 0x007C24, 0x007C29, 0x007C2D, 0x007C2E, 0x007C2F, 0x007C30,
	0x007C32, 0x007C35, 0x007C3B, 0x007C3C, 0x007C42, 0x007C44, 0x007C47, 0x007C48, 0x007C49, 0x007C4B, 0x007C4E, 0x007C62, 0x007C66, 0x007C68, 0x007C6A, 0x007C6B, 0x007C71, 0x007C76, 0x007C77, 0x007C78, 0x007C7A, 0x007C7F, 0x007C80, 0x007C84, 0x007C85, 0x007C88, 0x007C8A, 0x007C8C, 0x007C8E, 0x007C93, 0x007C96, 0x007C99,
	0x007C9A, 0x007CA3, 0x007CA9, 0x007CAC, 0x007CAF, 0x007CB4, 0x007CB5, 0x007CB8, 0x007CC3, 0x007CC6, 0x007CCB, 0x007CD0, 0x007CD1, 0x007CDB, 0x007CE1, 0x007CE3, 0x007CE4, 0x007CE5, 0x007CEA, 0x007CEC, 0x007CED, 0x007CEE, 0x007CF0, 0x007CF1, 0x007CF3, 0x007CF7, 0x007CF9, 0x007CFC, 0x007CFD, 0x007CFF, 0x007D01, 0x007D0C,
	0x007D0E, 0x007D1F, 0x007D24, 0x007D25, 0x007D28, 0x007D29, 0x007D34, 0x007D36, 0x007D37, 0x007D38, 0x007D3B, 0x007D49, 0x007D4A, 0x007D52, 0x007D54, 0x007D55, 0x007D58, 0x007D5F, 0x007D60, 0x007D64, 0x007D69, 0x007D6B, 0x007D6C, 0x007D6D, 0x007D6F, 0x007D74, 0x007D7C, 0x007D7E, 0x007D80, 0x007D84, 0x007D87, 0x007D8A,
	0x007D8E, 0x007D90, 0x007D92, 0x007D94, 0x007D95, 0x007D98, 0x007DA1, 0x007DA4, 0x007DA5, 0x007DA8, 0x007DA9, 0x007DBC, 0x007DC1, 0x007DC8, 0x007DC9, 0x007DD0, 0x007DD3, 0x007DD4, 0x007DDB, 0x007DDF, 0x007DE7, 0x007DEE, 0x007DF0, 0x007DF3, 0x007DF7, 0x007DF8, 0x007DFC, 0x007DFD, 0x007DFE, 0x007DFF, 0x007E02, 0x007E03,
	0x007E06, 0x007E07, 0x007E0C, 0x007E0D, 0x007E0E, 0x007E0F, 0x007E13, 0x007E14, 0x007E16, 0x007E18, 0x007E19, 0x007E1A, 0x007E24, 0x007E25, 0x007E29, 0x007E2A, 0x007E30, 0x007E34, 0x007E38, 0x007E3C, 0x007E40, 0x007E42, 0x007E48, 0x007E49, 0x007E4C, 0x007E4F, 0x007E51, 0x007E53, 0x007E57, 0x007E5B, 0x007E5C, 0x007E60,
	0x007E63, 0x007E64, 0x007E68, 0x007E6C, 0x007E71, 0x007E72, 0x007E74, 0x007E76, 0x007E77, 0x007E7A, 0x007E80, 0x007E84, 0x007E85, 0x007E8B, 0x007E97, 0x007E99, 0x007EAE, 0x007EB4, 0x007EBB, 0x007EBC, 0x007ED6, 0x007EE4, 0x007EEC, 0x007EF9, 0x007F0A, 0x007F10, 0x007F1E, 0x007F37, 0x007F39, 0x007F40, 0x007F41, 0x007F46,
	0x007F48, 0x007F49, 0x007F4A, 0x007F4B, 0x007F56, 0x007F59, 0x007F5E, 0x007F6C, 0x007F6F, 0x007F73, 0x007F76, 0x007F7A, 0x007F7B, 0x007F7C, 0x007F84, 0x007F89, 0x007F92, 0x007F93, 0x007F95, 0x007F98, 0x007F99, 0x007F9B, 0x007FA0, 0x007FA5, 0x007FAB, 0x007FAC, 0x007FB1, 0x007FB3, 0x007FB5, 0x007FB7, 0x007FBA, 0x007FBB,
	0x007FBE, 0x007FC2, 0x007FC4, 0x007FC7, 0x007FC9, 0x007FCB, 0x007FCD, 0x007FD0, 0x007FD1, 0x007FD3, 0x007FD6, 0x007FD7, 0x007FD9, 0x007FDA, 0x007FDC, 0x007FDD, 0x007FDE, 0x007FE2, 0x007FE4, 0x007FE7, 0x007FEA, 0x007FED, 0x007FF4, 0x007FF5, 0x007FF6, 0x007FF7, 0x007FF8, 0x008009, 0x00801A, 0x00801B, 0x008023, 0x00802B,
	0x00802D, 0x00802F, 0x008032, 0x008041, 0x008045, 0x008047, 0x008048, 0x008049, 0x00804E, 0x00804F, 0x008050, 0x008051, 0x008053, 0x008055, 0x008057, 0x008059, 0x00805B, 0x00805C, 0x00805D, 0x008063, 0x008065, 0x008067, 0x00806B, 0x00806C, 0x00806E, 0x008078, 0x00807A, 0x00807C, 0x008082, 0x00808A, 0x00808D, 0x00808F,
	0
};

static const unsigned map_kanji_gbk4_2[] = {
	0x008090, 0x008091, 0x008092, 0x008094, 0x008095, 0x008097, 0x0080A3, 0x0080A8, 0x0080B0, 0x0080B3, 0x0080B5, 0x0080B6, 0x0080BB, 0x0080C7, 0x0080C9, 0x0080CB, 0x0080D0, 0x0080D1, 0x0080D3, 0x0080DF, 0x0080E2, 0x0080E3, 0x0080F7, 0x0080FF, 0x008100, 0x008101, 0x008104, 0x00810C, 0x008115, 0x008119, 0x00811D, 0x00811F,
	0x008121, 0x008122, 0x008125, 0x008126, 0x008128, 0x00812A, 0x00812B, 0x00812D, 0x00812E, 0x008134, 0x008137, 0x00813B, 0x00813F, 0x008140, 0x008141, 0x008142, 0x008143, 0x008144, 0x008149, 0x00814D, 0x00814F, 0x008156, 0x008158, 0x00815B, 0x00815C, 0x00815D, 0x00815E, 0x008162, 0x008163, 0x008164, 0x00816A, 0x00816C,
	0x008172, 0x008173, 0x008175, 0x008176, 0x008187, 0x008189, 0x00818C, 0x00818D, 0x008192, 0x008194, 0x008197, 0x008199, 0x00819F, 0x0081A1, 0x0081A5, 0x0081A7, 0x0081AB, 0x0081AC, 0x0081AD, 0x0081AF, 0x0081B1, 0x0081B6, 0x0081B7, 0x0081B9, 0x0081BC, 0x0081C4, 0x0081C7, 0x0081D0, 0x0081D2, 0x0081D4, 0x0081DC, 0x0081E2,
	0x0081E6, 0x0081E9, 0x0081EE, 0x0081EF, 0x0081F7, 0x00820B, 0x008211, 0x008215, 0x008220, 0x008224, 0x008225, 0x008226, 0x008227, 0x00823D, 0x00823F, 0x008241, 0x008242, 0x008248, 0x00824A, 0x00824C, 0x00824D, 0x008250, 0x008252, 0x008253, 0x008254, 0x008255, 0x00825B, 0x00825E, 0x008261, 0x008265, 0x008269, 0x00826C,
	0x008275, 0x00827C, 0x008285, 0x008286, 0x00828C, 0x008290, 0x008295, 0x00829E, 0x0082A2, 0x0082B5, 0x0082B6, 0x0082C0, 0x0082C2, 0x0082C3, 0x0082C9, 0x0082D6, 0x0082DD, 0x0082E9, 0x0082EC, 0x0082EE, 0x0082F0, 0x0082F2, 0x0082F5, 0x0082F8, 0x0082FC, 0x0082FF, 0x00830D, 0x008310, 0x008312, 0x008313, 0x008319, 0x008320,
	0x008324, 0x008325, 0x008326, 0x008329, 0x00832A, 0x00833B, 0x00833E, 0x00833F, 0x008341, 0x008348, 0x00834B, 0x00834C, 0x008359, 0x00835D, 0x008371, 0x008372, 0x008374, 0x008376, 0x008379, 0x00837A, 0x00837E, 0x008381, 0x008383, 0x008388, 0x00838B, 0x00838C, 0x00838F, 0x008390, 0x008391, 0x008397, 0x0083A1, 0x0083A3,
	0x0083A4, 0x0083A5, 0x0083AE, 0x0083AF, 0x0083BB, 0x0083C2, 0x0083C3, 0x0083C4, 0x0083C6, 0x0083C8, 0x0083CB, 0x0083CD, 0x0083D2, 0x0083D5, 0x0083D7, 0x0083D9, 0x0083DA, 0x0083DB, 0x0083DE, 0x0083E2, 0x0083E3, 0x0083E4, 0x0083E6, 0x0083E7, 0x0083EC, 0x0083ED, 0x0083EE, 0x0083F3, 0x0083F5, 0x0083FA, 0x0083FE, 0x0083FF,
	0x008400, 0x008402, 0x008405, 0x008408, 0x008409, 0x008410, 0x008412, 0x008414, 0x008416, 0x00841A, 0x00841B, 0x00841E, 0x00841F, 0x008421, 0x008423, 0x00842B, 0x00842D, 0x00842E, 0x008430, 0x008432, 0x008433, 0x008434, 0x008436, 0x008437, 0x00843A, 0x00843B, 0x00843E, 0x00843F, 0x008440, 0x008441, 0x008442, 0x008443,
	0x008444, 0x00844B, 0x00844C, 0x008450, 0x008453, 0x008454, 0x008455, 0x00845D, 0x00845E, 0x008468, 0x008472, 0x00847B, 0x00847E, 0x00847F, 0x008480, 0x008483, 0x008486, 0x00848A, 0x00848D, 0x00848F, 0x008491, 0x008496, 0x008498, 0x00849A, 0x00849B, 0x00849D, 0x0084A0, 0x0084A2, 0x0084A3, 0x0084A4, 0x0084A5, 0x0084A7,
	0x0084AB, 0x0084AC, 0x0084AE, 0x0084B0, 0x0084B3, 0x0084B5, 0x0084B6, 0x0084B7, 0x0084C3, 0x0084C5, 0x0084D2, 0x0084D4, 0x0084D5, 0x0084D7, 0x0084D8, 0x0084DB, 0x0084DE, 0x0084E1, 0x0084E2, 0x0084E4, 0x0084E8, 0x0084E9, 0x0084EB, 0x0084ED, 0x0084F3, 0x0084F5, 0x0084F6, 0x0084F8, 0x0084F9, 0x0084FE, 0x008501, 0x008504,
	0x008505, 0x008508, 0x008509, 0x00850A, 0x00850B, 0x00850D, 0x00850F, 0x008512, 0x008516, 0x008519, 0x00851D, 0x008520, 0x008528, 0x008529, 0x00852E, 0x008530, 0x008531, 0x008542, 0x008544, 0x008545, 0x008547, 0x00854C, 0x00854D, 0x008554, 0x00855B, 0x008565, 0x008566, 0x008567, 0x00856C, 0x00856E, 0x008570, 0x008571,
	0x008573, 0x008575, 0x008576, 0x008578, 0x00857C, 0x008582, 0x008583, 0x00858D, 0x00858E, 0x008592, 0x008595, 0x008596, 0x00859A, 0x00859E, 0x0085A1, 0x0085A3, 0x0085B1, 0x0085B2, 0x0085B3, 0x0085B5, 0x0085BB, 0x0085C0, 0x0085C3, 0x0085C4, 0x0085C5, 0x0085C6, 0x0085C8, 0x0085CC, 0x0085D1, 0x0085D2, 0x0085D4, 0x0085D6,
	0x0085D7, 0x0085DB, 0x0085DE, 0x0085E2, 0x0085E3, 0x0085E7, 0x0085EB, 0x0085EC, 0x0085EE, 0x0085EF, 0x0085F0, 0x0085F1, 0x0085F2, 0x0085F4, 0x0085F5, 0x0085F8, 0x0085FD, 0x008601, 0x008603, 0x008608, 0x008609, 0x00860C, 0x00860F, 0x008614, 0x008615, 0x00861C, 0x00861D, 0x00861F, 0x008620, 0x008623, 0x008625, 0x008626,
	0x008628, 0x00862A, 0x00862B, 0x00862C, 0x00862E, 0x008631, 0x008632, 0x008633, 0x008634, 0x008635, 0x008637, 0x00863B, 0x00863E, 0x008643, 0x008644, 0x008645, 0x008647, 0x008648, 0x008649, 0x00864A, 0x00864B, 0x00864C, 0x008665, 0x008666, 0x008668, 0x00866A, 0x00866D, 0x008670, 0x008672, 0x008673, 0x008674, 0x008678,
	0x008683, 0x008684, 0x008685, 0x008686, 0x00868E, 0x00868F, 0x008690, 0x008692, 0x008694, 0x008697, 0x008699, 0x00869B, 0x00869E, 0x00869F, 0x0086A0, 0x0086A2, 0x0086A5, 0x0086AE, 0x0086B2, 0x0086BB, 0x0086BC, 0x0086BD, 0x0086BE, 0x0086C2, 0x0086C8, 0x0086CC, 0x0086D3, 0x0086D6, 0x0086DD, 0x0086E1, 0x0086E2, 0x0086E8,
	0x0086EA, 0x0086EB, 0x0086F5, 0x0086F6, 0x0086F7, 0x0086FF, 0x008701, 0x00870C, 0x008716, 0x00871B, 0x00871D, 0x008720, 0x008724, 0x008726, 0x008727, 0x00872A, 0x00872B, 0x00872C, 0x00872D, 0x008730, 0x008733, 0x008735, 0x008736, 0x008738, 0x008741, 0x008742, 0x008744, 0x008746, 0x00874A, 0x00874F, 0x008750, 0x008752,
	0x008754, 0x008756, 0x00875A, 0x00875B, 0x00875C, 0x00875E, 0x008762, 0x008767, 0x008769, 0x00876B, 0x00876C, 0x00876D, 0x008773, 0x008775, 0x008777, 0x008779, 0x00877A, 0x008780, 0x008781, 0x00878A, 0x00878E, 0x00878F, 0x008791, 0x008792, 0x008794, 0x008796, 0x00879A, 0x00879B, 0x00879C, 0x00879D, 0x0087A1, 0x0087A4,
	0x0087A5, 0x0087A6, 0x0087A9, 0x0087AA, 0x0087B0, 0x0087B2, 0x0087B4, 0x0087B6, 0x0087B7, 0x0087B8, 0x0087B9, 0x0087BC, 0x0087C2, 0x0087C3, 0x0087C5, 0x0087CC, 0x0087CD, 0x0087CF, 0x0087D4, 0x0087D7, 0x0087D8, 0x0087DD, 0x0087DE, 0x0087E1, 0x0087E7, 0x0087E8, 0x0087E9, 0x0087F0, 0x0087F4, 0x0087FC, 0x0087FD, 0x008800,
	0x008802, 0x008804, 0x008808, 0x00880C, 0x008817, 0x00881D, 0x008820, 0x008824, 0x008825, 0x008826, 0x008829, 0x00882A, 0x00882B, 0x00882C, 0x00882F, 0x008833, 0x008834, 0x008837, 0x008838, 0x00883D, 0x00883E, 0x00883F, 0x008847, 0x00884F, 0x008850, 0x008866, 0x008867, 0x00886A, 0x00886D, 0x008873, 0x008874, 0x008876,
	0x008878, 0x00887A, 0x00887C, 0x008883, 0x008886, 0x008887, 0x008889, 0x00888A, 0x00888C, 0x00888E, 0x00888F, 0x008890, 0x008891, 0x008893, 0x008894, 0x008895, 0x00889D, 0x0088A1, 0x0088A3, 0x0088A5, 0x0088A6, 0x0088A7, 0x0088A9, 0x0088AC, 0x0088AF, 0x0088B2, 0x0088B3, 0x0088B6, 0x0088B8, 0x0088B9, 0x0088BB, 0x0088C7,
	0x0088C8, 0x0088D0, 0x0088D6, 0x0088D7, 0x0088DA, 0x0088E0, 0x0088E6, 0x0088E9, 0x0088EA, 0x0088EB, 0x0088EC, 0x0088ED, 0x0088EE, 0x0088F6, 0x0088FA, 0x0088FB, 0x0088FF, 0x008900, 0x008903, 0x008905, 0x008908, 0x008909, 0x00890B, 0x008911, 0x008914, 0x008917, 0x00891F, 0x008922, 0x008923, 0x008924, 0x008929, 0x00892C,
	0x00892D, 0x00892E, 0x00892F, 0x008933, 0x00893C, 0x00893D, 0x00893F, 0x008947, 0x008948, 0x00894A, 0x00894B, 0x00894E, 0x008950, 0x008951, 0x008953, 0x008954, 0x008955, 0x008958, 0x008959, 0x00895D, 0x008965, 0x008967, 0x008968, 0x008969, 0x00896C, 0x008971, 0x008976, 0x008978, 0x008979, 0x008982, 0x008984, 0x008985,
	0x00898C, 0x00898E, 0x008991, 0x008992, 0x008999, 0x00899D, 0x00899E, 0x0089A2, 0x0089A3, 0x0089A4, 0x0089A8, 0x0089AB, 0x0089AD, 0x0089AE, 0x0089B1, 0x0089B8, 0x0089B9, 0x0089BB, 0x0089BE, 0x0089C3, 0x0089CD, 0x0089D3, 0x0089D9, 0x0089DB, 0x0089DF, 0x0089E0, 0x0089E1, 0x0089E2, 0x0089E4, 0x0089E8, 0x0089EA, 0x0089EC,
	0x0089EE, 0x0089F0, 0x0089F2, 0x0089F5, 0x0089F7, 0x0089FA, 0x0089FB, 0x0089FC, 0x0089FE, 0x008A01, 0x008A06, 0x008A09, 0x008A0B, 0x008A0D, 0x008A19, 0x008A1A, 0x008A1C, 0x008A27, 0x008A28, 0x008A29, 0x008A2E, 0x008A30, 0x008A32, 0x008A38, 0x008A39, 0x008A3F, 0x008A42, 0x008A44, 0x008A4A, 0x008A4B, 0x008A4C, 0x008A4F,
	0x008A59, 0x008A5A, 0x008A5F, 0x008A64, 0x008A68, 0x008A6A, 0x008A6F, 0x008A74, 0x008A78, 0x008A7D, 0x008A81, 0x008A88, 0x008A8E, 0x008A94, 0x008A9B, 0x008A9C, 0x008A9D, 0x008AA2, 0x008AAA, 0x008AAB, 0x008AB1, 0x008AB4, 0x008AB5, 0x008AB8, 0x008ABA, 0x008ABD, 0x008AC0, 0x008AC1, 0x008AC5, 0x008ACE, 0x008AD8, 0x008AD9,
	0x008AE3, 0x008AE5, 0x008AE8, 0x008AE9, 0x008AEA, 0x008AEF, 0x008AF2, 0x008AF9, 0x008AFB, 0x008AFD, 0x008B03, 0x008B08, 0x008B09, 0x008B0F, 0x008B12, 0x008B13, 0x008B15, 0x008B18, 0x008B22, 0x008B23, 0x008B24, 0x008B25, 0x008B27, 0x008B29, 0x008B2A, 0x008B2E, 0x008B2F, 0x008B31, 0x008B32, 0x008B34, 0x008B35, 0x008B36,
	0x008B38, 0x008B3A, 0x008B3B, 0x008B3D, 0x008B3F, 0x008B40, 0x008B47, 0x008B4A, 0x008B4B, 0x008B50, 0x008B55, 0x008B57, 0x008B5D, 0x008B60, 0x008B61, 0x008B62, 0x008B64, 0x008B65, 0x008B67, 0x008B68, 0x008B6A, 0x008B6E, 0x008B73, 0x008B75, 0x008B7A, 0x008B7B, 0x008B82, 0x008B86, 0x008B87, 0x008B88, 0x008B89, 0x008B91,
	0x008B97, 0x008B98, 0x008B9B, 0x008BAC, 0x008BB1, 0x008BBB, 0x008BC7, 0x008BD0, 0x008BEA, 0x008C09, 0x008C1E, 0x008C3B, 0x008C3C, 0x008C40, 0x008C42, 0x008C43, 0x008C44, 0x008C4D, 0x008C52, 0x008C56, 0x008C5C, 0x008C5E, 0x008C5F, 0x008C60, 0x008C65, 0x008C67, 0x008C6E, 0x008C6F, 0x008C70, 0x008C71, 0x008C72, 0x008C74,
	0x008C77, 0x008C7D, 0x008C7F, 0x008C80, 0x008C81, 0x008C83, 0x008C84, 0x008C88, 0x008C8F, 0x008C91, 0x008C95, 0x008C96, 0x008C97, 0x008C9A, 0x008C9F, 0x008CA3, 0x008CA5, 0x008CA6, 0x008CB1, 0x008CB5, 0x008CBE, 0x008CCC, 0x008CD0, 0x008CD2, 0x008CD4, 0x008CD7, 0x008CD8, 0x008CDF, 0x008CE5, 0x008CE7, 0x008CE9, 0x008CEB,
	0x008CEE, 0x008CF3, 0x008CF6, 0x008CF9, 0x008D00, 0x008D02, 0x008D06, 0x008D0C, 0x008D11, 0x008D15, 0x008D18, 0x008D19, 0x008D1A, 0x008D1C, 0x008D20, 0x008D51, 0x008D52, 0x008D57, 0x008D5F, 0x008D68, 0x008D6A, 0x008D6F, 0x008D72, 0x008D78, 0x008D79, 0x008D7A, 0x008D7B, 0x008D7C, 0x008D7D, 0x008D7E, 0x008D80, 0x008D83,
	0x008D86, 0x008D87, 0x008D89, 0x008D8C, 0x008D8E, 0x008D8F, 0x008D92, 0x008D93, 0x008D96, 0x008D97, 0x008D98, 0x008D9A, 0x008D9B, 0x008D9C, 0x008D9D, 0x008DA1, 0x008DA2, 0x008DA4, 0x008DA5, 0x008DA7, 0x008DA9, 0x008DAA, 0x008DAD, 0x008DAE, 0x008DB0, 0x008DB6, 0x008DBD, 0x008DC1, 0x008DC9, 0x008DCD, 0x008DD0, 0x008DD2,
	0x008DD3, 0x008DD8, 0x008DDC, 0x008DE0, 0x008DE2, 0x008DE6, 0x008DE9, 0x008DED, 0x008DEE, 0x008DF6, 0x008DFE, 0x008E00, 0x008E02, 0x008E03, 0x008E07, 0x008E0D, 0x008E0E, 0x008E12, 0x008E13, 0x008E15, 0x008E17, 0x008E18, 0x008E19, 0x008E1A, 0x008E1B, 0x008E1C, 0x008E24, 0x008E25, 0x008E28, 0x008E2B, 0x008E2D, 0x008E32,
	0
};

static const unsigned map_kanji_gbk4_3[] = {
	0x008E3B, 0x008E3C, 0x008E3E, 0x008E3F, 0x008E43, 0x008E45, 0x008E46, 0x008E53, 0x008E56, 0x008E57, 0x008E58, 0x008E5A, 0x008E65, 0x008E67, 0x008E68, 0x008E6A, 0x008E6B, 0x008E6E, 0x008E73, 0x008E75, 0x008E77, 0x008E78, 0x008E7D, 0x008E7E, 0x008E80, 0x008E86, 0x008E88, 0x008E8C, 0x008E8E, 0x008E96, 0x008E97, 0x008E9F,
	0x008EA0, 0x008EA3, 0x008EA4, 0x008EA5, 0x008EA6, 0x008EA8, 0x008EB4, 0x008EB7, 0x008EB8, 0x008EB9, 0x008EBC, 0x008EBD, 0x008EBF, 0x008EC2, 0x008EC9, 0x008ED0, 0x008ED3, 0x008ED5, 0x008ED6, 0x008ED7, 0x008ED8, 0x008ED9, 0x008EDA, 0x008EDD, 0x008EDE, 0x008EE0, 0x008EE1, 0x008EE4, 0x008EE5, 0x008EE6, 0x008EE7, 0x008EE9,
	0x008EEA, 0x008EEC, 0x008EEF, 0x008EF2, 0x008EF3, 0x008EF4, 0x008EF5, 0x008EF6, 0x008EFF, 0x008F01, 0x008F04, 0x008F06, 0x008F0B, 0x008F0D, 0x008F0E, 0x008F11, 0x008F1A, 0x008F22, 0x008F24, 0x008F30, 0x008F31, 0x008F32, 0x008F3C, 0x008F3D, 0x008F48, 0x008F4B, 0x008F50, 0x008F56, 0x008F59, 0x008F5A, 0x008F5B, 0x008F60,
	0x008F6A, 0x008F80, 0x008F8C, 0x008F92, 0x008FA2, 0x008FAA, 0x008FAC, 0x008FB3, 0x008FC3, 0x008FC9, 0x008FCC, 0x008FCF, 0x008FD6, 0x008FD7, 0x008FE1, 0x008FE7, 0x008FEC, 0x008FF2, 0x008FFC, 0x008FFF, 0x009007, 0x00901C, 0x009024, 0x009025, 0x00902B, 0x009030, 0x00903A, 0x00903D, 0x009040, 0x009046, 0x009048, 0x00905A,
	0x00905F, 0x009064, 0x00906A, 0x00906B, 0x009071, 0x009073, 0x00907B, 0x00907E, 0x009086, 0x00908D, 0x009092, 0x009094, 0x009096, 0x00909A, 0x00909C, 0x00909E, 0x00909F, 0x0090A4, 0x0090A7, 0x0090A9, 0x0090AB, 0x0090AD, 0x0090B7, 0x0090BC, 0x0090BF, 0x0090C0, 0x0090C2, 0x0090C6, 0x0090C9, 0x0090CB, 0x0090CD, 0x0090D4,
	0x0090D6, 0x0090DA, 0x0090E0, 0x0090E3, 0x0090E9, 0x0090EA, 0x0090EC, 0x0090EE, 0x0090F1, 0x0090F3, 0x0090F9, 0x0090FA, 0x0090FB, 0x0090FC, 0x009101, 0x009103, 0x009107, 0x009109, 0x00910A, 0x00910B, 0x00910C, 0x00910E, 0x00910F, 0x009111, 0x009113, 0x00911B, 0x00911D, 0x00911F, 0x009121, 0x009124, 0x009126, 0x009128,
	0x00912A, 0x00912B, 0x00912C, 0x009133, 0x009135, 0x009138, 0x00913B, 0x00913E, 0x00913F, 0x009140, 0x009141, 0x009142, 0x009144, 0x009145, 0x009151, 0x009155, 0x00915C, 0x00915F, 0x009160, 0x009166, 0x009168, 0x00916B, 0x00917C, 0x009180, 0x009184, 0x009188, 0x00918F, 0x009196, 0x009199, 0x00919D, 0x00919F, 0x0091A0,
	0x0091A5, 0x0091A7, 0x0091A9, 0x0091B7, 0x0091B9, 0x0091BE, 0x0091D2, 0x0091D5, 0x0091E0, 0x0091E2, 0x0091E8, 0x0091EB, 0x0091F2, 0x0091F3, 0x0091F4, 0x0091F8, 0x0091FA, 0x0091FE, 0x009202, 0x009203, 0x009208, 0x00920B, 0x00920F, 0x009219, 0x00921A, 0x00921B, 0x00921F, 0x009220, 0x009221, 0x009222, 0x009227, 0x00922A,
	0x00922B, 0x00922D, 0x009231, 0x009232, 0x00923B, 0x00923D, 0x009241, 0x00924C, 0x009252, 0x009253, 0x009254, 0x009255, 0x00925F, 0x009263, 0x00926A, 0x00926B, 0x00926C, 0x00926D, 0x009272, 0x009273, 0x00927A, 0x009281, 0x009282, 0x009284, 0x009286, 0x009287, 0x00928B, 0x00928C, 0x00928F, 0x009290, 0x009294, 0x00929D,
	0x00929E, 0x0092A1, 0x0092A2, 0x0092A3, 0x0092A6, 0x0092A9, 0x0092AA, 0x0092AC, 0x0092B0, 0x0092B1, 0x0092B3, 0x0092B4, 0x0092B5, 0x0092BE, 0x0092C4, 0x0092C9, 0x0092CA, 0x0092D1, 0x0092D4, 0x0092D6, 0x0092DA, 0x0092DB, 0x0092DE, 0x0092E2, 0x0092E6, 0x0092EB, 0x0092EF, 0x0092F1, 0x0092F4, 0x0092F5, 0x0092F6, 0x0092FD,
	0x0092FE, 0x009301, 0x009303, 0x009305, 0x009307, 0x009309, 0x00930A, 0x00930B, 0x00930C, 0x00930E, 0x009312, 0x009313, 0x009316, 0x009317, 0x00931B, 0x00932D, 0x009330, 0x009331, 0x009338, 0x009339, 0x00933C, 0x00933D, 0x00933F, 0x009340, 0x009341, 0x009342, 0x009343, 0x009345, 0x009346, 0x00934C, 0x00934E, 0x00934F,
	0x009353, 0x009359, 0x00935D, 0x00935F, 0x009361, 0x009362, 0x009363, 0x009366, 0x009368, 0x009372, 0x009377, 0x009378, 0x009379, 0x00937B, 0x009383, 0x009384, 0x009385, 0x009386, 0x009387, 0x009389, 0x00938E, 0x009390, 0x009391, 0x009393, 0x009399, 0x00939C, 0x00939D, 0x00939F, 0x0093A0, 0x0093A2, 0x0093A5, 0x0093AA,
	0x0093AF, 0x0093B1, 0x0093B2, 0x0093B3, 0x0093B7, 0x0093B8, 0x0093BB, 0x0093BC, 0x0093BD, 0x0093BE, 0x0093BF, 0x0093C0, 0x0093C2, 0x0093CE, 0x0093CF, 0x0093D2, 0x0093D4, 0x0093D5, 0x0093DA, 0x0093DB, 0x0093E0, 0x0093E3, 0x0093E9, 0x0093EA, 0x0093EB, 0x0093EC, 0x0093ED, 0x0093EE, 0x0093EF, 0x0093F0, 0x0093F2, 0x0093F3,
	0x0093F4, 0x0093F6, 0x0093FC, 0x0093FE, 0x0093FF, 0x009400, 0x009405, 0x009406, 0x00940A, 0x00940B, 0x00940C, 0x009411, 0x009412, 0x00941B, 0x00941C, 0x00941D, 0x00941E, 0x009420, 0x009422, 0x009423, 0x009424, 0x009425, 0x009426, 0x009427, 0x009428, 0x009429, 0x00942A, 0x00942C, 0x00942D, 0x009430, 0x009437, 0x009439,
	0x00943C, 0x009440, 0x009442, 0x009446, 0x009447, 0x009449, 0x00944B, 0x00944D, 0x00944E, 0x00944F, 0x009450, 0x009454, 0x009456, 0x009457, 0x009458, 0x00945D, 0x009464, 0x009465, 0x009466, 0x009467, 0x009469, 0x00946C, 0x009473, 0x009474, 0x009476, 0x009478, 0x009479, 0x00947A, 0x00947B, 0x009480, 0x009482, 0x009491,
	0x009496, 0x009498, 0x0094C7, 0x0094CF, 0x0094D3, 0x0094D4, 0x0094DA, 0x0094E6, 0x0094FB, 0x00951C, 0x009520, 0x009527, 0x009533, 0x00953D, 0x009543, 0x009548, 0x00954B, 0x009555, 0x00955A, 0x009560, 0x00956E, 0x009574, 0x009575, 0x00957A, 0x00957B, 0x00957C, 0x00957D, 0x009581, 0x009585, 0x009590, 0x009595, 0x009597,
	0x00959A, 0x00959B, 0x00959C, 0x0095AA, 0x0095AE, 0x0095AF, 0x0095B0, 0x0095B1, 0x0095B3, 0x0095B5, 0x0095B7, 0x0095B8, 0x0095C0, 0x0095C1, 0x0095C2, 0x0095C4, 0x0095C5, 0x0095CE, 0x0095CF, 0x0095D7, 0x0095DB, 0x0095E3, 0x0095E7, 0x0095EC, 0x0095FF, 0x009607, 0x009613, 0x009618, 0x00961B, 0x009620, 0x009623, 0x009627,
	0x009629, 0x00962B, 0x00962D, 0x009630, 0x00963E, 0x009643, 0x00964A, 0x00964E, 0x009651, 0x009653, 0x009659, 0x00965A, 0x009660, 0x00966B, 0x00966D, 0x00966F, 0x009671, 0x009679, 0x009680, 0x009687, 0x00968C, 0x009692, 0x009693, 0x00969E, 0x0096A1, 0x0096A2, 0x0096AB, 0x0096AC, 0x0096AD, 0x0096B5, 0x0096BF, 0x0096C2,
	0x0096C3, 0x0096C8, 0x0096D0, 0x0096D3, 0x0096D4, 0x0096D7, 0x0096E1, 0x0096E4, 0x0096E5, 0x0096E6, 0x0096E7, 0x0096EC, 0x0096ED, 0x0096EE, 0x0096F4, 0x0096F5, 0x0096F8, 0x0096FC, 0x0096FD, 0x0096FF, 0x00970B, 0x00970C, 0x009710, 0x009712, 0x009715, 0x009717, 0x009718, 0x00971F, 0x009720, 0x009725, 0x009726, 0x009729,
	0x00972B, 0x00972C, 0x00972E, 0x00972F, 0x009734, 0x009735, 0x009737, 0x00973A, 0x00973C, 0x00973F, 0x009740, 0x009745, 0x00974B, 0x00974C, 0x009750, 0x009754, 0x00975D, 0x00975F, 0x00976C, 0x00976F, 0x009770, 0x009772, 0x009775, 0x00977E, 0x009782, 0x009783, 0x009787, 0x009788, 0x00978A, 0x00978C, 0x00978E, 0x009793,
	0x00979B, 0x00979D, 0x0097A1, 0x0097A4, 0x0097A5, 0x0097A7, 0x0097A9, 0x0097AA, 0x0097B0, 0x0097B7, 0x0097BB, 0x0097BD, 0x0097C0, 0x0097C2, 0x0097CF, 0x0097D2, 0x0097D5, 0x0097D6, 0x0097DA, 0x0097DF, 0x0097E2, 0x0097E3, 0x0097E5, 0x0097E8, 0x0097F0, 0x0097F9, 0x0097FC, 0x0097FD, 0x0097FE, 0x009800, 0x009809, 0x00980B,
	0x009815, 0x00981B, 0x00981D, 0x00981F, 0x009822, 0x009827, 0x009828, 0x009829, 0x00982A, 0x009831, 0x009836, 0x009839, 0x00983A, 0x00983F, 0x009840, 0x009841, 0x009842, 0x009843, 0x009845, 0x009848, 0x009849, 0x009850, 0x00985C, 0x00985D, 0x00985F, 0x009860, 0x009861, 0x009864, 0x009868, 0x009869, 0x00986D, 0x00986E,
	0x009872, 0x00988B, 0x00988E, 0x009892, 0x009895, 0x009899, 0x0098A3, 0x0098A9, 0x0098AC, 0x0098B2, 0x0098B3, 0x0098B5, 0x0098B9, 0x0098BD, 0x0098BE, 0x0098C0, 0x0098C1, 0x0098C9, 0x0098CA, 0x0098CD, 0x0098CF, 0x0098D0, 0x0098D4, 0x0098D6, 0x0098D7, 0x0098DD, 0x0098E4, 0x0098EC, 0x0098F5, 0x0098F7, 0x0098F8, 0x0098F9,
	0x0098FA, 0x0098FB, 0x0098FF, 0x009900, 0x009901, 0x009904, 0x009906, 0x00990B, 0x00990E, 0x00990F, 0x009919, 0x009923, 0x009925, 0x009929, 0x00992A, 0x00992D, 0x00992F, 0x009930, 0x009936, 0x009937, 0x009938, 0x00993F, 0x009943, 0x009944, 0x00994A, 0x00994F, 0x009953, 0x009956, 0x00995A, 0x00995D, 0x009961, 0x009962,
	0x009964, 0x009966, 0x009973, 0x009978, 0x009979, 0x00997B, 0x00997E, 0x009982, 0x009983, 0x009989, 0x00998C, 0x00998E, 0x00999A, 0x00999C, 0x0099A0, 0x0099A1, 0x0099A2, 0x0099A4, 0x0099A7, 0x0099A9, 0x0099AA, 0x0099AB, 0x0099AF, 0x0099B6, 0x0099B7, 0x0099B8, 0x0099BB, 0x0099BE, 0x0099C0, 0x0099C2, 0x0099C7, 0x0099CA,
	0x0099CB, 0x0099CC, 0x0099CD, 0x0099CE, 0x0099CF, 0x0099D6, 0x0099D7, 0x0099E0, 0x0099E1, 0x0099E3, 0x0099E4, 0x0099E5, 0x0099E6, 0x0099E8, 0x0099E9, 0x0099EF, 0x0099F3, 0x0099F6, 0x0099F7, 0x0099FA, 0x009A00, 0x009A06, 0x009A07, 0x009A08, 0x009A09, 0x009A0D, 0x009A14, 0x009A15, 0x009A17, 0x009A18, 0x009A1B, 0x009A1C,
	0x009A1D, 0x009A1F, 0x009A21, 0x009A25, 0x009A26, 0x009A29, 0x009A2A, 0x009A2C, 0x009A2F, 0x009A32, 0x009A34, 0x009A39, 0x009A3A, 0x009A3B, 0x009A3C, 0x009A3D, 0x009A3F, 0x009A46, 0x009A48, 0x009A49, 0x009A4F, 0x009A50, 0x009A53, 0x009A59, 0x009A5C, 0x009A5E, 0x009A60, 0x009A61, 0x009A63, 0x009A66, 0x009A67, 0x009A68,
	0x009A72, 0x009A83, 0x009A89, 0x009A8D, 0x009A8E, 0x009A94, 0x009A95, 0x009A99, 0x009AA6, 0x009AA9, 0x009AAB, 0x009AB3, 0x009ABD, 0x009AC7, 0x009AC9, 0x009ACA, 0x009ACD, 0x009ADA, 0x009ADD, 0x009AE8, 0x009AF0, 0x009AF6, 0x009AF8, 0x009AFC, 0x009AFE, 0x009B07, 0x009B0A, 0x009B11, 0x009B14, 0x009B15, 0x009B17, 0x009B1D,
	0x009B1E, 0x009B21, 0x009B24, 0x009B2C, 0x009B30, 0x009B36, 0x009B38, 0x009B3E, 0x009B3F, 0x009B40, 0x009B46, 0x009B4A, 0x009B50, 0x009B52, 0x009B53, 0x009B59, 0x009B5C, 0x009B5D, 0x009B5F, 0x009B60, 0x009B62, 0x009B64, 0x009B67, 0x009B69, 0x009B70, 0x009B71, 0x009B7A, 0x009B7B, 0x009B7C, 0x009B7D, 0x009B7E, 0x009B81,
	0x009B82, 0x009B88, 0x009B8C, 0x009B95, 0x009B98, 0x009B99, 0x009B9B, 0x009B9C, 0x009BA1, 0x009BA2, 0x009BA3, 0x009BA4, 0x009BA5, 0x009BAF, 0x009BB3, 0x009BB5, 0x009BB6, 0x009BBA, 0x009BBD, 0x009BC2, 0x009BC3, 0x009BC4, 0x009BC5, 0x009BCB, 0x009BCC, 0x009BCD, 0x009BD3, 0x009BD5, 0x009BD9, 0x009BDA, 0x009BDC, 0x009BDE,
	0x009BE0, 0x009BE6, 0x009BE9, 0x009BEC, 0x009BED, 0x009BF4, 0x009BF6, 0x009BFB, 0x009BFC, 0x009BFE, 0x009C01, 0x009C03, 0x009C05, 0x009C07, 0x009C0E, 0x009C17, 0x009C1F, 0x009C20, 0x009C2B, 0x009C2C, 0x009C33, 0x009C34, 0x009C38, 0x009C3C, 0x009C3F, 0x009C40, 0x009C42, 0x009C4B, 0x009C4C, 0x009C4D, 0x009C51, 0x009C55,
	0x009C59, 0x009C61, 0x009C62, 0x009C64, 0x009C66, 0x009C6C, 0x009C6F, 0x009C71, 0x009C73, 0x009C74, 0x009C79, 0x009C7D, 0x009C7E, 0x009C80, 0x009C83, 0x009C84, 0x009C89, 0x009C8A, 0x009C8C, 0x009C8F, 0x009C93, 0x009C96, 0x009C97, 0x009C98, 0x009C99, 0x009C9D, 0x009CAA, 0x009CAC, 0x009CAF, 0x009CB9, 0x009CBE, 0x009CBF,
	0
};

static const unsigned map_kanji_gbk4_4[] = {
	0x009CC0, 0x009CC1, 0x009CC2, 0x009CC8, 0x009CC9, 0x009CD1, 0x009CD2, 0x009CDA, 0x009CDB, 0x009CE0, 0x009CE1, 0x009CE3, 0x009CE4, 0x009CE8, 0x009CEA, 0x009CED, 0x009CEE, 0x009CEF, 0x009CF1, 0x009CF5, 0x009CF8, 0x009CFA, 0x009CFB, 0x009CFC, 0x009CFD, 0x009CFE, 0x009CFF, 0x009D00, 0x009D01, 0x009D04, 0x009D05, 0x009D0A,
	0x009D0C, 0x009D0D, 0x009D0F, 0x009D10, 0x009D13, 0x009D14, 0x009D16, 0x009D19, 0x009D1A, 0x009D20, 0x009D21, 0x009D22, 0x009D24, 0x009D25, 0x009D27, 0x009D29, 0x009D2D, 0x009D2E, 0x009D31, 0x009D35, 0x009D36, 0x009D37, 0x009D38, 0x009D39, 0x009D40, 0x009D49, 0x009D4B, 0x009D4C, 0x009D4D, 0x009D4E, 0x009D4F, 0x009D55,
	0x009D56, 0x009D57, 0x009D58, 0x009D5A, 0x009D5B, 0x009D66, 0x009D67, 0x009D68, 0x009D6D, 0x009D6E, 0x009D71, 0x009D74, 0x009D75, 0x009D78, 0x009D79, 0x009D7D, 0x009D7F, 0x009D80, 0x009D81, 0x009D82, 0x009D85, 0x009D88, 0x009D8B, 0x009D8C, 0x009D90, 0x009D91, 0x009D94, 0x009D9B, 0x009D9C, 0x009D9D, 0x009D9E, 0x009D9F,
	0x009DA0, 0x009DA2, 0x009DA3, 0x009DA5, 0x009DA6, 0x009DA7, 0x009DA8, 0x009DAD, 0x009DB0, 0x009DB3, 0x009DB6, 0x009DB7, 0x009DBE, 0x009DC5, 0x009DC8, 0x009DCB, 0x009DCC, 0x009DCD, 0x009DCE, 0x009DD0, 0x009DD1, 0x009DD2, 0x009DD8, 0x009DDB, 0x009DDC, 0x009DDD, 0x009DE1, 0x009DE2, 0x009DE4, 0x009DE8, 0x009DEA, 0x009DEC,
	0x009DF1, 0x009DF5, 0x009DF6, 0x009DF7, 0x009DFB, 0x009DFC, 0x009DFF, 0x009E00, 0x009E01, 0x009E03, 0x009E04, 0x009E05, 0x009E06, 0x009E08, 0x009E09, 0x009E0B, 0x009E0C, 0x009E0F, 0x009E13, 0x009E14, 0x009E17, 0x009E18, 0x009E24, 0x009E27, 0x009E2E, 0x009E30, 0x009E34, 0x009E3B, 0x009E3C, 0x009E40, 0x009E4D, 0x009E50,
	0x009E52, 0x009E53, 0x009E54, 0x009E56, 0x009E59, 0x009E5D, 0x009E5F, 0x009E60, 0x009E61, 0x009E62, 0x009E65, 0x009E6E, 0x009E6F, 0x009E72, 0x009E74, 0x009E76, 0x009E77, 0x009E86, 0x009E89, 0x009E8A, 0x009E8D, 0x009E90, 0x009E94, 0x009E99, 0x009E9A, 0x009E9C, 0x009EA0, 0x009EA1, 0x009EA2, 0x009EA3, 0x009EA7, 0x009EAB,
	0x009EB1, 0x009EB2, 0x009EB6, 0x009EB7, 0x009EC0, 0x009EC1, 0x009EC2, 0x009EC5, 0x009EC7, 0x009ECA, 0x009ED3, 0x009ED6, 0x009ED7, 0x009EDA, 0x009EE1, 0x009EE3, 0x009EE6, 0x009EEB, 0x009EF3, 0x009EFA, 0x009F00, 0x009F01, 0x009F04, 0x009F05, 0x009F06, 0x009F0A, 0x009F0C, 0x009F18, 0x009F1C, 0x009F1D, 0x009F1E, 0x009F23,
	0x009F24, 0x009F25, 0x009F27, 0x009F28, 0x009F29, 0x009F2D, 0x009F2E, 0x009F30, 0x009F33, 0x009F35, 0x009F36, 0x009F38, 0x009F40, 0x009F42, 0x009F48, 0x009F49, 0x009F4C, 0x009F4D, 0x009F59, 0x009F5B, 0x009F5C, 0x009F64, 0x009F65, 0x009F6B, 0x009F74, 0x009F78, 0x009F79, 0x009F7B, 0x009F7C, 0x009F7E, 0x009F81, 0x009F82,
	0x009F8E, 0x009F93, 0x009F98, 0x009FA4, 0x00F92C, 0x00F979, 0x00F995, 0x00F9E7, 0x00F9F1, 0x00FA0C, 0x00FA0D,
	0
};

// GBK/5
static const unsigned map_symbol_gbk5[] = {
	0x0002CA, 0x0002CB, 0x002035, 0x002105, 0x002109, 0x002215, 0x002223, 0x002550, 0x002551, 0x002552, 0x002553, 0x002554, 0x002555, 0x002556, 0x002557, 0x002558, 0x002559, 0x00255A, 0x00255B, 0x00255C, 0x00255D, 0x00255E, 0x00255F, 0x002560, 0x002561, 0x002562, 0x002563, 0x002564, 0x002565, 0x002566, 0x002567, 0x002568,
	0x002569, 0x00256A, 0x00256B, 0x00256C, 0x00256D, 0x00256E, 0x00256F, 0x002570, 0x002571, 0x002572, 0x002573, 0x002581, 0x002582, 0x002583, 0x002584, 0x002585, 0x002586, 0x002587, 0x002588, 0x002589, 0x00258A, 0x00258B, 0x00258C, 0x00258D, 0x00258E, 0x00258F, 0x002593, 0x002594, 0x002595, 0x0025E2, 0x0025E3, 0x0025E4,
	0x0025E5, 0x002609, 0x0032A3, 0x0033CE, 0x0033D1, 0x0033D2, 0x0033D5, 0x00FE30, 0x00FE49, 0x00FE4A, 0x00FE4B, 0x00FE4C, 0x00FE4D, 0x00FE4E, 0x00FE4F, 0x00FE50, 0x00FE51, 0x00FE52, 0x00FE54, 0x00FE55, 0x00FE56, 0x00FE57, 0x00FE59, 0x00FE5A, 0x00FE5B, 0x00FE5C, 0x00FE5D, 0x00FE5E, 0x00FE5F, 0x00FE60, 0x00FE61, 0x00FE62,
	0x00FE63, 0x00FE64, 0x00FE65, 0x00FE66, 0x00FE68, 0x00FE69, 0x00FE6A, 0x00FE6B,
	0
};

// Hangul Jamo
static const unsigned map_jamo[] = {
	0x001100, 0x001101, 0x001102, 0x001103, 0x001104, 0x001105, 0x001106, 0x001107, 0x001108, 0x001109, 0x00110A, 0x00110B, 0x00110C, 0x00110D, 0x00110E, 0x00110F,
	0x001110, 0x001111, 0x001112, 0x001113, 0x001114, 0x001115, 0x001116, 0x001117, 0x001118, 0x001119, 0x00111A, 0x00111B, 0x00111C, 0x00111D, 0x00111E, 0x00111F,
	0x001120, 0x001121, 0x001122, 0x001123, 0x001124, 0x001125, 0x001126, 0x001127, 0x001128, 0x001129, 0x00112A, 0x00112B, 0x00112C, 0x00112D, 0x00112E, 0x00112F,
	0x001130, 0x001131, 0x001132, 0x001133, 0x001134, 0x001135, 0x001136, 0x001137, 0x001138, 0x001139, 0x00113A, 0x00113B, 0x00113C, 0x00113D, 0x00113E, 0x00113F,
	0x001140, 0x001141, 0x001142, 0x001143, 0x001144, 0x001145, 0x001146, 0x001147, 0x001148, 0x001149, 0x00114A, 0x00114B, 0x00114C, 0x00114D, 0x00114E, 0x00114F,
	0x001150, 0x001151, 0x001152, 0x001153, 0x001154, 0x001155, 0x001156, 0x001157, 0x001158, 0x001159, 0x00115A, 0x00115B, 0x00115C, 0x00115D, 0x00115E, 0x00115F,
	0x001160, 0x001161, 0x001162, 0x001163, 0x001164, 0x001165, 0x001166, 0x001167, 0x001168, 0x001169, 0x00116A, 0x00116B, 0x00116C, 0x00116D, 0x00116E, 0x00116F,
	0x001170, 0x001171, 0x001172, 0x001173, 0x001174, 0x001175, 0x001176, 0x001177, 0x001178, 0x001179, 0x00117A, 0x00117B, 0x00117C, 0x00117D, 0x00117E, 0x00117F,
	0x001180, 0x001181, 0x001182, 0x001183, 0x001184, 0x001185, 0x001186, 0x001187, 0x001188, 0x001189, 0x00118A, 0x00118B, 0x00118C, 0x00118D, 0x00118E, 0x00118F,
	0x001190, 0x001191, 0x001192, 0x001193, 0x001194, 0x001195, 0x001196, 0x001197, 0x001198, 0x001199, 0x00119A, 0x00119B, 0x00119C, 0x00119D, 0x00119E, 0x00119F,
	0x0011A0, 0x0011A1, 0x0011A2, 0x0011A3, 0x0011A4, 0x0011A5, 0x0011A6, 0x0011A7, 0x0011A8, 0x0011A9, 0x0011AA, 0x0011AB, 0x0011AC, 0x0011AD, 0x0011AE, 0x0011AF,
	0x0011B0, 0x0011B1, 0x0011B2, 0x0011B3, 0x0011B4, 0x0011B5, 0x0011B6, 0x0011B7, 0x0011B8, 0x0011B9, 0x0011BA, 0x0011BB, 0x0011BC, 0x0011BD, 0x0011BE, 0x0011BF,
	0x0011C0, 0x0011C1, 0x0011C2, 0x0011C3, 0x0011C4, 0x0011C5, 0x0011C6, 0x0011C7, 0x0011C8, 0x0011C9, 0x0011CA, 0x0011CB, 0x0011CC, 0x0011CD, 0x0011CE, 0x0011CF,
	0x0011D0, 0x0011D1, 0x0011D2, 0x0011D3, 0x0011D4, 0x0011D5, 0x0011D6, 0x0011D7, 0x0011D8, 0x0011D9, 0x0011DA, 0x0011DB, 0x0011DC, 0x0011DD, 0x0011DE, 0x0011DF,
	0x0011E0, 0x0011E1, 0x0011E2, 0x0011E3, 0x0011E4, 0x0011E5, 0x0011E6, 0x0011E7, 0x0011E8, 0x0011E9, 0x0011EA, 0x0011EB, 0x0011EC, 0x0011ED, 0x0011EE, 0x0011EF,
	0x0011F0, 0x0011F1, 0x0011F2, 0x0011F3, 0x0011F4, 0x0011F5, 0x0011F6, 0x0011F7, 0x0011F8, 0x0011F9, 0x0011FA, 0x0011FB, 0x0011FC, 0x0011FD, 0x0011FE, 0x0011FF,
	0
};

// Hangul Compatibility Jamo
static const unsigned map_jamo_ks[] = {
	0x003131, 0x003132, 0x003133, 0x003134, 0x003135, 0x003136, 0x003137, 0x003138, 0x003139, 0x00313A, 0x00313B, 0x00313C, 0x00313D, 0x00313E, 0x00313F, 0x003140,
	0x003141, 0x003142, 0x003143, 0x003144, 0x003145, 0x003146, 0x003147, 0x003148, 0x003149, 0x00314A, 0x00314B, 0x00314C, 0x00314D, 0x00314E, 0x00314F, 0x003150,
	0x003151, 0x003152, 0x003153, 0x003154, 0x003155, 0x003156, 0x003157, 0x003158, 0x003159, 0x00315A, 0x00315B, 0x00315C, 0x00315D, 0x00315E, 0x00315F, 0x003160,
	0x003161, 0x003162, 0x003163, 0x003164, 0x003165, 0x003166, 0x003167, 0x003168, 0x003169, 0x00316A, 0x00316B, 0x00316C, 0x00316D, 0x00316E, 0x00316F, 0x003170,
	0x003171, 0x003172, 0x003173, 0x003174, 0x003175, 0x003176, 0x003177, 0x003178, 0x003179, 0x00317A, 0x00317B, 0x00317C, 0x00317D, 0x00317E, 0x00317F, 0x003180,
	0x003181, 0x003182, 0x003183, 0x003184, 0x003185, 0x003186, 0x003187, 0x003188, 0x003189, 0x00318A, 0x00318B, 0x00318C, 0x00318D, 0x00318E,
	0
};

// Hangul Jamo Extended A
static const unsigned map_jamo_a[] = {
	0x00A960, 0x00A961, 0x00A962, 0x00A963, 0x00A964, 0x00A965, 0x00A966, 0x00A967,
	0x00A968, 0x00A969, 0x00A96A, 0x00A96B, 0x00A96C, 0x00A96D, 0x00A96E, 0x00A96F,
	0x00A970, 0x00A971, 0x00A972, 0x00A973, 0x00A974, 0x00A975, 0x00A976, 0x00A977,
	0x00A978, 0x00A979, 0x00A97A, 0x00A97B, 0x00A96C,
	0
};

// Hangul Jamo Extended B
static const unsigned map_jamo_b[] = {
	0x00D7B0, 0x00D7B1, 0x00D7B2, 0x00D7B3, 0x00D7B4, 0x00D7B5, 0x00D7B6, 0x00D7B7, 0x00D7B8, 0x00D7B9, 0x00D7BA, 0x00D7BB, 0x00D7BC, 0x00D7BD, 0x00D7BE, 0x00D7BF,
	0x00D7C0, 0x00D7C1, 0x00D7C2, 0x00D7C3, 0x00D7C4, 0x00D7C5, 0x00D7C6,                                         0x00D7CB, 0x00D7CC, 0x00D7CD, 0x00D7CE, 0x00D7CF,
	0x00D7D0, 0x00D7D1, 0x00D7D2, 0x00D7D3, 0x00D7D4, 0x00D7D5, 0x00D7D6, 0x00D7D7, 0x00D7D8, 0x00D7D9, 0x00D7DA, 0x00D7DB, 0x00D7DC, 0x00D7DD, 0x00D7DE, 0x00D7DF,
	0x00D7E0, 0x00D7E1, 0x00D7E2, 0x00D7E3, 0x00D7E4, 0x00D7E5, 0x00D7E6, 0x00D7E7, 0x00D7E8, 0x00D7E9, 0x00D7EA, 0x00D7EB, 0x00D7EC, 0x00D7ED, 0x00D7EE, 0x00D7EF,
	0x00D7F0, 0x00D7F1, 0x00D7F2, 0x00D7F3, 0x00D7F4, 0x00D7F5, 0x00D7F6, 0x00D7F7, 0x00D7F8, 0x00D7F9, 0x00D7FA, 0x00D7FB,
	0
};

// Hangul Syllables
static const unsigned map_hangul_1[] = {
	0x00AC00, 0x00AC01, 0x00AC02, 0x00AC03, 0x00AC04, 0x00AC05, 0x00AC06, 0x00AC07, 0x00AC08, 0x00AC09, 0x00AC0A, 0x00AC0B, 0x00AC0C, 0x00AC0D, 0x00AC0E, 0x00AC0F, 0x00AC10, 0x00AC11, 0x00AC12, 0x00AC13, 0x00AC14, 0x00AC15, 0x00AC16, 0x00AC17, 0x00AC18, 0x00AC19, 0x00AC1A, 0x00AC1B, 0x00AC1C, 0x00AC1D, 0x00AC1E, 0x00AC1F,
	0x00AC20, 0x00AC21, 0x00AC22, 0x00AC23, 0x00AC24, 0x00AC25, 0x00AC26, 0x00AC27, 0x00AC28, 0x00AC29, 0x00AC2A, 0x00AC2B, 0x00AC2C, 0x00AC2D, 0x00AC2E, 0x00AC2F, 0x00AC30, 0x00AC31, 0x00AC32, 0x00AC33, 0x00AC34, 0x00AC35, 0x00AC36, 0x00AC37, 0x00AC38, 0x00AC39, 0x00AC3A, 0x00AC3B, 0x00AC3C, 0x00AC3D, 0x00AC3E, 0x00AC3F,
	0x00AC40, 0x00AC41, 0x00AC42, 0x00AC43, 0x00AC44, 0x00AC45, 0x00AC46, 0x00AC47, 0x00AC48, 0x00AC49, 0x00AC4A, 0x00AC4B, 0x00AC4C, 0x00AC4D, 0x00AC4E, 0x00AC4F, 0x00AC50, 0x00AC51, 0x00AC52, 0x00AC53, 0x00AC54, 0x00AC55, 0x00AC56, 0x00AC57, 0x00AC58, 0x00AC59, 0x00AC5A, 0x00AC5B, 0x00AC5C, 0x00AC5D, 0x00AC5E, 0x00AC5F,
	0x00AC60, 0x00AC61, 0x00AC62, 0x00AC63, 0x00AC64, 0x00AC65, 0x00AC66, 0x00AC67, 0x00AC68, 0x00AC69, 0x00AC6A, 0x00AC6B, 0x00AC6C, 0x00AC6D, 0x00AC6E, 0x00AC6F, 0x00AC70, 0x00AC71, 0x00AC72, 0x00AC73, 0x00AC74, 0x00AC75, 0x00AC76, 0x00AC77, 0x00AC78, 0x00AC79, 0x00AC7A, 0x00AC7B, 0x00AC7C, 0x00AC7D, 0x00AC7E, 0x00AC7F,
	0x00AC80, 0x00AC81, 0x00AC82, 0x00AC83, 0x00AC84, 0x00AC85, 0x00AC86, 0x00AC87, 0x00AC88, 0x00AC89, 0x00AC8A, 0x00AC8B, 0x00AC8C, 0x00AC8D, 0x00AC8E, 0x00AC8F, 0x00AC90, 0x00AC91, 0x00AC92, 0x00AC93, 0x00AC94, 0x00AC95, 0x00AC96, 0x00AC97, 0x00AC98, 0x00AC99, 0x00AC9A, 0x00AC9B, 0x00AC9C, 0x00AC9D, 0x00AC9E, 0x00AC9F,
	0x00ACA0, 0x00ACA1, 0x00ACA2, 0x00ACA3, 0x00ACA4, 0x00ACA5, 0x00ACA6, 0x00ACA7, 0x00ACA8, 0x00ACA9, 0x00ACAA, 0x00ACAB, 0x00ACAC, 0x00ACAD, 0x00ACAE, 0x00ACAF, 0x00ACB0, 0x00ACB1, 0x00ACB2, 0x00ACB3, 0x00ACB4, 0x00ACB5, 0x00ACB6, 0x00ACB7, 0x00ACB8, 0x00ACB9, 0x00ACBA, 0x00ACBB, 0x00ACBC, 0x00ACBD, 0x00ACBE, 0x00ACBF,
	0x00ACC0, 0x00ACC1, 0x00ACC2, 0x00ACC3, 0x00ACC4, 0x00ACC5, 0x00ACC6, 0x00ACC7, 0x00ACC8, 0x00ACC9, 0x00ACCA, 0x00ACCB, 0x00ACCC, 0x00ACCD, 0x00ACCE, 0x00ACCF, 0x00ACD0, 0x00ACD1, 0x00ACD2, 0x00ACD3, 0x00ACD4, 0x00ACD5, 0x00ACD6, 0x00ACD7, 0x00ACD8, 0x00ACD9, 0x00ACDA, 0x00ACDB, 0x00ACDC, 0x00ACDD, 0x00ACDE, 0x00ACDF,
	0x00ACE0, 0x00ACE1, 0x00ACE2, 0x00ACE3, 0x00ACE4, 0x00ACE5, 0x00ACE6, 0x00ACE7, 0x00ACE8, 0x00ACE9, 0x00ACEA, 0x00ACEB, 0x00ACEC, 0x00ACED, 0x00ACEE, 0x00ACEF, 0x00ACF0, 0x00ACF1, 0x00ACF2, 0x00ACF3, 0x00ACF4, 0x00ACF5, 0x00ACF6, 0x00ACF7, 0x00ACF8, 0x00ACF9, 0x00ACFA, 0x00ACFB, 0x00ACFC, 0x00ACFD, 0x00ACFE, 0x00ACFF,
	0x00AD00, 0x00AD01, 0x00AD02, 0x00AD03, 0x00AD04, 0x00AD05, 0x00AD06, 0x00AD07, 0x00AD08, 0x00AD09, 0x00AD0A, 0x00AD0B, 0x00AD0C, 0x00AD0D, 0x00AD0E, 0x00AD0F, 0x00AD10, 0x00AD11, 0x00AD12, 0x00AD13, 0x00AD14, 0x00AD15, 0x00AD16, 0x00AD17, 0x00AD18, 0x00AD19, 0x00AD1A, 0x00AD1B, 0x00AD1C, 0x00AD1D, 0x00AD1E, 0x00AD1F,
	0x00AD20, 0x00AD21, 0x00AD22, 0x00AD23, 0x00AD24, 0x00AD25, 0x00AD26, 0x00AD27, 0x00AD28, 0x00AD29, 0x00AD2A, 0x00AD2B, 0x00AD2C, 0x00AD2D, 0x00AD2E, 0x00AD2F, 0x00AD30, 0x00AD31, 0x00AD32, 0x00AD33, 0x00AD34, 0x00AD35, 0x00AD36, 0x00AD37, 0x00AD38, 0x00AD39, 0x00AD3A, 0x00AD3B, 0x00AD3C, 0x00AD3D, 0x00AD3E, 0x00AD3F,
	0x00AD40, 0x00AD41, 0x00AD42, 0x00AD43, 0x00AD44, 0x00AD45, 0x00AD46, 0x00AD47, 0x00AD48, 0x00AD49, 0x00AD4A, 0x00AD4B, 0x00AD4C, 0x00AD4D, 0x00AD4E, 0x00AD4F, 0x00AD50, 0x00AD51, 0x00AD52, 0x00AD53, 0x00AD54, 0x00AD55, 0x00AD56, 0x00AD57, 0x00AD58, 0x00AD59, 0x00AD5A, 0x00AD5B, 0x00AD5C, 0x00AD5D, 0x00AD5E, 0x00AD5F,
	0x00AD60, 0x00AD61, 0x00AD62, 0x00AD63, 0x00AD64, 0x00AD65, 0x00AD66, 0x00AD67, 0x00AD68, 0x00AD69, 0x00AD6A, 0x00AD6B, 0x00AD6C, 0x00AD6D, 0x00AD6E, 0x00AD6F, 0x00AD70, 0x00AD71, 0x00AD72, 0x00AD73, 0x00AD74, 0x00AD75, 0x00AD76, 0x00AD77, 0x00AD78, 0x00AD79, 0x00AD7A, 0x00AD7B, 0x00AD7C, 0x00AD7D, 0x00AD7E, 0x00AD7F,
	0x00AD80, 0x00AD81, 0x00AD82, 0x00AD83, 0x00AD84, 0x00AD85, 0x00AD86, 0x00AD87, 0x00AD88, 0x00AD89, 0x00AD8A, 0x00AD8B, 0x00AD8C, 0x00AD8D, 0x00AD8E, 0x00AD8F, 0x00AD90, 0x00AD91, 0x00AD92, 0x00AD93, 0x00AD94, 0x00AD95, 0x00AD96, 0x00AD97, 0x00AD98, 0x00AD99, 0x00AD9A, 0x00AD9B, 0x00AD9C, 0x00AD9D, 0x00AD9E, 0x00AD9F,
	0x00ADA0, 0x00ADA1, 0x00ADA2, 0x00ADA3, 0x00ADA4, 0x00ADA5, 0x00ADA6, 0x00ADA7, 0x00ADA8, 0x00ADA9, 0x00ADAA, 0x00ADAB, 0x00ADAC, 0x00ADAD, 0x00ADAE, 0x00ADAF, 0x00ADB0, 0x00ADB1, 0x00ADB2, 0x00ADB3, 0x00ADB4, 0x00ADB5, 0x00ADB6, 0x00ADB7, 0x00ADB8, 0x00ADB9, 0x00ADBA, 0x00ADBB, 0x00ADBC, 0x00ADBD, 0x00ADBE, 0x00ADBF,
	0x00ADC0, 0x00ADC1, 0x00ADC2, 0x00ADC3, 0x00ADC4, 0x00ADC5, 0x00ADC6, 0x00ADC7, 0x00ADC8, 0x00ADC9, 0x00ADCA, 0x00ADCB, 0x00ADCC, 0x00ADCD, 0x00ADCE, 0x00ADCF, 0x00ADD0, 0x00ADD1, 0x00ADD2, 0x00ADD3, 0x00ADD4, 0x00ADD5, 0x00ADD6, 0x00ADD7, 0x00ADD8, 0x00ADD9, 0x00ADDA, 0x00ADDB, 0x00ADDC, 0x00ADDD, 0x00ADDE, 0x00ADDF,
	0x00ADE0, 0x00ADE1, 0x00ADE2, 0x00ADE3, 0x00ADE4, 0x00ADE5, 0x00ADE6, 0x00ADE7, 0x00ADE8, 0x00ADE9, 0x00ADEA, 0x00ADEB, 0x00ADEC, 0x00ADED, 0x00ADEE, 0x00ADEF, 0x00ADF0, 0x00ADF1, 0x00ADF2, 0x00ADF3, 0x00ADF4, 0x00ADF5, 0x00ADF6, 0x00ADF7, 0x00ADF8, 0x00ADF9, 0x00ADFA, 0x00ADFB, 0x00ADFC, 0x00ADFD, 0x00ADFE, 0x00ADFF,
	0x00AE00, 0x00AE01, 0x00AE02, 0x00AE03, 0x00AE04, 0x00AE05, 0x00AE06, 0x00AE07, 0x00AE08, 0x00AE09, 0x00AE0A, 0x00AE0B, 0x00AE0C, 0x00AE0D, 0x00AE0E, 0x00AE0F, 0x00AE10, 0x00AE11, 0x00AE12, 0x00AE13, 0x00AE14, 0x00AE15, 0x00AE16, 0x00AE17, 0x00AE18, 0x00AE19, 0x00AE1A, 0x00AE1B, 0x00AE1C, 0x00AE1D, 0x00AE1E, 0x00AE1F,
	0x00AE20, 0x00AE21, 0x00AE22, 0x00AE23, 0x00AE24, 0x00AE25, 0x00AE26, 0x00AE27, 0x00AE28, 0x00AE29, 0x00AE2A, 0x00AE2B, 0x00AE2C, 0x00AE2D, 0x00AE2E, 0x00AE2F, 0x00AE30, 0x00AE31, 0x00AE32, 0x00AE33, 0x00AE34, 0x00AE35, 0x00AE36, 0x00AE37, 0x00AE38, 0x00AE39, 0x00AE3A, 0x00AE3B, 0x00AE3C, 0x00AE3D, 0x00AE3E, 0x00AE3F,
	0x00AE40, 0x00AE41, 0x00AE42, 0x00AE43, 0x00AE44, 0x00AE45, 0x00AE46, 0x00AE47, 0x00AE48, 0x00AE49, 0x00AE4A, 0x00AE4B, 0x00AE4C, 0x00AE4D, 0x00AE4E, 0x00AE4F, 0x00AE50, 0x00AE51, 0x00AE52, 0x00AE53, 0x00AE54, 0x00AE55, 0x00AE56, 0x00AE57, 0x00AE58, 0x00AE59, 0x00AE5A, 0x00AE5B, 0x00AE5C, 0x00AE5D, 0x00AE5E, 0x00AE5F,
	0x00AE60, 0x00AE61, 0x00AE62, 0x00AE63, 0x00AE64, 0x00AE65, 0x00AE66, 0x00AE67, 0x00AE68, 0x00AE69, 0x00AE6A, 0x00AE6B, 0x00AE6C, 0x00AE6D, 0x00AE6E, 0x00AE6F, 0x00AE70, 0x00AE71, 0x00AE72, 0x00AE73, 0x00AE74, 0x00AE75, 0x00AE76, 0x00AE77, 0x00AE78, 0x00AE79, 0x00AE7A, 0x00AE7B, 0x00AE7C, 0x00AE7D, 0x00AE7E, 0x00AE7F,
	0x00AE80, 0x00AE81, 0x00AE82, 0x00AE83, 0x00AE84, 0x00AE85, 0x00AE86, 0x00AE87, 0x00AE88, 0x00AE89, 0x00AE8A, 0x00AE8B, 0x00AE8C, 0x00AE8D, 0x00AE8E, 0x00AE8F, 0x00AE90, 0x00AE91, 0x00AE92, 0x00AE93, 0x00AE94, 0x00AE95, 0x00AE96, 0x00AE97, 0x00AE98, 0x00AE99, 0x00AE9A, 0x00AE9B, 0x00AE9C, 0x00AE9D, 0x00AE9E, 0x00AE9F,
	0x00AEA0, 0x00AEA1, 0x00AEA2, 0x00AEA3, 0x00AEA4, 0x00AEA5, 0x00AEA6, 0x00AEA7, 0x00AEA8, 0x00AEA9, 0x00AEAA, 0x00AEAB, 0x00AEAC, 0x00AEAD, 0x00AEAE, 0x00AEAF, 0x00AEB0, 0x00AEB1, 0x00AEB2, 0x00AEB3, 0x00AEB4, 0x00AEB5, 0x00AEB6, 0x00AEB7, 0x00AEB8, 0x00AEB9, 0x00AEBA, 0x00AEBB, 0x00AEBC, 0x00AEBD, 0x00AEBE, 0x00AEBF,
	0x00AEC0, 0x00AEC1, 0x00AEC2, 0x00AEC3, 0x00AEC4, 0x00AEC5, 0x00AEC6, 0x00AEC7, 0x00AEC8, 0x00AEC9, 0x00AECA, 0x00AECB, 0x00AECC, 0x00AECD, 0x00AECE, 0x00AECF, 0x00AED0, 0x00AED1, 0x00AED2, 0x00AED3, 0x00AED4, 0x00AED5, 0x00AED6, 0x00AED7, 0x00AED8, 0x00AED9, 0x00AEDA, 0x00AEDB, 0x00AEDC, 0x00AEDD, 0x00AEDE, 0x00AEDF,
	0x00AEE0, 0x00AEE1, 0x00AEE2, 0x00AEE3, 0x00AEE4, 0x00AEE5, 0x00AEE6, 0x00AEE7, 0x00AEE8, 0x00AEE9, 0x00AEEA, 0x00AEEB, 0x00AEEC, 0x00AEED, 0x00AEEE, 0x00AEEF, 0x00AEF0, 0x00AEF1, 0x00AEF2, 0x00AEF3, 0x00AEF4, 0x00AEF5, 0x00AEF6, 0x00AEF7, 0x00AEF8, 0x00AEF9, 0x00AEFA, 0x00AEFB, 0x00AEFC, 0x00AEFD, 0x00AEFE, 0x00AEFF,
	0x00AF00, 0x00AF01, 0x00AF02, 0x00AF03, 0x00AF04, 0x00AF05, 0x00AF06, 0x00AF07, 0x00AF08, 0x00AF09, 0x00AF0A, 0x00AF0B, 0x00AF0C, 0x00AF0D, 0x00AF0E, 0x00AF0F, 0x00AF10, 0x00AF11, 0x00AF12, 0x00AF13, 0x00AF14, 0x00AF15, 0x00AF16, 0x00AF17, 0x00AF18, 0x00AF19, 0x00AF1A, 0x00AF1B, 0x00AF1C, 0x00AF1D, 0x00AF1E, 0x00AF1F,
	0x00AF20, 0x00AF21, 0x00AF22, 0x00AF23, 0x00AF24, 0x00AF25, 0x00AF26, 0x00AF27, 0x00AF28, 0x00AF29, 0x00AF2A, 0x00AF2B, 0x00AF2C, 0x00AF2D, 0x00AF2E, 0x00AF2F, 0x00AF30, 0x00AF31, 0x00AF32, 0x00AF33, 0x00AF34, 0x00AF35, 0x00AF36, 0x00AF37, 0x00AF38, 0x00AF39, 0x00AF3A, 0x00AF3B, 0x00AF3C, 0x00AF3D, 0x00AF3E, 0x00AF3F,
	0x00AF40, 0x00AF41, 0x00AF42, 0x00AF43, 0x00AF44, 0x00AF45, 0x00AF46, 0x00AF47, 0x00AF48, 0x00AF49, 0x00AF4A, 0x00AF4B, 0x00AF4C, 0x00AF4D, 0x00AF4E, 0x00AF4F, 0x00AF50, 0x00AF51, 0x00AF52, 0x00AF53, 0x00AF54, 0x00AF55, 0x00AF56, 0x00AF57, 0x00AF58, 0x00AF59, 0x00AF5A, 0x00AF5B, 0x00AF5C, 0x00AF5D, 0x00AF5E, 0x00AF5F,
	0x00AF60, 0x00AF61, 0x00AF62, 0x00AF63, 0x00AF64, 0x00AF65, 0x00AF66, 0x00AF67, 0x00AF68, 0x00AF69, 0x00AF6A, 0x00AF6B, 0x00AF6C, 0x00AF6D, 0x00AF6E, 0x00AF6F, 0x00AF70, 0x00AF71, 0x00AF72, 0x00AF73, 0x00AF74, 0x00AF75, 0x00AF76, 0x00AF77, 0x00AF78, 0x00AF79, 0x00AF7A, 0x00AF7B, 0x00AF7C, 0x00AF7D, 0x00AF7E, 0x00AF7F,
	0x00AF80, 0x00AF81, 0x00AF82, 0x00AF83, 0x00AF84, 0x00AF85, 0x00AF86, 0x00AF87, 0x00AF88, 0x00AF89, 0x00AF8A, 0x00AF8B, 0x00AF8C, 0x00AF8D, 0x00AF8E, 0x00AF8F, 0x00AF90, 0x00AF91, 0x00AF92, 0x00AF93, 0x00AF94, 0x00AF95, 0x00AF96, 0x00AF97, 0x00AF98, 0x00AF99, 0x00AF9A, 0x00AF9B, 0x00AF9C, 0x00AF9D, 0x00AF9E, 0x00AF9F,
	0x00AFA0, 0x00AFA1, 0x00AFA2, 0x00AFA3, 0x00AFA4, 0x00AFA5, 0x00AFA6, 0x00AFA7, 0x00AFA8, 0x00AFA9, 0x00AFAA, 0x00AFAB, 0x00AFAC, 0x00AFAD, 0x00AFAE, 0x00AFAF, 0x00AFB0, 0x00AFB1, 0x00AFB2, 0x00AFB3, 0x00AFB4, 0x00AFB5, 0x00AFB6, 0x00AFB7, 0x00AFB8, 0x00AFB9, 0x00AFBA, 0x00AFBB, 0x00AFBC, 0x00AFBD, 0x00AFBE, 0x00AFBF,
	0x00AFC0, 0x00AFC1, 0x00AFC2, 0x00AFC3, 0x00AFC4, 0x00AFC5, 0x00AFC6, 0x00AFC7, 0x00AFC8, 0x00AFC9, 0x00AFCA, 0x00AFCB, 0x00AFCC, 0x00AFCD, 0x00AFCE, 0x00AFCF, 0x00AFD0, 0x00AFD1, 0x00AFD2, 0x00AFD3, 0x00AFD4, 0x00AFD5, 0x00AFD6, 0x00AFD7, 0x00AFD8, 0x00AFD9, 0x00AFDA, 0x00AFDB, 0x00AFDC, 0x00AFDD, 0x00AFDE, 0x00AFDF,
	0x00AFE0, 0x00AFE1, 0x00AFE2, 0x00AFE3, 0x00AFE4, 0x00AFE5, 0x00AFE6, 0x00AFE7, 0x00AFE8, 0x00AFE9, 0x00AFEA, 0x00AFEB, 0x00AFEC, 0x00AFED, 0x00AFEE, 0x00AFEF, 0x00AFF0, 0x00AFF1, 0x00AFF2, 0x00AFF3, 0x00AFF4, 0x00AFF5, 0x00AFF6, 0x00AFF7, 0x00AFF8, 0x00AFF9, 0x00AFFA, 0x00AFFB, 0x00AFFC, 0x00AFFD, 0x00AFFE, 0x00AFFF,
	0
};

static const unsigned map_hangul_2[] = {
	0x00B000, 0x00B001, 0x00B002, 0x00B003, 0x00B004, 0x00B005, 0x00B006, 0x00B007, 0x00B008, 0x00B009, 0x00B00A, 0x00B00B, 0x00B00C, 0x00B00D, 0x00B00E, 0x00B00F, 0x00B010, 0x00B011, 0x00B012, 0x00B013, 0x00B014, 0x00B015, 0x00B016, 0x00B017, 0x00B018, 0x00B019, 0x00B01A, 0x00B01B, 0x00B01C, 0x00B01D, 0x00B01E, 0x00B01F,
	0x00B020, 0x00B021, 0x00B022, 0x00B023, 0x00B024, 0x00B025, 0x00B026, 0x00B027, 0x00B028, 0x00B029, 0x00B02A, 0x00B02B, 0x00B02C, 0x00B02D, 0x00B02E, 0x00B02F, 0x00B030, 0x00B031, 0x00B032, 0x00B033, 0x00B034, 0x00B035, 0x00B036, 0x00B037, 0x00B038, 0x00B039, 0x00B03A, 0x00B03B, 0x00B03C, 0x00B03D, 0x00B03E, 0x00B03F,
	0x00B040, 0x00B041, 0x00B042, 0x00B043, 0x00B044, 0x00B045, 0x00B046, 0x00B047, 0x00B048, 0x00B049, 0x00B04A, 0x00B04B, 0x00B04C, 0x00B04D, 0x00B04E, 0x00B04F, 0x00B050, 0x00B051, 0x00B052, 0x00B053, 0x00B054, 0x00B055, 0x00B056, 0x00B057, 0x00B058, 0x00B059, 0x00B05A, 0x00B05B, 0x00B05C, 0x00B05D, 0x00B05E, 0x00B05F,
	0x00B060, 0x00B061, 0x00B062, 0x00B063, 0x00B064, 0x00B065, 0x00B066, 0x00B067, 0x00B068, 0x00B069, 0x00B06A, 0x00B06B, 0x00B06C, 0x00B06D, 0x00B06E, 0x00B06F, 0x00B070, 0x00B071, 0x00B072, 0x00B073, 0x00B074, 0x00B075, 0x00B076, 0x00B077, 0x00B078, 0x00B079, 0x00B07A, 0x00B07B, 0x00B07C, 0x00B07D, 0x00B07E, 0x00B07F,
	0x00B080, 0x00B081, 0x00B082, 0x00B083, 0x00B084, 0x00B085, 0x00B086, 0x00B087, 0x00B088, 0x00B089, 0x00B08A, 0x00B08B, 0x00B08C, 0x00B08D, 0x00B08E, 0x00B08F, 0x00B090, 0x00B091, 0x00B092, 0x00B093, 0x00B094, 0x00B095, 0x00B096, 0x00B097, 0x00B098, 0x00B099, 0x00B09A, 0x00B09B, 0x00B09C, 0x00B09D, 0x00B09E, 0x00B09F,
	0x00B0A0, 0x00B0A1, 0x00B0A2, 0x00B0A3, 0x00B0A4, 0x00B0A5, 0x00B0A6, 0x00B0A7, 0x00B0A8, 0x00B0A9, 0x00B0AA, 0x00B0AB, 0x00B0AC, 0x00B0AD, 0x00B0AE, 0x00B0AF, 0x00B0B0, 0x00B0B1, 0x00B0B2, 0x00B0B3, 0x00B0B4, 0x00B0B5, 0x00B0B6, 0x00B0B7, 0x00B0B8, 0x00B0B9, 0x00B0BA, 0x00B0BB, 0x00B0BC, 0x00B0BD, 0x00B0BE, 0x00B0BF,
	0x00B0C0, 0x00B0C1, 0x00B0C2, 0x00B0C3, 0x00B0C4, 0x00B0C5, 0x00B0C6, 0x00B0C7, 0x00B0C8, 0x00B0C9, 0x00B0CA, 0x00B0CB, 0x00B0CC, 0x00B0CD, 0x00B0CE, 0x00B0CF, 0x00B0D0, 0x00B0D1, 0x00B0D2, 0x00B0D3, 0x00B0D4, 0x00B0D5, 0x00B0D6, 0x00B0D7, 0x00B0D8, 0x00B0D9, 0x00B0DA, 0x00B0DB, 0x00B0DC, 0x00B0DD, 0x00B0DE, 0x00B0DF,
	0x00B0E0, 0x00B0E1, 0x00B0E2, 0x00B0E3, 0x00B0E4, 0x00B0E5, 0x00B0E6, 0x00B0E7, 0x00B0E8, 0x00B0E9, 0x00B0EA, 0x00B0EB, 0x00B0EC, 0x00B0ED, 0x00B0EE, 0x00B0EF, 0x00B0F0, 0x00B0F1, 0x00B0F2, 0x00B0F3, 0x00B0F4, 0x00B0F5, 0x00B0F6, 0x00B0F7, 0x00B0F8, 0x00B0F9, 0x00B0FA, 0x00B0FB, 0x00B0FC, 0x00B0FD, 0x00B0FE, 0x00B0FF,
	0x00B100, 0x00B101, 0x00B102, 0x00B103, 0x00B104, 0x00B105, 0x00B106, 0x00B107, 0x00B108, 0x00B109, 0x00B10A, 0x00B10B, 0x00B10C, 0x00B10D, 0x00B10E, 0x00B10F, 0x00B110, 0x00B111, 0x00B112, 0x00B113, 0x00B114, 0x00B115, 0x00B116, 0x00B117, 0x00B118, 0x00B119, 0x00B11A, 0x00B11B, 0x00B11C, 0x00B11D, 0x00B11E, 0x00B11F,
	0x00B120, 0x00B121, 0x00B122, 0x00B123, 0x00B124, 0x00B125, 0x00B126, 0x00B127, 0x00B128, 0x00B129, 0x00B12A, 0x00B12B, 0x00B12C, 0x00B12D, 0x00B12E, 0x00B12F, 0x00B130, 0x00B131, 0x00B132, 0x00B133, 0x00B134, 0x00B135, 0x00B136, 0x00B137, 0x00B138, 0x00B139, 0x00B13A, 0x00B13B, 0x00B13C, 0x00B13D, 0x00B13E, 0x00B13F,
	0x00B140, 0x00B141, 0x00B142, 0x00B143, 0x00B144, 0x00B145, 0x00B146, 0x00B147, 0x00B148, 0x00B149, 0x00B14A, 0x00B14B, 0x00B14C, 0x00B14D, 0x00B14E, 0x00B14F, 0x00B150, 0x00B151, 0x00B152, 0x00B153, 0x00B154, 0x00B155, 0x00B156, 0x00B157, 0x00B158, 0x00B159, 0x00B15A, 0x00B15B, 0x00B15C, 0x00B15D, 0x00B15E, 0x00B15F,
	0x00B160, 0x00B161, 0x00B162, 0x00B163, 0x00B164, 0x00B165, 0x00B166, 0x00B167, 0x00B168, 0x00B169, 0x00B16A, 0x00B16B, 0x00B16C, 0x00B16D, 0x00B16E, 0x00B16F, 0x00B170, 0x00B171, 0x00B172, 0x00B173, 0x00B174, 0x00B175, 0x00B176, 0x00B177, 0x00B178, 0x00B179, 0x00B17A, 0x00B17B, 0x00B17C, 0x00B17D, 0x00B17E, 0x00B17F,
	0x00B180, 0x00B181, 0x00B182, 0x00B183, 0x00B184, 0x00B185, 0x00B186, 0x00B187, 0x00B188, 0x00B189, 0x00B18A, 0x00B18B, 0x00B18C, 0x00B18D, 0x00B18E, 0x00B18F, 0x00B190, 0x00B191, 0x00B192, 0x00B193, 0x00B194, 0x00B195, 0x00B196, 0x00B197, 0x00B198, 0x00B199, 0x00B19A, 0x00B19B, 0x00B19C, 0x00B19D, 0x00B19E, 0x00B19F,
	0x00B1A0, 0x00B1A1, 0x00B1A2, 0x00B1A3, 0x00B1A4, 0x00B1A5, 0x00B1A6, 0x00B1A7, 0x00B1A8, 0x00B1A9, 0x00B1AA, 0x00B1AB, 0x00B1AC, 0x00B1AD, 0x00B1AE, 0x00B1AF, 0x00B1B0, 0x00B1B1, 0x00B1B2, 0x00B1B3, 0x00B1B4, 0x00B1B5, 0x00B1B6, 0x00B1B7, 0x00B1B8, 0x00B1B9, 0x00B1BA, 0x00B1BB, 0x00B1BC, 0x00B1BD, 0x00B1BE, 0x00B1BF,
	0x00B1C0, 0x00B1C1, 0x00B1C2, 0x00B1C3, 0x00B1C4, 0x00B1C5, 0x00B1C6, 0x00B1C7, 0x00B1C8, 0x00B1C9, 0x00B1CA, 0x00B1CB, 0x00B1CC, 0x00B1CD, 0x00B1CE, 0x00B1CF, 0x00B1D0, 0x00B1D1, 0x00B1D2, 0x00B1D3, 0x00B1D4, 0x00B1D5, 0x00B1D6, 0x00B1D7, 0x00B1D8, 0x00B1D9, 0x00B1DA, 0x00B1DB, 0x00B1DC, 0x00B1DD, 0x00B1DE, 0x00B1DF,
	0x00B1E0, 0x00B1E1, 0x00B1E2, 0x00B1E3, 0x00B1E4, 0x00B1E5, 0x00B1E6, 0x00B1E7, 0x00B1E8, 0x00B1E9, 0x00B1EA, 0x00B1EB, 0x00B1EC, 0x00B1ED, 0x00B1EE, 0x00B1EF, 0x00B1F0, 0x00B1F1, 0x00B1F2, 0x00B1F3, 0x00B1F4, 0x00B1F5, 0x00B1F6, 0x00B1F7, 0x00B1F8, 0x00B1F9, 0x00B1FA, 0x00B1FB, 0x00B1FC, 0x00B1FD, 0x00B1FE, 0x00B1FF,
	0x00B200, 0x00B201, 0x00B202, 0x00B203, 0x00B204, 0x00B205, 0x00B206, 0x00B207, 0x00B208, 0x00B209, 0x00B20A, 0x00B20B, 0x00B20C, 0x00B20D, 0x00B20E, 0x00B20F, 0x00B210, 0x00B211, 0x00B212, 0x00B213, 0x00B214, 0x00B215, 0x00B216, 0x00B217, 0x00B218, 0x00B219, 0x00B21A, 0x00B21B, 0x00B21C, 0x00B21D, 0x00B21E, 0x00B21F,
	0x00B220, 0x00B221, 0x00B222, 0x00B223, 0x00B224, 0x00B225, 0x00B226, 0x00B227, 0x00B228, 0x00B229, 0x00B22A, 0x00B22B, 0x00B22C, 0x00B22D, 0x00B22E, 0x00B22F, 0x00B230, 0x00B231, 0x00B232, 0x00B233, 0x00B234, 0x00B235, 0x00B236, 0x00B237, 0x00B238, 0x00B239, 0x00B23A, 0x00B23B, 0x00B23C, 0x00B23D, 0x00B23E, 0x00B23F,
	0x00B240, 0x00B241, 0x00B242, 0x00B243, 0x00B244, 0x00B245, 0x00B246, 0x00B247, 0x00B248, 0x00B249, 0x00B24A, 0x00B24B, 0x00B24C, 0x00B24D, 0x00B24E, 0x00B24F, 0x00B250, 0x00B251, 0x00B252, 0x00B253, 0x00B254, 0x00B255, 0x00B256, 0x00B257, 0x00B258, 0x00B259, 0x00B25A, 0x00B25B, 0x00B25C, 0x00B25D, 0x00B25E, 0x00B25F,
	0x00B260, 0x00B261, 0x00B262, 0x00B263, 0x00B264, 0x00B265, 0x00B266, 0x00B267, 0x00B268, 0x00B269, 0x00B26A, 0x00B26B, 0x00B26C, 0x00B26D, 0x00B26E, 0x00B26F, 0x00B270, 0x00B271, 0x00B272, 0x00B273, 0x00B274, 0x00B275, 0x00B276, 0x00B277, 0x00B278, 0x00B279, 0x00B27A, 0x00B27B, 0x00B27C, 0x00B27D, 0x00B27E, 0x00B27F,
	0x00B280, 0x00B281, 0x00B282, 0x00B283, 0x00B284, 0x00B285, 0x00B286, 0x00B287, 0x00B288, 0x00B289, 0x00B28A, 0x00B28B, 0x00B28C, 0x00B28D, 0x00B28E, 0x00B28F, 0x00B290, 0x00B291, 0x00B292, 0x00B293, 0x00B294, 0x00B295, 0x00B296, 0x00B297, 0x00B298, 0x00B299, 0x00B29A, 0x00B29B, 0x00B29C, 0x00B29D, 0x00B29E, 0x00B29F,
	0x00B2A0, 0x00B2A1, 0x00B2A2, 0x00B2A3, 0x00B2A4, 0x00B2A5, 0x00B2A6, 0x00B2A7, 0x00B2A8, 0x00B2A9, 0x00B2AA, 0x00B2AB, 0x00B2AC, 0x00B2AD, 0x00B2AE, 0x00B2AF, 0x00B2B0, 0x00B2B1, 0x00B2B2, 0x00B2B3, 0x00B2B4, 0x00B2B5, 0x00B2B6, 0x00B2B7, 0x00B2B8, 0x00B2B9, 0x00B2BA, 0x00B2BB, 0x00B2BC, 0x00B2BD, 0x00B2BE, 0x00B2BF,
	0x00B2C0, 0x00B2C1, 0x00B2C2, 0x00B2C3, 0x00B2C4, 0x00B2C5, 0x00B2C6, 0x00B2C7, 0x00B2C8, 0x00B2C9, 0x00B2CA, 0x00B2CB, 0x00B2CC, 0x00B2CD, 0x00B2CE, 0x00B2CF, 0x00B2D0, 0x00B2D1, 0x00B2D2, 0x00B2D3, 0x00B2D4, 0x00B2D5, 0x00B2D6, 0x00B2D7, 0x00B2D8, 0x00B2D9, 0x00B2DA, 0x00B2DB, 0x00B2DC, 0x00B2DD, 0x00B2DE, 0x00B2DF,
	0x00B2E0, 0x00B2E1, 0x00B2E2, 0x00B2E3, 0x00B2E4, 0x00B2E5, 0x00B2E6, 0x00B2E7, 0x00B2E8, 0x00B2E9, 0x00B2EA, 0x00B2EB, 0x00B2EC, 0x00B2ED, 0x00B2EE, 0x00B2EF, 0x00B2F0, 0x00B2F1, 0x00B2F2, 0x00B2F3, 0x00B2F4, 0x00B2F5, 0x00B2F6, 0x00B2F7, 0x00B2F8, 0x00B2F9, 0x00B2FA, 0x00B2FB, 0x00B2FC, 0x00B2FD, 0x00B2FE, 0x00B2FF,
	0x00B300, 0x00B301, 0x00B302, 0x00B303, 0x00B304, 0x00B305, 0x00B306, 0x00B307, 0x00B308, 0x00B309, 0x00B30A, 0x00B30B, 0x00B30C, 0x00B30D, 0x00B30E, 0x00B30F, 0x00B310, 0x00B311, 0x00B312, 0x00B313, 0x00B314, 0x00B315, 0x00B316, 0x00B317, 0x00B318, 0x00B319, 0x00B31A, 0x00B31B, 0x00B31C, 0x00B31D, 0x00B31E, 0x00B31F,
	0x00B320, 0x00B321, 0x00B322, 0x00B323, 0x00B324, 0x00B325, 0x00B326, 0x00B327, 0x00B328, 0x00B329, 0x00B32A, 0x00B32B, 0x00B32C, 0x00B32D, 0x00B32E, 0x00B32F, 0x00B330, 0x00B331, 0x00B332, 0x00B333, 0x00B334, 0x00B335, 0x00B336, 0x00B337, 0x00B338, 0x00B339, 0x00B33A, 0x00B33B, 0x00B33C, 0x00B33D, 0x00B33E, 0x00B33F,
	0x00B340, 0x00B341, 0x00B342, 0x00B343, 0x00B344, 0x00B345, 0x00B346, 0x00B347, 0x00B348, 0x00B349, 0x00B34A, 0x00B34B, 0x00B34C, 0x00B34D, 0x00B34E, 0x00B34F, 0x00B350, 0x00B351, 0x00B352, 0x00B353, 0x00B354, 0x00B355, 0x00B356, 0x00B357, 0x00B358, 0x00B359, 0x00B35A, 0x00B35B, 0x00B35C, 0x00B35D, 0x00B35E, 0x00B35F,
	0x00B360, 0x00B361, 0x00B362, 0x00B363, 0x00B364, 0x00B365, 0x00B366, 0x00B367, 0x00B368, 0x00B369, 0x00B36A, 0x00B36B, 0x00B36C, 0x00B36D, 0x00B36E, 0x00B36F, 0x00B370, 0x00B371, 0x00B372, 0x00B373, 0x00B374, 0x00B375, 0x00B376, 0x00B377, 0x00B378, 0x00B379, 0x00B37A, 0x00B37B, 0x00B37C, 0x00B37D, 0x00B37E, 0x00B37F,
	0x00B380, 0x00B381, 0x00B382, 0x00B383, 0x00B384, 0x00B385, 0x00B386, 0x00B387, 0x00B388, 0x00B389, 0x00B38A, 0x00B38B, 0x00B38C, 0x00B38D, 0x00B38E, 0x00B38F, 0x00B390, 0x00B391, 0x00B392, 0x00B393, 0x00B394, 0x00B395, 0x00B396, 0x00B397, 0x00B398, 0x00B399, 0x00B39A, 0x00B39B, 0x00B39C, 0x00B39D, 0x00B39E, 0x00B39F,
	0x00B3A0, 0x00B3A1, 0x00B3A2, 0x00B3A3, 0x00B3A4, 0x00B3A5, 0x00B3A6, 0x00B3A7, 0x00B3A8, 0x00B3A9, 0x00B3AA, 0x00B3AB, 0x00B3AC, 0x00B3AD, 0x00B3AE, 0x00B3AF, 0x00B3B0, 0x00B3B1, 0x00B3B2, 0x00B3B3, 0x00B3B4, 0x00B3B5, 0x00B3B6, 0x00B3B7, 0x00B3B8, 0x00B3B9, 0x00B3BA, 0x00B3BB, 0x00B3BC, 0x00B3BD, 0x00B3BE, 0x00B3BF,
	0x00B3C0, 0x00B3C1, 0x00B3C2, 0x00B3C3, 0x00B3C4, 0x00B3C5, 0x00B3C6, 0x00B3C7, 0x00B3C8, 0x00B3C9, 0x00B3CA, 0x00B3CB, 0x00B3CC, 0x00B3CD, 0x00B3CE, 0x00B3CF, 0x00B3D0, 0x00B3D1, 0x00B3D2, 0x00B3D3, 0x00B3D4, 0x00B3D5, 0x00B3D6, 0x00B3D7, 0x00B3D8, 0x00B3D9, 0x00B3DA, 0x00B3DB, 0x00B3DC, 0x00B3DD, 0x00B3DE, 0x00B3DF,
	0x00B3E0, 0x00B3E1, 0x00B3E2, 0x00B3E3, 0x00B3E4, 0x00B3E5, 0x00B3E6, 0x00B3E7, 0x00B3E8, 0x00B3E9, 0x00B3EA, 0x00B3EB, 0x00B3EC, 0x00B3ED, 0x00B3EE, 0x00B3EF, 0x00B3F0, 0x00B3F1, 0x00B3F2, 0x00B3F3, 0x00B3F4, 0x00B3F5, 0x00B3F6, 0x00B3F7, 0x00B3F8, 0x00B3F9, 0x00B3FA, 0x00B3FB, 0x00B3FC, 0x00B3FD, 0x00B3FE, 0x00B3FF,
	0
};

static const unsigned map_hangul_3[] = {
	0x00B400, 0x00B401, 0x00B402, 0x00B403, 0x00B404, 0x00B405, 0x00B406, 0x00B407, 0x00B408, 0x00B409, 0x00B40A, 0x00B40B, 0x00B40C, 0x00B40D, 0x00B40E, 0x00B40F, 0x00B410, 0x00B411, 0x00B412, 0x00B413, 0x00B414, 0x00B415, 0x00B416, 0x00B417, 0x00B418, 0x00B419, 0x00B41A, 0x00B41B, 0x00B41C, 0x00B41D, 0x00B41E, 0x00B41F,
	0x00B420, 0x00B421, 0x00B422, 0x00B423, 0x00B424, 0x00B425, 0x00B426, 0x00B427, 0x00B428, 0x00B429, 0x00B42A, 0x00B42B, 0x00B42C, 0x00B42D, 0x00B42E, 0x00B42F, 0x00B430, 0x00B431, 0x00B432, 0x00B433, 0x00B434, 0x00B435, 0x00B436, 0x00B437, 0x00B438, 0x00B439, 0x00B43A, 0x00B43B, 0x00B43C, 0x00B43D, 0x00B43E, 0x00B43F,
	0x00B440, 0x00B441, 0x00B442, 0x00B443, 0x00B444, 0x00B445, 0x00B446, 0x00B447, 0x00B448, 0x00B449, 0x00B44A, 0x00B44B, 0x00B44C, 0x00B44D, 0x00B44E, 0x00B44F, 0x00B450, 0x00B451, 0x00B452, 0x00B453, 0x00B454, 0x00B455, 0x00B456, 0x00B457, 0x00B458, 0x00B459, 0x00B45A, 0x00B45B, 0x00B45C, 0x00B45D, 0x00B45E, 0x00B45F,
	0x00B460, 0x00B461, 0x00B462, 0x00B463, 0x00B464, 0x00B465, 0x00B466, 0x00B467, 0x00B468, 0x00B469, 0x00B46A, 0x00B46B, 0x00B46C, 0x00B46D, 0x00B46E, 0x00B46F, 0x00B470, 0x00B471, 0x00B472, 0x00B473, 0x00B474, 0x00B475, 0x00B476, 0x00B477, 0x00B478, 0x00B479, 0x00B47A, 0x00B47B, 0x00B47C, 0x00B47D, 0x00B47E, 0x00B47F,
	0x00B480, 0x00B481, 0x00B482, 0x00B483, 0x00B484, 0x00B485, 0x00B486, 0x00B487, 0x00B488, 0x00B489, 0x00B48A, 0x00B48B, 0x00B48C, 0x00B48D, 0x00B48E, 0x00B48F, 0x00B490, 0x00B491, 0x00B492, 0x00B493, 0x00B494, 0x00B495, 0x00B496, 0x00B497, 0x00B498, 0x00B499, 0x00B49A, 0x00B49B, 0x00B49C, 0x00B49D, 0x00B49E, 0x00B49F,
	0x00B4A0, 0x00B4A1, 0x00B4A2, 0x00B4A3, 0x00B4A4, 0x00B4A5, 0x00B4A6, 0x00B4A7, 0x00B4A8, 0x00B4A9, 0x00B4AA, 0x00B4AB, 0x00B4AC, 0x00B4AD, 0x00B4AE, 0x00B4AF, 0x00B4B0, 0x00B4B1, 0x00B4B2, 0x00B4B3, 0x00B4B4, 0x00B4B5, 0x00B4B6, 0x00B4B7, 0x00B4B8, 0x00B4B9, 0x00B4BA, 0x00B4BB, 0x00B4BC, 0x00B4BD, 0x00B4BE, 0x00B4BF,
	0x00B4C0, 0x00B4C1, 0x00B4C2, 0x00B4C3, 0x00B4C4, 0x00B4C5, 0x00B4C6, 0x00B4C7, 0x00B4C8, 0x00B4C9, 0x00B4CA, 0x00B4CB, 0x00B4CC, 0x00B4CD, 0x00B4CE, 0x00B4CF, 0x00B4D0, 0x00B4D1, 0x00B4D2, 0x00B4D3, 0x00B4D4, 0x00B4D5, 0x00B4D6, 0x00B4D7, 0x00B4D8, 0x00B4D9, 0x00B4DA, 0x00B4DB, 0x00B4DC, 0x00B4DD, 0x00B4DE, 0x00B4DF,
	0x00B4E0, 0x00B4E1, 0x00B4E2, 0x00B4E3, 0x00B4E4, 0x00B4E5, 0x00B4E6, 0x00B4E7, 0x00B4E8, 0x00B4E9, 0x00B4EA, 0x00B4EB, 0x00B4EC, 0x00B4ED, 0x00B4EE, 0x00B4EF, 0x00B4F0, 0x00B4F1, 0x00B4F2, 0x00B4F3, 0x00B4F4, 0x00B4F5, 0x00B4F6, 0x00B4F7, 0x00B4F8, 0x00B4F9, 0x00B4FA, 0x00B4FB, 0x00B4FC, 0x00B4FD, 0x00B4FE, 0x00B4FF,
	0x00B500, 0x00B501, 0x00B502, 0x00B503, 0x00B504, 0x00B505, 0x00B506, 0x00B507, 0x00B508, 0x00B509, 0x00B50A, 0x00B50B, 0x00B50C, 0x00B50D, 0x00B50E, 0x00B50F, 0x00B510, 0x00B511, 0x00B512, 0x00B513, 0x00B514, 0x00B515, 0x00B516, 0x00B517, 0x00B518, 0x00B519, 0x00B51A, 0x00B51B, 0x00B51C, 0x00B51D, 0x00B51E, 0x00B51F,
	0x00B520, 0x00B521, 0x00B522, 0x00B523, 0x00B524, 0x00B525, 0x00B526, 0x00B527, 0x00B528, 0x00B529, 0x00B52A, 0x00B52B, 0x00B52C, 0x00B52D, 0x00B52E, 0x00B52F, 0x00B530, 0x00B531, 0x00B532, 0x00B533, 0x00B534, 0x00B535, 0x00B536, 0x00B537, 0x00B538, 0x00B539, 0x00B53A, 0x00B53B, 0x00B53C, 0x00B53D, 0x00B53E, 0x00B53F,
	0x00B540, 0x00B541, 0x00B542, 0x00B543, 0x00B544, 0x00B545, 0x00B546, 0x00B547, 0x00B548, 0x00B549, 0x00B54A, 0x00B54B, 0x00B54C, 0x00B54D, 0x00B54E, 0x00B54F, 0x00B550, 0x00B551, 0x00B552, 0x00B553, 0x00B554, 0x00B555, 0x00B556, 0x00B557, 0x00B558, 0x00B559, 0x00B55A, 0x00B55B, 0x00B55C, 0x00B55D, 0x00B55E, 0x00B55F,
	0x00B560, 0x00B561, 0x00B562, 0x00B563, 0x00B564, 0x00B565, 0x00B566, 0x00B567, 0x00B568, 0x00B569, 0x00B56A, 0x00B56B, 0x00B56C, 0x00B56D, 0x00B56E, 0x00B56F, 0x00B570, 0x00B571, 0x00B572, 0x00B573, 0x00B574, 0x00B575, 0x00B576, 0x00B577, 0x00B578, 0x00B579, 0x00B57A, 0x00B57B, 0x00B57C, 0x00B57D, 0x00B57E, 0x00B57F,
	0x00B580, 0x00B581, 0x00B582, 0x00B583, 0x00B584, 0x00B585, 0x00B586, 0x00B587, 0x00B588, 0x00B589, 0x00B58A, 0x00B58B, 0x00B58C, 0x00B58D, 0x00B58E, 0x00B58F, 0x00B590, 0x00B591, 0x00B592, 0x00B593, 0x00B594, 0x00B595, 0x00B596, 0x00B597, 0x00B598, 0x00B599, 0x00B59A, 0x00B59B, 0x00B59C, 0x00B59D, 0x00B59E, 0x00B59F,
	0x00B5A0, 0x00B5A1, 0x00B5A2, 0x00B5A3, 0x00B5A4, 0x00B5A5, 0x00B5A6, 0x00B5A7, 0x00B5A8, 0x00B5A9, 0x00B5AA, 0x00B5AB, 0x00B5AC, 0x00B5AD, 0x00B5AE, 0x00B5AF, 0x00B5B0, 0x00B5B1, 0x00B5B2, 0x00B5B3, 0x00B5B4, 0x00B5B5, 0x00B5B6, 0x00B5B7, 0x00B5B8, 0x00B5B9, 0x00B5BA, 0x00B5BB, 0x00B5BC, 0x00B5BD, 0x00B5BE, 0x00B5BF,
	0x00B5C0, 0x00B5C1, 0x00B5C2, 0x00B5C3, 0x00B5C4, 0x00B5C5, 0x00B5C6, 0x00B5C7, 0x00B5C8, 0x00B5C9, 0x00B5CA, 0x00B5CB, 0x00B5CC, 0x00B5CD, 0x00B5CE, 0x00B5CF, 0x00B5D0, 0x00B5D1, 0x00B5D2, 0x00B5D3, 0x00B5D4, 0x00B5D5, 0x00B5D6, 0x00B5D7, 0x00B5D8, 0x00B5D9, 0x00B5DA, 0x00B5DB, 0x00B5DC, 0x00B5DD, 0x00B5DE, 0x00B5DF,
	0x00B5E0, 0x00B5E1, 0x00B5E2, 0x00B5E3, 0x00B5E4, 0x00B5E5, 0x00B5E6, 0x00B5E7, 0x00B5E8, 0x00B5E9, 0x00B5EA, 0x00B5EB, 0x00B5EC, 0x00B5ED, 0x00B5EE, 0x00B5EF, 0x00B5F0, 0x00B5F1, 0x00B5F2, 0x00B5F3, 0x00B5F4, 0x00B5F5, 0x00B5F6, 0x00B5F7, 0x00B5F8, 0x00B5F9, 0x00B5FA, 0x00B5FB, 0x00B5FC, 0x00B5FD, 0x00B5FE, 0x00B5FF,
	0x00B600, 0x00B601, 0x00B602, 0x00B603, 0x00B604, 0x00B605, 0x00B606, 0x00B607, 0x00B608, 0x00B609, 0x00B60A, 0x00B60B, 0x00B60C, 0x00B60D, 0x00B60E, 0x00B60F, 0x00B610, 0x00B611, 0x00B612, 0x00B613, 0x00B614, 0x00B615, 0x00B616, 0x00B617, 0x00B618, 0x00B619, 0x00B61A, 0x00B61B, 0x00B61C, 0x00B61D, 0x00B61E, 0x00B61F,
	0x00B620, 0x00B621, 0x00B622, 0x00B623, 0x00B624, 0x00B625, 0x00B626, 0x00B627, 0x00B628, 0x00B629, 0x00B62A, 0x00B62B, 0x00B62C, 0x00B62D, 0x00B62E, 0x00B62F, 0x00B630, 0x00B631, 0x00B632, 0x00B633, 0x00B634, 0x00B635, 0x00B636, 0x00B637, 0x00B638, 0x00B639, 0x00B63A, 0x00B63B, 0x00B63C, 0x00B63D, 0x00B63E, 0x00B63F,
	0x00B640, 0x00B641, 0x00B642, 0x00B643, 0x00B644, 0x00B645, 0x00B646, 0x00B647, 0x00B648, 0x00B649, 0x00B64A, 0x00B64B, 0x00B64C, 0x00B64D, 0x00B64E, 0x00B64F, 0x00B650, 0x00B651, 0x00B652, 0x00B653, 0x00B654, 0x00B655, 0x00B656, 0x00B657, 0x00B658, 0x00B659, 0x00B65A, 0x00B65B, 0x00B65C, 0x00B65D, 0x00B65E, 0x00B65F,
	0x00B660, 0x00B661, 0x00B662, 0x00B663, 0x00B664, 0x00B665, 0x00B666, 0x00B667, 0x00B668, 0x00B669, 0x00B66A, 0x00B66B, 0x00B66C, 0x00B66D, 0x00B66E, 0x00B66F, 0x00B670, 0x00B671, 0x00B672, 0x00B673, 0x00B674, 0x00B675, 0x00B676, 0x00B677, 0x00B678, 0x00B679, 0x00B67A, 0x00B67B, 0x00B67C, 0x00B67D, 0x00B67E, 0x00B67F,
	0x00B680, 0x00B681, 0x00B682, 0x00B683, 0x00B684, 0x00B685, 0x00B686, 0x00B687, 0x00B688, 0x00B689, 0x00B68A, 0x00B68B, 0x00B68C, 0x00B68D, 0x00B68E, 0x00B68F, 0x00B690, 0x00B691, 0x00B692, 0x00B693, 0x00B694, 0x00B695, 0x00B696, 0x00B697, 0x00B698, 0x00B699, 0x00B69A, 0x00B69B, 0x00B69C, 0x00B69D, 0x00B69E, 0x00B69F,
	0x00B6A0, 0x00B6A1, 0x00B6A2, 0x00B6A3, 0x00B6A4, 0x00B6A5, 0x00B6A6, 0x00B6A7, 0x00B6A8, 0x00B6A9, 0x00B6AA, 0x00B6AB, 0x00B6AC, 0x00B6AD, 0x00B6AE, 0x00B6AF, 0x00B6B0, 0x00B6B1, 0x00B6B2, 0x00B6B3, 0x00B6B4, 0x00B6B5, 0x00B6B6, 0x00B6B7, 0x00B6B8, 0x00B6B9, 0x00B6BA, 0x00B6BB, 0x00B6BC, 0x00B6BD, 0x00B6BE, 0x00B6BF,
	0x00B6C0, 0x00B6C1, 0x00B6C2, 0x00B6C3, 0x00B6C4, 0x00B6C5, 0x00B6C6, 0x00B6C7, 0x00B6C8, 0x00B6C9, 0x00B6CA, 0x00B6CB, 0x00B6CC, 0x00B6CD, 0x00B6CE, 0x00B6CF, 0x00B6D0, 0x00B6D1, 0x00B6D2, 0x00B6D3, 0x00B6D4, 0x00B6D5, 0x00B6D6, 0x00B6D7, 0x00B6D8, 0x00B6D9, 0x00B6DA, 0x00B6DB, 0x00B6DC, 0x00B6DD, 0x00B6DE, 0x00B6DF,
	0x00B6E0, 0x00B6E1, 0x00B6E2, 0x00B6E3, 0x00B6E4, 0x00B6E5, 0x00B6E6, 0x00B6E7, 0x00B6E8, 0x00B6E9, 0x00B6EA, 0x00B6EB, 0x00B6EC, 0x00B6ED, 0x00B6EE, 0x00B6EF, 0x00B6F0, 0x00B6F1, 0x00B6F2, 0x00B6F3, 0x00B6F4, 0x00B6F5, 0x00B6F6, 0x00B6F7, 0x00B6F8, 0x00B6F9, 0x00B6FA, 0x00B6FB, 0x00B6FC, 0x00B6FD, 0x00B6FE, 0x00B6FF,
	0x00B700, 0x00B701, 0x00B702, 0x00B703, 0x00B704, 0x00B705, 0x00B706, 0x00B707, 0x00B708, 0x00B709, 0x00B70A, 0x00B70B, 0x00B70C, 0x00B70D, 0x00B70E, 0x00B70F, 0x00B710, 0x00B711, 0x00B712, 0x00B713, 0x00B714, 0x00B715, 0x00B716, 0x00B717, 0x00B718, 0x00B719, 0x00B71A, 0x00B71B, 0x00B71C, 0x00B71D, 0x00B71E, 0x00B71F,
	0x00B720, 0x00B721, 0x00B722, 0x00B723, 0x00B724, 0x00B725, 0x00B726, 0x00B727, 0x00B728, 0x00B729, 0x00B72A, 0x00B72B, 0x00B72C, 0x00B72D, 0x00B72E, 0x00B72F, 0x00B730, 0x00B731, 0x00B732, 0x00B733, 0x00B734, 0x00B735, 0x00B736, 0x00B737, 0x00B738, 0x00B739, 0x00B73A, 0x00B73B, 0x00B73C, 0x00B73D, 0x00B73E, 0x00B73F,
	0x00B740, 0x00B741, 0x00B742, 0x00B743, 0x00B744, 0x00B745, 0x00B746, 0x00B747, 0x00B748, 0x00B749, 0x00B74A, 0x00B74B, 0x00B74C, 0x00B74D, 0x00B74E, 0x00B74F, 0x00B750, 0x00B751, 0x00B752, 0x00B753, 0x00B754, 0x00B755, 0x00B756, 0x00B757, 0x00B758, 0x00B759, 0x00B75A, 0x00B75B, 0x00B75C, 0x00B75D, 0x00B75E, 0x00B75F,
	0x00B760, 0x00B761, 0x00B762, 0x00B763, 0x00B764, 0x00B765, 0x00B766, 0x00B767, 0x00B768, 0x00B769, 0x00B76A, 0x00B76B, 0x00B76C, 0x00B76D, 0x00B76E, 0x00B76F, 0x00B770, 0x00B771, 0x00B772, 0x00B773, 0x00B774, 0x00B775, 0x00B776, 0x00B777, 0x00B778, 0x00B779, 0x00B77A, 0x00B77B, 0x00B77C, 0x00B77D, 0x00B77E, 0x00B77F,
	0x00B780, 0x00B781, 0x00B782, 0x00B783, 0x00B784, 0x00B785, 0x00B786, 0x00B787, 0x00B788, 0x00B789, 0x00B78A, 0x00B78B, 0x00B78C, 0x00B78D, 0x00B78E, 0x00B78F, 0x00B790, 0x00B791, 0x00B792, 0x00B793, 0x00B794, 0x00B795, 0x00B796, 0x00B797, 0x00B798, 0x00B799, 0x00B79A, 0x00B79B, 0x00B79C, 0x00B79D, 0x00B79E, 0x00B79F,
	0x00B7A0, 0x00B7A1, 0x00B7A2, 0x00B7A3, 0x00B7A4, 0x00B7A5, 0x00B7A6, 0x00B7A7, 0x00B7A8, 0x00B7A9, 0x00B7AA, 0x00B7AB, 0x00B7AC, 0x00B7AD, 0x00B7AE, 0x00B7AF, 0x00B7B0, 0x00B7B1, 0x00B7B2, 0x00B7B3, 0x00B7B4, 0x00B7B5, 0x00B7B6, 0x00B7B7, 0x00B7B8, 0x00B7B9, 0x00B7BA, 0x00B7BB, 0x00B7BC, 0x00B7BD, 0x00B7BE, 0x00B7BF,
	0x00B7C0, 0x00B7C1, 0x00B7C2, 0x00B7C3, 0x00B7C4, 0x00B7C5, 0x00B7C6, 0x00B7C7, 0x00B7C8, 0x00B7C9, 0x00B7CA, 0x00B7CB, 0x00B7CC, 0x00B7CD, 0x00B7CE, 0x00B7CF, 0x00B7D0, 0x00B7D1, 0x00B7D2, 0x00B7D3, 0x00B7D4, 0x00B7D5, 0x00B7D6, 0x00B7D7, 0x00B7D8, 0x00B7D9, 0x00B7DA, 0x00B7DB, 0x00B7DC, 0x00B7DD, 0x00B7DE, 0x00B7DF,
	0x00B7E0, 0x00B7E1, 0x00B7E2, 0x00B7E3, 0x00B7E4, 0x00B7E5, 0x00B7E6, 0x00B7E7, 0x00B7E8, 0x00B7E9, 0x00B7EA, 0x00B7EB, 0x00B7EC, 0x00B7ED, 0x00B7EE, 0x00B7EF, 0x00B7F0, 0x00B7F1, 0x00B7F2, 0x00B7F3, 0x00B7F4, 0x00B7F5, 0x00B7F6, 0x00B7F7, 0x00B7F8, 0x00B7F9, 0x00B7FA, 0x00B7FB, 0x00B7FC, 0x00B7FD, 0x00B7FE, 0x00B7FF,
	0
};

static const unsigned map_hangul_4[] = {
	0x00B800, 0x00B801, 0x00B802, 0x00B803, 0x00B804, 0x00B805, 0x00B806, 0x00B807, 0x00B808, 0x00B809, 0x00B80A, 0x00B80B, 0x00B80C, 0x00B80D, 0x00B80E, 0x00B80F, 0x00B810, 0x00B811, 0x00B812, 0x00B813, 0x00B814, 0x00B815, 0x00B816, 0x00B817, 0x00B818, 0x00B819, 0x00B81A, 0x00B81B, 0x00B81C, 0x00B81D, 0x00B81E, 0x00B81F,
	0x00B820, 0x00B821, 0x00B822, 0x00B823, 0x00B824, 0x00B825, 0x00B826, 0x00B827, 0x00B828, 0x00B829, 0x00B82A, 0x00B82B, 0x00B82C, 0x00B82D, 0x00B82E, 0x00B82F, 0x00B830, 0x00B831, 0x00B832, 0x00B833, 0x00B834, 0x00B835, 0x00B836, 0x00B837, 0x00B838, 0x00B839, 0x00B83A, 0x00B83B, 0x00B83C, 0x00B83D, 0x00B83E, 0x00B83F,
	0x00B840, 0x00B841, 0x00B842, 0x00B843, 0x00B844, 0x00B845, 0x00B846, 0x00B847, 0x00B848, 0x00B849, 0x00B84A, 0x00B84B, 0x00B84C, 0x00B84D, 0x00B84E, 0x00B84F, 0x00B850, 0x00B851, 0x00B852, 0x00B853, 0x00B854, 0x00B855, 0x00B856, 0x00B857, 0x00B858, 0x00B859, 0x00B85A, 0x00B85B, 0x00B85C, 0x00B85D, 0x00B85E, 0x00B85F,
	0x00B860, 0x00B861, 0x00B862, 0x00B863, 0x00B864, 0x00B865, 0x00B866, 0x00B867, 0x00B868, 0x00B869, 0x00B86A, 0x00B86B, 0x00B86C, 0x00B86D, 0x00B86E, 0x00B86F, 0x00B870, 0x00B871, 0x00B872, 0x00B873, 0x00B874, 0x00B875, 0x00B876, 0x00B877, 0x00B878, 0x00B879, 0x00B87A, 0x00B87B, 0x00B87C, 0x00B87D, 0x00B87E, 0x00B87F,
	0x00B880, 0x00B881, 0x00B882, 0x00B883, 0x00B884, 0x00B885, 0x00B886, 0x00B887, 0x00B888, 0x00B889, 0x00B88A, 0x00B88B, 0x00B88C, 0x00B88D, 0x00B88E, 0x00B88F, 0x00B890, 0x00B891, 0x00B892, 0x00B893, 0x00B894, 0x00B895, 0x00B896, 0x00B897, 0x00B898, 0x00B899, 0x00B89A, 0x00B89B, 0x00B89C, 0x00B89D, 0x00B89E, 0x00B89F,
	0x00B8A0, 0x00B8A1, 0x00B8A2, 0x00B8A3, 0x00B8A4, 0x00B8A5, 0x00B8A6, 0x00B8A7, 0x00B8A8, 0x00B8A9, 0x00B8AA, 0x00B8AB, 0x00B8AC, 0x00B8AD, 0x00B8AE, 0x00B8AF, 0x00B8B0, 0x00B8B1, 0x00B8B2, 0x00B8B3, 0x00B8B4, 0x00B8B5, 0x00B8B6, 0x00B8B7, 0x00B8B8, 0x00B8B9, 0x00B8BA, 0x00B8BB, 0x00B8BC, 0x00B8BD, 0x00B8BE, 0x00B8BF,
	0x00B8C0, 0x00B8C1, 0x00B8C2, 0x00B8C3, 0x00B8C4, 0x00B8C5, 0x00B8C6, 0x00B8C7, 0x00B8C8, 0x00B8C9, 0x00B8CA, 0x00B8CB, 0x00B8CC, 0x00B8CD, 0x00B8CE, 0x00B8CF, 0x00B8D0, 0x00B8D1, 0x00B8D2, 0x00B8D3, 0x00B8D4, 0x00B8D5, 0x00B8D6, 0x00B8D7, 0x00B8D8, 0x00B8D9, 0x00B8DA, 0x00B8DB, 0x00B8DC, 0x00B8DD, 0x00B8DE, 0x00B8DF,
	0x00B8E0, 0x00B8E1, 0x00B8E2, 0x00B8E3, 0x00B8E4, 0x00B8E5, 0x00B8E6, 0x00B8E7, 0x00B8E8, 0x00B8E9, 0x00B8EA, 0x00B8EB, 0x00B8EC, 0x00B8ED, 0x00B8EE, 0x00B8EF, 0x00B8F0, 0x00B8F1, 0x00B8F2, 0x00B8F3, 0x00B8F4, 0x00B8F5, 0x00B8F6, 0x00B8F7, 0x00B8F8, 0x00B8F9, 0x00B8FA, 0x00B8FB, 0x00B8FC, 0x00B8FD, 0x00B8FE, 0x00B8FF,
	0x00B900, 0x00B901, 0x00B902, 0x00B903, 0x00B904, 0x00B905, 0x00B906, 0x00B907, 0x00B908, 0x00B909, 0x00B90A, 0x00B90B, 0x00B90C, 0x00B90D, 0x00B90E, 0x00B90F, 0x00B910, 0x00B911, 0x00B912, 0x00B913, 0x00B914, 0x00B915, 0x00B916, 0x00B917, 0x00B918, 0x00B919, 0x00B91A, 0x00B91B, 0x00B91C, 0x00B91D, 0x00B91E, 0x00B91F,
	0x00B920, 0x00B921, 0x00B922, 0x00B923, 0x00B924, 0x00B925, 0x00B926, 0x00B927, 0x00B928, 0x00B929, 0x00B92A, 0x00B92B, 0x00B92C, 0x00B92D, 0x00B92E, 0x00B92F, 0x00B930, 0x00B931, 0x00B932, 0x00B933, 0x00B934, 0x00B935, 0x00B936, 0x00B937, 0x00B938, 0x00B939, 0x00B93A, 0x00B93B, 0x00B93C, 0x00B93D, 0x00B93E, 0x00B93F,
	0x00B940, 0x00B941, 0x00B942, 0x00B943, 0x00B944, 0x00B945, 0x00B946, 0x00B947, 0x00B948, 0x00B949, 0x00B94A, 0x00B94B, 0x00B94C, 0x00B94D, 0x00B94E, 0x00B94F, 0x00B950, 0x00B951, 0x00B952, 0x00B953, 0x00B954, 0x00B955, 0x00B956, 0x00B957, 0x00B958, 0x00B959, 0x00B95A, 0x00B95B, 0x00B95C, 0x00B95D, 0x00B95E, 0x00B95F,
	0x00B960, 0x00B961, 0x00B962, 0x00B963, 0x00B964, 0x00B965, 0x00B966, 0x00B967, 0x00B968, 0x00B969, 0x00B96A, 0x00B96B, 0x00B96C, 0x00B96D, 0x00B96E, 0x00B96F, 0x00B970, 0x00B971, 0x00B972, 0x00B973, 0x00B974, 0x00B975, 0x00B976, 0x00B977, 0x00B978, 0x00B979, 0x00B97A, 0x00B97B, 0x00B97C, 0x00B97D, 0x00B97E, 0x00B97F,
	0x00B980, 0x00B981, 0x00B982, 0x00B983, 0x00B984, 0x00B985, 0x00B986, 0x00B987, 0x00B988, 0x00B989, 0x00B98A, 0x00B98B, 0x00B98C, 0x00B98D, 0x00B98E, 0x00B98F, 0x00B990, 0x00B991, 0x00B992, 0x00B993, 0x00B994, 0x00B995, 0x00B996, 0x00B997, 0x00B998, 0x00B999, 0x00B99A, 0x00B99B, 0x00B99C, 0x00B99D, 0x00B99E, 0x00B99F,
	0x00B9A0, 0x00B9A1, 0x00B9A2, 0x00B9A3, 0x00B9A4, 0x00B9A5, 0x00B9A6, 0x00B9A7, 0x00B9A8, 0x00B9A9, 0x00B9AA, 0x00B9AB, 0x00B9AC, 0x00B9AD, 0x00B9AE, 0x00B9AF, 0x00B9B0, 0x00B9B1, 0x00B9B2, 0x00B9B3, 0x00B9B4, 0x00B9B5, 0x00B9B6, 0x00B9B7, 0x00B9B8, 0x00B9B9, 0x00B9BA, 0x00B9BB, 0x00B9BC, 0x00B9BD, 0x00B9BE, 0x00B9BF,
	0x00B9C0, 0x00B9C1, 0x00B9C2, 0x00B9C3, 0x00B9C4, 0x00B9C5, 0x00B9C6, 0x00B9C7, 0x00B9C8, 0x00B9C9, 0x00B9CA, 0x00B9CB, 0x00B9CC, 0x00B9CD, 0x00B9CE, 0x00B9CF, 0x00B9D0, 0x00B9D1, 0x00B9D2, 0x00B9D3, 0x00B9D4, 0x00B9D5, 0x00B9D6, 0x00B9D7, 0x00B9D8, 0x00B9D9, 0x00B9DA, 0x00B9DB, 0x00B9DC, 0x00B9DD, 0x00B9DE, 0x00B9DF,
	0x00B9E0, 0x00B9E1, 0x00B9E2, 0x00B9E3, 0x00B9E4, 0x00B9E5, 0x00B9E6, 0x00B9E7, 0x00B9E8, 0x00B9E9, 0x00B9EA, 0x00B9EB, 0x00B9EC, 0x00B9ED, 0x00B9EE, 0x00B9EF, 0x00B9F0, 0x00B9F1, 0x00B9F2, 0x00B9F3, 0x00B9F4, 0x00B9F5, 0x00B9F6, 0x00B9F7, 0x00B9F8, 0x00B9F9, 0x00B9FA, 0x00B9FB, 0x00B9FC, 0x00B9FD, 0x00B9FE, 0x00B9FF,
	0x00BA00, 0x00BA01, 0x00BA02, 0x00BA03, 0x00BA04, 0x00BA05, 0x00BA06, 0x00BA07, 0x00BA08, 0x00BA09, 0x00BA0A, 0x00BA0B, 0x00BA0C, 0x00BA0D, 0x00BA0E, 0x00BA0F, 0x00BA10, 0x00BA11, 0x00BA12, 0x00BA13, 0x00BA14, 0x00BA15, 0x00BA16, 0x00BA17, 0x00BA18, 0x00BA19, 0x00BA1A, 0x00BA1B, 0x00BA1C, 0x00BA1D, 0x00BA1E, 0x00BA1F,
	0x00BA20, 0x00BA21, 0x00BA22, 0x00BA23, 0x00BA24, 0x00BA25, 0x00BA26, 0x00BA27, 0x00BA28, 0x00BA29, 0x00BA2A, 0x00BA2B, 0x00BA2C, 0x00BA2D, 0x00BA2E, 0x00BA2F, 0x00BA30, 0x00BA31, 0x00BA32, 0x00BA33, 0x00BA34, 0x00BA35, 0x00BA36, 0x00BA37, 0x00BA38, 0x00BA39, 0x00BA3A, 0x00BA3B, 0x00BA3C, 0x00BA3D, 0x00BA3E, 0x00BA3F,
	0x00BA40, 0x00BA41, 0x00BA42, 0x00BA43, 0x00BA44, 0x00BA45, 0x00BA46, 0x00BA47, 0x00BA48, 0x00BA49, 0x00BA4A, 0x00BA4B, 0x00BA4C, 0x00BA4D, 0x00BA4E, 0x00BA4F, 0x00BA50, 0x00BA51, 0x00BA52, 0x00BA53, 0x00BA54, 0x00BA55, 0x00BA56, 0x00BA57, 0x00BA58, 0x00BA59, 0x00BA5A, 0x00BA5B, 0x00BA5C, 0x00BA5D, 0x00BA5E, 0x00BA5F,
	0x00BA60, 0x00BA61, 0x00BA62, 0x00BA63, 0x00BA64, 0x00BA65, 0x00BA66, 0x00BA67, 0x00BA68, 0x00BA69, 0x00BA6A, 0x00BA6B, 0x00BA6C, 0x00BA6D, 0x00BA6E, 0x00BA6F, 0x00BA70, 0x00BA71, 0x00BA72, 0x00BA73, 0x00BA74, 0x00BA75, 0x00BA76, 0x00BA77, 0x00BA78, 0x00BA79, 0x00BA7A, 0x00BA7B, 0x00BA7C, 0x00BA7D, 0x00BA7E, 0x00BA7F,
	0x00BA80, 0x00BA81, 0x00BA82, 0x00BA83, 0x00BA84, 0x00BA85, 0x00BA86, 0x00BA87, 0x00BA88, 0x00BA89, 0x00BA8A, 0x00BA8B, 0x00BA8C, 0x00BA8D, 0x00BA8E, 0x00BA8F, 0x00BA90, 0x00BA91, 0x00BA92, 0x00BA93, 0x00BA94, 0x00BA95, 0x00BA96, 0x00BA97, 0x00BA98, 0x00BA99, 0x00BA9A, 0x00BA9B, 0x00BA9C, 0x00BA9D, 0x00BA9E, 0x00BA9F,
	0x00BAA0, 0x00BAA1, 0x00BAA2, 0x00BAA3, 0x00BAA4, 0x00BAA5, 0x00BAA6, 0x00BAA7, 0x00BAA8, 0x00BAA9, 0x00BAAA, 0x00BAAB, 0x00BAAC, 0x00BAAD, 0x00BAAE, 0x00BAAF, 0x00BAB0, 0x00BAB1, 0x00BAB2, 0x00BAB3, 0x00BAB4, 0x00BAB5, 0x00BAB6, 0x00BAB7, 0x00BAB8, 0x00BAB9, 0x00BABA, 0x00BABB, 0x00BABC, 0x00BABD, 0x00BABE, 0x00BABF,
	0x00BAC0, 0x00BAC1, 0x00BAC2, 0x00BAC3, 0x00BAC4, 0x00BAC5, 0x00BAC6, 0x00BAC7, 0x00BAC8, 0x00BAC9, 0x00BACA, 0x00BACB, 0x00BACC, 0x00BACD, 0x00BACE, 0x00BACF, 0x00BAD0, 0x00BAD1, 0x00BAD2, 0x00BAD3, 0x00BAD4, 0x00BAD5, 0x00BAD6, 0x00BAD7, 0x00BAD8, 0x00BAD9, 0x00BADA, 0x00BADB, 0x00BADC, 0x00BADD, 0x00BADE, 0x00BADF,
	0x00BAE0, 0x00BAE1, 0x00BAE2, 0x00BAE3, 0x00BAE4, 0x00BAE5, 0x00BAE6, 0x00BAE7, 0x00BAE8, 0x00BAE9, 0x00BAEA, 0x00BAEB, 0x00BAEC, 0x00BAED, 0x00BAEE, 0x00BAEF, 0x00BAF0, 0x00BAF1, 0x00BAF2, 0x00BAF3, 0x00BAF4, 0x00BAF5, 0x00BAF6, 0x00BAF7, 0x00BAF8, 0x00BAF9, 0x00BAFA, 0x00BAFB, 0x00BAFC, 0x00BAFD, 0x00BAFE, 0x00BAFF,
	0x00BB00, 0x00BB01, 0x00BB02, 0x00BB03, 0x00BB04, 0x00BB05, 0x00BB06, 0x00BB07, 0x00BB08, 0x00BB09, 0x00BB0A, 0x00BB0B, 0x00BB0C, 0x00BB0D, 0x00BB0E, 0x00BB0F, 0x00BB10, 0x00BB11, 0x00BB12, 0x00BB13, 0x00BB14, 0x00BB15, 0x00BB16, 0x00BB17, 0x00BB18, 0x00BB19, 0x00BB1A, 0x00BB1B, 0x00BB1C, 0x00BB1D, 0x00BB1E, 0x00BB1F,
	0x00BB20, 0x00BB21, 0x00BB22, 0x00BB23, 0x00BB24, 0x00BB25, 0x00BB26, 0x00BB27, 0x00BB28, 0x00BB29, 0x00BB2A, 0x00BB2B, 0x00BB2C, 0x00BB2D, 0x00BB2E, 0x00BB2F, 0x00BB30, 0x00BB31, 0x00BB32, 0x00BB33, 0x00BB34, 0x00BB35, 0x00BB36, 0x00BB37, 0x00BB38, 0x00BB39, 0x00BB3A, 0x00BB3B, 0x00BB3C, 0x00BB3D, 0x00BB3E, 0x00BB3F,
	0x00BB40, 0x00BB41, 0x00BB42, 0x00BB43, 0x00BB44, 0x00BB45, 0x00BB46, 0x00BB47, 0x00BB48, 0x00BB49, 0x00BB4A, 0x00BB4B, 0x00BB4C, 0x00BB4D, 0x00BB4E, 0x00BB4F, 0x00BB50, 0x00BB51, 0x00BB52, 0x00BB53, 0x00BB54, 0x00BB55, 0x00BB56, 0x00BB57, 0x00BB58, 0x00BB59, 0x00BB5A, 0x00BB5B, 0x00BB5C, 0x00BB5D, 0x00BB5E, 0x00BB5F,
	0x00BB60, 0x00BB61, 0x00BB62, 0x00BB63, 0x00BB64, 0x00BB65, 0x00BB66, 0x00BB67, 0x00BB68, 0x00BB69, 0x00BB6A, 0x00BB6B, 0x00BB6C, 0x00BB6D, 0x00BB6E, 0x00BB6F, 0x00BB70, 0x00BB71, 0x00BB72, 0x00BB73, 0x00BB74, 0x00BB75, 0x00BB76, 0x00BB77, 0x00BB78, 0x00BB79, 0x00BB7A, 0x00BB7B, 0x00BB7C, 0x00BB7D, 0x00BB7E, 0x00BB7F,
	0x00BB80, 0x00BB81, 0x00BB82, 0x00BB83, 0x00BB84, 0x00BB85, 0x00BB86, 0x00BB87, 0x00BB88, 0x00BB89, 0x00BB8A, 0x00BB8B, 0x00BB8C, 0x00BB8D, 0x00BB8E, 0x00BB8F, 0x00BB90, 0x00BB91, 0x00BB92, 0x00BB93, 0x00BB94, 0x00BB95, 0x00BB96, 0x00BB97, 0x00BB98, 0x00BB99, 0x00BB9A, 0x00BB9B, 0x00BB9C, 0x00BB9D, 0x00BB9E, 0x00BB9F,
	0x00BBA0, 0x00BBA1, 0x00BBA2, 0x00BBA3, 0x00BBA4, 0x00BBA5, 0x00BBA6, 0x00BBA7, 0x00BBA8, 0x00BBA9, 0x00BBAA, 0x00BBAB, 0x00BBAC, 0x00BBAD, 0x00BBAE, 0x00BBAF, 0x00BBB0, 0x00BBB1, 0x00BBB2, 0x00BBB3, 0x00BBB4, 0x00BBB5, 0x00BBB6, 0x00BBB7, 0x00BBB8, 0x00BBB9, 0x00BBBA, 0x00BBBB, 0x00BBBC, 0x00BBBD, 0x00BBBE, 0x00BBBF,
	0x00BBC0, 0x00BBC1, 0x00BBC2, 0x00BBC3, 0x00BBC4, 0x00BBC5, 0x00BBC6, 0x00BBC7, 0x00BBC8, 0x00BBC9, 0x00BBCA, 0x00BBCB, 0x00BBCC, 0x00BBCD, 0x00BBCE, 0x00BBCF, 0x00BBD0, 0x00BBD1, 0x00BBD2, 0x00BBD3, 0x00BBD4, 0x00BBD5, 0x00BBD6, 0x00BBD7, 0x00BBD8, 0x00BBD9, 0x00BBDA, 0x00BBDB, 0x00BBDC, 0x00BBDD, 0x00BBDE, 0x00BBDF,
	0x00BBE0, 0x00BBE1, 0x00BBE2, 0x00BBE3, 0x00BBE4, 0x00BBE5, 0x00BBE6, 0x00BBE7, 0x00BBE8, 0x00BBE9, 0x00BBEA, 0x00BBEB, 0x00BBEC, 0x00BBED, 0x00BBEE, 0x00BBEF, 0x00BBF0, 0x00BBF1, 0x00BBF2, 0x00BBF3, 0x00BBF4, 0x00BBF5, 0x00BBF6, 0x00BBF7, 0x00BBF8, 0x00BBF9, 0x00BBFA, 0x00BBFB, 0x00BBFC, 0x00BBFD, 0x00BBFE, 0x00BBFF,
	0
};

static const unsigned map_hangul_5[] = {
	0x00BC00, 0x00BC01, 0x00BC02, 0x00BC03, 0x00BC04, 0x00BC05, 0x00BC06, 0x00BC07, 0x00BC08, 0x00BC09, 0x00BC0A, 0x00BC0B, 0x00BC0C, 0x00BC0D, 0x00BC0E, 0x00BC0F, 0x00BC10, 0x00BC11, 0x00BC12, 0x00BC13, 0x00BC14, 0x00BC15, 0x00BC16, 0x00BC17, 0x00BC18, 0x00BC19, 0x00BC1A, 0x00BC1B, 0x00BC1C, 0x00BC1D, 0x00BC1E, 0x00BC1F,
	0x00BC20, 0x00BC21, 0x00BC22, 0x00BC23, 0x00BC24, 0x00BC25, 0x00BC26, 0x00BC27, 0x00BC28, 0x00BC29, 0x00BC2A, 0x00BC2B, 0x00BC2C, 0x00BC2D, 0x00BC2E, 0x00BC2F, 0x00BC30, 0x00BC31, 0x00BC32, 0x00BC33, 0x00BC34, 0x00BC35, 0x00BC36, 0x00BC37, 0x00BC38, 0x00BC39, 0x00BC3A, 0x00BC3B, 0x00BC3C, 0x00BC3D, 0x00BC3E, 0x00BC3F,
	0x00BC40, 0x00BC41, 0x00BC42, 0x00BC43, 0x00BC44, 0x00BC45, 0x00BC46, 0x00BC47, 0x00BC48, 0x00BC49, 0x00BC4A, 0x00BC4B, 0x00BC4C, 0x00BC4D, 0x00BC4E, 0x00BC4F, 0x00BC50, 0x00BC51, 0x00BC52, 0x00BC53, 0x00BC54, 0x00BC55, 0x00BC56, 0x00BC57, 0x00BC58, 0x00BC59, 0x00BC5A, 0x00BC5B, 0x00BC5C, 0x00BC5D, 0x00BC5E, 0x00BC5F,
	0x00BC60, 0x00BC61, 0x00BC62, 0x00BC63, 0x00BC64, 0x00BC65, 0x00BC66, 0x00BC67, 0x00BC68, 0x00BC69, 0x00BC6A, 0x00BC6B, 0x00BC6C, 0x00BC6D, 0x00BC6E, 0x00BC6F, 0x00BC70, 0x00BC71, 0x00BC72, 0x00BC73, 0x00BC74, 0x00BC75, 0x00BC76, 0x00BC77, 0x00BC78, 0x00BC79, 0x00BC7A, 0x00BC7B, 0x00BC7C, 0x00BC7D, 0x00BC7E, 0x00BC7F,
	0x00BC80, 0x00BC81, 0x00BC82, 0x00BC83, 0x00BC84, 0x00BC85, 0x00BC86, 0x00BC87, 0x00BC88, 0x00BC89, 0x00BC8A, 0x00BC8B, 0x00BC8C, 0x00BC8D, 0x00BC8E, 0x00BC8F, 0x00BC90, 0x00BC91, 0x00BC92, 0x00BC93, 0x00BC94, 0x00BC95, 0x00BC96, 0x00BC97, 0x00BC98, 0x00BC99, 0x00BC9A, 0x00BC9B, 0x00BC9C, 0x00BC9D, 0x00BC9E, 0x00BC9F,
	0x00BCA0, 0x00BCA1, 0x00BCA2, 0x00BCA3, 0x00BCA4, 0x00BCA5, 0x00BCA6, 0x00BCA7, 0x00BCA8, 0x00BCA9, 0x00BCAA, 0x00BCAB, 0x00BCAC, 0x00BCAD, 0x00BCAE, 0x00BCAF, 0x00BCB0, 0x00BCB1, 0x00BCB2, 0x00BCB3, 0x00BCB4, 0x00BCB5, 0x00BCB6, 0x00BCB7, 0x00BCB8, 0x00BCB9, 0x00BCBA, 0x00BCBB, 0x00BCBC, 0x00BCBD, 0x00BCBE, 0x00BCBF,
	0x00BCC0, 0x00BCC1, 0x00BCC2, 0x00BCC3, 0x00BCC4, 0x00BCC5, 0x00BCC6, 0x00BCC7, 0x00BCC8, 0x00BCC9, 0x00BCCA, 0x00BCCB, 0x00BCCC, 0x00BCCD, 0x00BCCE, 0x00BCCF, 0x00BCD0, 0x00BCD1, 0x00BCD2, 0x00BCD3, 0x00BCD4, 0x00BCD5, 0x00BCD6, 0x00BCD7, 0x00BCD8, 0x00BCD9, 0x00BCDA, 0x00BCDB, 0x00BCDC, 0x00BCDD, 0x00BCDE, 0x00BCDF,
	0x00BCE0, 0x00BCE1, 0x00BCE2, 0x00BCE3, 0x00BCE4, 0x00BCE5, 0x00BCE6, 0x00BCE7, 0x00BCE8, 0x00BCE9, 0x00BCEA, 0x00BCEB, 0x00BCEC, 0x00BCED, 0x00BCEE, 0x00BCEF, 0x00BCF0, 0x00BCF1, 0x00BCF2, 0x00BCF3, 0x00BCF4, 0x00BCF5, 0x00BCF6, 0x00BCF7, 0x00BCF8, 0x00BCF9, 0x00BCFA, 0x00BCFB, 0x00BCFC, 0x00BCFD, 0x00BCFE, 0x00BCFF,
	0x00BD00, 0x00BD01, 0x00BD02, 0x00BD03, 0x00BD04, 0x00BD05, 0x00BD06, 0x00BD07, 0x00BD08, 0x00BD09, 0x00BD0A, 0x00BD0B, 0x00BD0C, 0x00BD0D, 0x00BD0E, 0x00BD0F, 0x00BD10, 0x00BD11, 0x00BD12, 0x00BD13, 0x00BD14, 0x00BD15, 0x00BD16, 0x00BD17, 0x00BD18, 0x00BD19, 0x00BD1A, 0x00BD1B, 0x00BD1C, 0x00BD1D, 0x00BD1E, 0x00BD1F,
	0x00BD20, 0x00BD21, 0x00BD22, 0x00BD23, 0x00BD24, 0x00BD25, 0x00BD26, 0x00BD27, 0x00BD28, 0x00BD29, 0x00BD2A, 0x00BD2B, 0x00BD2C, 0x00BD2D, 0x00BD2E, 0x00BD2F, 0x00BD30, 0x00BD31, 0x00BD32, 0x00BD33, 0x00BD34, 0x00BD35, 0x00BD36, 0x00BD37, 0x00BD38, 0x00BD39, 0x00BD3A, 0x00BD3B, 0x00BD3C, 0x00BD3D, 0x00BD3E, 0x00BD3F,
	0x00BD40, 0x00BD41, 0x00BD42, 0x00BD43, 0x00BD44, 0x00BD45, 0x00BD46, 0x00BD47, 0x00BD48, 0x00BD49, 0x00BD4A, 0x00BD4B, 0x00BD4C, 0x00BD4D, 0x00BD4E, 0x00BD4F, 0x00BD50, 0x00BD51, 0x00BD52, 0x00BD53, 0x00BD54, 0x00BD55, 0x00BD56, 0x00BD57, 0x00BD58, 0x00BD59, 0x00BD5A, 0x00BD5B, 0x00BD5C, 0x00BD5D, 0x00BD5E, 0x00BD5F,
	0x00BD60, 0x00BD61, 0x00BD62, 0x00BD63, 0x00BD64, 0x00BD65, 0x00BD66, 0x00BD67, 0x00BD68, 0x00BD69, 0x00BD6A, 0x00BD6B, 0x00BD6C, 0x00BD6D, 0x00BD6E, 0x00BD6F, 0x00BD70, 0x00BD71, 0x00BD72, 0x00BD73, 0x00BD74, 0x00BD75, 0x00BD76, 0x00BD77, 0x00BD78, 0x00BD79, 0x00BD7A, 0x00BD7B, 0x00BD7C, 0x00BD7D, 0x00BD7E, 0x00BD7F,
	0x00BD80, 0x00BD81, 0x00BD82, 0x00BD83, 0x00BD84, 0x00BD85, 0x00BD86, 0x00BD87, 0x00BD88, 0x00BD89, 0x00BD8A, 0x00BD8B, 0x00BD8C, 0x00BD8D, 0x00BD8E, 0x00BD8F, 0x00BD90, 0x00BD91, 0x00BD92, 0x00BD93, 0x00BD94, 0x00BD95, 0x00BD96, 0x00BD97, 0x00BD98, 0x00BD99, 0x00BD9A, 0x00BD9B, 0x00BD9C, 0x00BD9D, 0x00BD9E, 0x00BD9F,
	0x00BDA0, 0x00BDA1, 0x00BDA2, 0x00BDA3, 0x00BDA4, 0x00BDA5, 0x00BDA6, 0x00BDA7, 0x00BDA8, 0x00BDA9, 0x00BDAA, 0x00BDAB, 0x00BDAC, 0x00BDAD, 0x00BDAE, 0x00BDAF, 0x00BDB0, 0x00BDB1, 0x00BDB2, 0x00BDB3, 0x00BDB4, 0x00BDB5, 0x00BDB6, 0x00BDB7, 0x00BDB8, 0x00BDB9, 0x00BDBA, 0x00BDBB, 0x00BDBC, 0x00BDBD, 0x00BDBE, 0x00BDBF,
	0x00BDC0, 0x00BDC1, 0x00BDC2, 0x00BDC3, 0x00BDC4, 0x00BDC5, 0x00BDC6, 0x00BDC7, 0x00BDC8, 0x00BDC9, 0x00BDCA, 0x00BDCB, 0x00BDCC, 0x00BDCD, 0x00BDCE, 0x00BDCF, 0x00BDD0, 0x00BDD1, 0x00BDD2, 0x00BDD3, 0x00BDD4, 0x00BDD5, 0x00BDD6, 0x00BDD7, 0x00BDD8, 0x00BDD9, 0x00BDDA, 0x00BDDB, 0x00BDDC, 0x00BDDD, 0x00BDDE, 0x00BDDF,
	0x00BDE0, 0x00BDE1, 0x00BDE2, 0x00BDE3, 0x00BDE4, 0x00BDE5, 0x00BDE6, 0x00BDE7, 0x00BDE8, 0x00BDE9, 0x00BDEA, 0x00BDEB, 0x00BDEC, 0x00BDED, 0x00BDEE, 0x00BDEF, 0x00BDF0, 0x00BDF1, 0x00BDF2, 0x00BDF3, 0x00BDF4, 0x00BDF5, 0x00BDF6, 0x00BDF7, 0x00BDF8, 0x00BDF9, 0x00BDFA, 0x00BDFB, 0x00BDFC, 0x00BDFD, 0x00BDFE, 0x00BDFF,
	0x00BE00, 0x00BE01, 0x00BE02, 0x00BE03, 0x00BE04, 0x00BE05, 0x00BE06, 0x00BE07, 0x00BE08, 0x00BE09, 0x00BE0A, 0x00BE0B, 0x00BE0C, 0x00BE0D, 0x00BE0E, 0x00BE0F, 0x00BE10, 0x00BE11, 0x00BE12, 0x00BE13, 0x00BE14, 0x00BE15, 0x00BE16, 0x00BE17, 0x00BE18, 0x00BE19, 0x00BE1A, 0x00BE1B, 0x00BE1C, 0x00BE1D, 0x00BE1E, 0x00BE1F,
	0x00BE20, 0x00BE21, 0x00BE22, 0x00BE23, 0x00BE24, 0x00BE25, 0x00BE26, 0x00BE27, 0x00BE28, 0x00BE29, 0x00BE2A, 0x00BE2B, 0x00BE2C, 0x00BE2D, 0x00BE2E, 0x00BE2F, 0x00BE30, 0x00BE31, 0x00BE32, 0x00BE33, 0x00BE34, 0x00BE35, 0x00BE36, 0x00BE37, 0x00BE38, 0x00BE39, 0x00BE3A, 0x00BE3B, 0x00BE3C, 0x00BE3D, 0x00BE3E, 0x00BE3F,
	0x00BE40, 0x00BE41, 0x00BE42, 0x00BE43, 0x00BE44, 0x00BE45, 0x00BE46, 0x00BE47, 0x00BE48, 0x00BE49, 0x00BE4A, 0x00BE4B, 0x00BE4C, 0x00BE4D, 0x00BE4E, 0x00BE4F, 0x00BE50, 0x00BE51, 0x00BE52, 0x00BE53, 0x00BE54, 0x00BE55, 0x00BE56, 0x00BE57, 0x00BE58, 0x00BE59, 0x00BE5A, 0x00BE5B, 0x00BE5C, 0x00BE5D, 0x00BE5E, 0x00BE5F,
	0x00BE60, 0x00BE61, 0x00BE62, 0x00BE63, 0x00BE64, 0x00BE65, 0x00BE66, 0x00BE67, 0x00BE68, 0x00BE69, 0x00BE6A, 0x00BE6B, 0x00BE6C, 0x00BE6D, 0x00BE6E, 0x00BE6F, 0x00BE70, 0x00BE71, 0x00BE72, 0x00BE73, 0x00BE74, 0x00BE75, 0x00BE76, 0x00BE77, 0x00BE78, 0x00BE79, 0x00BE7A, 0x00BE7B, 0x00BE7C, 0x00BE7D, 0x00BE7E, 0x00BE7F,
	0x00BE80, 0x00BE81, 0x00BE82, 0x00BE83, 0x00BE84, 0x00BE85, 0x00BE86, 0x00BE87, 0x00BE88, 0x00BE89, 0x00BE8A, 0x00BE8B, 0x00BE8C, 0x00BE8D, 0x00BE8E, 0x00BE8F, 0x00BE90, 0x00BE91, 0x00BE92, 0x00BE93, 0x00BE94, 0x00BE95, 0x00BE96, 0x00BE97, 0x00BE98, 0x00BE99, 0x00BE9A, 0x00BE9B, 0x00BE9C, 0x00BE9D, 0x00BE9E, 0x00BE9F,
	0x00BEA0, 0x00BEA1, 0x00BEA2, 0x00BEA3, 0x00BEA4, 0x00BEA5, 0x00BEA6, 0x00BEA7, 0x00BEA8, 0x00BEA9, 0x00BEAA, 0x00BEAB, 0x00BEAC, 0x00BEAD, 0x00BEAE, 0x00BEAF, 0x00BEB0, 0x00BEB1, 0x00BEB2, 0x00BEB3, 0x00BEB4, 0x00BEB5, 0x00BEB6, 0x00BEB7, 0x00BEB8, 0x00BEB9, 0x00BEBA, 0x00BEBB, 0x00BEBC, 0x00BEBD, 0x00BEBE, 0x00BEBF,
	0x00BEC0, 0x00BEC1, 0x00BEC2, 0x00BEC3, 0x00BEC4, 0x00BEC5, 0x00BEC6, 0x00BEC7, 0x00BEC8, 0x00BEC9, 0x00BECA, 0x00BECB, 0x00BECC, 0x00BECD, 0x00BECE, 0x00BECF, 0x00BED0, 0x00BED1, 0x00BED2, 0x00BED3, 0x00BED4, 0x00BED5, 0x00BED6, 0x00BED7, 0x00BED8, 0x00BED9, 0x00BEDA, 0x00BEDB, 0x00BEDC, 0x00BEDD, 0x00BEDE, 0x00BEDF,
	0x00BEE0, 0x00BEE1, 0x00BEE2, 0x00BEE3, 0x00BEE4, 0x00BEE5, 0x00BEE6, 0x00BEE7, 0x00BEE8, 0x00BEE9, 0x00BEEA, 0x00BEEB, 0x00BEEC, 0x00BEED, 0x00BEEE, 0x00BEEF, 0x00BEF0, 0x00BEF1, 0x00BEF2, 0x00BEF3, 0x00BEF4, 0x00BEF5, 0x00BEF6, 0x00BEF7, 0x00BEF8, 0x00BEF9, 0x00BEFA, 0x00BEFB, 0x00BEFC, 0x00BEFD, 0x00BEFE, 0x00BEFF,
	0x00BF00, 0x00BF01, 0x00BF02, 0x00BF03, 0x00BF04, 0x00BF05, 0x00BF06, 0x00BF07, 0x00BF08, 0x00BF09, 0x00BF0A, 0x00BF0B, 0x00BF0C, 0x00BF0D, 0x00BF0E, 0x00BF0F, 0x00BF10, 0x00BF11, 0x00BF12, 0x00BF13, 0x00BF14, 0x00BF15, 0x00BF16, 0x00BF17, 0x00BF18, 0x00BF19, 0x00BF1A, 0x00BF1B, 0x00BF1C, 0x00BF1D, 0x00BF1E, 0x00BF1F,
	0x00BF20, 0x00BF21, 0x00BF22, 0x00BF23, 0x00BF24, 0x00BF25, 0x00BF26, 0x00BF27, 0x00BF28, 0x00BF29, 0x00BF2A, 0x00BF2B, 0x00BF2C, 0x00BF2D, 0x00BF2E, 0x00BF2F, 0x00BF30, 0x00BF31, 0x00BF32, 0x00BF33, 0x00BF34, 0x00BF35, 0x00BF36, 0x00BF37, 0x00BF38, 0x00BF39, 0x00BF3A, 0x00BF3B, 0x00BF3C, 0x00BF3D, 0x00BF3E, 0x00BF3F,
	0x00BF40, 0x00BF41, 0x00BF42, 0x00BF43, 0x00BF44, 0x00BF45, 0x00BF46, 0x00BF47, 0x00BF48, 0x00BF49, 0x00BF4A, 0x00BF4B, 0x00BF4C, 0x00BF4D, 0x00BF4E, 0x00BF4F, 0x00BF50, 0x00BF51, 0x00BF52, 0x00BF53, 0x00BF54, 0x00BF55, 0x00BF56, 0x00BF57, 0x00BF58, 0x00BF59, 0x00BF5A, 0x00BF5B, 0x00BF5C, 0x00BF5D, 0x00BF5E, 0x00BF5F,
	0x00BF60, 0x00BF61, 0x00BF62, 0x00BF63, 0x00BF64, 0x00BF65, 0x00BF66, 0x00BF67, 0x00BF68, 0x00BF69, 0x00BF6A, 0x00BF6B, 0x00BF6C, 0x00BF6D, 0x00BF6E, 0x00BF6F, 0x00BF70, 0x00BF71, 0x00BF72, 0x00BF73, 0x00BF74, 0x00BF75, 0x00BF76, 0x00BF77, 0x00BF78, 0x00BF79, 0x00BF7A, 0x00BF7B, 0x00BF7C, 0x00BF7D, 0x00BF7E, 0x00BF7F,
	0x00BF80, 0x00BF81, 0x00BF82, 0x00BF83, 0x00BF84, 0x00BF85, 0x00BF86, 0x00BF87, 0x00BF88, 0x00BF89, 0x00BF8A, 0x00BF8B, 0x00BF8C, 0x00BF8D, 0x00BF8E, 0x00BF8F, 0x00BF90, 0x00BF91, 0x00BF92, 0x00BF93, 0x00BF94, 0x00BF95, 0x00BF96, 0x00BF97, 0x00BF98, 0x00BF99, 0x00BF9A, 0x00BF9B, 0x00BF9C, 0x00BF9D, 0x00BF9E, 0x00BF9F,
	0x00BFA0, 0x00BFA1, 0x00BFA2, 0x00BFA3, 0x00BFA4, 0x00BFA5, 0x00BFA6, 0x00BFA7, 0x00BFA8, 0x00BFA9, 0x00BFAA, 0x00BFAB, 0x00BFAC, 0x00BFAD, 0x00BFAE, 0x00BFAF, 0x00BFB0, 0x00BFB1, 0x00BFB2, 0x00BFB3, 0x00BFB4, 0x00BFB5, 0x00BFB6, 0x00BFB7, 0x00BFB8, 0x00BFB9, 0x00BFBA, 0x00BFBB, 0x00BFBC, 0x00BFBD, 0x00BFBE, 0x00BFBF,
	0x00BFC0, 0x00BFC1, 0x00BFC2, 0x00BFC3, 0x00BFC4, 0x00BFC5, 0x00BFC6, 0x00BFC7, 0x00BFC8, 0x00BFC9, 0x00BFCA, 0x00BFCB, 0x00BFCC, 0x00BFCD, 0x00BFCE, 0x00BFCF, 0x00BFD0, 0x00BFD1, 0x00BFD2, 0x00BFD3, 0x00BFD4, 0x00BFD5, 0x00BFD6, 0x00BFD7, 0x00BFD8, 0x00BFD9, 0x00BFDA, 0x00BFDB, 0x00BFDC, 0x00BFDD, 0x00BFDE, 0x00BFDF,
	0x00BFE0, 0x00BFE1, 0x00BFE2, 0x00BFE3, 0x00BFE4, 0x00BFE5, 0x00BFE6, 0x00BFE7, 0x00BFE8, 0x00BFE9, 0x00BFEA, 0x00BFEB, 0x00BFEC, 0x00BFED, 0x00BFEE, 0x00BFEF, 0x00BFF0, 0x00BFF1, 0x00BFF2, 0x00BFF3, 0x00BFF4, 0x00BFF5, 0x00BFF6, 0x00BFF7, 0x00BFF8, 0x00BFF9, 0x00BFFA, 0x00BFFB, 0x00BFFC, 0x00BFFD, 0x00BFFE, 0x00BFFF,
	0
};

static const unsigned map_hangul_6[] = {
	0x00C000, 0x00C001, 0x00C002, 0x00C003, 0x00C004, 0x00C005, 0x00C006, 0x00C007, 0x00C008, 0x00C009, 0x00C00A, 0x00C00B, 0x00C00C, 0x00C00D, 0x00C00E, 0x00C00F, 0x00C010, 0x00C011, 0x00C012, 0x00C013, 0x00C014, 0x00C015, 0x00C016, 0x00C017, 0x00C018, 0x00C019, 0x00C01A, 0x00C01B, 0x00C01C, 0x00C01D, 0x00C01E, 0x00C01F,
	0x00C020, 0x00C021, 0x00C022, 0x00C023, 0x00C024, 0x00C025, 0x00C026, 0x00C027, 0x00C028, 0x00C029, 0x00C02A, 0x00C02B, 0x00C02C, 0x00C02D, 0x00C02E, 0x00C02F, 0x00C030, 0x00C031, 0x00C032, 0x00C033, 0x00C034, 0x00C035, 0x00C036, 0x00C037, 0x00C038, 0x00C039, 0x00C03A, 0x00C03B, 0x00C03C, 0x00C03D, 0x00C03E, 0x00C03F,
	0x00C040, 0x00C041, 0x00C042, 0x00C043, 0x00C044, 0x00C045, 0x00C046, 0x00C047, 0x00C048, 0x00C049, 0x00C04A, 0x00C04B, 0x00C04C, 0x00C04D, 0x00C04E, 0x00C04F, 0x00C050, 0x00C051, 0x00C052, 0x00C053, 0x00C054, 0x00C055, 0x00C056, 0x00C057, 0x00C058, 0x00C059, 0x00C05A, 0x00C05B, 0x00C05C, 0x00C05D, 0x00C05E, 0x00C05F,
	0x00C060, 0x00C061, 0x00C062, 0x00C063, 0x00C064, 0x00C065, 0x00C066, 0x00C067, 0x00C068, 0x00C069, 0x00C06A, 0x00C06B, 0x00C06C, 0x00C06D, 0x00C06E, 0x00C06F, 0x00C070, 0x00C071, 0x00C072, 0x00C073, 0x00C074, 0x00C075, 0x00C076, 0x00C077, 0x00C078, 0x00C079, 0x00C07A, 0x00C07B, 0x00C07C, 0x00C07D, 0x00C07E, 0x00C07F,
	0x00C080, 0x00C081, 0x00C082, 0x00C083, 0x00C084, 0x00C085, 0x00C086, 0x00C087, 0x00C088, 0x00C089, 0x00C08A, 0x00C08B, 0x00C08C, 0x00C08D, 0x00C08E, 0x00C08F, 0x00C090, 0x00C091, 0x00C092, 0x00C093, 0x00C094, 0x00C095, 0x00C096, 0x00C097, 0x00C098, 0x00C099, 0x00C09A, 0x00C09B, 0x00C09C, 0x00C09D, 0x00C09E, 0x00C09F,
	0x00C0A0, 0x00C0A1, 0x00C0A2, 0x00C0A3, 0x00C0A4, 0x00C0A5, 0x00C0A6, 0x00C0A7, 0x00C0A8, 0x00C0A9, 0x00C0AA, 0x00C0AB, 0x00C0AC, 0x00C0AD, 0x00C0AE, 0x00C0AF, 0x00C0B0, 0x00C0B1, 0x00C0B2, 0x00C0B3, 0x00C0B4, 0x00C0B5, 0x00C0B6, 0x00C0B7, 0x00C0B8, 0x00C0B9, 0x00C0BA, 0x00C0BB, 0x00C0BC, 0x00C0BD, 0x00C0BE, 0x00C0BF,
	0x00C0C0, 0x00C0C1, 0x00C0C2, 0x00C0C3, 0x00C0C4, 0x00C0C5, 0x00C0C6, 0x00C0C7, 0x00C0C8, 0x00C0C9, 0x00C0CA, 0x00C0CB, 0x00C0CC, 0x00C0CD, 0x00C0CE, 0x00C0CF, 0x00C0D0, 0x00C0D1, 0x00C0D2, 0x00C0D3, 0x00C0D4, 0x00C0D5, 0x00C0D6, 0x00C0D7, 0x00C0D8, 0x00C0D9, 0x00C0DA, 0x00C0DB, 0x00C0DC, 0x00C0DD, 0x00C0DE, 0x00C0DF,
	0x00C0E0, 0x00C0E1, 0x00C0E2, 0x00C0E3, 0x00C0E4, 0x00C0E5, 0x00C0E6, 0x00C0E7, 0x00C0E8, 0x00C0E9, 0x00C0EA, 0x00C0EB, 0x00C0EC, 0x00C0ED, 0x00C0EE, 0x00C0EF, 0x00C0F0, 0x00C0F1, 0x00C0F2, 0x00C0F3, 0x00C0F4, 0x00C0F5, 0x00C0F6, 0x00C0F7, 0x00C0F8, 0x00C0F9, 0x00C0FA, 0x00C0FB, 0x00C0FC, 0x00C0FD, 0x00C0FE, 0x00C0FF,
	0x00C100, 0x00C101, 0x00C102, 0x00C103, 0x00C104, 0x00C105, 0x00C106, 0x00C107, 0x00C108, 0x00C109, 0x00C10A, 0x00C10B, 0x00C10C, 0x00C10D, 0x00C10E, 0x00C10F, 0x00C110, 0x00C111, 0x00C112, 0x00C113, 0x00C114, 0x00C115, 0x00C116, 0x00C117, 0x00C118, 0x00C119, 0x00C11A, 0x00C11B, 0x00C11C, 0x00C11D, 0x00C11E, 0x00C11F,
	0x00C120, 0x00C121, 0x00C122, 0x00C123, 0x00C124, 0x00C125, 0x00C126, 0x00C127, 0x00C128, 0x00C129, 0x00C12A, 0x00C12B, 0x00C12C, 0x00C12D, 0x00C12E, 0x00C12F, 0x00C130, 0x00C131, 0x00C132, 0x00C133, 0x00C134, 0x00C135, 0x00C136, 0x00C137, 0x00C138, 0x00C139, 0x00C13A, 0x00C13B, 0x00C13C, 0x00C13D, 0x00C13E, 0x00C13F,
	0x00C140, 0x00C141, 0x00C142, 0x00C143, 0x00C144, 0x00C145, 0x00C146, 0x00C147, 0x00C148, 0x00C149, 0x00C14A, 0x00C14B, 0x00C14C, 0x00C14D, 0x00C14E, 0x00C14F, 0x00C150, 0x00C151, 0x00C152, 0x00C153, 0x00C154, 0x00C155, 0x00C156, 0x00C157, 0x00C158, 0x00C159, 0x00C15A, 0x00C15B, 0x00C15C, 0x00C15D, 0x00C15E, 0x00C15F,
	0x00C160, 0x00C161, 0x00C162, 0x00C163, 0x00C164, 0x00C165, 0x00C166, 0x00C167, 0x00C168, 0x00C169, 0x00C16A, 0x00C16B, 0x00C16C, 0x00C16D, 0x00C16E, 0x00C16F, 0x00C170, 0x00C171, 0x00C172, 0x00C173, 0x00C174, 0x00C175, 0x00C176, 0x00C177, 0x00C178, 0x00C179, 0x00C17A, 0x00C17B, 0x00C17C, 0x00C17D, 0x00C17E, 0x00C17F,
	0x00C180, 0x00C181, 0x00C182, 0x00C183, 0x00C184, 0x00C185, 0x00C186, 0x00C187, 0x00C188, 0x00C189, 0x00C18A, 0x00C18B, 0x00C18C, 0x00C18D, 0x00C18E, 0x00C18F, 0x00C190, 0x00C191, 0x00C192, 0x00C193, 0x00C194, 0x00C195, 0x00C196, 0x00C197, 0x00C198, 0x00C199, 0x00C19A, 0x00C19B, 0x00C19C, 0x00C19D, 0x00C19E, 0x00C19F,
	0x00C1A0, 0x00C1A1, 0x00C1A2, 0x00C1A3, 0x00C1A4, 0x00C1A5, 0x00C1A6, 0x00C1A7, 0x00C1A8, 0x00C1A9, 0x00C1AA, 0x00C1AB, 0x00C1AC, 0x00C1AD, 0x00C1AE, 0x00C1AF, 0x00C1B0, 0x00C1B1, 0x00C1B2, 0x00C1B3, 0x00C1B4, 0x00C1B5, 0x00C1B6, 0x00C1B7, 0x00C1B8, 0x00C1B9, 0x00C1BA, 0x00C1BB, 0x00C1BC, 0x00C1BD, 0x00C1BE, 0x00C1BF,
	0x00C1C0, 0x00C1C1, 0x00C1C2, 0x00C1C3, 0x00C1C4, 0x00C1C5, 0x00C1C6, 0x00C1C7, 0x00C1C8, 0x00C1C9, 0x00C1CA, 0x00C1CB, 0x00C1CC, 0x00C1CD, 0x00C1CE, 0x00C1CF, 0x00C1D0, 0x00C1D1, 0x00C1D2, 0x00C1D3, 0x00C1D4, 0x00C1D5, 0x00C1D6, 0x00C1D7, 0x00C1D8, 0x00C1D9, 0x00C1DA, 0x00C1DB, 0x00C1DC, 0x00C1DD, 0x00C1DE, 0x00C1DF,
	0x00C1E0, 0x00C1E1, 0x00C1E2, 0x00C1E3, 0x00C1E4, 0x00C1E5, 0x00C1E6, 0x00C1E7, 0x00C1E8, 0x00C1E9, 0x00C1EA, 0x00C1EB, 0x00C1EC, 0x00C1ED, 0x00C1EE, 0x00C1EF, 0x00C1F0, 0x00C1F1, 0x00C1F2, 0x00C1F3, 0x00C1F4, 0x00C1F5, 0x00C1F6, 0x00C1F7, 0x00C1F8, 0x00C1F9, 0x00C1FA, 0x00C1FB, 0x00C1FC, 0x00C1FD, 0x00C1FE, 0x00C1FF,
	0x00C200, 0x00C201, 0x00C202, 0x00C203, 0x00C204, 0x00C205, 0x00C206, 0x00C207, 0x00C208, 0x00C209, 0x00C20A, 0x00C20B, 0x00C20C, 0x00C20D, 0x00C20E, 0x00C20F, 0x00C210, 0x00C211, 0x00C212, 0x00C213, 0x00C214, 0x00C215, 0x00C216, 0x00C217, 0x00C218, 0x00C219, 0x00C21A, 0x00C21B, 0x00C21C, 0x00C21D, 0x00C21E, 0x00C21F,
	0x00C220, 0x00C221, 0x00C222, 0x00C223, 0x00C224, 0x00C225, 0x00C226, 0x00C227, 0x00C228, 0x00C229, 0x00C22A, 0x00C22B, 0x00C22C, 0x00C22D, 0x00C22E, 0x00C22F, 0x00C230, 0x00C231, 0x00C232, 0x00C233, 0x00C234, 0x00C235, 0x00C236, 0x00C237, 0x00C238, 0x00C239, 0x00C23A, 0x00C23B, 0x00C23C, 0x00C23D, 0x00C23E, 0x00C23F,
	0x00C240, 0x00C241, 0x00C242, 0x00C243, 0x00C244, 0x00C245, 0x00C246, 0x00C247, 0x00C248, 0x00C249, 0x00C24A, 0x00C24B, 0x00C24C, 0x00C24D, 0x00C24E, 0x00C24F, 0x00C250, 0x00C251, 0x00C252, 0x00C253, 0x00C254, 0x00C255, 0x00C256, 0x00C257, 0x00C258, 0x00C259, 0x00C25A, 0x00C25B, 0x00C25C, 0x00C25D, 0x00C25E, 0x00C25F,
	0x00C260, 0x00C261, 0x00C262, 0x00C263, 0x00C264, 0x00C265, 0x00C266, 0x00C267, 0x00C268, 0x00C269, 0x00C26A, 0x00C26B, 0x00C26C, 0x00C26D, 0x00C26E, 0x00C26F, 0x00C270, 0x00C271, 0x00C272, 0x00C273, 0x00C274, 0x00C275, 0x00C276, 0x00C277, 0x00C278, 0x00C279, 0x00C27A, 0x00C27B, 0x00C27C, 0x00C27D, 0x00C27E, 0x00C27F,
	0x00C280, 0x00C281, 0x00C282, 0x00C283, 0x00C284, 0x00C285, 0x00C286, 0x00C287, 0x00C288, 0x00C289, 0x00C28A, 0x00C28B, 0x00C28C, 0x00C28D, 0x00C28E, 0x00C28F, 0x00C290, 0x00C291, 0x00C292, 0x00C293, 0x00C294, 0x00C295, 0x00C296, 0x00C297, 0x00C298, 0x00C299, 0x00C29A, 0x00C29B, 0x00C29C, 0x00C29D, 0x00C29E, 0x00C29F,
	0x00C2A0, 0x00C2A1, 0x00C2A2, 0x00C2A3, 0x00C2A4, 0x00C2A5, 0x00C2A6, 0x00C2A7, 0x00C2A8, 0x00C2A9, 0x00C2AA, 0x00C2AB, 0x00C2AC, 0x00C2AD, 0x00C2AE, 0x00C2AF, 0x00C2B0, 0x00C2B1, 0x00C2B2, 0x00C2B3, 0x00C2B4, 0x00C2B5, 0x00C2B6, 0x00C2B7, 0x00C2B8, 0x00C2B9, 0x00C2BA, 0x00C2BB, 0x00C2BC, 0x00C2BD, 0x00C2BE, 0x00C2BF,
	0x00C2C0, 0x00C2C1, 0x00C2C2, 0x00C2C3, 0x00C2C4, 0x00C2C5, 0x00C2C6, 0x00C2C7, 0x00C2C8, 0x00C2C9, 0x00C2CA, 0x00C2CB, 0x00C2CC, 0x00C2CD, 0x00C2CE, 0x00C2CF, 0x00C2D0, 0x00C2D1, 0x00C2D2, 0x00C2D3, 0x00C2D4, 0x00C2D5, 0x00C2D6, 0x00C2D7, 0x00C2D8, 0x00C2D9, 0x00C2DA, 0x00C2DB, 0x00C2DC, 0x00C2DD, 0x00C2DE, 0x00C2DF,
	0x00C2E0, 0x00C2E1, 0x00C2E2, 0x00C2E3, 0x00C2E4, 0x00C2E5, 0x00C2E6, 0x00C2E7, 0x00C2E8, 0x00C2E9, 0x00C2EA, 0x00C2EB, 0x00C2EC, 0x00C2ED, 0x00C2EE, 0x00C2EF, 0x00C2F0, 0x00C2F1, 0x00C2F2, 0x00C2F3, 0x00C2F4, 0x00C2F5, 0x00C2F6, 0x00C2F7, 0x00C2F8, 0x00C2F9, 0x00C2FA, 0x00C2FB, 0x00C2FC, 0x00C2FD, 0x00C2FE, 0x00C2FF,
	0x00C300, 0x00C301, 0x00C302, 0x00C303, 0x00C304, 0x00C305, 0x00C306, 0x00C307, 0x00C308, 0x00C309, 0x00C30A, 0x00C30B, 0x00C30C, 0x00C30D, 0x00C30E, 0x00C30F, 0x00C310, 0x00C311, 0x00C312, 0x00C313, 0x00C314, 0x00C315, 0x00C316, 0x00C317, 0x00C318, 0x00C319, 0x00C31A, 0x00C31B, 0x00C31C, 0x00C31D, 0x00C31E, 0x00C31F,
	0x00C320, 0x00C321, 0x00C322, 0x00C323, 0x00C324, 0x00C325, 0x00C326, 0x00C327, 0x00C328, 0x00C329, 0x00C32A, 0x00C32B, 0x00C32C, 0x00C32D, 0x00C32E, 0x00C32F, 0x00C330, 0x00C331, 0x00C332, 0x00C333, 0x00C334, 0x00C335, 0x00C336, 0x00C337, 0x00C338, 0x00C339, 0x00C33A, 0x00C33B, 0x00C33C, 0x00C33D, 0x00C33E, 0x00C33F,
	0x00C340, 0x00C341, 0x00C342, 0x00C343, 0x00C344, 0x00C345, 0x00C346, 0x00C347, 0x00C348, 0x00C349, 0x00C34A, 0x00C34B, 0x00C34C, 0x00C34D, 0x00C34E, 0x00C34F, 0x00C350, 0x00C351, 0x00C352, 0x00C353, 0x00C354, 0x00C355, 0x00C356, 0x00C357, 0x00C358, 0x00C359, 0x00C35A, 0x00C35B, 0x00C35C, 0x00C35D, 0x00C35E, 0x00C35F,
	0x00C360, 0x00C361, 0x00C362, 0x00C363, 0x00C364, 0x00C365, 0x00C366, 0x00C367, 0x00C368, 0x00C369, 0x00C36A, 0x00C36B, 0x00C36C, 0x00C36D, 0x00C36E, 0x00C36F, 0x00C370, 0x00C371, 0x00C372, 0x00C373, 0x00C374, 0x00C375, 0x00C376, 0x00C377, 0x00C378, 0x00C379, 0x00C37A, 0x00C37B, 0x00C37C, 0x00C37D, 0x00C37E, 0x00C37F,
	0x00C380, 0x00C381, 0x00C382, 0x00C383, 0x00C384, 0x00C385, 0x00C386, 0x00C387, 0x00C388, 0x00C389, 0x00C38A, 0x00C38B, 0x00C38C, 0x00C38D, 0x00C38E, 0x00C38F, 0x00C390, 0x00C391, 0x00C392, 0x00C393, 0x00C394, 0x00C395, 0x00C396, 0x00C397, 0x00C398, 0x00C399, 0x00C39A, 0x00C39B, 0x00C39C, 0x00C39D, 0x00C39E, 0x00C39F,
	0x00C3A0, 0x00C3A1, 0x00C3A2, 0x00C3A3, 0x00C3A4, 0x00C3A5, 0x00C3A6, 0x00C3A7, 0x00C3A8, 0x00C3A9, 0x00C3AA, 0x00C3AB, 0x00C3AC, 0x00C3AD, 0x00C3AE, 0x00C3AF, 0x00C3B0, 0x00C3B1, 0x00C3B2, 0x00C3B3, 0x00C3B4, 0x00C3B5, 0x00C3B6, 0x00C3B7, 0x00C3B8, 0x00C3B9, 0x00C3BA, 0x00C3BB, 0x00C3BC, 0x00C3BD, 0x00C3BE, 0x00C3BF,
	0x00C3C0, 0x00C3C1, 0x00C3C2, 0x00C3C3, 0x00C3C4, 0x00C3C5, 0x00C3C6, 0x00C3C7, 0x00C3C8, 0x00C3C9, 0x00C3CA, 0x00C3CB, 0x00C3CC, 0x00C3CD, 0x00C3CE, 0x00C3CF, 0x00C3D0, 0x00C3D1, 0x00C3D2, 0x00C3D3, 0x00C3D4, 0x00C3D5, 0x00C3D6, 0x00C3D7, 0x00C3D8, 0x00C3D9, 0x00C3DA, 0x00C3DB, 0x00C3DC, 0x00C3DD, 0x00C3DE, 0x00C3DF,
	0x00C3E0, 0x00C3E1, 0x00C3E2, 0x00C3E3, 0x00C3E4, 0x00C3E5, 0x00C3E6, 0x00C3E7, 0x00C3E8, 0x00C3E9, 0x00C3EA, 0x00C3EB, 0x00C3EC, 0x00C3ED, 0x00C3EE, 0x00C3EF, 0x00C3F0, 0x00C3F1, 0x00C3F2, 0x00C3F3, 0x00C3F4, 0x00C3F5, 0x00C3F6, 0x00C3F7, 0x00C3F8, 0x00C3F9, 0x00C3FA, 0x00C3FB, 0x00C3FC, 0x00C3FD, 0x00C3FE, 0x00C3FF,
	0
};

static const unsigned map_hangul_7[] = {
	0x00C400, 0x00C401, 0x00C402, 0x00C403, 0x00C404, 0x00C405, 0x00C406, 0x00C407, 0x00C408, 0x00C409, 0x00C40A, 0x00C40B, 0x00C40C, 0x00C40D, 0x00C40E, 0x00C40F, 0x00C410, 0x00C411, 0x00C412, 0x00C413, 0x00C414, 0x00C415, 0x00C416, 0x00C417, 0x00C418, 0x00C419, 0x00C41A, 0x00C41B, 0x00C41C, 0x00C41D, 0x00C41E, 0x00C41F,
	0x00C420, 0x00C421, 0x00C422, 0x00C423, 0x00C424, 0x00C425, 0x00C426, 0x00C427, 0x00C428, 0x00C429, 0x00C42A, 0x00C42B, 0x00C42C, 0x00C42D, 0x00C42E, 0x00C42F, 0x00C430, 0x00C431, 0x00C432, 0x00C433, 0x00C434, 0x00C435, 0x00C436, 0x00C437, 0x00C438, 0x00C439, 0x00C43A, 0x00C43B, 0x00C43C, 0x00C43D, 0x00C43E, 0x00C43F,
	0x00C440, 0x00C441, 0x00C442, 0x00C443, 0x00C444, 0x00C445, 0x00C446, 0x00C447, 0x00C448, 0x00C449, 0x00C44A, 0x00C44B, 0x00C44C, 0x00C44D, 0x00C44E, 0x00C44F, 0x00C450, 0x00C451, 0x00C452, 0x00C453, 0x00C454, 0x00C455, 0x00C456, 0x00C457, 0x00C458, 0x00C459, 0x00C45A, 0x00C45B, 0x00C45C, 0x00C45D, 0x00C45E, 0x00C45F,
	0x00C460, 0x00C461, 0x00C462, 0x00C463, 0x00C464, 0x00C465, 0x00C466, 0x00C467, 0x00C468, 0x00C469, 0x00C46A, 0x00C46B, 0x00C46C, 0x00C46D, 0x00C46E, 0x00C46F, 0x00C470, 0x00C471, 0x00C472, 0x00C473, 0x00C474, 0x00C475, 0x00C476, 0x00C477, 0x00C478, 0x00C479, 0x00C47A, 0x00C47B, 0x00C47C, 0x00C47D, 0x00C47E, 0x00C47F,
	0x00C480, 0x00C481, 0x00C482, 0x00C483, 0x00C484, 0x00C485, 0x00C486, 0x00C487, 0x00C488, 0x00C489, 0x00C48A, 0x00C48B, 0x00C48C, 0x00C48D, 0x00C48E, 0x00C48F, 0x00C490, 0x00C491, 0x00C492, 0x00C493, 0x00C494, 0x00C495, 0x00C496, 0x00C497, 0x00C498, 0x00C499, 0x00C49A, 0x00C49B, 0x00C49C, 0x00C49D, 0x00C49E, 0x00C49F,
	0x00C4A0, 0x00C4A1, 0x00C4A2, 0x00C4A3, 0x00C4A4, 0x00C4A5, 0x00C4A6, 0x00C4A7, 0x00C4A8, 0x00C4A9, 0x00C4AA, 0x00C4AB, 0x00C4AC, 0x00C4AD, 0x00C4AE, 0x00C4AF, 0x00C4B0, 0x00C4B1, 0x00C4B2, 0x00C4B3, 0x00C4B4, 0x00C4B5, 0x00C4B6, 0x00C4B7, 0x00C4B8, 0x00C4B9, 0x00C4BA, 0x00C4BB, 0x00C4BC, 0x00C4BD, 0x00C4BE, 0x00C4BF,
	0x00C4C0, 0x00C4C1, 0x00C4C2, 0x00C4C3, 0x00C4C4, 0x00C4C5, 0x00C4C6, 0x00C4C7, 0x00C4C8, 0x00C4C9, 0x00C4CA, 0x00C4CB, 0x00C4CC, 0x00C4CD, 0x00C4CE, 0x00C4CF, 0x00C4D0, 0x00C4D1, 0x00C4D2, 0x00C4D3, 0x00C4D4, 0x00C4D5, 0x00C4D6, 0x00C4D7, 0x00C4D8, 0x00C4D9, 0x00C4DA, 0x00C4DB, 0x00C4DC, 0x00C4DD, 0x00C4DE, 0x00C4DF,
	0x00C4E0, 0x00C4E1, 0x00C4E2, 0x00C4E3, 0x00C4E4, 0x00C4E5, 0x00C4E6, 0x00C4E7, 0x00C4E8, 0x00C4E9, 0x00C4EA, 0x00C4EB, 0x00C4EC, 0x00C4ED, 0x00C4EE, 0x00C4EF, 0x00C4F0, 0x00C4F1, 0x00C4F2, 0x00C4F3, 0x00C4F4, 0x00C4F5, 0x00C4F6, 0x00C4F7, 0x00C4F8, 0x00C4F9, 0x00C4FA, 0x00C4FB, 0x00C4FC, 0x00C4FD, 0x00C4FE, 0x00C4FF,
	0x00C500, 0x00C501, 0x00C502, 0x00C503, 0x00C504, 0x00C505, 0x00C506, 0x00C507, 0x00C508, 0x00C509, 0x00C50A, 0x00C50B, 0x00C50C, 0x00C50D, 0x00C50E, 0x00C50F, 0x00C510, 0x00C511, 0x00C512, 0x00C513, 0x00C514, 0x00C515, 0x00C516, 0x00C517, 0x00C518, 0x00C519, 0x00C51A, 0x00C51B, 0x00C51C, 0x00C51D, 0x00C51E, 0x00C51F,
	0x00C520, 0x00C521, 0x00C522, 0x00C523, 0x00C524, 0x00C525, 0x00C526, 0x00C527, 0x00C528, 0x00C529, 0x00C52A, 0x00C52B, 0x00C52C, 0x00C52D, 0x00C52E, 0x00C52F, 0x00C530, 0x00C531, 0x00C532, 0x00C533, 0x00C534, 0x00C535, 0x00C536, 0x00C537, 0x00C538, 0x00C539, 0x00C53A, 0x00C53B, 0x00C53C, 0x00C53D, 0x00C53E, 0x00C53F,
	0x00C540, 0x00C541, 0x00C542, 0x00C543, 0x00C544, 0x00C545, 0x00C546, 0x00C547, 0x00C548, 0x00C549, 0x00C54A, 0x00C54B, 0x00C54C, 0x00C54D, 0x00C54E, 0x00C54F, 0x00C550, 0x00C551, 0x00C552, 0x00C553, 0x00C554, 0x00C555, 0x00C556, 0x00C557, 0x00C558, 0x00C559, 0x00C55A, 0x00C55B, 0x00C55C, 0x00C55D, 0x00C55E, 0x00C55F,
	0x00C560, 0x00C561, 0x00C562, 0x00C563, 0x00C564, 0x00C565, 0x00C566, 0x00C567, 0x00C568, 0x00C569, 0x00C56A, 0x00C56B, 0x00C56C, 0x00C56D, 0x00C56E, 0x00C56F, 0x00C570, 0x00C571, 0x00C572, 0x00C573, 0x00C574, 0x00C575, 0x00C576, 0x00C577, 0x00C578, 0x00C579, 0x00C57A, 0x00C57B, 0x00C57C, 0x00C57D, 0x00C57E, 0x00C57F,
	0x00C580, 0x00C581, 0x00C582, 0x00C583, 0x00C584, 0x00C585, 0x00C586, 0x00C587, 0x00C588, 0x00C589, 0x00C58A, 0x00C58B, 0x00C58C, 0x00C58D, 0x00C58E, 0x00C58F, 0x00C590, 0x00C591, 0x00C592, 0x00C593, 0x00C594, 0x00C595, 0x00C596, 0x00C597, 0x00C598, 0x00C599, 0x00C59A, 0x00C59B, 0x00C59C, 0x00C59D, 0x00C59E, 0x00C59F,
	0x00C5A0, 0x00C5A1, 0x00C5A2, 0x00C5A3, 0x00C5A4, 0x00C5A5, 0x00C5A6, 0x00C5A7, 0x00C5A8, 0x00C5A9, 0x00C5AA, 0x00C5AB, 0x00C5AC, 0x00C5AD, 0x00C5AE, 0x00C5AF, 0x00C5B0, 0x00C5B1, 0x00C5B2, 0x00C5B3, 0x00C5B4, 0x00C5B5, 0x00C5B6, 0x00C5B7, 0x00C5B8, 0x00C5B9, 0x00C5BA, 0x00C5BB, 0x00C5BC, 0x00C5BD, 0x00C5BE, 0x00C5BF,
	0x00C5C0, 0x00C5C1, 0x00C5C2, 0x00C5C3, 0x00C5C4, 0x00C5C5, 0x00C5C6, 0x00C5C7, 0x00C5C8, 0x00C5C9, 0x00C5CA, 0x00C5CB, 0x00C5CC, 0x00C5CD, 0x00C5CE, 0x00C5CF, 0x00C5D0, 0x00C5D1, 0x00C5D2, 0x00C5D3, 0x00C5D4, 0x00C5D5, 0x00C5D6, 0x00C5D7, 0x00C5D8, 0x00C5D9, 0x00C5DA, 0x00C5DB, 0x00C5DC, 0x00C5DD, 0x00C5DE, 0x00C5DF,
	0x00C5E0, 0x00C5E1, 0x00C5E2, 0x00C5E3, 0x00C5E4, 0x00C5E5, 0x00C5E6, 0x00C5E7, 0x00C5E8, 0x00C5E9, 0x00C5EA, 0x00C5EB, 0x00C5EC, 0x00C5ED, 0x00C5EE, 0x00C5EF, 0x00C5F0, 0x00C5F1, 0x00C5F2, 0x00C5F3, 0x00C5F4, 0x00C5F5, 0x00C5F6, 0x00C5F7, 0x00C5F8, 0x00C5F9, 0x00C5FA, 0x00C5FB, 0x00C5FC, 0x00C5FD, 0x00C5FE, 0x00C5FF,
	0x00C600, 0x00C601, 0x00C602, 0x00C603, 0x00C604, 0x00C605, 0x00C606, 0x00C607, 0x00C608, 0x00C609, 0x00C60A, 0x00C60B, 0x00C60C, 0x00C60D, 0x00C60E, 0x00C60F, 0x00C610, 0x00C611, 0x00C612, 0x00C613, 0x00C614, 0x00C615, 0x00C616, 0x00C617, 0x00C618, 0x00C619, 0x00C61A, 0x00C61B, 0x00C61C, 0x00C61D, 0x00C61E, 0x00C61F,
	0x00C620, 0x00C621, 0x00C622, 0x00C623, 0x00C624, 0x00C625, 0x00C626, 0x00C627, 0x00C628, 0x00C629, 0x00C62A, 0x00C62B, 0x00C62C, 0x00C62D, 0x00C62E, 0x00C62F, 0x00C630, 0x00C631, 0x00C632, 0x00C633, 0x00C634, 0x00C635, 0x00C636, 0x00C637, 0x00C638, 0x00C639, 0x00C63A, 0x00C63B, 0x00C63C, 0x00C63D, 0x00C63E, 0x00C63F,
	0x00C640, 0x00C641, 0x00C642, 0x00C643, 0x00C644, 0x00C645, 0x00C646, 0x00C647, 0x00C648, 0x00C649, 0x00C64A, 0x00C64B, 0x00C64C, 0x00C64D, 0x00C64E, 0x00C64F, 0x00C650, 0x00C651, 0x00C652, 0x00C653, 0x00C654, 0x00C655, 0x00C656, 0x00C657, 0x00C658, 0x00C659, 0x00C65A, 0x00C65B, 0x00C65C, 0x00C65D, 0x00C65E, 0x00C65F,
	0x00C660, 0x00C661, 0x00C662, 0x00C663, 0x00C664, 0x00C665, 0x00C666, 0x00C667, 0x00C668, 0x00C669, 0x00C66A, 0x00C66B, 0x00C66C, 0x00C66D, 0x00C66E, 0x00C66F, 0x00C670, 0x00C671, 0x00C672, 0x00C673, 0x00C674, 0x00C675, 0x00C676, 0x00C677, 0x00C678, 0x00C679, 0x00C67A, 0x00C67B, 0x00C67C, 0x00C67D, 0x00C67E, 0x00C67F,
	0x00C680, 0x00C681, 0x00C682, 0x00C683, 0x00C684, 0x00C685, 0x00C686, 0x00C687, 0x00C688, 0x00C689, 0x00C68A, 0x00C68B, 0x00C68C, 0x00C68D, 0x00C68E, 0x00C68F, 0x00C690, 0x00C691, 0x00C692, 0x00C693, 0x00C694, 0x00C695, 0x00C696, 0x00C697, 0x00C698, 0x00C699, 0x00C69A, 0x00C69B, 0x00C69C, 0x00C69D, 0x00C69E, 0x00C69F,
	0x00C6A0, 0x00C6A1, 0x00C6A2, 0x00C6A3, 0x00C6A4, 0x00C6A5, 0x00C6A6, 0x00C6A7, 0x00C6A8, 0x00C6A9, 0x00C6AA, 0x00C6AB, 0x00C6AC, 0x00C6AD, 0x00C6AE, 0x00C6AF, 0x00C6B0, 0x00C6B1, 0x00C6B2, 0x00C6B3, 0x00C6B4, 0x00C6B5, 0x00C6B6, 0x00C6B7, 0x00C6B8, 0x00C6B9, 0x00C6BA, 0x00C6BB, 0x00C6BC, 0x00C6BD, 0x00C6BE, 0x00C6BF,
	0x00C6C0, 0x00C6C1, 0x00C6C2, 0x00C6C3, 0x00C6C4, 0x00C6C5, 0x00C6C6, 0x00C6C7, 0x00C6C8, 0x00C6C9, 0x00C6CA, 0x00C6CB, 0x00C6CC, 0x00C6CD, 0x00C6CE, 0x00C6CF, 0x00C6D0, 0x00C6D1, 0x00C6D2, 0x00C6D3, 0x00C6D4, 0x00C6D5, 0x00C6D6, 0x00C6D7, 0x00C6D8, 0x00C6D9, 0x00C6DA, 0x00C6DB, 0x00C6DC, 0x00C6DD, 0x00C6DE, 0x00C6DF,
	0x00C6E0, 0x00C6E1, 0x00C6E2, 0x00C6E3, 0x00C6E4, 0x00C6E5, 0x00C6E6, 0x00C6E7, 0x00C6E8, 0x00C6E9, 0x00C6EA, 0x00C6EB, 0x00C6EC, 0x00C6ED, 0x00C6EE, 0x00C6EF, 0x00C6F0, 0x00C6F1, 0x00C6F2, 0x00C6F3, 0x00C6F4, 0x00C6F5, 0x00C6F6, 0x00C6F7, 0x00C6F8, 0x00C6F9, 0x00C6FA, 0x00C6FB, 0x00C6FC, 0x00C6FD, 0x00C6FE, 0x00C6FF,
	0x00C700, 0x00C701, 0x00C702, 0x00C703, 0x00C704, 0x00C705, 0x00C706, 0x00C707, 0x00C708, 0x00C709, 0x00C70A, 0x00C70B, 0x00C70C, 0x00C70D, 0x00C70E, 0x00C70F, 0x00C710, 0x00C711, 0x00C712, 0x00C713, 0x00C714, 0x00C715, 0x00C716, 0x00C717, 0x00C718, 0x00C719, 0x00C71A, 0x00C71B, 0x00C71C, 0x00C71D, 0x00C71E, 0x00C71F,
	0x00C720, 0x00C721, 0x00C722, 0x00C723, 0x00C724, 0x00C725, 0x00C726, 0x00C727, 0x00C728, 0x00C729, 0x00C72A, 0x00C72B, 0x00C72C, 0x00C72D, 0x00C72E, 0x00C72F, 0x00C730, 0x00C731, 0x00C732, 0x00C733, 0x00C734, 0x00C735, 0x00C736, 0x00C737, 0x00C738, 0x00C739, 0x00C73A, 0x00C73B, 0x00C73C, 0x00C73D, 0x00C73E, 0x00C73F,
	0x00C740, 0x00C741, 0x00C742, 0x00C743, 0x00C744, 0x00C745, 0x00C746, 0x00C747, 0x00C748, 0x00C749, 0x00C74A, 0x00C74B, 0x00C74C, 0x00C74D, 0x00C74E, 0x00C74F, 0x00C750, 0x00C751, 0x00C752, 0x00C753, 0x00C754, 0x00C755, 0x00C756, 0x00C757, 0x00C758, 0x00C759, 0x00C75A, 0x00C75B, 0x00C75C, 0x00C75D, 0x00C75E, 0x00C75F,
	0x00C760, 0x00C761, 0x00C762, 0x00C763, 0x00C764, 0x00C765, 0x00C766, 0x00C767, 0x00C768, 0x00C769, 0x00C76A, 0x00C76B, 0x00C76C, 0x00C76D, 0x00C76E, 0x00C76F, 0x00C770, 0x00C771, 0x00C772, 0x00C773, 0x00C774, 0x00C775, 0x00C776, 0x00C777, 0x00C778, 0x00C779, 0x00C77A, 0x00C77B, 0x00C77C, 0x00C77D, 0x00C77E, 0x00C77F,
	0x00C780, 0x00C781, 0x00C782, 0x00C783, 0x00C784, 0x00C785, 0x00C786, 0x00C787, 0x00C788, 0x00C789, 0x00C78A, 0x00C78B, 0x00C78C, 0x00C78D, 0x00C78E, 0x00C78F, 0x00C790, 0x00C791, 0x00C792, 0x00C793, 0x00C794, 0x00C795, 0x00C796, 0x00C797, 0x00C798, 0x00C799, 0x00C79A, 0x00C79B, 0x00C79C, 0x00C79D, 0x00C79E, 0x00C79F,
	0x00C7A0, 0x00C7A1, 0x00C7A2, 0x00C7A3, 0x00C7A4, 0x00C7A5, 0x00C7A6, 0x00C7A7, 0x00C7A8, 0x00C7A9, 0x00C7AA, 0x00C7AB, 0x00C7AC, 0x00C7AD, 0x00C7AE, 0x00C7AF, 0x00C7B0, 0x00C7B1, 0x00C7B2, 0x00C7B3, 0x00C7B4, 0x00C7B5, 0x00C7B6, 0x00C7B7, 0x00C7B8, 0x00C7B9, 0x00C7BA, 0x00C7BB, 0x00C7BC, 0x00C7BD, 0x00C7BE, 0x00C7BF,
	0x00C7C0, 0x00C7C1, 0x00C7C2, 0x00C7C3, 0x00C7C4, 0x00C7C5, 0x00C7C6, 0x00C7C7, 0x00C7C8, 0x00C7C9, 0x00C7CA, 0x00C7CB, 0x00C7CC, 0x00C7CD, 0x00C7CE, 0x00C7CF, 0x00C7D0, 0x00C7D1, 0x00C7D2, 0x00C7D3, 0x00C7D4, 0x00C7D5, 0x00C7D6, 0x00C7D7, 0x00C7D8, 0x00C7D9, 0x00C7DA, 0x00C7DB, 0x00C7DC, 0x00C7DD, 0x00C7DE, 0x00C7DF,
	0x00C7E0, 0x00C7E1, 0x00C7E2, 0x00C7E3, 0x00C7E4, 0x00C7E5, 0x00C7E6, 0x00C7E7, 0x00C7E8, 0x00C7E9, 0x00C7EA, 0x00C7EB, 0x00C7EC, 0x00C7ED, 0x00C7EE, 0x00C7EF, 0x00C7F0, 0x00C7F1, 0x00C7F2, 0x00C7F3, 0x00C7F4, 0x00C7F5, 0x00C7F6, 0x00C7F7, 0x00C7F8, 0x00C7F9, 0x00C7FA, 0x00C7FB, 0x00C7FC, 0x00C7FD, 0x00C7FE, 0x00C7FF,
	0
};

static const unsigned map_hangul_8[] = {
	0x00C800, 0x00C801, 0x00C802, 0x00C803, 0x00C804, 0x00C805, 0x00C806, 0x00C807, 0x00C808, 0x00C809, 0x00C80A, 0x00C80B, 0x00C80C, 0x00C80D, 0x00C80E, 0x00C80F, 0x00C810, 0x00C811, 0x00C812, 0x00C813, 0x00C814, 0x00C815, 0x00C816, 0x00C817, 0x00C818, 0x00C819, 0x00C81A, 0x00C81B, 0x00C81C, 0x00C81D, 0x00C81E, 0x00C81F,
	0x00C820, 0x00C821, 0x00C822, 0x00C823, 0x00C824, 0x00C825, 0x00C826, 0x00C827, 0x00C828, 0x00C829, 0x00C82A, 0x00C82B, 0x00C82C, 0x00C82D, 0x00C82E, 0x00C82F, 0x00C830, 0x00C831, 0x00C832, 0x00C833, 0x00C834, 0x00C835, 0x00C836, 0x00C837, 0x00C838, 0x00C839, 0x00C83A, 0x00C83B, 0x00C83C, 0x00C83D, 0x00C83E, 0x00C83F,
	0x00C840, 0x00C841, 0x00C842, 0x00C843, 0x00C844, 0x00C845, 0x00C846, 0x00C847, 0x00C848, 0x00C849, 0x00C84A, 0x00C84B, 0x00C84C, 0x00C84D, 0x00C84E, 0x00C84F, 0x00C850, 0x00C851, 0x00C852, 0x00C853, 0x00C854, 0x00C855, 0x00C856, 0x00C857, 0x00C858, 0x00C859, 0x00C85A, 0x00C85B, 0x00C85C, 0x00C85D, 0x00C85E, 0x00C85F,
	0x00C860, 0x00C861, 0x00C862, 0x00C863, 0x00C864, 0x00C865, 0x00C866, 0x00C867, 0x00C868, 0x00C869, 0x00C86A, 0x00C86B, 0x00C86C, 0x00C86D, 0x00C86E, 0x00C86F, 0x00C870, 0x00C871, 0x00C872, 0x00C873, 0x00C874, 0x00C875, 0x00C876, 0x00C877, 0x00C878, 0x00C879, 0x00C87A, 0x00C87B, 0x00C87C, 0x00C87D, 0x00C87E, 0x00C87F,
	0x00C880, 0x00C881, 0x00C882, 0x00C883, 0x00C884, 0x00C885, 0x00C886, 0x00C887, 0x00C888, 0x00C889, 0x00C88A, 0x00C88B, 0x00C88C, 0x00C88D, 0x00C88E, 0x00C88F, 0x00C890, 0x00C891, 0x00C892, 0x00C893, 0x00C894, 0x00C895, 0x00C896, 0x00C897, 0x00C898, 0x00C899, 0x00C89A, 0x00C89B, 0x00C89C, 0x00C89D, 0x00C89E, 0x00C89F,
	0x00C8A0, 0x00C8A1, 0x00C8A2, 0x00C8A3, 0x00C8A4, 0x00C8A5, 0x00C8A6, 0x00C8A7, 0x00C8A8, 0x00C8A9, 0x00C8AA, 0x00C8AB, 0x00C8AC, 0x00C8AD, 0x00C8AE, 0x00C8AF, 0x00C8B0, 0x00C8B1, 0x00C8B2, 0x00C8B3, 0x00C8B4, 0x00C8B5, 0x00C8B6, 0x00C8B7, 0x00C8B8, 0x00C8B9, 0x00C8BA, 0x00C8BB, 0x00C8BC, 0x00C8BD, 0x00C8BE, 0x00C8BF,
	0x00C8C0, 0x00C8C1, 0x00C8C2, 0x00C8C3, 0x00C8C4, 0x00C8C5, 0x00C8C6, 0x00C8C7, 0x00C8C8, 0x00C8C9, 0x00C8CA, 0x00C8CB, 0x00C8CC, 0x00C8CD, 0x00C8CE, 0x00C8CF, 0x00C8D0, 0x00C8D1, 0x00C8D2, 0x00C8D3, 0x00C8D4, 0x00C8D5, 0x00C8D6, 0x00C8D7, 0x00C8D8, 0x00C8D9, 0x00C8DA, 0x00C8DB, 0x00C8DC, 0x00C8DD, 0x00C8DE, 0x00C8DF,
	0x00C8E0, 0x00C8E1, 0x00C8E2, 0x00C8E3, 0x00C8E4, 0x00C8E5, 0x00C8E6, 0x00C8E7, 0x00C8E8, 0x00C8E9, 0x00C8EA, 0x00C8EB, 0x00C8EC, 0x00C8ED, 0x00C8EE, 0x00C8EF, 0x00C8F0, 0x00C8F1, 0x00C8F2, 0x00C8F3, 0x00C8F4, 0x00C8F5, 0x00C8F6, 0x00C8F7, 0x00C8F8, 0x00C8F9, 0x00C8FA, 0x00C8FB, 0x00C8FC, 0x00C8FD, 0x00C8FE, 0x00C8FF,
	0x00C900, 0x00C901, 0x00C902, 0x00C903, 0x00C904, 0x00C905, 0x00C906, 0x00C907, 0x00C908, 0x00C909, 0x00C90A, 0x00C90B, 0x00C90C, 0x00C90D, 0x00C90E, 0x00C90F, 0x00C910, 0x00C911, 0x00C912, 0x00C913, 0x00C914, 0x00C915, 0x00C916, 0x00C917, 0x00C918, 0x00C919, 0x00C91A, 0x00C91B, 0x00C91C, 0x00C91D, 0x00C91E, 0x00C91F,
	0x00C920, 0x00C921, 0x00C922, 0x00C923, 0x00C924, 0x00C925, 0x00C926, 0x00C927, 0x00C928, 0x00C929, 0x00C92A, 0x00C92B, 0x00C92C, 0x00C92D, 0x00C92E, 0x00C92F, 0x00C930, 0x00C931, 0x00C932, 0x00C933, 0x00C934, 0x00C935, 0x00C936, 0x00C937, 0x00C938, 0x00C939, 0x00C93A, 0x00C93B, 0x00C93C, 0x00C93D, 0x00C93E, 0x00C93F,
	0x00C940, 0x00C941, 0x00C942, 0x00C943, 0x00C944, 0x00C945, 0x00C946, 0x00C947, 0x00C948, 0x00C949, 0x00C94A, 0x00C94B, 0x00C94C, 0x00C94D, 0x00C94E, 0x00C94F, 0x00C950, 0x00C951, 0x00C952, 0x00C953, 0x00C954, 0x00C955, 0x00C956, 0x00C957, 0x00C958, 0x00C959, 0x00C95A, 0x00C95B, 0x00C95C, 0x00C95D, 0x00C95E, 0x00C95F,
	0x00C960, 0x00C961, 0x00C962, 0x00C963, 0x00C964, 0x00C965, 0x00C966, 0x00C967, 0x00C968, 0x00C969, 0x00C96A, 0x00C96B, 0x00C96C, 0x00C96D, 0x00C96E, 0x00C96F, 0x00C970, 0x00C971, 0x00C972, 0x00C973, 0x00C974, 0x00C975, 0x00C976, 0x00C977, 0x00C978, 0x00C979, 0x00C97A, 0x00C97B, 0x00C97C, 0x00C97D, 0x00C97E, 0x00C97F,
	0x00C980, 0x00C981, 0x00C982, 0x00C983, 0x00C984, 0x00C985, 0x00C986, 0x00C987, 0x00C988, 0x00C989, 0x00C98A, 0x00C98B, 0x00C98C, 0x00C98D, 0x00C98E, 0x00C98F, 0x00C990, 0x00C991, 0x00C992, 0x00C993, 0x00C994, 0x00C995, 0x00C996, 0x00C997, 0x00C998, 0x00C999, 0x00C99A, 0x00C99B, 0x00C99C, 0x00C99D, 0x00C99E, 0x00C99F,
	0x00C9A0, 0x00C9A1, 0x00C9A2, 0x00C9A3, 0x00C9A4, 0x00C9A5, 0x00C9A6, 0x00C9A7, 0x00C9A8, 0x00C9A9, 0x00C9AA, 0x00C9AB, 0x00C9AC, 0x00C9AD, 0x00C9AE, 0x00C9AF, 0x00C9B0, 0x00C9B1, 0x00C9B2, 0x00C9B3, 0x00C9B4, 0x00C9B5, 0x00C9B6, 0x00C9B7, 0x00C9B8, 0x00C9B9, 0x00C9BA, 0x00C9BB, 0x00C9BC, 0x00C9BD, 0x00C9BE, 0x00C9BF,
	0x00C9C0, 0x00C9C1, 0x00C9C2, 0x00C9C3, 0x00C9C4, 0x00C9C5, 0x00C9C6, 0x00C9C7, 0x00C9C8, 0x00C9C9, 0x00C9CA, 0x00C9CB, 0x00C9CC, 0x00C9CD, 0x00C9CE, 0x00C9CF, 0x00C9D0, 0x00C9D1, 0x00C9D2, 0x00C9D3, 0x00C9D4, 0x00C9D5, 0x00C9D6, 0x00C9D7, 0x00C9D8, 0x00C9D9, 0x00C9DA, 0x00C9DB, 0x00C9DC, 0x00C9DD, 0x00C9DE, 0x00C9DF,
	0x00C9E0, 0x00C9E1, 0x00C9E2, 0x00C9E3, 0x00C9E4, 0x00C9E5, 0x00C9E6, 0x00C9E7, 0x00C9E8, 0x00C9E9, 0x00C9EA, 0x00C9EB, 0x00C9EC, 0x00C9ED, 0x00C9EE, 0x00C9EF, 0x00C9F0, 0x00C9F1, 0x00C9F2, 0x00C9F3, 0x00C9F4, 0x00C9F5, 0x00C9F6, 0x00C9F7, 0x00C9F8, 0x00C9F9, 0x00C9FA, 0x00C9FB, 0x00C9FC, 0x00C9FD, 0x00C9FE, 0x00C9FF,
	0x00CA00, 0x00CA01, 0x00CA02, 0x00CA03, 0x00CA04, 0x00CA05, 0x00CA06, 0x00CA07, 0x00CA08, 0x00CA09, 0x00CA0A, 0x00CA0B, 0x00CA0C, 0x00CA0D, 0x00CA0E, 0x00CA0F, 0x00CA10, 0x00CA11, 0x00CA12, 0x00CA13, 0x00CA14, 0x00CA15, 0x00CA16, 0x00CA17, 0x00CA18, 0x00CA19, 0x00CA1A, 0x00CA1B, 0x00CA1C, 0x00CA1D, 0x00CA1E, 0x00CA1F,
	0x00CA20, 0x00CA21, 0x00CA22, 0x00CA23, 0x00CA24, 0x00CA25, 0x00CA26, 0x00CA27, 0x00CA28, 0x00CA29, 0x00CA2A, 0x00CA2B, 0x00CA2C, 0x00CA2D, 0x00CA2E, 0x00CA2F, 0x00CA30, 0x00CA31, 0x00CA32, 0x00CA33, 0x00CA34, 0x00CA35, 0x00CA36, 0x00CA37, 0x00CA38, 0x00CA39, 0x00CA3A, 0x00CA3B, 0x00CA3C, 0x00CA3D, 0x00CA3E, 0x00CA3F,
	0x00CA40, 0x00CA41, 0x00CA42, 0x00CA43, 0x00CA44, 0x00CA45, 0x00CA46, 0x00CA47, 0x00CA48, 0x00CA49, 0x00CA4A, 0x00CA4B, 0x00CA4C, 0x00CA4D, 0x00CA4E, 0x00CA4F, 0x00CA50, 0x00CA51, 0x00CA52, 0x00CA53, 0x00CA54, 0x00CA55, 0x00CA56, 0x00CA57, 0x00CA58, 0x00CA59, 0x00CA5A, 0x00CA5B, 0x00CA5C, 0x00CA5D, 0x00CA5E, 0x00CA5F,
	0x00CA60, 0x00CA61, 0x00CA62, 0x00CA63, 0x00CA64, 0x00CA65, 0x00CA66, 0x00CA67, 0x00CA68, 0x00CA69, 0x00CA6A, 0x00CA6B, 0x00CA6C, 0x00CA6D, 0x00CA6E, 0x00CA6F, 0x00CA70, 0x00CA71, 0x00CA72, 0x00CA73, 0x00CA74, 0x00CA75, 0x00CA76, 0x00CA77, 0x00CA78, 0x00CA79, 0x00CA7A, 0x00CA7B, 0x00CA7C, 0x00CA7D, 0x00CA7E, 0x00CA7F,
	0x00CA80, 0x00CA81, 0x00CA82, 0x00CA83, 0x00CA84, 0x00CA85, 0x00CA86, 0x00CA87, 0x00CA88, 0x00CA89, 0x00CA8A, 0x00CA8B, 0x00CA8C, 0x00CA8D, 0x00CA8E, 0x00CA8F, 0x00CA90, 0x00CA91, 0x00CA92, 0x00CA93, 0x00CA94, 0x00CA95, 0x00CA96, 0x00CA97, 0x00CA98, 0x00CA99, 0x00CA9A, 0x00CA9B, 0x00CA9C, 0x00CA9D, 0x00CA9E, 0x00CA9F,
	0x00CAA0, 0x00CAA1, 0x00CAA2, 0x00CAA3, 0x00CAA4, 0x00CAA5, 0x00CAA6, 0x00CAA7, 0x00CAA8, 0x00CAA9, 0x00CAAA, 0x00CAAB, 0x00CAAC, 0x00CAAD, 0x00CAAE, 0x00CAAF, 0x00CAB0, 0x00CAB1, 0x00CAB2, 0x00CAB3, 0x00CAB4, 0x00CAB5, 0x00CAB6, 0x00CAB7, 0x00CAB8, 0x00CAB9, 0x00CABA, 0x00CABB, 0x00CABC, 0x00CABD, 0x00CABE, 0x00CABF,
	0x00CAC0, 0x00CAC1, 0x00CAC2, 0x00CAC3, 0x00CAC4, 0x00CAC5, 0x00CAC6, 0x00CAC7, 0x00CAC8, 0x00CAC9, 0x00CACA, 0x00CACB, 0x00CACC, 0x00CACD, 0x00CACE, 0x00CACF, 0x00CAD0, 0x00CAD1, 0x00CAD2, 0x00CAD3, 0x00CAD4, 0x00CAD5, 0x00CAD6, 0x00CAD7, 0x00CAD8, 0x00CAD9, 0x00CADA, 0x00CADB, 0x00CADC, 0x00CADD, 0x00CADE, 0x00CADF,
	0x00CAE0, 0x00CAE1, 0x00CAE2, 0x00CAE3, 0x00CAE4, 0x00CAE5, 0x00CAE6, 0x00CAE7, 0x00CAE8, 0x00CAE9, 0x00CAEA, 0x00CAEB, 0x00CAEC, 0x00CAED, 0x00CAEE, 0x00CAEF, 0x00CAF0, 0x00CAF1, 0x00CAF2, 0x00CAF3, 0x00CAF4, 0x00CAF5, 0x00CAF6, 0x00CAF7, 0x00CAF8, 0x00CAF9, 0x00CAFA, 0x00CAFB, 0x00CAFC, 0x00CAFD, 0x00CAFE, 0x00CAFF,
	0x00CB00, 0x00CB01, 0x00CB02, 0x00CB03, 0x00CB04, 0x00CB05, 0x00CB06, 0x00CB07, 0x00CB08, 0x00CB09, 0x00CB0A, 0x00CB0B, 0x00CB0C, 0x00CB0D, 0x00CB0E, 0x00CB0F, 0x00CB10, 0x00CB11, 0x00CB12, 0x00CB13, 0x00CB14, 0x00CB15, 0x00CB16, 0x00CB17, 0x00CB18, 0x00CB19, 0x00CB1A, 0x00CB1B, 0x00CB1C, 0x00CB1D, 0x00CB1E, 0x00CB1F,
	0x00CB20, 0x00CB21, 0x00CB22, 0x00CB23, 0x00CB24, 0x00CB25, 0x00CB26, 0x00CB27, 0x00CB28, 0x00CB29, 0x00CB2A, 0x00CB2B, 0x00CB2C, 0x00CB2D, 0x00CB2E, 0x00CB2F, 0x00CB30, 0x00CB31, 0x00CB32, 0x00CB33, 0x00CB34, 0x00CB35, 0x00CB36, 0x00CB37, 0x00CB38, 0x00CB39, 0x00CB3A, 0x00CB3B, 0x00CB3C, 0x00CB3D, 0x00CB3E, 0x00CB3F,
	0x00CB40, 0x00CB41, 0x00CB42, 0x00CB43, 0x00CB44, 0x00CB45, 0x00CB46, 0x00CB47, 0x00CB48, 0x00CB49, 0x00CB4A, 0x00CB4B, 0x00CB4C, 0x00CB4D, 0x00CB4E, 0x00CB4F, 0x00CB50, 0x00CB51, 0x00CB52, 0x00CB53, 0x00CB54, 0x00CB55, 0x00CB56, 0x00CB57, 0x00CB58, 0x00CB59, 0x00CB5A, 0x00CB5B, 0x00CB5C, 0x00CB5D, 0x00CB5E, 0x00CB5F,
	0x00CB60, 0x00CB61, 0x00CB62, 0x00CB63, 0x00CB64, 0x00CB65, 0x00CB66, 0x00CB67, 0x00CB68, 0x00CB69, 0x00CB6A, 0x00CB6B, 0x00CB6C, 0x00CB6D, 0x00CB6E, 0x00CB6F, 0x00CB70, 0x00CB71, 0x00CB72, 0x00CB73, 0x00CB74, 0x00CB75, 0x00CB76, 0x00CB77, 0x00CB78, 0x00CB79, 0x00CB7A, 0x00CB7B, 0x00CB7C, 0x00CB7D, 0x00CB7E, 0x00CB7F,
	0x00CB80, 0x00CB81, 0x00CB82, 0x00CB83, 0x00CB84, 0x00CB85, 0x00CB86, 0x00CB87, 0x00CB88, 0x00CB89, 0x00CB8A, 0x00CB8B, 0x00CB8C, 0x00CB8D, 0x00CB8E, 0x00CB8F, 0x00CB90, 0x00CB91, 0x00CB92, 0x00CB93, 0x00CB94, 0x00CB95, 0x00CB96, 0x00CB97, 0x00CB98, 0x00CB99, 0x00CB9A, 0x00CB9B, 0x00CB9C, 0x00CB9D, 0x00CB9E, 0x00CB9F,
	0x00CBA0, 0x00CBA1, 0x00CBA2, 0x00CBA3, 0x00CBA4, 0x00CBA5, 0x00CBA6, 0x00CBA7, 0x00CBA8, 0x00CBA9, 0x00CBAA, 0x00CBAB, 0x00CBAC, 0x00CBAD, 0x00CBAE, 0x00CBAF, 0x00CBB0, 0x00CBB1, 0x00CBB2, 0x00CBB3, 0x00CBB4, 0x00CBB5, 0x00CBB6, 0x00CBB7, 0x00CBB8, 0x00CBB9, 0x00CBBA, 0x00CBBB, 0x00CBBC, 0x00CBBD, 0x00CBBE, 0x00CBBF,
	0x00CBC0, 0x00CBC1, 0x00CBC2, 0x00CBC3, 0x00CBC4, 0x00CBC5, 0x00CBC6, 0x00CBC7, 0x00CBC8, 0x00CBC9, 0x00CBCA, 0x00CBCB, 0x00CBCC, 0x00CBCD, 0x00CBCE, 0x00CBCF, 0x00CBD0, 0x00CBD1, 0x00CBD2, 0x00CBD3, 0x00CBD4, 0x00CBD5, 0x00CBD6, 0x00CBD7, 0x00CBD8, 0x00CBD9, 0x00CBDA, 0x00CBDB, 0x00CBDC, 0x00CBDD, 0x00CBDE, 0x00CBDF,
	0x00CBE0, 0x00CBE1, 0x00CBE2, 0x00CBE3, 0x00CBE4, 0x00CBE5, 0x00CBE6, 0x00CBE7, 0x00CBE8, 0x00CBE9, 0x00CBEA, 0x00CBEB, 0x00CBEC, 0x00CBED, 0x00CBEE, 0x00CBEF, 0x00CBF0, 0x00CBF1, 0x00CBF2, 0x00CBF3, 0x00CBF4, 0x00CBF5, 0x00CBF6, 0x00CBF7, 0x00CBF8, 0x00CBF9, 0x00CBFA, 0x00CBFB, 0x00CBFC, 0x00CBFD, 0x00CBFE, 0x00CBFF,
	0
};

static const unsigned map_hangul_9[] = {
	0x00CC00, 0x00CC01, 0x00CC02, 0x00CC03, 0x00CC04, 0x00CC05, 0x00CC06, 0x00CC07, 0x00CC08, 0x00CC09, 0x00CC0A, 0x00CC0B, 0x00CC0C, 0x00CC0D, 0x00CC0E, 0x00CC0F, 0x00CC10, 0x00CC11, 0x00CC12, 0x00CC13, 0x00CC14, 0x00CC15, 0x00CC16, 0x00CC17, 0x00CC18, 0x00CC19, 0x00CC1A, 0x00CC1B, 0x00CC1C, 0x00CC1D, 0x00CC1E, 0x00CC1F,
	0x00CC20, 0x00CC21, 0x00CC22, 0x00CC23, 0x00CC24, 0x00CC25, 0x00CC26, 0x00CC27, 0x00CC28, 0x00CC29, 0x00CC2A, 0x00CC2B, 0x00CC2C, 0x00CC2D, 0x00CC2E, 0x00CC2F, 0x00CC30, 0x00CC31, 0x00CC32, 0x00CC33, 0x00CC34, 0x00CC35, 0x00CC36, 0x00CC37, 0x00CC38, 0x00CC39, 0x00CC3A, 0x00CC3B, 0x00CC3C, 0x00CC3D, 0x00CC3E, 0x00CC3F,
	0x00CC40, 0x00CC41, 0x00CC42, 0x00CC43, 0x00CC44, 0x00CC45, 0x00CC46, 0x00CC47, 0x00CC48, 0x00CC49, 0x00CC4A, 0x00CC4B, 0x00CC4C, 0x00CC4D, 0x00CC4E, 0x00CC4F, 0x00CC50, 0x00CC51, 0x00CC52, 0x00CC53, 0x00CC54, 0x00CC55, 0x00CC56, 0x00CC57, 0x00CC58, 0x00CC59, 0x00CC5A, 0x00CC5B, 0x00CC5C, 0x00CC5D, 0x00CC5E, 0x00CC5F,
	0x00CC60, 0x00CC61, 0x00CC62, 0x00CC63, 0x00CC64, 0x00CC65, 0x00CC66, 0x00CC67, 0x00CC68, 0x00CC69, 0x00CC6A, 0x00CC6B, 0x00CC6C, 0x00CC6D, 0x00CC6E, 0x00CC6F, 0x00CC70, 0x00CC71, 0x00CC72, 0x00CC73, 0x00CC74, 0x00CC75, 0x00CC76, 0x00CC77, 0x00CC78, 0x00CC79, 0x00CC7A, 0x00CC7B, 0x00CC7C, 0x00CC7D, 0x00CC7E, 0x00CC7F,
	0x00CC80, 0x00CC81, 0x00CC82, 0x00CC83, 0x00CC84, 0x00CC85, 0x00CC86, 0x00CC87, 0x00CC88, 0x00CC89, 0x00CC8A, 0x00CC8B, 0x00CC8C, 0x00CC8D, 0x00CC8E, 0x00CC8F, 0x00CC90, 0x00CC91, 0x00CC92, 0x00CC93, 0x00CC94, 0x00CC95, 0x00CC96, 0x00CC97, 0x00CC98, 0x00CC99, 0x00CC9A, 0x00CC9B, 0x00CC9C, 0x00CC9D, 0x00CC9E, 0x00CC9F,
	0x00CCA0, 0x00CCA1, 0x00CCA2, 0x00CCA3, 0x00CCA4, 0x00CCA5, 0x00CCA6, 0x00CCA7, 0x00CCA8, 0x00CCA9, 0x00CCAA, 0x00CCAB, 0x00CCAC, 0x00CCAD, 0x00CCAE, 0x00CCAF, 0x00CCB0, 0x00CCB1, 0x00CCB2, 0x00CCB3, 0x00CCB4, 0x00CCB5, 0x00CCB6, 0x00CCB7, 0x00CCB8, 0x00CCB9, 0x00CCBA, 0x00CCBB, 0x00CCBC, 0x00CCBD, 0x00CCBE, 0x00CCBF,
	0x00CCC0, 0x00CCC1, 0x00CCC2, 0x00CCC3, 0x00CCC4, 0x00CCC5, 0x00CCC6, 0x00CCC7, 0x00CCC8, 0x00CCC9, 0x00CCCA, 0x00CCCB, 0x00CCCC, 0x00CCCD, 0x00CCCE, 0x00CCCF, 0x00CCD0, 0x00CCD1, 0x00CCD2, 0x00CCD3, 0x00CCD4, 0x00CCD5, 0x00CCD6, 0x00CCD7, 0x00CCD8, 0x00CCD9, 0x00CCDA, 0x00CCDB, 0x00CCDC, 0x00CCDD, 0x00CCDE, 0x00CCDF,
	0x00CCE0, 0x00CCE1, 0x00CCE2, 0x00CCE3, 0x00CCE4, 0x00CCE5, 0x00CCE6, 0x00CCE7, 0x00CCE8, 0x00CCE9, 0x00CCEA, 0x00CCEB, 0x00CCEC, 0x00CCED, 0x00CCEE, 0x00CCEF, 0x00CCF0, 0x00CCF1, 0x00CCF2, 0x00CCF3, 0x00CCF4, 0x00CCF5, 0x00CCF6, 0x00CCF7, 0x00CCF8, 0x00CCF9, 0x00CCFA, 0x00CCFB, 0x00CCFC, 0x00CCFD, 0x00CCFE, 0x00CCFF,
	0x00CD00, 0x00CD01, 0x00CD02, 0x00CD03, 0x00CD04, 0x00CD05, 0x00CD06, 0x00CD07, 0x00CD08, 0x00CD09, 0x00CD0A, 0x00CD0B, 0x00CD0C, 0x00CD0D, 0x00CD0E, 0x00CD0F, 0x00CD10, 0x00CD11, 0x00CD12, 0x00CD13, 0x00CD14, 0x00CD15, 0x00CD16, 0x00CD17, 0x00CD18, 0x00CD19, 0x00CD1A, 0x00CD1B, 0x00CD1C, 0x00CD1D, 0x00CD1E, 0x00CD1F,
	0x00CD20, 0x00CD21, 0x00CD22, 0x00CD23, 0x00CD24, 0x00CD25, 0x00CD26, 0x00CD27, 0x00CD28, 0x00CD29, 0x00CD2A, 0x00CD2B, 0x00CD2C, 0x00CD2D, 0x00CD2E, 0x00CD2F, 0x00CD30, 0x00CD31, 0x00CD32, 0x00CD33, 0x00CD34, 0x00CD35, 0x00CD36, 0x00CD37, 0x00CD38, 0x00CD39, 0x00CD3A, 0x00CD3B, 0x00CD3C, 0x00CD3D, 0x00CD3E, 0x00CD3F,
	0x00CD40, 0x00CD41, 0x00CD42, 0x00CD43, 0x00CD44, 0x00CD45, 0x00CD46, 0x00CD47, 0x00CD48, 0x00CD49, 0x00CD4A, 0x00CD4B, 0x00CD4C, 0x00CD4D, 0x00CD4E, 0x00CD4F, 0x00CD50, 0x00CD51, 0x00CD52, 0x00CD53, 0x00CD54, 0x00CD55, 0x00CD56, 0x00CD57, 0x00CD58, 0x00CD59, 0x00CD5A, 0x00CD5B, 0x00CD5C, 0x00CD5D, 0x00CD5E, 0x00CD5F,
	0x00CD60, 0x00CD61, 0x00CD62, 0x00CD63, 0x00CD64, 0x00CD65, 0x00CD66, 0x00CD67, 0x00CD68, 0x00CD69, 0x00CD6A, 0x00CD6B, 0x00CD6C, 0x00CD6D, 0x00CD6E, 0x00CD6F, 0x00CD70, 0x00CD71, 0x00CD72, 0x00CD73, 0x00CD74, 0x00CD75, 0x00CD76, 0x00CD77, 0x00CD78, 0x00CD79, 0x00CD7A, 0x00CD7B, 0x00CD7C, 0x00CD7D, 0x00CD7E, 0x00CD7F,
	0x00CD80, 0x00CD81, 0x00CD82, 0x00CD83, 0x00CD84, 0x00CD85, 0x00CD86, 0x00CD87, 0x00CD88, 0x00CD89, 0x00CD8A, 0x00CD8B, 0x00CD8C, 0x00CD8D, 0x00CD8E, 0x00CD8F, 0x00CD90, 0x00CD91, 0x00CD92, 0x00CD93, 0x00CD94, 0x00CD95, 0x00CD96, 0x00CD97, 0x00CD98, 0x00CD99, 0x00CD9A, 0x00CD9B, 0x00CD9C, 0x00CD9D, 0x00CD9E, 0x00CD9F,
	0x00CDA0, 0x00CDA1, 0x00CDA2, 0x00CDA3, 0x00CDA4, 0x00CDA5, 0x00CDA6, 0x00CDA7, 0x00CDA8, 0x00CDA9, 0x00CDAA, 0x00CDAB, 0x00CDAC, 0x00CDAD, 0x00CDAE, 0x00CDAF, 0x00CDB0, 0x00CDB1, 0x00CDB2, 0x00CDB3, 0x00CDB4, 0x00CDB5, 0x00CDB6, 0x00CDB7, 0x00CDB8, 0x00CDB9, 0x00CDBA, 0x00CDBB, 0x00CDBC, 0x00CDBD, 0x00CDBE, 0x00CDBF,
	0x00CDC0, 0x00CDC1, 0x00CDC2, 0x00CDC3, 0x00CDC4, 0x00CDC5, 0x00CDC6, 0x00CDC7, 0x00CDC8, 0x00CDC9, 0x00CDCA, 0x00CDCB, 0x00CDCC, 0x00CDCD, 0x00CDCE, 0x00CDCF, 0x00CDD0, 0x00CDD1, 0x00CDD2, 0x00CDD3, 0x00CDD4, 0x00CDD5, 0x00CDD6, 0x00CDD7, 0x00CDD8, 0x00CDD9, 0x00CDDA, 0x00CDDB, 0x00CDDC, 0x00CDDD, 0x00CDDE, 0x00CDDF,
	0x00CDE0, 0x00CDE1, 0x00CDE2, 0x00CDE3, 0x00CDE4, 0x00CDE5, 0x00CDE6, 0x00CDE7, 0x00CDE8, 0x00CDE9, 0x00CDEA, 0x00CDEB, 0x00CDEC, 0x00CDED, 0x00CDEE, 0x00CDEF, 0x00CDF0, 0x00CDF1, 0x00CDF2, 0x00CDF3, 0x00CDF4, 0x00CDF5, 0x00CDF6, 0x00CDF7, 0x00CDF8, 0x00CDF9, 0x00CDFA, 0x00CDFB, 0x00CDFC, 0x00CDFD, 0x00CDFE, 0x00CDFF,
	0x00CE00, 0x00CE01, 0x00CE02, 0x00CE03, 0x00CE04, 0x00CE05, 0x00CE06, 0x00CE07, 0x00CE08, 0x00CE09, 0x00CE0A, 0x00CE0B, 0x00CE0C, 0x00CE0D, 0x00CE0E, 0x00CE0F, 0x00CE10, 0x00CE11, 0x00CE12, 0x00CE13, 0x00CE14, 0x00CE15, 0x00CE16, 0x00CE17, 0x00CE18, 0x00CE19, 0x00CE1A, 0x00CE1B, 0x00CE1C, 0x00CE1D, 0x00CE1E, 0x00CE1F,
	0x00CE20, 0x00CE21, 0x00CE22, 0x00CE23, 0x00CE24, 0x00CE25, 0x00CE26, 0x00CE27, 0x00CE28, 0x00CE29, 0x00CE2A, 0x00CE2B, 0x00CE2C, 0x00CE2D, 0x00CE2E, 0x00CE2F, 0x00CE30, 0x00CE31, 0x00CE32, 0x00CE33, 0x00CE34, 0x00CE35, 0x00CE36, 0x00CE37, 0x00CE38, 0x00CE39, 0x00CE3A, 0x00CE3B, 0x00CE3C, 0x00CE3D, 0x00CE3E, 0x00CE3F,
	0x00CE40, 0x00CE41, 0x00CE42, 0x00CE43, 0x00CE44, 0x00CE45, 0x00CE46, 0x00CE47, 0x00CE48, 0x00CE49, 0x00CE4A, 0x00CE4B, 0x00CE4C, 0x00CE4D, 0x00CE4E, 0x00CE4F, 0x00CE50, 0x00CE51, 0x00CE52, 0x00CE53, 0x00CE54, 0x00CE55, 0x00CE56, 0x00CE57, 0x00CE58, 0x00CE59, 0x00CE5A, 0x00CE5B, 0x00CE5C, 0x00CE5D, 0x00CE5E, 0x00CE5F,
	0x00CE60, 0x00CE61, 0x00CE62, 0x00CE63, 0x00CE64, 0x00CE65, 0x00CE66, 0x00CE67, 0x00CE68, 0x00CE69, 0x00CE6A, 0x00CE6B, 0x00CE6C, 0x00CE6D, 0x00CE6E, 0x00CE6F, 0x00CE70, 0x00CE71, 0x00CE72, 0x00CE73, 0x00CE74, 0x00CE75, 0x00CE76, 0x00CE77, 0x00CE78, 0x00CE79, 0x00CE7A, 0x00CE7B, 0x00CE7C, 0x00CE7D, 0x00CE7E, 0x00CE7F,
	0x00CE80, 0x00CE81, 0x00CE82, 0x00CE83, 0x00CE84, 0x00CE85, 0x00CE86, 0x00CE87, 0x00CE88, 0x00CE89, 0x00CE8A, 0x00CE8B, 0x00CE8C, 0x00CE8D, 0x00CE8E, 0x00CE8F, 0x00CE90, 0x00CE91, 0x00CE92, 0x00CE93, 0x00CE94, 0x00CE95, 0x00CE96, 0x00CE97, 0x00CE98, 0x00CE99, 0x00CE9A, 0x00CE9B, 0x00CE9C, 0x00CE9D, 0x00CE9E, 0x00CE9F,
	0x00CEA0, 0x00CEA1, 0x00CEA2, 0x00CEA3, 0x00CEA4, 0x00CEA5, 0x00CEA6, 0x00CEA7, 0x00CEA8, 0x00CEA9, 0x00CEAA, 0x00CEAB, 0x00CEAC, 0x00CEAD, 0x00CEAE, 0x00CEAF, 0x00CEB0, 0x00CEB1, 0x00CEB2, 0x00CEB3, 0x00CEB4, 0x00CEB5, 0x00CEB6, 0x00CEB7, 0x00CEB8, 0x00CEB9, 0x00CEBA, 0x00CEBB, 0x00CEBC, 0x00CEBD, 0x00CEBE, 0x00CEBF,
	0x00CEC0, 0x00CEC1, 0x00CEC2, 0x00CEC3, 0x00CEC4, 0x00CEC5, 0x00CEC6, 0x00CEC7, 0x00CEC8, 0x00CEC9, 0x00CECA, 0x00CECB, 0x00CECC, 0x00CECD, 0x00CECE, 0x00CECF, 0x00CED0, 0x00CED1, 0x00CED2, 0x00CED3, 0x00CED4, 0x00CED5, 0x00CED6, 0x00CED7, 0x00CED8, 0x00CED9, 0x00CEDA, 0x00CEDB, 0x00CEDC, 0x00CEDD, 0x00CEDE, 0x00CEDF,
	0x00CEE0, 0x00CEE1, 0x00CEE2, 0x00CEE3, 0x00CEE4, 0x00CEE5, 0x00CEE6, 0x00CEE7, 0x00CEE8, 0x00CEE9, 0x00CEEA, 0x00CEEB, 0x00CEEC, 0x00CEED, 0x00CEEE, 0x00CEEF, 0x00CEF0, 0x00CEF1, 0x00CEF2, 0x00CEF3, 0x00CEF4, 0x00CEF5, 0x00CEF6, 0x00CEF7, 0x00CEF8, 0x00CEF9, 0x00CEFA, 0x00CEFB, 0x00CEFC, 0x00CEFD, 0x00CEFE, 0x00CEFF,
	0x00CF00, 0x00CF01, 0x00CF02, 0x00CF03, 0x00CF04, 0x00CF05, 0x00CF06, 0x00CF07, 0x00CF08, 0x00CF09, 0x00CF0A, 0x00CF0B, 0x00CF0C, 0x00CF0D, 0x00CF0E, 0x00CF0F, 0x00CF10, 0x00CF11, 0x00CF12, 0x00CF13, 0x00CF14, 0x00CF15, 0x00CF16, 0x00CF17, 0x00CF18, 0x00CF19, 0x00CF1A, 0x00CF1B, 0x00CF1C, 0x00CF1D, 0x00CF1E, 0x00CF1F,
	0x00CF20, 0x00CF21, 0x00CF22, 0x00CF23, 0x00CF24, 0x00CF25, 0x00CF26, 0x00CF27, 0x00CF28, 0x00CF29, 0x00CF2A, 0x00CF2B, 0x00CF2C, 0x00CF2D, 0x00CF2E, 0x00CF2F, 0x00CF30, 0x00CF31, 0x00CF32, 0x00CF33, 0x00CF34, 0x00CF35, 0x00CF36, 0x00CF37, 0x00CF38, 0x00CF39, 0x00CF3A, 0x00CF3B, 0x00CF3C, 0x00CF3D, 0x00CF3E, 0x00CF3F,
	0x00CF40, 0x00CF41, 0x00CF42, 0x00CF43, 0x00CF44, 0x00CF45, 0x00CF46, 0x00CF47, 0x00CF48, 0x00CF49, 0x00CF4A, 0x00CF4B, 0x00CF4C, 0x00CF4D, 0x00CF4E, 0x00CF4F, 0x00CF50, 0x00CF51, 0x00CF52, 0x00CF53, 0x00CF54, 0x00CF55, 0x00CF56, 0x00CF57, 0x00CF58, 0x00CF59, 0x00CF5A, 0x00CF5B, 0x00CF5C, 0x00CF5D, 0x00CF5E, 0x00CF5F,
	0x00CF60, 0x00CF61, 0x00CF62, 0x00CF63, 0x00CF64, 0x00CF65, 0x00CF66, 0x00CF67, 0x00CF68, 0x00CF69, 0x00CF6A, 0x00CF6B, 0x00CF6C, 0x00CF6D, 0x00CF6E, 0x00CF6F, 0x00CF70, 0x00CF71, 0x00CF72, 0x00CF73, 0x00CF74, 0x00CF75, 0x00CF76, 0x00CF77, 0x00CF78, 0x00CF79, 0x00CF7A, 0x00CF7B, 0x00CF7C, 0x00CF7D, 0x00CF7E, 0x00CF7F,
	0x00CF80, 0x00CF81, 0x00CF82, 0x00CF83, 0x00CF84, 0x00CF85, 0x00CF86, 0x00CF87, 0x00CF88, 0x00CF89, 0x00CF8A, 0x00CF8B, 0x00CF8C, 0x00CF8D, 0x00CF8E, 0x00CF8F, 0x00CF90, 0x00CF91, 0x00CF92, 0x00CF93, 0x00CF94, 0x00CF95, 0x00CF96, 0x00CF97, 0x00CF98, 0x00CF99, 0x00CF9A, 0x00CF9B, 0x00CF9C, 0x00CF9D, 0x00CF9E, 0x00CF9F,
	0x00CFA0, 0x00CFA1, 0x00CFA2, 0x00CFA3, 0x00CFA4, 0x00CFA5, 0x00CFA6, 0x00CFA7, 0x00CFA8, 0x00CFA9, 0x00CFAA, 0x00CFAB, 0x00CFAC, 0x00CFAD, 0x00CFAE, 0x00CFAF, 0x00CFB0, 0x00CFB1, 0x00CFB2, 0x00CFB3, 0x00CFB4, 0x00CFB5, 0x00CFB6, 0x00CFB7, 0x00CFB8, 0x00CFB9, 0x00CFBA, 0x00CFBB, 0x00CFBC, 0x00CFBD, 0x00CFBE, 0x00CFBF,
	0x00CFC0, 0x00CFC1, 0x00CFC2, 0x00CFC3, 0x00CFC4, 0x00CFC5, 0x00CFC6, 0x00CFC7, 0x00CFC8, 0x00CFC9, 0x00CFCA, 0x00CFCB, 0x00CFCC, 0x00CFCD, 0x00CFCE, 0x00CFCF, 0x00CFD0, 0x00CFD1, 0x00CFD2, 0x00CFD3, 0x00CFD4, 0x00CFD5, 0x00CFD6, 0x00CFD7, 0x00CFD8, 0x00CFD9, 0x00CFDA, 0x00CFDB, 0x00CFDC, 0x00CFDD, 0x00CFDE, 0x00CFDF,
	0x00CFE0, 0x00CFE1, 0x00CFE2, 0x00CFE3, 0x00CFE4, 0x00CFE5, 0x00CFE6, 0x00CFE7, 0x00CFE8, 0x00CFE9, 0x00CFEA, 0x00CFEB, 0x00CFEC, 0x00CFED, 0x00CFEE, 0x00CFEF, 0x00CFF0, 0x00CFF1, 0x00CFF2, 0x00CFF3, 0x00CFF4, 0x00CFF5, 0x00CFF6, 0x00CFF7, 0x00CFF8, 0x00CFF9, 0x00CFFA, 0x00CFFB, 0x00CFFC, 0x00CFFD, 0x00CFFE, 0x00CFFF,
	0
};

static const unsigned map_hangul_10[] = {
	0x00D000, 0x00D001, 0x00D002, 0x00D003, 0x00D004, 0x00D005, 0x00D006, 0x00D007, 0x00D008, 0x00D009, 0x00D00A, 0x00D00B, 0x00D00C, 0x00D00D, 0x00D00E, 0x00D00F, 0x00D010, 0x00D011, 0x00D012, 0x00D013, 0x00D014, 0x00D015, 0x00D016, 0x00D017, 0x00D018, 0x00D019, 0x00D01A, 0x00D01B, 0x00D01C, 0x00D01D, 0x00D01E, 0x00D01F,
	0x00D020, 0x00D021, 0x00D022, 0x00D023, 0x00D024, 0x00D025, 0x00D026, 0x00D027, 0x00D028, 0x00D029, 0x00D02A, 0x00D02B, 0x00D02C, 0x00D02D, 0x00D02E, 0x00D02F, 0x00D030, 0x00D031, 0x00D032, 0x00D033, 0x00D034, 0x00D035, 0x00D036, 0x00D037, 0x00D038, 0x00D039, 0x00D03A, 0x00D03B, 0x00D03C, 0x00D03D, 0x00D03E, 0x00D03F,
	0x00D040, 0x00D041, 0x00D042, 0x00D043, 0x00D044, 0x00D045, 0x00D046, 0x00D047, 0x00D048, 0x00D049, 0x00D04A, 0x00D04B, 0x00D04C, 0x00D04D, 0x00D04E, 0x00D04F, 0x00D050, 0x00D051, 0x00D052, 0x00D053, 0x00D054, 0x00D055, 0x00D056, 0x00D057, 0x00D058, 0x00D059, 0x00D05A, 0x00D05B, 0x00D05C, 0x00D05D, 0x00D05E, 0x00D05F,
	0x00D060, 0x00D061, 0x00D062, 0x00D063, 0x00D064, 0x00D065, 0x00D066, 0x00D067, 0x00D068, 0x00D069, 0x00D06A, 0x00D06B, 0x00D06C, 0x00D06D, 0x00D06E, 0x00D06F, 0x00D070, 0x00D071, 0x00D072, 0x00D073, 0x00D074, 0x00D075, 0x00D076, 0x00D077, 0x00D078, 0x00D079, 0x00D07A, 0x00D07B, 0x00D07C, 0x00D07D, 0x00D07E, 0x00D07F,
	0x00D080, 0x00D081, 0x00D082, 0x00D083, 0x00D084, 0x00D085, 0x00D086, 0x00D087, 0x00D088, 0x00D089, 0x00D08A, 0x00D08B, 0x00D08C, 0x00D08D, 0x00D08E, 0x00D08F, 0x00D090, 0x00D091, 0x00D092, 0x00D093, 0x00D094, 0x00D095, 0x00D096, 0x00D097, 0x00D098, 0x00D099, 0x00D09A, 0x00D09B, 0x00D09C, 0x00D09D, 0x00D09E, 0x00D09F,
	0x00D0A0, 0x00D0A1, 0x00D0A2, 0x00D0A3, 0x00D0A4, 0x00D0A5, 0x00D0A6, 0x00D0A7, 0x00D0A8, 0x00D0A9, 0x00D0AA, 0x00D0AB, 0x00D0AC, 0x00D0AD, 0x00D0AE, 0x00D0AF, 0x00D0B0, 0x00D0B1, 0x00D0B2, 0x00D0B3, 0x00D0B4, 0x00D0B5, 0x00D0B6, 0x00D0B7, 0x00D0B8, 0x00D0B9, 0x00D0BA, 0x00D0BB, 0x00D0BC, 0x00D0BD, 0x00D0BE, 0x00D0BF,
	0x00D0C0, 0x00D0C1, 0x00D0C2, 0x00D0C3, 0x00D0C4, 0x00D0C5, 0x00D0C6, 0x00D0C7, 0x00D0C8, 0x00D0C9, 0x00D0CA, 0x00D0CB, 0x00D0CC, 0x00D0CD, 0x00D0CE, 0x00D0CF, 0x00D0D0, 0x00D0D1, 0x00D0D2, 0x00D0D3, 0x00D0D4, 0x00D0D5, 0x00D0D6, 0x00D0D7, 0x00D0D8, 0x00D0D9, 0x00D0DA, 0x00D0DB, 0x00D0DC, 0x00D0DD, 0x00D0DE, 0x00D0DF,
	0x00D0E0, 0x00D0E1, 0x00D0E2, 0x00D0E3, 0x00D0E4, 0x00D0E5, 0x00D0E6, 0x00D0E7, 0x00D0E8, 0x00D0E9, 0x00D0EA, 0x00D0EB, 0x00D0EC, 0x00D0ED, 0x00D0EE, 0x00D0EF, 0x00D0F0, 0x00D0F1, 0x00D0F2, 0x00D0F3, 0x00D0F4, 0x00D0F5, 0x00D0F6, 0x00D0F7, 0x00D0F8, 0x00D0F9, 0x00D0FA, 0x00D0FB, 0x00D0FC, 0x00D0FD, 0x00D0FE, 0x00D0FF,
	0x00D100, 0x00D101, 0x00D102, 0x00D103, 0x00D104, 0x00D105, 0x00D106, 0x00D107, 0x00D108, 0x00D109, 0x00D10A, 0x00D10B, 0x00D10C, 0x00D10D, 0x00D10E, 0x00D10F, 0x00D110, 0x00D111, 0x00D112, 0x00D113, 0x00D114, 0x00D115, 0x00D116, 0x00D117, 0x00D118, 0x00D119, 0x00D11A, 0x00D11B, 0x00D11C, 0x00D11D, 0x00D11E, 0x00D11F,
	0x00D120, 0x00D121, 0x00D122, 0x00D123, 0x00D124, 0x00D125, 0x00D126, 0x00D127, 0x00D128, 0x00D129, 0x00D12A, 0x00D12B, 0x00D12C, 0x00D12D, 0x00D12E, 0x00D12F, 0x00D130, 0x00D131, 0x00D132, 0x00D133, 0x00D134, 0x00D135, 0x00D136, 0x00D137, 0x00D138, 0x00D139, 0x00D13A, 0x00D13B, 0x00D13C, 0x00D13D, 0x00D13E, 0x00D13F,
	0x00D140, 0x00D141, 0x00D142, 0x00D143, 0x00D144, 0x00D145, 0x00D146, 0x00D147, 0x00D148, 0x00D149, 0x00D14A, 0x00D14B, 0x00D14C, 0x00D14D, 0x00D14E, 0x00D14F, 0x00D150, 0x00D151, 0x00D152, 0x00D153, 0x00D154, 0x00D155, 0x00D156, 0x00D157, 0x00D158, 0x00D159, 0x00D15A, 0x00D15B, 0x00D15C, 0x00D15D, 0x00D15E, 0x00D15F,
	0x00D160, 0x00D161, 0x00D162, 0x00D163, 0x00D164, 0x00D165, 0x00D166, 0x00D167, 0x00D168, 0x00D169, 0x00D16A, 0x00D16B, 0x00D16C, 0x00D16D, 0x00D16E, 0x00D16F, 0x00D170, 0x00D171, 0x00D172, 0x00D173, 0x00D174, 0x00D175, 0x00D176, 0x00D177, 0x00D178, 0x00D179, 0x00D17A, 0x00D17B, 0x00D17C, 0x00D17D, 0x00D17E, 0x00D17F,
	0x00D180, 0x00D181, 0x00D182, 0x00D183, 0x00D184, 0x00D185, 0x00D186, 0x00D187, 0x00D188, 0x00D189, 0x00D18A, 0x00D18B, 0x00D18C, 0x00D18D, 0x00D18E, 0x00D18F, 0x00D190, 0x00D191, 0x00D192, 0x00D193, 0x00D194, 0x00D195, 0x00D196, 0x00D197, 0x00D198, 0x00D199, 0x00D19A, 0x00D19B, 0x00D19C, 0x00D19D, 0x00D19E, 0x00D19F,
	0x00D1A0, 0x00D1A1, 0x00D1A2, 0x00D1A3, 0x00D1A4, 0x00D1A5, 0x00D1A6, 0x00D1A7, 0x00D1A8, 0x00D1A9, 0x00D1AA, 0x00D1AB, 0x00D1AC, 0x00D1AD, 0x00D1AE, 0x00D1AF, 0x00D1B0, 0x00D1B1, 0x00D1B2, 0x00D1B3, 0x00D1B4, 0x00D1B5, 0x00D1B6, 0x00D1B7, 0x00D1B8, 0x00D1B9, 0x00D1BA, 0x00D1BB, 0x00D1BC, 0x00D1BD, 0x00D1BE, 0x00D1BF,
	0x00D1C0, 0x00D1C1, 0x00D1C2, 0x00D1C3, 0x00D1C4, 0x00D1C5, 0x00D1C6, 0x00D1C7, 0x00D1C8, 0x00D1C9, 0x00D1CA, 0x00D1CB, 0x00D1CC, 0x00D1CD, 0x00D1CE, 0x00D1CF, 0x00D1D0, 0x00D1D1, 0x00D1D2, 0x00D1D3, 0x00D1D4, 0x00D1D5, 0x00D1D6, 0x00D1D7, 0x00D1D8, 0x00D1D9, 0x00D1DA, 0x00D1DB, 0x00D1DC, 0x00D1DD, 0x00D1DE, 0x00D1DF,
	0x00D1E0, 0x00D1E1, 0x00D1E2, 0x00D1E3, 0x00D1E4, 0x00D1E5, 0x00D1E6, 0x00D1E7, 0x00D1E8, 0x00D1E9, 0x00D1EA, 0x00D1EB, 0x00D1EC, 0x00D1ED, 0x00D1EE, 0x00D1EF, 0x00D1F0, 0x00D1F1, 0x00D1F2, 0x00D1F3, 0x00D1F4, 0x00D1F5, 0x00D1F6, 0x00D1F7, 0x00D1F8, 0x00D1F9, 0x00D1FA, 0x00D1FB, 0x00D1FC, 0x00D1FD, 0x00D1FE, 0x00D1FF,
	0x00D200, 0x00D201, 0x00D202, 0x00D203, 0x00D204, 0x00D205, 0x00D206, 0x00D207, 0x00D208, 0x00D209, 0x00D20A, 0x00D20B, 0x00D20C, 0x00D20D, 0x00D20E, 0x00D20F, 0x00D210, 0x00D211, 0x00D212, 0x00D213, 0x00D214, 0x00D215, 0x00D216, 0x00D217, 0x00D218, 0x00D219, 0x00D21A, 0x00D21B, 0x00D21C, 0x00D21D, 0x00D21E, 0x00D21F,
	0x00D220, 0x00D221, 0x00D222, 0x00D223, 0x00D224, 0x00D225, 0x00D226, 0x00D227, 0x00D228, 0x00D229, 0x00D22A, 0x00D22B, 0x00D22C, 0x00D22D, 0x00D22E, 0x00D22F, 0x00D230, 0x00D231, 0x00D232, 0x00D233, 0x00D234, 0x00D235, 0x00D236, 0x00D237, 0x00D238, 0x00D239, 0x00D23A, 0x00D23B, 0x00D23C, 0x00D23D, 0x00D23E, 0x00D23F,
	0x00D240, 0x00D241, 0x00D242, 0x00D243, 0x00D244, 0x00D245, 0x00D246, 0x00D247, 0x00D248, 0x00D249, 0x00D24A, 0x00D24B, 0x00D24C, 0x00D24D, 0x00D24E, 0x00D24F, 0x00D250, 0x00D251, 0x00D252, 0x00D253, 0x00D254, 0x00D255, 0x00D256, 0x00D257, 0x00D258, 0x00D259, 0x00D25A, 0x00D25B, 0x00D25C, 0x00D25D, 0x00D25E, 0x00D25F,
	0x00D260, 0x00D261, 0x00D262, 0x00D263, 0x00D264, 0x00D265, 0x00D266, 0x00D267, 0x00D268, 0x00D269, 0x00D26A, 0x00D26B, 0x00D26C, 0x00D26D, 0x00D26E, 0x00D26F, 0x00D270, 0x00D271, 0x00D272, 0x00D273, 0x00D274, 0x00D275, 0x00D276, 0x00D277, 0x00D278, 0x00D279, 0x00D27A, 0x00D27B, 0x00D27C, 0x00D27D, 0x00D27E, 0x00D27F,
	0x00D280, 0x00D281, 0x00D282, 0x00D283, 0x00D284, 0x00D285, 0x00D286, 0x00D287, 0x00D288, 0x00D289, 0x00D28A, 0x00D28B, 0x00D28C, 0x00D28D, 0x00D28E, 0x00D28F, 0x00D290, 0x00D291, 0x00D292, 0x00D293, 0x00D294, 0x00D295, 0x00D296, 0x00D297, 0x00D298, 0x00D299, 0x00D29A, 0x00D29B, 0x00D29C, 0x00D29D, 0x00D29E, 0x00D29F,
	0x00D2A0, 0x00D2A1, 0x00D2A2, 0x00D2A3, 0x00D2A4, 0x00D2A5, 0x00D2A6, 0x00D2A7, 0x00D2A8, 0x00D2A9, 0x00D2AA, 0x00D2AB, 0x00D2AC, 0x00D2AD, 0x00D2AE, 0x00D2AF, 0x00D2B0, 0x00D2B1, 0x00D2B2, 0x00D2B3, 0x00D2B4, 0x00D2B5, 0x00D2B6, 0x00D2B7, 0x00D2B8, 0x00D2B9, 0x00D2BA, 0x00D2BB, 0x00D2BC, 0x00D2BD, 0x00D2BE, 0x00D2BF,
	0x00D2C0, 0x00D2C1, 0x00D2C2, 0x00D2C3, 0x00D2C4, 0x00D2C5, 0x00D2C6, 0x00D2C7, 0x00D2C8, 0x00D2C9, 0x00D2CA, 0x00D2CB, 0x00D2CC, 0x00D2CD, 0x00D2CE, 0x00D2CF, 0x00D2D0, 0x00D2D1, 0x00D2D2, 0x00D2D3, 0x00D2D4, 0x00D2D5, 0x00D2D6, 0x00D2D7, 0x00D2D8, 0x00D2D9, 0x00D2DA, 0x00D2DB, 0x00D2DC, 0x00D2DD, 0x00D2DE, 0x00D2DF,
	0x00D2E0, 0x00D2E1, 0x00D2E2, 0x00D2E3, 0x00D2E4, 0x00D2E5, 0x00D2E6, 0x00D2E7, 0x00D2E8, 0x00D2E9, 0x00D2EA, 0x00D2EB, 0x00D2EC, 0x00D2ED, 0x00D2EE, 0x00D2EF, 0x00D2F0, 0x00D2F1, 0x00D2F2, 0x00D2F3, 0x00D2F4, 0x00D2F5, 0x00D2F6, 0x00D2F7, 0x00D2F8, 0x00D2F9, 0x00D2FA, 0x00D2FB, 0x00D2FC, 0x00D2FD, 0x00D2FE, 0x00D2FF,
	0x00D300, 0x00D301, 0x00D302, 0x00D303, 0x00D304, 0x00D305, 0x00D306, 0x00D307, 0x00D308, 0x00D309, 0x00D30A, 0x00D30B, 0x00D30C, 0x00D30D, 0x00D30E, 0x00D30F, 0x00D310, 0x00D311, 0x00D312, 0x00D313, 0x00D314, 0x00D315, 0x00D316, 0x00D317, 0x00D318, 0x00D319, 0x00D31A, 0x00D31B, 0x00D31C, 0x00D31D, 0x00D31E, 0x00D31F,
	0x00D320, 0x00D321, 0x00D322, 0x00D323, 0x00D324, 0x00D325, 0x00D326, 0x00D327, 0x00D328, 0x00D329, 0x00D32A, 0x00D32B, 0x00D32C, 0x00D32D, 0x00D32E, 0x00D32F, 0x00D330, 0x00D331, 0x00D332, 0x00D333, 0x00D334, 0x00D335, 0x00D336, 0x00D337, 0x00D338, 0x00D339, 0x00D33A, 0x00D33B, 0x00D33C, 0x00D33D, 0x00D33E, 0x00D33F,
	0x00D340, 0x00D341, 0x00D342, 0x00D343, 0x00D344, 0x00D345, 0x00D346, 0x00D347, 0x00D348, 0x00D349, 0x00D34A, 0x00D34B, 0x00D34C, 0x00D34D, 0x00D34E, 0x00D34F, 0x00D350, 0x00D351, 0x00D352, 0x00D353, 0x00D354, 0x00D355, 0x00D356, 0x00D357, 0x00D358, 0x00D359, 0x00D35A, 0x00D35B, 0x00D35C, 0x00D35D, 0x00D35E, 0x00D35F,
	0x00D360, 0x00D361, 0x00D362, 0x00D363, 0x00D364, 0x00D365, 0x00D366, 0x00D367, 0x00D368, 0x00D369, 0x00D36A, 0x00D36B, 0x00D36C, 0x00D36D, 0x00D36E, 0x00D36F, 0x00D370, 0x00D371, 0x00D372, 0x00D373, 0x00D374, 0x00D375, 0x00D376, 0x00D377, 0x00D378, 0x00D379, 0x00D37A, 0x00D37B, 0x00D37C, 0x00D37D, 0x00D37E, 0x00D37F,
	0x00D380, 0x00D381, 0x00D382, 0x00D383, 0x00D384, 0x00D385, 0x00D386, 0x00D387, 0x00D388, 0x00D389, 0x00D38A, 0x00D38B, 0x00D38C, 0x00D38D, 0x00D38E, 0x00D38F, 0x00D390, 0x00D391, 0x00D392, 0x00D393, 0x00D394, 0x00D395, 0x00D396, 0x00D397, 0x00D398, 0x00D399, 0x00D39A, 0x00D39B, 0x00D39C, 0x00D39D, 0x00D39E, 0x00D39F,
	0x00D3A0, 0x00D3A1, 0x00D3A2, 0x00D3A3, 0x00D3A4, 0x00D3A5, 0x00D3A6, 0x00D3A7, 0x00D3A8, 0x00D3A9, 0x00D3AA, 0x00D3AB, 0x00D3AC, 0x00D3AD, 0x00D3AE, 0x00D3AF, 0x00D3B0, 0x00D3B1, 0x00D3B2, 0x00D3B3, 0x00D3B4, 0x00D3B5, 0x00D3B6, 0x00D3B7, 0x00D3B8, 0x00D3B9, 0x00D3BA, 0x00D3BB, 0x00D3BC, 0x00D3BD, 0x00D3BE, 0x00D3BF,
	0x00D3C0, 0x00D3C1, 0x00D3C2, 0x00D3C3, 0x00D3C4, 0x00D3C5, 0x00D3C6, 0x00D3C7, 0x00D3C8, 0x00D3C9, 0x00D3CA, 0x00D3CB, 0x00D3CC, 0x00D3CD, 0x00D3CE, 0x00D3CF, 0x00D3D0, 0x00D3D1, 0x00D3D2, 0x00D3D3, 0x00D3D4, 0x00D3D5, 0x00D3D6, 0x00D3D7, 0x00D3D8, 0x00D3D9, 0x00D3DA, 0x00D3DB, 0x00D3DC, 0x00D3DD, 0x00D3DE, 0x00D3DF,
	0x00D3E0, 0x00D3E1, 0x00D3E2, 0x00D3E3, 0x00D3E4, 0x00D3E5, 0x00D3E6, 0x00D3E7, 0x00D3E8, 0x00D3E9, 0x00D3EA, 0x00D3EB, 0x00D3EC, 0x00D3ED, 0x00D3EE, 0x00D3EF, 0x00D3F0, 0x00D3F1, 0x00D3F2, 0x00D3F3, 0x00D3F4, 0x00D3F5, 0x00D3F6, 0x00D3F7, 0x00D3F8, 0x00D3F9, 0x00D3FA, 0x00D3FB, 0x00D3FC, 0x00D3FD, 0x00D3FE, 0x00D3FF,
	0
};

static const unsigned map_hangul_11[] = {
	0x00D400, 0x00D401, 0x00D402, 0x00D403, 0x00D404, 0x00D405, 0x00D406, 0x00D407, 0x00D408, 0x00D409, 0x00D40A, 0x00D40B, 0x00D40C, 0x00D40D, 0x00D40E, 0x00D40F, 0x00D410, 0x00D411, 0x00D412, 0x00D413, 0x00D414, 0x00D415, 0x00D416, 0x00D417, 0x00D418, 0x00D419, 0x00D41A, 0x00D41B, 0x00D41C, 0x00D41D, 0x00D41E, 0x00D41F,
	0x00D420, 0x00D421, 0x00D422, 0x00D423, 0x00D424, 0x00D425, 0x00D426, 0x00D427, 0x00D428, 0x00D429, 0x00D42A, 0x00D42B, 0x00D42C, 0x00D42D, 0x00D42E, 0x00D42F, 0x00D430, 0x00D431, 0x00D432, 0x00D433, 0x00D434, 0x00D435, 0x00D436, 0x00D437, 0x00D438, 0x00D439, 0x00D43A, 0x00D43B, 0x00D43C, 0x00D43D, 0x00D43E, 0x00D43F,
	0x00D440, 0x00D441, 0x00D442, 0x00D443, 0x00D444, 0x00D445, 0x00D446, 0x00D447, 0x00D448, 0x00D449, 0x00D44A, 0x00D44B, 0x00D44C, 0x00D44D, 0x00D44E, 0x00D44F, 0x00D450, 0x00D451, 0x00D452, 0x00D453, 0x00D454, 0x00D455, 0x00D456, 0x00D457, 0x00D458, 0x00D459, 0x00D45A, 0x00D45B, 0x00D45C, 0x00D45D, 0x00D45E, 0x00D45F,
	0x00D460, 0x00D461, 0x00D462, 0x00D463, 0x00D464, 0x00D465, 0x00D466, 0x00D467, 0x00D468, 0x00D469, 0x00D46A, 0x00D46B, 0x00D46C, 0x00D46D, 0x00D46E, 0x00D46F, 0x00D470, 0x00D471, 0x00D472, 0x00D473, 0x00D474, 0x00D475, 0x00D476, 0x00D477, 0x00D478, 0x00D479, 0x00D47A, 0x00D47B, 0x00D47C, 0x00D47D, 0x00D47E, 0x00D47F,
	0x00D480, 0x00D481, 0x00D482, 0x00D483, 0x00D484, 0x00D485, 0x00D486, 0x00D487, 0x00D488, 0x00D489, 0x00D48A, 0x00D48B, 0x00D48C, 0x00D48D, 0x00D48E, 0x00D48F, 0x00D490, 0x00D491, 0x00D492, 0x00D493, 0x00D494, 0x00D495, 0x00D496, 0x00D497, 0x00D498, 0x00D499, 0x00D49A, 0x00D49B, 0x00D49C, 0x00D49D, 0x00D49E, 0x00D49F,
	0x00D4A0, 0x00D4A1, 0x00D4A2, 0x00D4A3, 0x00D4A4, 0x00D4A5, 0x00D4A6, 0x00D4A7, 0x00D4A8, 0x00D4A9, 0x00D4AA, 0x00D4AB, 0x00D4AC, 0x00D4AD, 0x00D4AE, 0x00D4AF, 0x00D4B0, 0x00D4B1, 0x00D4B2, 0x00D4B3, 0x00D4B4, 0x00D4B5, 0x00D4B6, 0x00D4B7, 0x00D4B8, 0x00D4B9, 0x00D4BA, 0x00D4BB, 0x00D4BC, 0x00D4BD, 0x00D4BE, 0x00D4BF,
	0x00D4C0, 0x00D4C1, 0x00D4C2, 0x00D4C3, 0x00D4C4, 0x00D4C5, 0x00D4C6, 0x00D4C7, 0x00D4C8, 0x00D4C9, 0x00D4CA, 0x00D4CB, 0x00D4CC, 0x00D4CD, 0x00D4CE, 0x00D4CF, 0x00D4D0, 0x00D4D1, 0x00D4D2, 0x00D4D3, 0x00D4D4, 0x00D4D5, 0x00D4D6, 0x00D4D7, 0x00D4D8, 0x00D4D9, 0x00D4DA, 0x00D4DB, 0x00D4DC, 0x00D4DD, 0x00D4DE, 0x00D4DF,
	0x00D4E0, 0x00D4E1, 0x00D4E2, 0x00D4E3, 0x00D4E4, 0x00D4E5, 0x00D4E6, 0x00D4E7, 0x00D4E8, 0x00D4E9, 0x00D4EA, 0x00D4EB, 0x00D4EC, 0x00D4ED, 0x00D4EE, 0x00D4EF, 0x00D4F0, 0x00D4F1, 0x00D4F2, 0x00D4F3, 0x00D4F4, 0x00D4F5, 0x00D4F6, 0x00D4F7, 0x00D4F8, 0x00D4F9, 0x00D4FA, 0x00D4FB, 0x00D4FC, 0x00D4FD, 0x00D4FE, 0x00D4FF,
	0x00D500, 0x00D501, 0x00D502, 0x00D503, 0x00D504, 0x00D505, 0x00D506, 0x00D507, 0x00D508, 0x00D509, 0x00D50A, 0x00D50B, 0x00D50C, 0x00D50D, 0x00D50E, 0x00D50F, 0x00D510, 0x00D511, 0x00D512, 0x00D513, 0x00D514, 0x00D515, 0x00D516, 0x00D517, 0x00D518, 0x00D519, 0x00D51A, 0x00D51B, 0x00D51C, 0x00D51D, 0x00D51E, 0x00D51F,
	0x00D520, 0x00D521, 0x00D522, 0x00D523, 0x00D524, 0x00D525, 0x00D526, 0x00D527, 0x00D528, 0x00D529, 0x00D52A, 0x00D52B, 0x00D52C, 0x00D52D, 0x00D52E, 0x00D52F, 0x00D530, 0x00D531, 0x00D532, 0x00D533, 0x00D534, 0x00D535, 0x00D536, 0x00D537, 0x00D538, 0x00D539, 0x00D53A, 0x00D53B, 0x00D53C, 0x00D53D, 0x00D53E, 0x00D53F,
	0x00D540, 0x00D541, 0x00D542, 0x00D543, 0x00D544, 0x00D545, 0x00D546, 0x00D547, 0x00D548, 0x00D549, 0x00D54A, 0x00D54B, 0x00D54C, 0x00D54D, 0x00D54E, 0x00D54F, 0x00D550, 0x00D551, 0x00D552, 0x00D553, 0x00D554, 0x00D555, 0x00D556, 0x00D557, 0x00D558, 0x00D559, 0x00D55A, 0x00D55B, 0x00D55C, 0x00D55D, 0x00D55E, 0x00D55F,
	0x00D560, 0x00D561, 0x00D562, 0x00D563, 0x00D564, 0x00D565, 0x00D566, 0x00D567, 0x00D568, 0x00D569, 0x00D56A, 0x00D56B, 0x00D56C, 0x00D56D, 0x00D56E, 0x00D56F, 0x00D570, 0x00D571, 0x00D572, 0x00D573, 0x00D574, 0x00D575, 0x00D576, 0x00D577, 0x00D578, 0x00D579, 0x00D57A, 0x00D57B, 0x00D57C, 0x00D57D, 0x00D57E, 0x00D57F,
	0x00D580, 0x00D581, 0x00D582, 0x00D583, 0x00D584, 0x00D585, 0x00D586, 0x00D587, 0x00D588, 0x00D589, 0x00D58A, 0x00D58B, 0x00D58C, 0x00D58D, 0x00D58E, 0x00D58F, 0x00D590, 0x00D591, 0x00D592, 0x00D593, 0x00D594, 0x00D595, 0x00D596, 0x00D597, 0x00D598, 0x00D599, 0x00D59A, 0x00D59B, 0x00D59C, 0x00D59D, 0x00D59E, 0x00D59F,
	0x00D5A0, 0x00D5A1, 0x00D5A2, 0x00D5A3, 0x00D5A4, 0x00D5A5, 0x00D5A6, 0x00D5A7, 0x00D5A8, 0x00D5A9, 0x00D5AA, 0x00D5AB, 0x00D5AC, 0x00D5AD, 0x00D5AE, 0x00D5AF, 0x00D5B0, 0x00D5B1, 0x00D5B2, 0x00D5B3, 0x00D5B4, 0x00D5B5, 0x00D5B6, 0x00D5B7, 0x00D5B8, 0x00D5B9, 0x00D5BA, 0x00D5BB, 0x00D5BC, 0x00D5BD, 0x00D5BE, 0x00D5BF,
	0x00D5C0, 0x00D5C1, 0x00D5C2, 0x00D5C3, 0x00D5C4, 0x00D5C5, 0x00D5C6, 0x00D5C7, 0x00D5C8, 0x00D5C9, 0x00D5CA, 0x00D5CB, 0x00D5CC, 0x00D5CD, 0x00D5CE, 0x00D5CF, 0x00D5D0, 0x00D5D1, 0x00D5D2, 0x00D5D3, 0x00D5D4, 0x00D5D5, 0x00D5D6, 0x00D5D7, 0x00D5D8, 0x00D5D9, 0x00D5DA, 0x00D5DB, 0x00D5DC, 0x00D5DD, 0x00D5DE, 0x00D5DF,
	0x00D5E0, 0x00D5E1, 0x00D5E2, 0x00D5E3, 0x00D5E4, 0x00D5E5, 0x00D5E6, 0x00D5E7, 0x00D5E8, 0x00D5E9, 0x00D5EA, 0x00D5EB, 0x00D5EC, 0x00D5ED, 0x00D5EE, 0x00D5EF, 0x00D5F0, 0x00D5F1, 0x00D5F2, 0x00D5F3, 0x00D5F4, 0x00D5F5, 0x00D5F6, 0x00D5F7, 0x00D5F8, 0x00D5F9, 0x00D5FA, 0x00D5FB, 0x00D5FC, 0x00D5FD, 0x00D5FE, 0x00D5FF,
	0x00D600, 0x00D601, 0x00D602, 0x00D603, 0x00D604, 0x00D605, 0x00D606, 0x00D607, 0x00D608, 0x00D609, 0x00D60A, 0x00D60B, 0x00D60C, 0x00D60D, 0x00D60E, 0x00D60F, 0x00D610, 0x00D611, 0x00D612, 0x00D613, 0x00D614, 0x00D615, 0x00D616, 0x00D617, 0x00D618, 0x00D619, 0x00D61A, 0x00D61B, 0x00D61C, 0x00D61D, 0x00D61E, 0x00D61F,
	0x00D620, 0x00D621, 0x00D622, 0x00D623, 0x00D624, 0x00D625, 0x00D626, 0x00D627, 0x00D628, 0x00D629, 0x00D62A, 0x00D62B, 0x00D62C, 0x00D62D, 0x00D62E, 0x00D62F, 0x00D630, 0x00D631, 0x00D632, 0x00D633, 0x00D634, 0x00D635, 0x00D636, 0x00D637, 0x00D638, 0x00D639, 0x00D63A, 0x00D63B, 0x00D63C, 0x00D63D, 0x00D63E, 0x00D63F,
	0x00D640, 0x00D641, 0x00D642, 0x00D643, 0x00D644, 0x00D645, 0x00D646, 0x00D647, 0x00D648, 0x00D649, 0x00D64A, 0x00D64B, 0x00D64C, 0x00D64D, 0x00D64E, 0x00D64F, 0x00D650, 0x00D651, 0x00D652, 0x00D653, 0x00D654, 0x00D655, 0x00D656, 0x00D657, 0x00D658, 0x00D659, 0x00D65A, 0x00D65B, 0x00D65C, 0x00D65D, 0x00D65E, 0x00D65F,
	0x00D660, 0x00D661, 0x00D662, 0x00D663, 0x00D664, 0x00D665, 0x00D666, 0x00D667, 0x00D668, 0x00D669, 0x00D66A, 0x00D66B, 0x00D66C, 0x00D66D, 0x00D66E, 0x00D66F, 0x00D670, 0x00D671, 0x00D672, 0x00D673, 0x00D674, 0x00D675, 0x00D676, 0x00D677, 0x00D678, 0x00D679, 0x00D67A, 0x00D67B, 0x00D67C, 0x00D67D, 0x00D67E, 0x00D67F,
	0x00D680, 0x00D681, 0x00D682, 0x00D683, 0x00D684, 0x00D685, 0x00D686, 0x00D687, 0x00D688, 0x00D689, 0x00D68A, 0x00D68B, 0x00D68C, 0x00D68D, 0x00D68E, 0x00D68F, 0x00D690, 0x00D691, 0x00D692, 0x00D693, 0x00D694, 0x00D695, 0x00D696, 0x00D697, 0x00D698, 0x00D699, 0x00D69A, 0x00D69B, 0x00D69C, 0x00D69D, 0x00D69E, 0x00D69F,
	0x00D6A0, 0x00D6A1, 0x00D6A2, 0x00D6A3, 0x00D6A4, 0x00D6A5, 0x00D6A6, 0x00D6A7, 0x00D6A8, 0x00D6A9, 0x00D6AA, 0x00D6AB, 0x00D6AC, 0x00D6AD, 0x00D6AE, 0x00D6AF, 0x00D6B0, 0x00D6B1, 0x00D6B2, 0x00D6B3, 0x00D6B4, 0x00D6B5, 0x00D6B6, 0x00D6B7, 0x00D6B8, 0x00D6B9, 0x00D6BA, 0x00D6BB, 0x00D6BC, 0x00D6BD, 0x00D6BE, 0x00D6BF,
	0x00D6C0, 0x00D6C1, 0x00D6C2, 0x00D6C3, 0x00D6C4, 0x00D6C5, 0x00D6C6, 0x00D6C7, 0x00D6C8, 0x00D6C9, 0x00D6CA, 0x00D6CB, 0x00D6CC, 0x00D6CD, 0x00D6CE, 0x00D6CF, 0x00D6D0, 0x00D6D1, 0x00D6D2, 0x00D6D3, 0x00D6D4, 0x00D6D5, 0x00D6D6, 0x00D6D7, 0x00D6D8, 0x00D6D9, 0x00D6DA, 0x00D6DB, 0x00D6DC, 0x00D6DD, 0x00D6DE, 0x00D6DF,
	0x00D6E0, 0x00D6E1, 0x00D6E2, 0x00D6E3, 0x00D6E4, 0x00D6E5, 0x00D6E6, 0x00D6E7, 0x00D6E8, 0x00D6E9, 0x00D6EA, 0x00D6EB, 0x00D6EC, 0x00D6ED, 0x00D6EE, 0x00D6EF, 0x00D6F0, 0x00D6F1, 0x00D6F2, 0x00D6F3, 0x00D6F4, 0x00D6F5, 0x00D6F6, 0x00D6F7, 0x00D6F8, 0x00D6F9, 0x00D6FA, 0x00D6FB, 0x00D6FC, 0x00D6FD, 0x00D6FE, 0x00D6FF,
	0x00D700, 0x00D701, 0x00D702, 0x00D703, 0x00D704, 0x00D705, 0x00D706, 0x00D707, 0x00D708, 0x00D709, 0x00D70A, 0x00D70B, 0x00D70C, 0x00D70D, 0x00D70E, 0x00D70F, 0x00D710, 0x00D711, 0x00D712, 0x00D713, 0x00D714, 0x00D715, 0x00D716, 0x00D717, 0x00D718, 0x00D719, 0x00D71A, 0x00D71B, 0x00D71C, 0x00D71D, 0x00D71E, 0x00D71F,
	0x00D720, 0x00D721, 0x00D722, 0x00D723, 0x00D724, 0x00D725, 0x00D726, 0x00D727, 0x00D728, 0x00D729, 0x00D72A, 0x00D72B, 0x00D72C, 0x00D72D, 0x00D72E, 0x00D72F, 0x00D730, 0x00D731, 0x00D732, 0x00D733, 0x00D734, 0x00D735, 0x00D736, 0x00D737, 0x00D738, 0x00D739, 0x00D73A, 0x00D73B, 0x00D73C, 0x00D73D, 0x00D73E, 0x00D73F,
	0x00D740, 0x00D741, 0x00D742, 0x00D743, 0x00D744, 0x00D745, 0x00D746, 0x00D747, 0x00D748, 0x00D749, 0x00D74A, 0x00D74B, 0x00D74C, 0x00D74D, 0x00D74E, 0x00D74F, 0x00D750, 0x00D751, 0x00D752, 0x00D753, 0x00D754, 0x00D755, 0x00D756, 0x00D757, 0x00D758, 0x00D759, 0x00D75A, 0x00D75B, 0x00D75C, 0x00D75D, 0x00D75E, 0x00D75F,
	0x00D760, 0x00D761, 0x00D762, 0x00D763, 0x00D764, 0x00D765, 0x00D766, 0x00D767, 0x00D768, 0x00D769, 0x00D76A, 0x00D76B, 0x00D76C, 0x00D76D, 0x00D76E, 0x00D76F, 0x00D770, 0x00D771, 0x00D772, 0x00D773, 0x00D774, 0x00D775, 0x00D776, 0x00D777, 0x00D778, 0x00D779, 0x00D77A, 0x00D77B, 0x00D77C, 0x00D77D, 0x00D77E, 0x00D77F,
	0x00D780, 0x00D781, 0x00D782, 0x00D783, 0x00D784, 0x00D785, 0x00D786, 0x00D787, 0x00D788, 0x00D789, 0x00D78A, 0x00D78B, 0x00D78C, 0x00D78D, 0x00D78E, 0x00D78F, 0x00D790, 0x00D791, 0x00D792, 0x00D793, 0x00D794, 0x00D795, 0x00D796, 0x00D797, 0x00D798, 0x00D799, 0x00D79A, 0x00D79B, 0x00D79C, 0x00D79D, 0x00D79E, 0x00D79F,
	0x00D7A0, 0x00D7A1, 0x00D7A2, 0x00D7A3,
	0
};

// KS X 1001 Symbol
static const unsigned map_symbol_ks[] = {
	0x00223C, 0x0002D0, 0x0025C8, 0x0025A3, 0x002592, 0x0025A4, 0x0025A5, 0x0025A8, 0x0025A7, 0x0025A6, 0x0025A9, 0x00260F, 0x00261C, 0x002195, 0x00327F, 0x00321C,
	0x0033C7, 0x0033C2, 0x0033D8, 0x00FFE6, 0x0003C6, 0x003395, 0x003396, 0x003397, 0x003398, 0x0033A3, 0x0033A4, 0x0033A5, 0x0033A6, 0x003399, 0x00339A, 0x00339B,
	0x00339F, 0x0033A0, 0x0033A2, 0x0033CA, 0x00338D, 0x0033CF, 0x003388, 0x003389, 0x0033C8, 0x0033A7, 0x0033A8, 0x0033B0, 0x0033B1, 0x0033B2, 0x0033B3, 0x0033B4,
	0x0033B5, 0x0033B6, 0x0033B7, 0x0033B8, 0x0033B9, 0x003380, 0x003381, 0x003382, 0x003383, 0x003384, 0x0033BA, 0x0033BB, 0x0033BC, 0x0033BD, 0x0033BE, 0x0033BF,
	0x003390, 0x003391, 0x003392, 0x003393, 0x003394, 0x002126, 0x0033C0, 0x0033C1, 0x00338A, 0x00338B, 0x00338C, 0x0033D6, 0x0033C5, 0x0033AD, 0x0033AE, 0x0033AF,
	0x0033DB, 0x0033A9, 0x0033AA, 0x0033AB, 0x0033AC, 0x0033DD, 0x0033D0, 0x0033D3, 0x0033C3, 0x0033C9, 0x0033DC, 0x0033C6, 0x003260, 0x003261, 0x003262, 0x003263,
	0x003264, 0x003265, 0x003266, 0x003267, 0x003268, 0x003269, 0x00326A, 0x00326B, 0x00326C, 0x00326D, 0x00326E, 0x00326F, 0x003270, 0x003271, 0x003272, 0x003273,
	0x003274, 0x003275, 0x003276, 0x003277, 0x003278, 0x003279, 0x00327A, 0x00327B, 0x00215B, 0x00215C, 0x00215D, 0x00215E, 0x003200, 0x003201, 0x003202, 0x003203,
	0x003204, 0x003205, 0x003206, 0x003207, 0x003208, 0x003209, 0x00320A, 0x00320B, 0x00320C, 0x00320D, 0x00320E, 0x00320F, 0x003210, 0x003211, 0x003212, 0x003213,
	0x003214, 0x003215, 0x003216, 0x003217, 0x003218, 0x003219, 0x00321A, 0x00321B, 0x00249C, 0x00249D, 0x00249E, 0x00249F, 0x0024A0, 0x0024A1, 0x0024A2, 0x0024A3,
	0x0024A4, 0x0024A5, 0x0024A6, 0x0024A7, 0x0024A8, 0x0024A9, 0x0024AA, 0x0024AB, 0x0024AC, 0x0024AD, 0x0024AE, 0x0024AF, 0x0024B0, 0x0024B1, 0x0024B2, 0x0024B3,
	0x0024B4, 0x0024B5, 0x002074, 0x00207F, 0x002081, 0x002082, 0x002083, 0x002084,
	0
};

// Compatibility Ideograph for KS X 1001, NOT INCLUDING JIS X 0213 CHARACTERS
static const unsigned map_kanji_ks[] = {
	0x00F900, 0x00F901, 0x00F902, 0x00F903, 0x00F904, 0x00F905, 0x00F906, 0x00F907, 0x00F908, 0x00F909, 0x00F90A, 0x00F90B, 0x00F90C, 0x00F90D, 0x00F90E, 0x00F90F,
	0x00F910, 0x00F911, 0x00F912, 0x00F913, 0x00F914, 0x00F915, 0x00F916, 0x00F917, 0x00F918, 0x00F919, 0x00F91A, 0x00F91B, 0x00F91C, 0x00F91E, 0x00F91F, 0x00F920,
	0x00F921, 0x00F922, 0x00F923, 0x00F924, 0x00F925, 0x00F926, 0x00F927, 0x00F92A, 0x00F92B, 0x00F92D, 0x00F92E, 0x00F92F, 0x00F930, 0x00F931, 0x00F932, 0x00F933,
	0x00F934, 0x00F935, 0x00F937, 0x00F938, 0x00F939, 0x00F93A, 0x00F93B, 0x00F93C, 0x00F93D, 0x00F93E, 0x00F93F, 0x00F940, 0x00F941, 0x00F942, 0x00F943, 0x00F944,
	0x00F945, 0x00F946, 0x00F947, 0x00F948, 0x00F949, 0x00F94A, 0x00F94B, 0x00F94C, 0x00F94D, 0x00F94E, 0x00F94F, 0x00F950, 0x00F951, 0x00F952, 0x00F953, 0x00F954,
	0x00F955, 0x00F956, 0x00F957, 0x00F958, 0x00F959, 0x00F95A, 0x00F95B, 0x00F95C, 0x00F95D, 0x00F95E, 0x00F95F, 0x00F960, 0x00F961, 0x00F962, 0x00F963, 0x00F964,
	0x00F965, 0x00F966, 0x00F967, 0x00F968, 0x00F969, 0x00F96A, 0x00F96B, 0x00F96C, 0x00F96D, 0x00F96E, 0x00F96F, 0x00F971, 0x00F972, 0x00F973, 0x00F974, 0x00F975,
	0x00F976, 0x00F977, 0x00F978, 0x00F97A, 0x00F97B, 0x00F97C, 0x00F97D, 0x00F97E, 0x00F97F, 0x00F980, 0x00F981, 0x00F982, 0x00F983, 0x00F984, 0x00F985, 0x00F986,
	0x00F987, 0x00F988, 0x00F989, 0x00F98A, 0x00F98B, 0x00F98C, 0x00F98D, 0x00F98E, 0x00F98F, 0x00F990, 0x00F991, 0x00F992, 0x00F993, 0x00F994, 0x00F996, 0x00F997,
	0x00F998, 0x00F999, 0x00F99A, 0x00F99B, 0x00F99C, 0x00F99D, 0x00F99E, 0x00F99F, 0x00F9A0, 0x00F9A1, 0x00F9A2, 0x00F9A3, 0x00F9A4, 0x00F9A5, 0x00F9A6, 0x00F9A7,
	0x00F9A8, 0x00F9A9, 0x00F9AA, 0x00F9AB, 0x00F9AC, 0x00F9AD, 0x00F9AE, 0x00F9AF, 0x00F9B0, 0x00F9B1, 0x00F9B2, 0x00F9B3, 0x00F9B4, 0x00F9B5, 0x00F9B6, 0x00F9B7,
	0x00F9B8, 0x00F9B9, 0x00F9BA, 0x00F9BB, 0x00F9BC, 0x00F9BD, 0x00F9BE, 0x00F9BF, 0x00F9C0, 0x00F9C1, 0x00F9C2, 0x00F9C3, 0x00F9C4, 0x00F9C5, 0x00F9C6, 0x00F9C7,
	0x00F9C8, 0x00F9C9, 0x00F9CA, 0x00F9CB, 0x00F9CC, 0x00F9CD, 0x00F9CE, 0x00F9CF, 0x00F9D1, 0x00F9D2, 0x00F9D3, 0x00F9D4, 0x00F9D5, 0x00F9D6, 0x00F9D7, 0x00F9D8,
	0x00F9D9, 0x00F9DA, 0x00F9DB, 0x00F9DD, 0x00F9DE, 0x00F9DF, 0x00F9E0, 0x00F9E1, 0x00F9E2, 0x00F9E3, 0x00F9E4, 0x00F9E5, 0x00F9E6, 0x00F9E8, 0x00F9E9, 0x00F9EA,
	0x00F9EB, 0x00F9EC, 0x00F9ED, 0x00F9EE, 0x00F9EF, 0x00F9F0, 0x00F9F2, 0x00F9F3, 0x00F9F4, 0x00F9F5, 0x00F9F6, 0x00F9F7, 0x00F9F8, 0x00F9F9, 0x00F9FA, 0x00F9FB,
	0x00F9FC, 0x00F9FD, 0x00F9FE, 0x00F9FF, 0x00FA00, 0x00FA01, 0x00FA02, 0x00FA03, 0x00FA04, 0x00FA05, 0x00FA06, 0x00FA07, 0x00FA08, 0x00FA09, 0x00FA0A, 0x00FA0B,
	0
};

// Compatibility Ideograph for CNS 11643, Unicode 3.1
// NOT USED DUE TO STEPMANIA'S SPECIFICATION.
static const unsigned map_kanji_cns[] = {
	0x02F800, 0x02F801, 0x02F802, 0x02F803, 0x02F804, 0x02F805, 0x02F806, 0x02F807, 0x02F808, 0x02F809, 0x02F80A, 0x02F80B, 0x02F80C, 0x02F80D, 0x02F80E, 0x02F80F, 0x02F810, 0x02F811, 0x02F812, 0x02F813, 0x02F814, 0x02F815, 0x02F816, 0x02F817, 0x02F818, 0x02F819, 0x02F81A, 0x02F81B, 0x02F81C, 0x02F81D, 0x02F81E, 0x02F81F,
	0x02F820, 0x02F821, 0x02F822, 0x02F823, 0x02F824, 0x02F825, 0x02F826, 0x02F827, 0x02F828, 0x02F829, 0x02F82A, 0x02F82B, 0x02F82C, 0x02F82D, 0x02F82E, 0x02F82F, 0x02F830, 0x02F831, 0x02F832, 0x02F833, 0x02F834, 0x02F835, 0x02F836, 0x02F837, 0x02F838, 0x02F839, 0x02F83A, 0x02F83B, 0x02F83C, 0x02F83D, 0x02F83E, 0x02F83F,
	0x02F840, 0x02F841, 0x02F842, 0x02F843, 0x02F844, 0x02F845, 0x02F846, 0x02F847, 0x02F848, 0x02F849, 0x02F84A, 0x02F84B, 0x02F84C, 0x02F84D, 0x02F84E, 0x02F84F, 0x02F850, 0x02F851, 0x02F852, 0x02F853, 0x02F854, 0x02F855, 0x02F856, 0x02F857, 0x02F858, 0x02F859, 0x02F85A, 0x02F85B, 0x02F85C, 0x02F85D, 0x02F85E, 0x02F85F,
	0x02F860, 0x02F861, 0x02F862, 0x02F863, 0x02F864, 0x02F865, 0x02F866, 0x02F867, 0x02F868, 0x02F869, 0x02F86A, 0x02F86B, 0x02F86C, 0x02F86D, 0x02F86E, 0x02F86F, 0x02F870, 0x02F871, 0x02F872, 0x02F873, 0x02F874, 0x02F875, 0x02F876, 0x02F877, 0x02F878, 0x02F879, 0x02F87A, 0x02F87B, 0x02F87C, 0x02F87D, 0x02F87E, 0x02F87F,
	0x02F880, 0x02F881, 0x02F882, 0x02F883, 0x02F884, 0x02F885, 0x02F886, 0x02F887, 0x02F888, 0x02F889, 0x02F88A, 0x02F88B, 0x02F88C, 0x02F88D, 0x02F88E, 0x02F88F, 0x02F890, 0x02F891, 0x02F892, 0x02F893, 0x02F894, 0x02F895, 0x02F896, 0x02F897, 0x02F898, 0x02F899, 0x02F89A, 0x02F89B, 0x02F89C, 0x02F89D, 0x02F89E, 0x02F89F,
	0x02F8A0, 0x02F8A1, 0x02F8A2, 0x02F8A3, 0x02F8A4, 0x02F8A5, 0x02F8A6, 0x02F8A7, 0x02F8A8, 0x02F8A9, 0x02F8AA, 0x02F8AB, 0x02F8AC, 0x02F8AD, 0x02F8AE, 0x02F8AF, 0x02F8B0, 0x02F8B1, 0x02F8B2, 0x02F8B3, 0x02F8B4, 0x02F8B5, 0x02F8B6, 0x02F8B7, 0x02F8B8, 0x02F8B9, 0x02F8BA, 0x02F8BB, 0x02F8BC, 0x02F8BD, 0x02F8BE, 0x02F8BF,
	0x02F8C0, 0x02F8C1, 0x02F8C2, 0x02F8C3, 0x02F8C4, 0x02F8C5, 0x02F8C6, 0x02F8C7, 0x02F8C8, 0x02F8C9, 0x02F8CA, 0x02F8CB, 0x02F8CC, 0x02F8CD, 0x02F8CE, 0x02F8CF, 0x02F8D0, 0x02F8D1, 0x02F8D2, 0x02F8D3, 0x02F8D4, 0x02F8D5, 0x02F8D6, 0x02F8D7, 0x02F8D8, 0x02F8D9, 0x02F8DA, 0x02F8DB, 0x02F8DC, 0x02F8DD, 0x02F8DE, 0x02F8DF,
	0x02F8E0, 0x02F8E1, 0x02F8E2, 0x02F8E3, 0x02F8E4, 0x02F8E5, 0x02F8E6, 0x02F8E7, 0x02F8E8, 0x02F8E9, 0x02F8EA, 0x02F8EB, 0x02F8EC, 0x02F8ED, 0x02F8EE, 0x02F8EF, 0x02F8F0, 0x02F8F1, 0x02F8F2, 0x02F8F3, 0x02F8F4, 0x02F8F5, 0x02F8F6, 0x02F8F7, 0x02F8F8, 0x02F8F9, 0x02F8FA, 0x02F8FB, 0x02F8FC, 0x02F8FD, 0x02F8FE, 0x02F8FF,
	0x02F900, 0x02F901, 0x02F902, 0x02F903, 0x02F904, 0x02F905, 0x02F906, 0x02F907, 0x02F908, 0x02F909, 0x02F90A, 0x02F90B, 0x02F90C, 0x02F90D, 0x02F90E, 0x02F90F, 0x02F910, 0x02F911, 0x02F912, 0x02F913, 0x02F914, 0x02F915, 0x02F916, 0x02F917, 0x02F918, 0x02F919, 0x02F91A, 0x02F91B, 0x02F91C, 0x02F91D, 0x02F91E, 0x02F91F,
	0x02F920, 0x02F921, 0x02F922, 0x02F923, 0x02F924, 0x02F925, 0x02F926, 0x02F927, 0x02F928, 0x02F929, 0x02F92A, 0x02F92B, 0x02F92C, 0x02F92D, 0x02F92E, 0x02F92F, 0x02F930, 0x02F931, 0x02F932, 0x02F933, 0x02F934, 0x02F935, 0x02F936, 0x02F937, 0x02F938, 0x02F939, 0x02F93A, 0x02F93B, 0x02F93C, 0x02F93D, 0x02F93E, 0x02F93F,
	0x02F940, 0x02F941, 0x02F942, 0x02F943, 0x02F944, 0x02F945, 0x02F946, 0x02F947, 0x02F948, 0x02F949, 0x02F94A, 0x02F94B, 0x02F94C, 0x02F94D, 0x02F94E, 0x02F94F, 0x02F950, 0x02F951, 0x02F952, 0x02F953, 0x02F954, 0x02F955, 0x02F956, 0x02F957, 0x02F958, 0x02F959, 0x02F95A, 0x02F95B, 0x02F95C, 0x02F95D, 0x02F95E, 0x02F95F,
	0x02F960, 0x02F961, 0x02F962, 0x02F963, 0x02F964, 0x02F965, 0x02F966, 0x02F967, 0x02F968, 0x02F969, 0x02F96A, 0x02F96B, 0x02F96C, 0x02F96D, 0x02F96E, 0x02F96F, 0x02F970, 0x02F971, 0x02F972, 0x02F973, 0x02F974, 0x02F975, 0x02F976, 0x02F977, 0x02F978, 0x02F979, 0x02F97A, 0x02F97B, 0x02F97C, 0x02F97D, 0x02F97E, 0x02F97F,
	0x02F980, 0x02F981, 0x02F982, 0x02F983, 0x02F984, 0x02F985, 0x02F986, 0x02F987, 0x02F988, 0x02F989, 0x02F98A, 0x02F98B, 0x02F98C, 0x02F98D, 0x02F98E, 0x02F98F, 0x02F990, 0x02F991, 0x02F992, 0x02F993, 0x02F994, 0x02F995, 0x02F996, 0x02F997, 0x02F998, 0x02F999, 0x02F99A, 0x02F99B, 0x02F99C, 0x02F99D, 0x02F99E, 0x02F99F,
	0x02F9A0, 0x02F9A1, 0x02F9A2, 0x02F9A3, 0x02F9A4, 0x02F9A5, 0x02F9A6, 0x02F9A7, 0x02F9A8, 0x02F9A9, 0x02F9AA, 0x02F9AB, 0x02F9AC, 0x02F9AD, 0x02F9AE, 0x02F9AF, 0x02F9B0, 0x02F9B1, 0x02F9B2, 0x02F9B3, 0x02F9B4, 0x02F9B5, 0x02F9B6, 0x02F9B7, 0x02F9B8, 0x02F9B9, 0x02F9BA, 0x02F9BB, 0x02F9BC, 0x02F9BD, 0x02F9BE, 0x02F9BF,
	0x02F9C0, 0x02F9C1, 0x02F9C2, 0x02F9C3, 0x02F9C4, 0x02F9C5, 0x02F9C6, 0x02F9C7, 0x02F9C8, 0x02F9C9, 0x02F9CA, 0x02F9CB, 0x02F9CC, 0x02F9CD, 0x02F9CE, 0x02F9CF, 0x02F9D0, 0x02F9D1, 0x02F9D2, 0x02F9D3, 0x02F9D4, 0x02F9D5, 0x02F9D6, 0x02F9D7, 0x02F9D8, 0x02F9D9, 0x02F9DA, 0x02F9DB, 0x02F9DC, 0x02F9DD, 0x02F9DE, 0x02F9DF,
	0x02F9E0, 0x02F9E1, 0x02F9E2, 0x02F9E3, 0x02F9E4, 0x02F9E5, 0x02F9E6, 0x02F9E7, 0x02F9E8, 0x02F9E9, 0x02F9EA, 0x02F9EB, 0x02F9EC, 0x02F9ED, 0x02F9EE, 0x02F9EF, 0x02F9F0, 0x02F9F1, 0x02F9F2, 0x02F9F3, 0x02F9F4, 0x02F9F5, 0x02F9F6, 0x02F9F7, 0x02F9F8, 0x02F9F9, 0x02F9FA, 0x02F9FB, 0x02F9FC, 0x02F9FD, 0x02F9FE, 0x02F9FF,
	0x02FA00, 0x02FA01, 0x02FA02, 0x02FA03, 0x02FA04, 0x02FA05, 0x02FA06, 0x02FA07, 0x02FA08, 0x02FA09, 0x02FA0A, 0x02FA0B, 0x02FA0C, 0x02FA0D, 0x02FA0E, 0x02FA0F, 0x02FA10, 0x02FA11, 0x02FA12, 0x02FA13, 0x02FA14, 0x02FA15, 0x02FA16, 0x02FA17, 0x02FA18, 0x02FA19, 0x02FA1A, 0x02FA1B, 0x02FA1C, 0x02FA1D,
	0
};

// Compatibility Ideograph for KPS 10721, Unicode 4.1
static const unsigned map_kanji_kps[] = {
	0x00FA70, 0x00FA71, 0x00FA72, 0x00FA73, 0x00FA74, 0x00FA75, 0x00FA76, 0x00FA77, 0x00FA78, 0x00FA79, 0x00FA7A, 0x00FA7B, 0x00FA7C, 0x00FA7D, 0x00FA7E, 0x00FA7F, 
	0x00FA80, 0x00FA81, 0x00FA82, 0x00FA83, 0x00FA84, 0x00FA85, 0x00FA86, 0x00FA87, 0x00FA88, 0x00FA89, 0x00FA8A, 0x00FA8B, 0x00FA8C, 0x00FA8D, 0x00FA8E, 0x00FA8F, 
	0x00FA90, 0x00FA91, 0x00FA92, 0x00FA93, 0x00FA94, 0x00FA95, 0x00FA96, 0x00FA97, 0x00FA98, 0x00FA99, 0x00FA9A, 0x00FA9B, 0x00FA9C, 0x00FA9D, 0x00FA9E, 0x00FA9F, 
	0x00FAA0, 0x00FAA1, 0x00FAA2, 0x00FAA3, 0x00FAA4, 0x00FAA5, 0x00FAA6, 0x00FAA7, 0x00FAA8, 0x00FAA9, 0x00FAAA, 0x00FAAB, 0x00FAAC, 0x00FAAD, 0x00FAAE, 0x00FAAF, 
	0x00FAB0, 0x00FAB1, 0x00FAB2, 0x00FAB3, 0x00FAB4, 0x00FAB5, 0x00FAB6, 0x00FAB7, 0x00FAB8, 0x00FAB9, 0x00FABA, 0x00FABB, 0x00FABC, 0x00FABD, 0x00FABE, 0x00FABF, 
	0x00FAC0, 0x00FAC1, 0x00FAC2, 0x00FAC3, 0x00FAC4, 0x00FAC5, 0x00FAC6, 0x00FAC7, 0x00FAC8, 0x00FAC9, 0x00FACA, 0x00FACB, 0x00FACC, 0x00FACD, 0x00FACE, 0x00FACF, 
	0x00FAD0, 0x00FAD1, 0x00FAD2, 0x00FAD3, 0x00FAD4, 0x00FAD5, 0x00FAD6, 0x00FAD7, 0x00FAD8, 0x00FAD9,
	0
};

// Compatibility Ideograph for ARIB STD-B24, Unicode 5.2
static const unsigned map_kanji_arib[] = {
	0x00FA6B, 0x00FA6C, 0x00FA6D,
	0
};

// Compatibility Ideograph Eratta, IRG
static const unsigned map_kanji_errata[] = {
	0x00FA2E, 0x00FA2F,
	0
};

static const unsigned basic_latin[] = {
	0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027, 0x0028, 0x0029, 0x002A, 0x002B, 0x002C, 0x002D,
	0x002E, 0x002F, 0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037, 0x0038, 0x0039, 0x003A, 0x003B,
	0x003C, 0x003D, 0x003E, 0x003F, 0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047, 0x0048, 0x0049,
	0x004A, 0x004B, 0x004C, 0x004D, 0x004E, 0x004F, 0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057,
	0x0058, 0x0059, 0x005A, 0x005B, 0x005C, 0x005D, 0x005E, 0x005F, 0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065,
	0x0066, 0x0067, 0x0068, 0x0069, 0x006A, 0x006B, 0x006C, 0x006D, 0x006E, 0x006F, 0x0070, 0x0071, 0x0072, 0x0073,
	0x0074, 0x0075, 0x0076, 0x0077, 0x0078, 0x0079, 0x007A, 0x007B, 0x007C, 0x007D, 0x007E, 0x007F,
	0
};

static const unsigned latin_supplement[] = {
	0x00A0, 0x00A1, 0x00A2, 0x00A3, 0x00A4, 0x00A5, 0x00A6, 0x00A7, 0x00A8, 0x00A9, 0x00AA, 0x00AB, 0x00AC, 0x00AD,
	0x00AE, 0x00AF, 0x00B0, 0x00B1, 0x00B2, 0x00B3, 0x00B4, 0x00B5, 0x00B6, 0x00B7, 0x00B8, 0x00B9, 0x00BA, 0x00BB,
	0x00BC, 0x00BD, 0x00BE, 0x00BF, 0x00C0, 0x00C1, 0x00C2, 0x00C3, 0x00C4, 0x00C5, 0x00C6, 0x00C7, 0x00C8, 0x00C9,
	0x00CA, 0x00CB, 0x00CC, 0x00CD, 0x00CE, 0x00CF, 0x00D0, 0x00D1, 0x00D2, 0x00D3, 0x00D4, 0x00D5, 0x00D6, 0x00D7,
	0x00D8, 0x00D9, 0x00DA, 0x00DB, 0x00DC, 0x00DD, 0x00DE, 0x00DF, 0x00E0, 0x00E1, 0x00E2, 0x00E3, 0x00E4, 0x00E5,
	0x00E6, 0x00E7, 0x00E8, 0x00E9, 0x00EA, 0x00EB, 0x00EC, 0x00ED, 0x00EE, 0x00EF, 0x00F0, 0x00F1, 0x00F2, 0x00F3,
	0x00F4, 0x00F5, 0x00F6, 0x00F7, 0x00F8, 0x00F9, 0x00FA, 0x00FB, 0x00FC, 0x00FD, 0x00FE, 0x00FF,
	0
};

static const unsigned latin_extended_a[] = {
	0x0100, 0x0101, 0x0102, 0x0103, 0x0104, 0x0105, 0x0106, 0x0107, 0x0108, 0x0109, 0x010A, 0x010B, 0x010C, 0x010D,
	0x010E, 0x010F, 0x0110, 0x0111, 0x0112, 0x0113, 0x0114, 0x0115, 0x0116, 0x0117, 0x0118, 0x0119, 0x011A, 0x011B,
	0x011C, 0x011D, 0x011E, 0x011F, 0x0120, 0x0121, 0x0122, 0x0123, 0x0124, 0x0125, 0x0126, 0x0127, 0x0128, 0x0129,
	0x012A, 0x012B, 0x012C, 0x012D, 0x012E, 0x012F, 0x0130, 0x0131, 0x0132, 0x0133, 0x0134, 0x0135, 0x0136, 0x0137,
	0x0138, 0x0139, 0x013A, 0x013B, 0x013C, 0x013D, 0x013E, 0x013F, 0x0140, 0x0141, 0x0142, 0x0143, 0x0144, 0x0145,
	0x0146, 0x0147, 0x0148, 0x0149, 0x014A, 0x014B, 0x014C, 0x014D, 0x014E, 0x014F, 0x0150, 0x0151, 0x0152, 0x0153,
	0x0154, 0x0155, 0x0156, 0x0157, 0x0158, 0x0159, 0x015A, 0x015B, 0x015C, 0x015D, 0x015E, 0x015F, 0x0160, 0x0161,
	0x0162, 0x0163, 0x0164, 0x0165, 0x0166, 0x0167, 0x0168, 0x0169, 0x016A, 0x016B, 0x016C, 0x016D, 0x016E, 0x016F,
	0x0170, 0x0171, 0x0172, 0x0173, 0x0174, 0x0175, 0x0176, 0x0177, 0x0178, 0x0179, 0x017A, 0x017B, 0x017C, 0x017D,
	0x017E, 0x017F,
	0
};

static const unsigned latin_extended_b1[] = {
	0x0180, 0x0181, 0x0182, 0x0183, 0x0184, 0x0185, 0x0186, 0x0187, 0x0188, 0x0189, 0x018A, 0x018B, 0x018C, 0x018D,
	0x018E, 0x018F, 0x0190, 0x0191, 0x0192, 0x0193, 0x0194, 0x0195, 0x0196, 0x0197, 0x0198, 0x0199, 0x019A, 0x019B,
	0x019C, 0x019D, 0x019E, 0x019F, 0x01A0, 0x01A1, 0x01A2, 0x01A3, 0x01A4, 0x01A5, 0x01A6, 0x01A7, 0x01A8, 0x01A9,
	0x01AA, 0x01AB, 0x01AC, 0x01AD, 0x01AE, 0x01AF, 0x01B0, 0x01B1, 0x01B2, 0x01B3, 0x01B4, 0x01B5, 0x01B6, 0x01B7,
	0x01B8, 0x01B9, 0x01BA, 0x01BB, 0x01BC, 0x01BD, 0x01BE, 0x01BF, 0x01C0, 0x01C1, 0x01C2, 0x01C3, 0x01C4, 0x01C5,
	0x01C6, 0x01C7, 0x01C8, 0x01C9, 0x01CA, 0x01CB, 0x01CC, 0x01CD, 0x01CE, 0x01CF, 0x01D0, 0x01D1, 0x01D2, 0x01D3,
	0x01D4, 0x01D5, 0x01D6, 0x01D7, 0x01D8, 0x01D9, 0x01DA, 0x01DB, 0x01DC, 0x01DD, 0x01DE, 0x01DF,
	0
};

static const unsigned latin_extended_b2[] = {
	0x01E0, 0x01E1, 0x01E2, 0x01E3, 0x01E4, 0x01E5, 0x01E6, 0x01E7, 0x01E8, 0x01E9, 0x01EA, 0x01EB, 0x01EC, 0x01ED,
	0x01EE, 0x01EF, 0x01F0, 0x01F1, 0x01F2, 0x01F3, 0x01F4, 0x01F5, 0x01F6, 0x01F7, 0x01F8, 0x01F9, 0x01FA, 0x01FB,
	0x01FC, 0x01FD, 0x01FE, 0x01FF, 0x0200, 0x0201, 0x0202, 0x0203, 0x0204, 0x0205, 0x0206, 0x0207, 0x0208, 0x0209,
	0x020A, 0x020B, 0x020C, 0x020D, 0x020E, 0x020F, 0x0210, 0x0211, 0x0212, 0x0213, 0x0214, 0x0215, 0x0216, 0x0217,
	0x0218, 0x0219, 0x021A, 0x021B, 0x021C, 0x021D, 0x021E, 0x021F, 0x0220, 0x0221, 0x0222, 0x0223, 0x0224, 0x0225,
	0x0226, 0x0227, 0x0228, 0x0229, 0x022A, 0x022B, 0x022C, 0x022D, 0x022E, 0x022F, 0x0230, 0x0231, 0x0232, 0x0233,
	0x0234, 0x0235, 0x0236, 0x0237, 0x0238, 0x0239, 0x023A, 0x023B, 0x023C, 0x023D, 0x023E, 0x023F, 0x0240, 0x0241,
	0x0242, 0x0243, 0x0244, 0x0245, 0x0246, 0x0247, 0x0248, 0x0249, 0x024A, 0x024B, 0x024C, 0x024D, 0x024E, 0x024F,
	0
};

static const unsigned combining_diacritical[] = {
	0x0300, 0x0301, 0x0302, 0x0303, 0x0304, 0x0305, 0x0306, 0x0307, 0x0308, 0x0309, 0x030A, 0x030B, 0x030C, 0x030D,
	0x030E, 0x030F, 0x0310, 0x0311, 0x0312, 0x0313, 0x0314, 0x0315, 0x0316, 0x0317, 0x0318, 0x0319, 0x031A, 0x031B,
	0x031C, 0x031D, 0x031E, 0x031F, 0x0320, 0x0321, 0x0322, 0x0323, 0x0324, 0x0325, 0x0326, 0x0327, 0x0328, 0x0329,
	0x032A, 0x032B, 0x032C, 0x032D, 0x032E, 0x032F, 0x0330, 0x0331, 0x0332, 0x0333, 0x0334, 0x0335, 0x0336, 0x0337,
	0x0338, 0x0339, 0x033A, 0x033B, 0x033C, 0x033D, 0x033E, 0x033F, 0x0340, 0x0341, 0x0342, 0x0343, 0x0344, 0x0345,
	0x0346, 0x0347, 0x0348, 0x0349, 0x034A, 0x034B, 0x034C, 0x034D, 0x034E, 0x0360, 0x0361, 0x0362, 0x0363, 0x0364,
	0x0365, 0x0366, 0x0367, 0x0368, 0x0369, 0x036A, 0x036B, 0x036C, 0x036D, 0x036E, 0x036F,
	0
};

static const unsigned greek[] = {
	0x0374, 0x0375, 0x0376, 0x0377, 0x037A, 0x037E, 0x0384, 0x0385, 0x0386, 0x0387, 0x0388, 0x0389, 0x038A, 0x038C,
	0x038E, 0x038F, 0x0390, 0x0391, 0x0392, 0x0393, 0x0394, 0x0395, 0x0396, 0x0397, 0x0398, 0x0399, 0x039A, 0x039B,
	0x039C, 0x039D, 0x039E,	0x039F, 0x03A0, 0x03A1, 0x03A3, 0x03A4, 0x03A5, 0x03A6, 0x03A7, 0x03A8, 0x03A9, 0x03AA,
	0x03AB, 0x03AC, 0x03AD, 0x03AE, 0x03AF, 0x03B0, 0x03B1, 0x03B2, 0x03B3, 0x03B4, 0x03B5, 0x03B6, 0x03B7, 0x03B8,
	0x03B9, 0x03BA, 0x03BB, 0x03BC, 0x03BD, 0x03BE, 0x03BF, 0x03C0, 0x03C1, 0x03C2, 0x03C3, 0x03C4, 0x03C5, 0x03C6,
	0x03C7, 0x03C8, 0x03C9, 0x03CA, 0x03CB, 0x03CC, 0x03CD, 0x03CE, 0x03D0, 0x03D1, 0x03D2, 0x03D3, 0x03D4, 0x03D5,
	0x03D6, 0x03D7, 0x03DA, 0x03DB, 0x03DC, 0x03DD, 0x03DE, 0x03DF, 0x03E0, 0x03E1, 0x03E2, 0x03E3, 0x03E4, 0x03E5,
	0x03E6, 0x03E7, 0x03E8, 0x03E9, 0x03EA, 0x03EB, 0x03EC, 0x03ED, 0x03EE, 0x03EF, 0x03F0, 0x03F1, 0x03F2, 0x03F3,
	0x03F4, 0x03F5,	0x03F6, 0x03F7, 0x03F8, 0x03F9, 0x03FA, 0x03FB, 0x03FC, 0x03FD, 0x03FE, 0x03FF,
	0
};

static const unsigned cyrillic_a[] = {
	0x0400, 0x0401, 0x0402, 0x0403, 0x0404, 0x0405, 0x0406, 0x0407, 0x0408, 0x0409, 0x040A, 0x040B, 0x040C, 0x040D,
	0x040E, 0x040F, 0x0410, 0x0411, 0x0412, 0x0413, 0x0414, 0x0415, 0x0416, 0x0417, 0x0418, 0x0419, 0x041A, 0x041B,
	0x041C, 0x041D, 0x041E, 0x041F, 0x0420, 0x0421, 0x0422, 0x0423, 0x0424, 0x0425, 0x0426, 0x0427, 0x0428, 0x0429,
	0x042A,	0x042B,	0x042C,	0x042D,	0x042E,	0x042F,	0x0430,	0x0431,	0x0432,	0x0433,	0x0434,	0x0435,	0x0436, 0x0437,
	0x0438, 0x0439, 0x043A, 0x043B, 0x043C, 0x043D, 0x043E, 0x043F, 0x0440, 0x0441, 0x0442, 0x0443, 0x0444, 0x0445,
	0x0446, 0x0447, 0x0448, 0x0449, 0x044A, 0x044B, 0x044C, 0x044D, 0x044E, 0x044F, 0x0450, 0x0451, 0x0452, 0x0453,
	0x0454, 0x0455, 0x0456, 0x0457, 0x0458, 0x0459, 0x045A, 0x045B, 0x045C, 0x045D, 0x045E, 0x045F, 0x0460, 0x0461,
	0x0462, 0x0463, 0x0464, 0x0465, 0x0466, 0x0467, 0x0468, 0x0469, 0x046A, 0x046B, 0x046C, 0x046D, 0x046E, 0x046F,
	0x0470,	0x0471,	0x0472,	0x0473,	0x0474,	0x0475,	0x0476,	0x0477,	0x0478,	0x0479,	0x047A,	0x047B,	0x047C,	0x047D,
	0x047E, 0x047F,
	0
};

static const unsigned cyrillic_b[] = {
	0x0480, 0x0481, 0x0482, 0x0483, 0x0484, 0x0485, 0x0486, 0x0488, 0x0489, 0x048C, 0x048D, 0x048E, 0x048F, 0x0490,
	0x0491,	0x0492,	0x0493,	0x0494,	0x0495,	0x0496,	0x0497,	0x0498,	0x0499,	0x049A,	0x049B,	0x049C,	0x049D,	0x049E,
	0x049F,	0x04A0,	0x04A1,	0x04A2,	0x04A3,	0x04A4,	0x04A5,	0x04A6,	0x04A7,	0x04A8,	0x04A9,	0x04AA,	0x04AB,	0x04AC,
	0x04AD,	0x04AE,	0x04AF,	0x04B0,	0x04B1,	0x04B2,	0x04B3,	0x04B4,	0x04B5,	0x04B6,	0x04B7,	0x04B8,	0x04B9,	0x04BA,
	0x04BB, 0x04BC, 0x04BD, 0x04BE, 0x04BF, 0x04C0, 0x04C1, 0x04C2, 0x04C3, 0x04C4, 0x04C7, 0x04C8, 0x04CB, 0x04CC,
	0x04D0, 0x04D1, 0x04D2, 0x04D3, 0x04D4, 0x04D5, 0x04D6, 0x04D7, 0x04D8, 0x04D9, 0x04DA, 0x04DB, 0x04DC, 0x04DD,
	0x04DE, 0x04DF, 0x04E0, 0x04E1, 0x04E2, 0x04E3, 0x04E4, 0x04E5, 0x04E6, 0x04E7, 0x04E8, 0x04E9, 0x04EA, 0x04EB,
	0x04EC, 0x04ED, 0x04EE, 0x04EF, 0x04F0, 0x04F1, 0x04F2, 0x04F3, 0x04F4, 0x04F5, 0x04F8, 0x04F9, 0x04FA, 0x04FB,
	0x04FC, 0x04FD, 0x04FE, 0x04FF,
	0
};

static const unsigned misc_1[] = {
	0x2600, 0x2601, 0x2602, 0x2603, 0x2604, 0x2605, 0x2606, 0x2607, 0x2608, 0x2609, 0x260A, 0x260B, 0x260C, 0x260D,
    0x260E, 0x260F, 0x2610, 0x2611, 0x2612, 0x2613, 0x2614, 0x2615, 0x2616, 0x2617, 0x2618, 0x2619, 0x261A, 0x261B,
    0x261C, 0x261D, 0x261E, 0x261F, 0x2620, 0x2621, 0x2622, 0x2623, 0x2624, 0x2625, 0x2626, 0x2627, 0x2628, 0x2629,
    0x262A, 0x262B, 0x262C, 0x262D, 0x262E, 0x262F, 0x2630, 0x2631, 0x2632, 0x2633, 0x2634, 0x2635, 0x2636, 0x2637,
    0x2638, 0x2639, 0x263A, 0x263B, 0x263C, 0x263D, 0x263E, 0x263F, 0x2640, 0x2641, 0x2642, 0x2643, 0x2644, 0x2645,
    0x2646, 0x2647, 0x2648, 0x2649, 0x264A, 0x264B, 0x264C, 0x264D, 0x264E, 0x264F, 0x2650, 0x2651, 0x2652, 0x2653,
    0x2654, 0x2655, 0x2656, 0x2657, 0x2658, 0x2659, 0x265A, 0x265B, 0x265C, 0x265D, 0x265E, 0x265F, 0x2660, 0x2661,
    0x2662, 0x2663, 0x2664, 0x2665, 0x2666, 0x2667, 0x2668, 0x2669, 0x266A, 0x266B, 0x266C, 0x266D, 0x266E, 0x266F,
    0x2670, 0x2671, 0x2672, 0x2673, 0x2674, 0x2675, 0x2676, 0x2677, 0x2678, 0x2679, 0x267A, 0x267B, 0x267C, 0x267D,
    0x267E, 0x267F,
    0
};

static const unsigned misc_2[] = {
	0x2680, 0x2681, 0x2682, 0x2683, 0x2684, 0x2685, 0x2686, 0x2687, 0x2688, 0x2689, 0x268A, 0x268B, 0x268C, 0x268D,
	0x268E, 0x268F, 0x2690, 0x2691, 0x2692, 0x2693, 0x2694, 0x2695, 0x2696, 0x2697, 0x2698, 0x2699, 0x269A, 0x269B,
	0x269C, 0x269D, 0x269E, 0x269F, 0x26A0, 0x26A1, 0x26A2, 0x26A3, 0x26A4, 0x26A5, 0x26A6, 0x26A7, 0x26A8, 0x26A9,
	0x26AA, 0x26AB, 0x26AC, 0x26AD, 0x26AE, 0x26AF, 0x26B0, 0x26B1, 0x26B2, 0x26B3, 0x26B4, 0x26B5, 0x26B6, 0x26B7,
	0x26B8, 0x26B9, 0x26BA, 0x26BB, 0x26BC, 0x26BD, 0x26BE, 0x26BF, 0x26C0, 0x26C1, 0x26C2, 0x26C3, 0x26C4, 0x26C5,
	0x26C6, 0x26C7, 0x26C8, 0x26C9, 0x26CA, 0x26CB, 0x26CC, 0x26CD, 0x26CE, 0x26CF, 0x26D0, 0x26D1, 0x26D2, 0x26D3,
	0x26D4, 0x26D5, 0x26D6, 0x26D7, 0x26D8, 0x26D9, 0x26DA, 0x26DB, 0x26DC, 0x26DD, 0x26DE, 0x26DF, 0x26E0, 0x26E1,
	0x26E2, 0x26E3, 0x26E4, 0x26E5, 0x26E6, 0x26E7, 0x26E8, 0x26E9, 0x26EA, 0x26EB, 0x26EC, 0x26ED, 0x26EE, 0x26EF,
	0x26F0, 0x26F1, 0x26F2, 0x26F3, 0x26F4, 0x26F5, 0x26F6, 0x26F7, 0x26F8, 0x26F9, 0x26FA, 0x26FB, 0x26FC, 0x26FD,
	0x26FE, 0x26FF,
	0
};

static const unsigned arrows[] = {
	0x2190, 0x2191, 0x2192, 0x2193, 0x2194, 0x2195, 0x2196, 0x2197, 0x2198, 0x2199, 0x219A, 0x219B, 0x219C, 0x219D, 
	0x219E, 0x219F, 0x21A0, 0x21A1, 0x21A2, 0x21A3, 0x21A4, 0x21A5, 0x21A6, 0x21A7, 0x21A8, 0x21A9, 0x21AA, 0x21AB,
	0x21AC, 0x21AD, 0x21AE, 0x21AF, 0x21B0, 0x21B1, 0x21B2, 0x21B3, 0x21B4, 0x21B5, 0x21B6, 0x21B7, 0x21B8, 0x21B9,
	0x21BA, 0x21BB, 0x21BC, 0x21BD, 0x21BE, 0x21BF, 0x21C0, 0x21C1, 0x21C2, 0x21C3, 0x21C4, 0x21C5, 0x21C6, 0x21C7,
	0x21C8, 0x21C9, 0x21CA, 0x21CB, 0x21CC, 0x21CD, 0x21CE, 0x21CF, 0x21D0, 0x21D1, 0x21D2, 0x21D3, 0x21D4, 0x21D5,
	0x21D6, 0x21D7, 0x21D8, 0x21D9, 0x21DA, 0x21DB, 0x21DC, 0x21DD, 0x21DE, 0x21DF, 0x21E0, 0x21E1, 0x21E2, 0x21E3,
	0x21E4, 0x21E5, 0x21E6, 0x21E7, 0x21E8, 0x21E9, 0x21EA, 0x21EB, 0x21EC, 0x21ED, 0x21EE, 0x21EF, 0x21F0, 0x21F1,
	0x21F2, 0x21F3, 0x21F4, 0x21F5, 0x21F6, 0x21F7, 0x21F8, 0x21F9, 0x21FA, 0x21FB, 0x21FC, 0x21FD, 0x21FE, 0x21FF,
	0
};

/* Regenerate the font, pulling in settings from widgets. */
void CTextureFontGeneratorDlg::UpdateFont( bool bSavingDoubleRes )
{
	m_bUpdateFontNeeded = false;

	m_FontView.SetBitmap(nullptr );

	CString sOld;
	{
		int iOldSel = m_ShownPage.GetCurSel();
		if( iOldSel != -1 && iOldSel < static_cast<int>(g_pTextureFont->m_PagesToGenerate.size()) )
			sOld = g_pTextureFont->m_PagesToGenerate[iOldSel].name;
	}

	/* Get the selected font m_sFamily. */
	CString sText;
	m_FamilyList.GetWindowText( sText );
	g_pTextureFont->m_sFamily = sText;
	
	m_FontSize.GetWindowText(sText);
	g_pTextureFont->m_fFontSizePixels = static_cast<float>(atof(sText));
	if( bSavingDoubleRes )
		g_pTextureFont->m_fFontSizePixels *= 2;
	
	m_Padding.GetWindowText(sText);
	g_pTextureFont->m_iPadding = atoi(sText);
	if( bSavingDoubleRes )
		g_pTextureFont->m_iPadding *= 2;

	CMenu *pMenu = GetMenu();
	g_pTextureFont->m_bBold = !!( pMenu->GetMenuState(ID_STYLE_BOLD, 0) & MF_CHECKED );
	g_pTextureFont->m_bItalic = !!( pMenu->GetMenuState(ID_STYLE_ITALIC, 0) & MF_CHECKED );
	g_pTextureFont->m_bAntiAlias = !!( pMenu->GetMenuState(ID_STYLE_ANTIALIASED, 0) & MF_CHECKED );

	g_pTextureFont->m_PagesToGenerate.clear();
	FontPageDescription desc;

	bool bNumbersOnly = !!( pMenu->GetMenuState(ID_OPTIONS_NUMBERSONLY, 0) & MF_CHECKED );

	desc.name = "main";
	for( int i = 0; map_main[i]; ++i )
	{
		unsigned wc = map_main[i];
		desc.chars.push_back(wc);
	}
	g_pTextureFont->m_PagesToGenerate.push_back( desc );

#ifdef _TFG_JAPANESE_SUPPLEMENTAL
	// JIS X 0212 characters not in JIS X 0208, CP932, JIS X 0213
	if( !bNumbersOnly )
	{
		desc.name = "symbol-hojo";
		desc.chars.clear();
		for( int i = 0; map_symbol_hojo[i]; ++i )
		{
			unsigned wc = map_symbol_hojo[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji-hojo 1";
		desc.chars.clear();
		for( int i = 0; map_kanji_hojo_1[i]; ++i )
		{
			unsigned wc = map_kanji_hojo_1[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji-hojo 2";
		desc.chars.clear();
		for( int i = 0; map_kanji_hojo_2[i]; ++i )
		{
			unsigned wc = map_kanji_hojo_2[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji-hojo 3";
		desc.chars.clear();
		for( int i = 0; map_kanji_hojo_3[i]; ++i )
		{
			unsigned wc = map_kanji_hojo_3[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );
	}
	else
	{
		desc.name = "kanji-hojo 4";
		desc.chars.clear();
		for( int i = 0; map_kanji_hojo_4[i]; ++i )
		{
			unsigned wc = map_kanji_hojo_4[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji-hojo 5";
		desc.chars.clear();
		for( int i = 0; map_kanji_hojo_5[i]; ++i )
		{
			unsigned wc = map_kanji_hojo_5[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji-hojo 6";
		desc.chars.clear();
		for( int i = 0; map_kanji_hojo_6[i]; ++i )
		{
			unsigned wc = map_kanji_hojo_6[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );
	}
#elif defined _TFG_CHINESE_ADDITION // _TFG_JAPANESE_SUPPLEMENTAL
	// GBK characters not in JIS X 0208, CP932, JIS X 0213, JIS X 0212
	if( !bNumbersOnly )
	{
		// GB 2132
		desc.name = "symbol-gbk1";
		desc.chars.clear();
		for( int i = 0; map_symbol_gbk1[i]; ++i )
		{
			unsigned wc = map_symbol_gbk1[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji-gbk2 1";
		desc.chars.clear();
		for( int i = 0; map_kanji_gbk2_1[i]; ++i )
		{
			unsigned wc = map_kanji_gbk2_1[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji-gbk2 2";
		desc.chars.clear();
		for( int i = 0; map_kanji_gbk2_2[i]; ++i )
		{
			unsigned wc = map_kanji_gbk2_2[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji-gbk2 3";
		desc.chars.clear();
		for( int i = 0; map_kanji_gbk2_3[i]; ++i )
		{
			unsigned wc = map_kanji_gbk2_3[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );
	}
	else
	{
		// GBK Extension and Compatible Ideograph
		desc.name = "kanji-gbk3 1";
		desc.chars.clear();
		for( int i = 0; map_kanji_gbk3_1[i]; ++i )
		{
			unsigned wc = map_kanji_gbk3_1[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji-gbk3 2";
		desc.chars.clear();
		for( int i = 0; map_kanji_gbk3_2[i]; ++i )
		{
			unsigned wc = map_kanji_gbk3_2[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji-gbk3 3";
		desc.chars.clear();
		for( int i = 0; map_kanji_gbk3_3[i]; ++i )
		{
			unsigned wc = map_kanji_gbk3_3[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji-gbk4 1";
		desc.chars.clear();
		for( int i = 0; map_kanji_gbk4_1[i]; ++i )
		{
			unsigned wc = map_kanji_gbk4_1[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji-gbk4 2";
		desc.chars.clear();
		for( int i = 0; map_kanji_gbk4_2[i]; ++i )
		{
			unsigned wc = map_kanji_gbk4_2[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji-gbk4 3";
		desc.chars.clear();
		for( int i = 0; map_kanji_gbk4_3[i]; ++i )
		{
			unsigned wc = map_kanji_gbk4_3[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji-gbk4 4";
		desc.chars.clear();
		for( int i = 0; map_kanji_gbk4_4[i]; ++i )
		{
			unsigned wc = map_kanji_gbk4_4[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "symbol-gbk5";
		desc.chars.clear();
		for( int i = 0; map_symbol_gbk5[i]; ++i )
		{
			unsigned wc = map_symbol_gbk5[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

// DO NOT GENERATE DUE TO STEPMANIA'S SPECIFICATION.
#ifdef _TFG_GENERATE_SIP
		desc.name = "kanji-cns";
		desc.chars.clear();
		for( int i = 0; map_kanji_cns[i]; ++i )
		{
			unsigned wc = map_kanji_cns[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );
#endif
	}
#elif defined _TFG_KOREAN_ADDITION // _TFG_CHINESE_ADDITION
	// KS X 1001 characters not in JIS X 0208, CP932, JIS X 0213, JIS X 0212, GBK
	if( !bNumbersOnly )
	{
		// Hangul Jamo, Compatible Ideograph, KS X 1001 Symbol
		desc.name = "jamo";
		desc.chars.clear();
		for( int i = 0; map_jamo[i]; ++i )
		{
			unsigned wc = map_jamo[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "jamo-ks";
		desc.chars.clear();
		for( int i = 0; map_jamo_ks[i]; ++i )
		{
			unsigned wc = map_jamo_ks[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "jamo-a";
		desc.chars.clear();
		for( int i = 0; map_jamo_a[i]; ++i )
		{
			unsigned wc = map_jamo_a[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "jamo-b";
		desc.chars.clear();
		for( int i = 0; map_jamo_b[i]; ++i )
		{
			unsigned wc = map_jamo_b[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji-ks";
		desc.chars.clear();
		for( int i = 0; map_kanji_ks[i]; ++i )
		{
			unsigned wc = map_kanji_ks[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "symbol-ks";
		desc.chars.clear();
		for( int i = 0; map_symbol_ks[i]; ++i )
		{
			unsigned wc = map_symbol_ks[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		// Hangul Syllables, first half
		desc.name = "hangul 1";
		desc.chars.clear();
		for( int i = 0; map_hangul_1[i]; ++i )
		{
			unsigned wc = map_hangul_1[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "hangul 2";
		desc.chars.clear();
		for( int i = 0; map_hangul_2[i]; ++i )
		{
			unsigned wc = map_hangul_2[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "hangul 3";
		desc.chars.clear();
		for( int i = 0; map_hangul_3[i]; ++i )
		{
			unsigned wc = map_hangul_3[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "hangul 4";
		desc.chars.clear();
		for( int i = 0; map_hangul_4[i]; ++i )
		{
			unsigned wc = map_hangul_4[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "hangul 5";
		desc.chars.clear();
		for( int i = 0; map_hangul_5[i]; ++i )
		{
			unsigned wc = map_hangul_5[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );
	}
	else
	{
		// Hangul Syllables, last half
		desc.name = "hangul 6";
		desc.chars.clear();
		for( int i = 0; map_hangul_6[i]; ++i )
		{
			unsigned wc = map_hangul_6[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "hangul 7";
		desc.chars.clear();
		for( int i = 0; map_hangul_7[i]; ++i )
		{
			unsigned wc = map_hangul_7[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "hangul 8";
		desc.chars.clear();
		for( int i = 0; map_hangul_8[i]; ++i )
		{
			unsigned wc = map_hangul_8[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "hangul 9";
		desc.chars.clear();
		for( int i = 0; map_hangul_9[i]; ++i )
		{
			unsigned wc = map_hangul_9[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "hangul 10";
		desc.chars.clear();
		for( int i = 0; map_hangul_10[i]; ++i )
		{
			unsigned wc = map_hangul_10[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "hangul 11";
		desc.chars.clear();
		for( int i = 0; map_hangul_11[i]; ++i )
		{
			unsigned wc = map_hangul_11[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );
	}
#elif defined _TFG_EURO_ADDITION
if (!bNumbersOnly)
{
	desc.name = "basic-latin";
	desc.chars.clear();
	for (int i = 0; basic_latin[i]; ++i)
	{
		unsigned wc = basic_latin[i];
		desc.chars.push_back(wc);
	}
	g_pTextureFont->m_PagesToGenerate.push_back(desc);

	desc.name = "latin-supplement";
	desc.chars.clear();
	for (int i = 0; latin_supplement[i]; ++i)
	{
		unsigned wc = latin_supplement[i];
		desc.chars.push_back(wc);
	}
	g_pTextureFont->m_PagesToGenerate.push_back(desc);
	
	desc.name = "latin-extended-a";
	desc.chars.clear();
	for (int i = 0; latin_extended_a[i]; ++i)
	{
		unsigned wc = latin_extended_a[i];
		desc.chars.push_back(wc);
	}
	g_pTextureFont->m_PagesToGenerate.push_back(desc);

	desc.name = "latin-extended-b1";
	desc.chars.clear();
	for (int i = 0; latin_extended_b1[i]; ++i)
	{
		unsigned wc = latin_extended_b1[i];
		desc.chars.push_back(wc);
	}
	g_pTextureFont->m_PagesToGenerate.push_back(desc);
	
	desc.name = "latin-extended-b2";
	desc.chars.clear();
	for (int i = 0; latin_extended_b2[i]; ++i)
	{
		unsigned wc = latin_extended_b2[i];
		desc.chars.push_back(wc);
	}
	g_pTextureFont->m_PagesToGenerate.push_back(desc);

	desc.name = "combining-diacritical";
	desc.chars.clear();
	for (int i = 0; combining_diacritical[i]; ++i)
	{
		unsigned wc = combining_diacritical[i];
		desc.chars.push_back(wc);
	}
	g_pTextureFont->m_PagesToGenerate.push_back(desc);

	desc.name = "greek";
	desc.chars.clear();
	for (int i = 0; greek[i]; ++i)
	{
		unsigned wc = greek[i];
		desc.chars.push_back(wc);
	}
	g_pTextureFont->m_PagesToGenerate.push_back(desc);

	desc.name = "cyrillic-a";
	desc.chars.clear();
	for (int i = 0; cyrillic_a[i]; ++i)
	{
		unsigned wc = cyrillic_a[i];
		desc.chars.push_back(wc);
	}
	g_pTextureFont->m_PagesToGenerate.push_back(desc);

	desc.name = "cyrillic-b";
	desc.chars.clear();
	for (int i = 0; cyrillic_b[i]; ++i)
	{
		unsigned wc = cyrillic_b[i];
		desc.chars.push_back(wc);
	}
	g_pTextureFont->m_PagesToGenerate.push_back(desc);
}
else
{
	desc.name = "misc-1";
	desc.chars.clear();
	for (int i = 0; misc_1[i]; ++i)
	{
		unsigned wc = misc_1[i];
		desc.chars.push_back(wc);
	}
	g_pTextureFont->m_PagesToGenerate.push_back(desc);

	desc.name = "misc-2";
	desc.chars.clear();
	for (int i = 0; misc_2[i]; ++i)
	{
		unsigned wc = misc_2[i];
		desc.chars.push_back(wc);
	}
	g_pTextureFont->m_PagesToGenerate.push_back(desc);

	desc.name = "arrows";
	desc.chars.clear();
	for (int i = 0; arrows[i]; ++i)
	{
		unsigned wc = arrows[i];
		desc.chars.push_back(wc);
	}
	g_pTextureFont->m_PagesToGenerate.push_back(desc);
}
#else // _TFG_CHINESE_ADDITION, _TFG_KOREAN_ADDITION
	if( !bNumbersOnly )
	{
	/* // Don't generate duplicate
		desc.name = "kanji 1";
		desc.chars.clear();
		for( int i = 0; map_kanji_1[i]; ++i )
		{
			unsigned wc = map_kanji_1[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji 2";
		desc.chars.clear();
		for( int i = 0; map_kanji_2[i]; ++i )
		{
			unsigned wc = map_kanji_2[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji 3";
		desc.chars.clear();
		for( int i = 0; map_kanji_3[i]; ++i )
		{
			unsigned wc = map_kanji_3[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji 4";
		desc.chars.clear();
		for( int i = 0; map_kanji_4[i]; ++i )
		{
			unsigned wc = map_kanji_4[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji 5";
		desc.chars.clear();
		for( int i = 0; map_kanji_5[i]; ++i )
		{
			unsigned wc = map_kanji_5[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji 6";
		desc.chars.clear();
		for( int i = 0; map_kanji_6[i]; ++i )
		{
			unsigned wc = map_kanji_6[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji 7";
		desc.chars.clear();
		for( int i = 0; map_kanji_7[i]; ++i )
		{
			unsigned wc = map_kanji_7[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji 8";
		desc.chars.clear();
		for( int i = 0; map_kanji_8[i]; ++i )
		{
			unsigned wc = map_kanji_8[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji 9";
		desc.chars.clear();
		for( int i = 0; map_kanji_9[i]; ++i )
		{
			unsigned wc = map_kanji_9[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );
	*/

		// JIS X 0201/0208 and IBM/NEC Extension
		desc.name = "ank";
		desc.chars.clear();
		for( int i = 0; map_ank[i]; ++i )
		{
			unsigned wc = map_ank[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "symbol";
		desc.chars.clear();
		for( int i = 0; map_symbol[i]; ++i )
		{
			unsigned wc = map_symbol[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "symbol2";
		desc.chars.clear();
		for( int i = 0; map_symbol2[i]; ++i )
		{
			unsigned wc = map_symbol2[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "symbol3";
		desc.chars.clear();
		for( int i = 0; map_symbol3[i]; ++i )
		{
			unsigned wc = map_symbol3[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji-regular 1";
		desc.chars.clear();
		for( int i = 0; map_kanji_regular_1[i]; ++i )
		{
			unsigned wc = map_kanji_regular_1[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji-regular 2";
		desc.chars.clear();
		for( int i = 0; map_kanji_regular_2[i]; ++i )
		{
			unsigned wc = map_kanji_regular_2[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji-regular 3";
		desc.chars.clear();
		for( int i = 0; map_kanji_regular_3[i]; ++i )
		{
			unsigned wc = map_kanji_regular_3[i];
#ifndef _TFG_GENERATE_SIP
			if (wc < 0x010000)
#endif
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji-jis1";
		desc.chars.clear();
		for( int i = 0; map_kanji_jis1[i]; ++i )
		{
			unsigned wc = map_kanji_jis1[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji-jis2 1";
		desc.chars.clear();
		for( int i = 0; map_kanji_jis2_1[i]; ++i )
		{
			unsigned wc = map_kanji_jis2_1[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji-jis2 2";
		desc.chars.clear();
		for( int i = 0; map_kanji_jis2_2[i]; ++i )
		{
			unsigned wc = map_kanji_jis2_2[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji-jis2 3";
		desc.chars.clear();
		for( int i = 0; map_kanji_jis2_3[i]; ++i )
		{
			unsigned wc = map_kanji_jis2_3[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji-jis2 4";
		desc.chars.clear();
		for( int i = 0; map_kanji_jis2_4[i]; ++i )
		{
			unsigned wc = map_kanji_jis2_4[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji-ibm";
		desc.chars.clear();
		for( int i = 0; map_kanji_ibm[i]; ++i )
		{
			unsigned wc = map_kanji_ibm[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "compatible";
		desc.chars.clear();
		for( int i = 0; map_compatible[i]; ++i )
		{
			unsigned wc = map_compatible[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );
	}
	else
	{
		// JIS X 0213 Extension and ARIB Extension
		desc.name = "symbol4";
		desc.chars.clear();
		for( int i = 0; map_symbol4[i]; ++i )
		{
			unsigned wc = map_symbol4[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "symbol5";
		desc.chars.clear();
		for( int i = 0; map_symbol5[i]; ++i )
		{
			unsigned wc = map_symbol5[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji-jis3 1";
		desc.chars.clear();
		for( int i = 0; map_kanji_jis3_1[i]; ++i )
		{
			unsigned wc = map_kanji_jis3_1[i];
#ifndef _TFG_GENERATE_SIP
			if (wc < 0x010000)
#endif
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji-jis3 2";
		desc.chars.clear();
		for( int i = 0; map_kanji_jis3_2[i]; ++i )
		{
			unsigned wc = map_kanji_jis3_2[i];
#ifndef _TFG_GENERATE_SIP
			if (wc < 0x010000)
#endif
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji-jis4 1";
		desc.chars.clear();
		for( int i = 0; map_kanji_jis4_1[i]; ++i )
		{
			unsigned wc = map_kanji_jis4_1[i];
#ifndef _TFG_GENERATE_SIP
			if (wc < 0x010000)
#endif
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji-jis4 2";
		desc.chars.clear();
		for( int i = 0; map_kanji_jis4_2[i]; ++i )
		{
			unsigned wc = map_kanji_jis4_2[i];
#ifndef _TFG_GENERATE_SIP
			if (wc < 0x010000)
#endif
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji-jis4 3";
		desc.chars.clear();
		for( int i = 0; map_kanji_jis4_3[i]; ++i )
		{
			unsigned wc = map_kanji_jis4_3[i];
#ifndef _TFG_GENERATE_SIP
			if (wc < 0x010000)
#endif
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );

		desc.name = "kanji-arib";
		desc.chars.clear();
		for( int i = 0; map_kanji_arib[i]; ++i )
		{
			unsigned wc = map_kanji_arib[i];
			desc.chars.push_back(wc);
		}
		g_pTextureFont->m_PagesToGenerate.push_back( desc );
	}
#endif

	/* Go: */
	g_pTextureFont->FormatFontPages();

	m_SpinTop.SetPos( g_pTextureFont->m_iCharTop );
	m_SpinBaseline.SetPos( g_pTextureFont->m_iCharBaseline );

	m_ShownPage.ResetContent();
	for( unsigned p = 0; p < g_pTextureFont->m_PagesToGenerate.size(); ++p )
		m_ShownPage.AddString( g_pTextureFont->m_PagesToGenerate[p].name.GetString() );
	int iRet = m_ShownPage.FindStringExact( -1, sOld );
	if( iRet == CB_ERR )
		iRet = 0;
	m_ShownPage.SetCurSel( iRet );
}

void CTextureFontGeneratorDlg::UpdateCloseUp()
{
	const HBITMAP hBitmap = m_CloseUp.GetBitmap();
	HBITMAP hSample;
	m_CloseUp.SetBitmap(nullptr );
	if( hBitmap )
		DeleteObject( hBitmap );

	hSample = g_pTextureFont->m_Characters[0x003042];

	if( hSample == nullptr )
		return;

	const HDC hCharacterDC = CreateCompatibleDC(nullptr );
	const HGDIOBJ hOldCharacterBitmap = SelectObject( hCharacterDC, hSample );

	BITMAPINFO CharacterInfo;
	memset( &CharacterInfo, 0, sizeof(CharacterInfo) );
	CharacterInfo.bmiHeader.biSize = sizeof(CharacterInfo.bmiHeader);
	GetDIBits( hCharacterDC, hSample, 0, 0, nullptr, &CharacterInfo, DIB_RGB_COLORS );

	const int iWidth = CharacterInfo.bmiHeader.biWidth;
	const int iHeight = CharacterInfo.bmiHeader.biHeight;

	/* Set up a bitmap to zoom the image into. */
	const int iSourceHeight = g_pTextureFont->m_BoundingRect.bottom;
	const int iZoomFactor = 4;

	HBITMAP hZoom;
	{
		const HDC hTempDC = ::GetDC(nullptr);
		hZoom = CreateCompatibleBitmap( hTempDC, iWidth * iZoomFactor, iSourceHeight * iZoomFactor );
		::ReleaseDC(nullptr, hTempDC );
	}
	const HDC hZoomDC = CreateCompatibleDC(nullptr );
	const HGDIOBJ hOldZoomBitmap = SelectObject( hZoomDC, hZoom );

	StretchBlt(
		hZoomDC, 0, 0,
		iWidth * iZoomFactor, iHeight * iZoomFactor,
		hCharacterDC, 0, 0,
		iWidth, iHeight,
		SRCCOPY
	);
	SelectObject( hCharacterDC, hOldCharacterBitmap );
	::DeleteDC( hCharacterDC );

	HPEN hPen = CreatePen( PS_SOLID, 1, RGB(0, 255, 255) );
	HGDIOBJ hOldPen = SelectObject( hZoomDC, hPen );

	/* Align the line with the top of the unit, so, when correct, the font
	 * "sits" on the line. */
	int iY = iZoomFactor*g_pTextureFont->m_iCharBaseline - 1;
	MoveToEx( hZoomDC, 0, iY, nullptr );
	LineTo( hZoomDC, iWidth * iZoomFactor, iY );

	SelectObject( hZoomDC, hOldPen );
	DeleteObject( hPen );

	hPen = CreatePen( PS_SOLID, 1, RGB(255, 255, 50) );
	hOldPen = SelectObject( hZoomDC, hPen );
	
	/* Align the line with the bottom of the unit, so, when correct, the line
	 * "sits" on the font. */
	iY = iZoomFactor*g_pTextureFont->m_iCharTop - 1;
	MoveToEx( hZoomDC, 0, iY, nullptr );
	LineTo( hZoomDC, iWidth * iZoomFactor, iY );

	SelectObject( hZoomDC, hOldPen );
	DeleteObject( hPen );

	SelectObject( hZoomDC, hOldZoomBitmap );
	::DeleteDC( hZoomDC );

	m_CloseUp.SetBitmap( hZoom );
}


int CALLBACK CTextureFontGeneratorDlg::EnumFontFamiliesCallback( const LOGFONTA *pLogicalFontData, const TEXTMETRICA *pPhysicalFontData, DWORD FontType, LPARAM lParam )
{
	CTextureFontGeneratorDlg *pThis = reinterpret_cast<CTextureFontGeneratorDlg*>(lParam);
	set<CString> *pSet = reinterpret_cast<set<CString>*>(lParam);
	pSet->insert( pLogicalFontData->lfFaceName );
	return 1;
}

BOOL CTextureFontGeneratorDlg::OnInitDialog()
{
	CDialog::OnInitDialog();

	// Set the icon for this dialog.  The framework does this automatically
	//  when the application's main window is not a dialog
	SetIcon(m_hIcon, TRUE);			// Set big icon
	SetIcon(m_hIcon, FALSE);		// Set small icon

	// XXX: AddFontResourceEx
	// AddFontMemResourceEx
	// drag drop
	{
		LOGFONT font;
		memset( &font, 0, sizeof(font) );
		font.lfCharSet = DEFAULT_CHARSET;
		const CPaintDC dc(this);
		set<CString> setFamilies;
		EnumFontFamiliesEx( dc.GetSafeHdc(), &font, EnumFontFamiliesCallback, reinterpret_cast<LPARAM>(&setFamilies), 0 );
		for( set<CString>::const_iterator it = setFamilies.begin(); it != setFamilies.end(); ++it )
			m_FamilyList.AddString( *it );
		m_FamilyList.SetCurSel( 0 );
	}

	CMenu *pMenu = GetMenu();
	pMenu->CheckMenuItem( ID_STYLE_ANTIALIASED, MF_CHECKED );

	m_FontSize.SetWindowText( "20" );
	m_Padding.SetWindowText( "2" );
	m_SpinTop.SetRange( 64, 0 );
	m_SpinBaseline.SetRange( 64, 0 );

	m_bUpdateFontNeeded = true;
	m_bUpdateFontViewAndCloseUpNeeded = true;

	return TRUE;  // return TRUE  unless you set the focus to a control
}

void CTextureFontGeneratorDlg::UpdateFontViewAndCloseUp()
{
	m_bUpdateFontViewAndCloseUpNeeded = false;
	m_FontView.SetBitmap(nullptr );

	m_SpinTop.EnableWindow( true );
	m_SpinBaseline.EnableWindow( true );
	GetMenu()->EnableMenuItem( ID_FILE_SAVE, MF_ENABLED );
	
	if( g_pTextureFont->m_sError != "" )
	{
		m_SpinTop.EnableWindow(false);
		m_SpinBaseline.EnableWindow(false);
		GetMenu()->EnableMenuItem( ID_FILE_SAVE, MF_GRAYED );

		m_ErrorOrWarning.SetWindowText( "Error: " + g_pTextureFont->m_sError );
		return;
	}

	m_ErrorOrWarning.SetWindowText( g_pTextureFont->m_sWarnings );

	const int iSelectedPage = m_ShownPage.GetCurSel();
	ASSERT( iSelectedPage < static_cast<int>(g_pTextureFont->m_apPages.size()) );

	g_pTextureFont->m_iCharTop = LOWORD(m_SpinTop.GetPos());
	g_pTextureFont->m_iCharBaseline = LOWORD(m_SpinBaseline.GetPos());

	// HBITMAP hBitmap = g_pTextureFont->m_apPages[0]->m_hPage;
	ASSERT( g_pTextureFont->m_apPages[iSelectedPage] );
	ASSERT( g_pTextureFont->m_apPages[iSelectedPage]->m_hPage );
	const HBITMAP hBitmap = g_pTextureFont->m_apPages[iSelectedPage]->m_hPage;
	m_FontView.SetBitmap( hBitmap );
	
	UpdateCloseUp();

	CString sStr;
	sStr.Format("Overlap: %i, %i\nMaximum size: %ix%i", 
		g_pTextureFont->m_iCharLeftOverlap, g_pTextureFont->m_iCharRightOverlap,
		g_pTextureFont->m_BoundingRect.right - g_pTextureFont->m_BoundingRect.left,
		g_pTextureFont->m_BoundingRect.bottom - g_pTextureFont->m_BoundingRect.top
		);
	m_TextOverlap.SetWindowText( sStr );
}

void CTextureFontGeneratorDlg::OnDestroy()
{
	WinHelp(0L, HELP_QUIT);
	CDialog::OnDestroy();
}

void CTextureFontGeneratorDlg::OnPaint() 
{
	if( IsIconic() )
	{
		CPaintDC dc(this); // device context for painting

		SendMessage(WM_ICONERASEBKGND, reinterpret_cast<WPARAM>(dc.GetSafeHdc()), 0);

		// Center icon in client rectangle
		const int cxIcon = GetSystemMetrics(SM_CXICON);
		const int cyIcon = GetSystemMetrics(SM_CYICON);
		CRect rect;
		GetClientRect(&rect);
		const int x = (rect.Width() - cxIcon + 1) / 2;
		const int y = (rect.Height() - cyIcon + 1) / 2;

		// Draw the icon
		dc.DrawIcon(x, y, m_hIcon);
		return;
	}

	if( m_bUpdateFontNeeded )
	{
		m_bUpdateFontViewAndCloseUpNeeded = true;
		UpdateFont( false );
	}

	if( m_bUpdateFontViewAndCloseUpNeeded )
		UpdateFontViewAndCloseUp();

	CDialog::OnPaint();
}

// The system calls this function to obtain the cursor to display while the user drags
//  the minimized window.
HCURSOR CTextureFontGeneratorDlg::OnQueryDragIcon()
{
	return static_cast<HCURSOR>(m_hIcon);
}


void CTextureFontGeneratorDlg::OnClose() 
{
	if( CanExit() )
		CDialog::OnClose();
}

void CTextureFontGeneratorDlg::OnOK() 
{
	if( CanExit() )
		CDialog::OnOK();
}

void CTextureFontGeneratorDlg::OnCancel() 
{
	if (CanExit())
		CDialog::OnCancel();
}

BOOL CTextureFontGeneratorDlg::OnMouseWheel( UINT nFlags, short zDelta, CPoint pt )
{
	CWnd *pFocus = this->GetFocus();
	if( nFlags & MK_CONTROL )
		zDelta *= 5;
	if( pFocus == &m_Padding || pFocus == &m_FontSize )
	{
		CEdit *pFocusedEdit = dynamic_cast<CEdit *>(pFocus);
		CString sText;
		pFocusedEdit->GetWindowText( sText );
		int i = atoi( sText );
		i += zDelta / WHEEL_DELTA;
		if( pFocus == &m_Padding )
			i = max( i, 0 );
		else if( pFocus == &m_FontSize )
			i = max( i, 1 );

		sText.Format( "%i", i );
		pFocusedEdit->SetWindowText( sText );
		return FALSE;
	}

	return CDialog::OnMouseWheel( nFlags, zDelta, pt );
}

BOOL CTextureFontGeneratorDlg::CanExit()
{
	return TRUE;
}

void CTextureFontGeneratorDlg::OnCbnSelchangeShownPage()
{
	m_bUpdateFontViewAndCloseUpNeeded = true;
	Invalidate( FALSE );
	UpdateWindow();
}

void CTextureFontGeneratorDlg::OnCbnSelchangeFamilyList()
{
	m_bUpdateFontNeeded = true;
	Invalidate( FALSE );
	UpdateWindow();
}

void CTextureFontGeneratorDlg::OnEnChangeFontSize()
{
	m_bUpdateFontNeeded = true;
	Invalidate( FALSE );
	UpdateWindow();
}

void CTextureFontGeneratorDlg::OnEnChangePadding()
{
	m_bUpdateFontNeeded = true;
	Invalidate( FALSE );
}

void CTextureFontGeneratorDlg::OnStyleAntialiased()
{
	CMenu *pMenu = GetMenu();
	int Checked = pMenu->GetMenuState(ID_STYLE_ANTIALIASED, 0) & MF_CHECKED;
	Checked ^= MF_CHECKED;
	pMenu->CheckMenuItem( ID_STYLE_ANTIALIASED, Checked );
	m_bUpdateFontNeeded = true;
	Invalidate( FALSE );
}

void CTextureFontGeneratorDlg::OnStyleBold()
{
	CMenu *pMenu = GetMenu();
	int Checked = pMenu->GetMenuState(ID_STYLE_BOLD, 0) & MF_CHECKED;
	Checked ^= MF_CHECKED;
	pMenu->CheckMenuItem( ID_STYLE_BOLD, Checked );
	m_bUpdateFontNeeded = true;
	Invalidate( FALSE );
}

void CTextureFontGeneratorDlg::OnStyleItalic()
{
	CMenu *pMenu = GetMenu();
	int Checked = pMenu->GetMenuState(ID_STYLE_ITALIC, 0) & MF_CHECKED;
	Checked ^= MF_CHECKED;
	pMenu->CheckMenuItem( ID_STYLE_ITALIC, Checked );
	m_bUpdateFontNeeded = true;
	Invalidate( FALSE );
}

void CTextureFontGeneratorDlg::OnOptionsDoubleres()
{
	CMenu *pMenu = GetMenu();
	int Checked = pMenu->GetMenuState(ID_OPTIONS_DOUBLERES, 0) & MF_CHECKED;
	Checked ^= MF_CHECKED;
	pMenu->CheckMenuItem( ID_OPTIONS_DOUBLERES, Checked );
	m_bUpdateFontNeeded = true;
	Invalidate( FALSE );
}

void CTextureFontGeneratorDlg::OnOptionsExportstroketemplates()
{
	CMenu *pMenu = GetMenu();
	int Checked = pMenu->GetMenuState(ID_OPTIONS_EXPORTSTROKETEMPLATES, 0) & MF_CHECKED;
	Checked ^= MF_CHECKED;
	pMenu->CheckMenuItem( ID_OPTIONS_EXPORTSTROKETEMPLATES, Checked );
}

void CTextureFontGeneratorDlg::OnOptionsNumbersonly()
{
	CMenu *pMenu = GetMenu();
	int Checked = pMenu->GetMenuState(ID_OPTIONS_NUMBERSONLY, 0) & MF_CHECKED;
	Checked ^= MF_CHECKED;
	pMenu->CheckMenuItem( ID_OPTIONS_NUMBERSONLY, Checked );
	m_bUpdateFontNeeded = true;
	Invalidate( FALSE );
}

void CTextureFontGeneratorDlg::OnDeltaposSpinTop(NMHDR *pNMHDR, LRESULT *pResult)
{
	LPNMUPDOWN pNMUpDown = reinterpret_cast<LPNMUPDOWN>(pNMHDR);
	*pResult = 0;

	m_bUpdateFontViewAndCloseUpNeeded = true;
	Invalidate( FALSE );
}

void CTextureFontGeneratorDlg::OnDeltaposSpinBaseline(NMHDR *pNMHDR, LRESULT *pResult)
{
	LPNMUPDOWN pNMUpDown = reinterpret_cast<LPNMUPDOWN>(pNMHDR);
	*pResult = 0;

	m_bUpdateFontViewAndCloseUpNeeded = true;
	Invalidate( FALSE );
}

void CTextureFontGeneratorDlg::OnFileSave()
{
	// XXX
	OPENFILENAME ofn;
	memset( &ofn, 0, sizeof(ofn) );
	ofn.lStructSize = sizeof(ofn);
	ofn.hwndOwner = m_hWnd;
	ofn.lpstrFilter = "*.ini\000Font control files\000\000";

	char szFile[1024] = "";
	ofn.lpstrFile = szFile;
	{
		CString sName = g_pTextureFont->m_sFamily;
		sName.MakeLower();
		_snprintf( szFile, 1023, "_%s%s %gpx", 
			static_cast<const char*>(sName), 
			g_pTextureFont->m_bBold ? " Bold" : "",
			g_pTextureFont->m_fFontSizePixels );
	}
	ofn.nMaxFile = sizeof(szFile);

	// lpstrInitialDir
	ofn.Flags = OFN_HIDEREADONLY|OFN_LONGNAMES;

	if( !GetSaveFileName(&ofn) )
		return;

	/* If the filename provided has an extension, remove it. */
	{
		char *pExt = strrchr( szFile, '.' );
		if( pExt )
			*pExt = 0;
	}


/*	{
		vector<CString> asOldFiles;
		WIN32_FIND_DATA fd;

		CString sPath = szFile;
		HANDLE hFind = FindFirstFile( sPath+"*", &fd );

		if( hFind != INVALID_HANDLE_VALUE )
		{
			do {
				if( fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY )
					continue;
				asOldFiles.push_back( fd.cFileName );
			} while( FindNextFile(hFind, &fd) );
			FindClose( hFind );
		}
	}
*/

	CMenu *pMenu = GetMenu();
	const bool bExportStrokeTemplates = !!( pMenu->GetMenuState(ID_OPTIONS_EXPORTSTROKETEMPLATES, 0) & MF_CHECKED );
	const bool bDoubleRes = !!( pMenu->GetMenuState(ID_OPTIONS_DOUBLERES, 0) & MF_CHECKED );
	if( bDoubleRes )
	{
		g_pTextureFont->Save( szFile, "", true, false, bExportStrokeTemplates );	// save metrics
		UpdateFont( true );	// generate DoubleRes bitmaps
		g_pTextureFont->Save( szFile, " (doubleres)", false, true, bExportStrokeTemplates );	// save bitmaps
		// reset to normal, non-DoubleRes font
		m_bUpdateFontNeeded = true;
		Invalidate( FALSE );
		UpdateWindow();
	}
	else	// normal res
	{
		g_pTextureFont->Save( szFile, "", true, true, bExportStrokeTemplates );	// save metrics and bitmaps
	}
}

void CTextureFontGeneratorDlg::OnFileExit()
{
	DestroyWindow();
}

/*
 * Copyright (c) 2003-2007 Glenn Maynard
 * All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, and/or sell copies of the Software, and to permit persons to
 * whom the Software is furnished to do so, provided that the above
 * copyright notice(s) and this permission notice appear in all copies of
 * the Software and that both the above copyright notice(s) and this
 * permission notice appear in supporting documentation.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT OF
 * THIRD PARTY RIGHTS. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR HOLDERS
 * INCLUDED IN THIS NOTICE BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL INDIRECT
 * OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */
